@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -13,11 +13,11 @@ package net.sf.saxon;
 
 public final class Version {
 
-    private static final int[] STRUCTURED_VERSION = {9, 7, 0, 10};
-    private static final String VERSION = "9.7.0.10";
-    private static final String BUILD = "102116"; //mmddhh
-    private static final String RELEASE_DATE = "2016-10-21";
-    private static final String MAJOR_RELEASE_DATE = "2015-11-27";
+    private static final int[] STRUCTURED_VERSION = {9, 8, 0, 8};
+    private static final String VERSION = "9.8.0.8";
+    private static final String BUILD = "020515"; //mmddhh
+    private static final String RELEASE_DATE = "2018-02-05"; //yyy-mm-dd
+    private static final String MAJOR_RELEASE_DATE = "2017-06-08";
 
     private Version() {
         // class is never instantiated
@@ -158,11 +158,18 @@ public final class Version {
         //#endif
 //#else
             platform = new net.sf.saxon.java.JavaPlatform();
-        //#if PE==true
+
+        //#if PE==true && DOTNET != true && SAXONC == true
+                platform = new com.saxonica.config.JavaPlatformC();
+        //#endif
+        //#if EE==true && DOTNET != true && SAXONC==true
+                platform = new com.saxonica.config.JavaPlatformC();
+        //#endif
+        //#if PE==true && SAXONC != true
             platform = new com.saxonica.config.JavaPlatformPE();
         //#endif
-        //#if EE==true
-            platform = new com.saxonica.config.JavaPlatformPE();
+        //#if EE==true  && SAXONC != true
+            platform = new com.saxonica.config.JavaPlatformEE();
         //#endif
 //#endif
     }
