@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,16 +7,18 @@
 
 package net.sf.saxon.event;
 
-import net.sf.saxon.expr.parser.Location;
-import net.sf.saxon.ma.arrays.ArrayItem;
-import net.sf.saxon.ma.map.MapItem;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.parser.ExplicitLocation;
+import net.sf.saxon.expr.parser.Location;
+import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.lib.ParseOptions;
 import net.sf.saxon.lib.Validation;
+import net.sf.saxon.ma.arrays.ArrayItem;
+import net.sf.saxon.ma.map.MapItem;
 import net.sf.saxon.om.*;
 import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.XPathException;
+import net.sf.saxon.tree.util.Orphan;
 import net.sf.saxon.type.BuiltInAtomicType;
 import net.sf.saxon.type.SchemaType;
 import net.sf.saxon.type.SimpleType;
@@ -24,6 +26,8 @@ import net.sf.saxon.type.Type;
 import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.ObjectValue;
 
+import java.util.Arrays;
+
 /**
  * This class is used for generating complex content, that is, the content of an
  * element or document node. It enforces the rules on the order of events within
@@ -53,6 +57,7 @@ public final class ComplexContentOutputter extends SequenceReceiver {
     // are ignored unless they occur at the outermost level, except that they
     // still change the level number
     private boolean[] currentLevelIsDocument = new boolean[20];
+    private InScopeNamespaces[] copyNamespacesStack = new InScopeNamespaces[20];
     private Boolean elementIsInNullNamespace;
     private NodeName[] pendingAttCode = new NodeName[20];
     private SimpleType[] pendingAttType = new SimpleType[20];
@@ -98,6 +103,14 @@ public final class ComplexContentOutputter extends SequenceReceiver {
     public static SequenceReceiver makeComplexContentReceiver(Receiver receiver, ParseOptions options)
         throws XPathException {
 //        System.err.println("CHANGE OUTPUT DESTINATION new=" + receiver);
+
+        String systemId = receiver.getSystemId();
+        boolean validate = options != null && options.getSchemaValidationMode() != Validation.PRESERVE;
+
+        if (receiver instanceof ComplexContentOutputter && !validate) {
+            return (ComplexContentOutputter) receiver;
+        }
+
         PipelineConfiguration pipe = receiver.getPipelineConfiguration();
         ComplexContentOutputter out = new ComplexContentOutputter(pipe);
         out.setHostLanguage(pipe.getHostLanguage());
@@ -110,12 +123,13 @@ public final class ComplexContentOutputter extends SequenceReceiver {
 
         // add a validator to the pipeline if required
 
-        if (options != null && options.getSchemaValidationMode() != Validation.PRESERVE) {
+        if (validate) {
             Configuration config = pipe.getConfiguration();
-            receiver = config.getDocumentValidator(ne, receiver.getSystemId(), options);
+            receiver = config.getDocumentValidator(ne, receiver.getSystemId(), options, null);
         }
 
         out.setReceiver(receiver);
+        out.setSystemId(systemId);
         return out;
     }
 
@@ -210,9 +224,7 @@ public final class ComplexContentOutputter extends SequenceReceiver {
         }
         previousAtomic = false;
         if (currentLevelIsDocument.length < level + 1) {
-            boolean[] b2 = new boolean[level * 2];
-            System.arraycopy(currentLevelIsDocument, 0, b2, 0, level);
-            currentLevelIsDocument = b2;
+            currentLevelIsDocument = Arrays.copyOf(currentLevelIsDocument, level * 2);
         }
         currentLevelIsDocument[level] = true;
     }
@@ -229,6 +241,17 @@ public final class ComplexContentOutputter extends SequenceReceiver {
         level--;
     }
 
+    /**
+     * Notify an unparsed entity URI.
+     *
+     * @param name     The name of the unparsed entity
+     * @param systemID The system identifier of the unparsed entity
+     * @param publicID The public identifier of the unparsed entity
+     */
+    @Override
+    public void setUnparsedEntity(String name, String systemID, String publicID) throws XPathException {
+        nextReceiver.setUnparsedEntity(name, systemID, publicID);
+    }
 
     /**
      * Produce text content output. <BR>
@@ -281,97 +304,108 @@ public final class ComplexContentOutputter extends SequenceReceiver {
         currentSimpleType = typeCode;
         previousAtomic = false;
         if (currentLevelIsDocument.length < level + 1) {
-            boolean[] b2 = new boolean[level * 2];
-            System.arraycopy(currentLevelIsDocument, 0, b2, 0, level);
-            currentLevelIsDocument = b2;
+            currentLevelIsDocument = Arrays.copyOf(currentLevelIsDocument, level * 2);
+        }
+        if (copyNamespacesStack.length < level + 1) {
+            copyNamespacesStack = Arrays.copyOf(copyNamespacesStack, level * 2);
         }
         currentLevelIsDocument[level] = false;
     }
 
 
     /**
-     * Output a namespace declaration. <br>
+     * Output one or more namespace declarations. <br>
      * This is added to a list of pending namespaces for the current start tag.
      * If there is already another declaration of the same prefix, this one is
      * ignored, unless the REJECT_DUPLICATES flag is set, in which case this is an error.
      * Note that unlike SAX2 startPrefixMapping(), this call is made AFTER writing the start tag.
      *
-     * @param nsBinding The namespace binding
+     * @param nsBindings The namespace bindings
      * @throws XPathException if there is no start tag to write to (created using writeStartTag),
      *                        or if character content has been written since the start tag was written.
      */
 
-    public void namespace(NamespaceBinding nsBinding, int properties)
+    public void namespace(NamespaceBindingSet nsBindings, int properties)
             throws XPathException {
 
-        // System.err.println("Write namespace prefix=" + (nscode>>16) + " uri=" + (nscode&0xffff));
-        if (pendingStartTagDepth < 0) {
-            throw NoOpenStartTagException.makeNoOpenStartTagException(
-                    Type.NAMESPACE,
-                    nsBinding.getPrefix(),
-                    hostLanguage,
-                    pendingStartTagDepth == -2,
-                    isSerializing(),
-                startElementLocationId);
-        }
-
-        // elimination of namespaces already present on an outer element of the
-        // result tree is done by the NamespaceReducer.
-
-        // Handle declarations whose prefix is duplicated for this element.
-
-        boolean rejectDuplicates = (properties & ReceiverOptions.REJECT_DUPLICATES) != 0;
-
-        for (int i = 0; i < pendingNSListSize; i++) {
-            if (nsBinding.equals(pendingNSList[i])) {
-                // same prefix and URI: ignore this duplicate
+        // Optimization for recursive shallow-copy added in 9.8 - see bug 3011.
+        if (nsBindings instanceof InScopeNamespaces) {
+            copyNamespacesStack[level] = (InScopeNamespaces)nsBindings;
+            if (level > 1 && copyNamespacesStack[level-1] instanceof InScopeNamespaces &&
+                    copyNamespacesStack[level - 1].getElement().isSameNodeInfo(((InScopeNamespaces)nsBindings).getElement())) {
+                // Ignore these namespaces if they are the same as the namespaces for the parent element
                 return;
             }
-            if (nsBinding.getPrefix().equals(pendingNSList[i].getPrefix())) {
-                if (pendingNSList[i].isDefaultUndeclaration() || nsBinding.isDefaultUndeclaration()) {
-                    // xmlns="" overridden by xmlns="abc"
-                    pendingNSList[i] = nsBinding;
-                } else if (rejectDuplicates) {
-                    String prefix = nsBinding.getPrefix();
-                    String uri1 = nsBinding.getURI();
-                    String uri2 = pendingNSList[i].getURI();
-                    XPathException err = new XPathException("Cannot create two namespace nodes with the same prefix mapped to different URIs (prefix=" +
-                            (prefix.isEmpty() ? "\"\"" : prefix) + ", URI=" +
-                            (uri1.isEmpty() ? "\"\"" : uri1) + ", URI=" +
-                            (uri2.isEmpty() ? "\"\"" : uri2) + ")");
-                    err.setErrorCode(hostLanguage == Configuration.XSLT ? "XTDE0430" : "XQDY0102");
-                    throw err;
-                } else {
-                    // same prefix, do a quick exit
+        }
+
+        // System.err.println("Write namespace prefix=" + (nscode>>16) + " uri=" + (nscode&0xffff));
+        for (NamespaceBinding ns : nsBindings) {
+            if (pendingStartTagDepth < 0) {
+                throw NoOpenStartTagException.makeNoOpenStartTagException(
+                        Type.NAMESPACE,
+                        ns.getPrefix(),
+                        hostLanguage,
+                        pendingStartTagDepth == -2,
+                        isSerializing(),
+                    startElementLocationId);
+            }
+
+            // elimination of namespaces already present on an outer element of the
+            // result tree is done by the NamespaceReducer.
+
+            // Handle declarations whose prefix is duplicated for this element.
+
+            boolean rejectDuplicates = (properties & ReceiverOptions.REJECT_DUPLICATES) != 0;
+
+            for (int i = 0; i < pendingNSListSize; i++) {
+                if (nsBindings.equals(pendingNSList[i])) {
+                    // same prefix and URI: ignore this duplicate
                     return;
                 }
+                if (ns.getPrefix().equals(pendingNSList[i].getPrefix())) {
+                    if (pendingNSList[i].isDefaultUndeclaration() || ns.isDefaultUndeclaration()) {
+                        // xmlns="" overridden by xmlns="abc"
+                        pendingNSList[i] = ns;
+                    } else if (rejectDuplicates) {
+                        String prefix = ns.getPrefix();
+                        String uri1 = ns.getURI();
+                        String uri2 = pendingNSList[i].getURI();
+                        XPathException err = new XPathException("Cannot create two namespace nodes with the same prefix mapped to different URIs (prefix=" +
+                                (prefix.isEmpty() ? "\"\"" : prefix) + ", URI=" +
+                                (uri1.isEmpty() ? "\"\"" : uri1) + ", URI=" +
+                                (uri2.isEmpty() ? "\"\"" : uri2) + ")");
+                        err.setErrorCode(hostLanguage == Configuration.XSLT ? "XTDE0430" : "XQDY0102");
+                        throw err;
+                    } else {
+                        // same prefix, do a quick exit
+                        return;
+                    }
+                }
             }
-        }
 
-        // It is an error to output a namespace node for the default namespace if the element
-        // itself is in the null namespace, as the resulting element could not be serialized
+            // It is an error to output a namespace node for the default namespace if the element
+            // itself is in the null namespace, as the resulting element could not be serialized
 
-        if (nsBinding.getPrefix().isEmpty() && (nsBinding.getURI().length() != 0)) {
-            declaresDefaultNamespace = true;
-            if (elementIsInNullNamespace == null) {
-                elementIsInNullNamespace = pendingStartTag.hasURI("");
+            if (ns.getPrefix().isEmpty() && (ns.getURI().length() != 0)) {
+                declaresDefaultNamespace = true;
+                if (elementIsInNullNamespace == null) {
+                    elementIsInNullNamespace = pendingStartTag.hasURI("");
+                }
+                if (elementIsInNullNamespace) {
+                    XPathException err = new XPathException("Cannot output a namespace node for the default namespace when the element is in no namespace");
+                    err.setErrorCode(hostLanguage == Configuration.XSLT ? "XTDE0440" : "XQDY0102");
+                    throw err;
+                }
             }
-            if (elementIsInNullNamespace) {
-                XPathException err = new XPathException("Cannot output a namespace node for the default namespace when the element is in no namespace");
-                err.setErrorCode(hostLanguage == Configuration.XSLT ? "XTDE0440" : "XQDY0102");
-                throw err;
+
+            // if it's not a duplicate namespace, add it to the list for this start tag
+
+            if (pendingNSListSize + 1 > pendingNSList.length) {
+                pendingNSList = Arrays.copyOf(pendingNSList, pendingNSListSize * 2);
             }
+            pendingNSList[pendingNSListSize++] = ns;
+            previousAtomic = false;
         }
-
-        // if it's not a duplicate namespace, add it to the list for this start tag
-
-        if (pendingNSListSize + 1 > pendingNSList.length) {
-            NamespaceBinding[] newlist = new NamespaceBinding[pendingNSListSize * 2];
-            System.arraycopy(pendingNSList, 0, newlist, 0, pendingNSListSize);
-            pendingNSList = newlist;
-        }
-        pendingNSList[pendingNSListSize++] = nsBinding;
-        previousAtomic = false;
     }
 
 
@@ -443,21 +477,11 @@ public final class ComplexContentOutputter extends SequenceReceiver {
         // otherwise, add this one to the list
 
         if (pendingAttListSize >= pendingAttCode.length) {
-            NodeName[] attCode2 = new NodeName[pendingAttListSize * 2];
-            SimpleType[] attType2 = new SimpleType[pendingAttListSize * 2];
-            String[] attValue2 = new String[pendingAttListSize * 2];
-            Location[] attLoc2 = new Location[pendingAttListSize * 2];
-            int[] attProp2 = new int[pendingAttListSize * 2];
-            System.arraycopy(pendingAttCode, 0, attCode2, 0, pendingAttListSize);
-            System.arraycopy(pendingAttType, 0, attType2, 0, pendingAttListSize);
-            System.arraycopy(pendingAttValue, 0, attValue2, 0, pendingAttListSize);
-            System.arraycopy(pendingAttLocation, 0, attLoc2, 0, pendingAttListSize);
-            System.arraycopy(pendingAttProp, 0, attProp2, 0, pendingAttListSize);
-            pendingAttCode = attCode2;
-            pendingAttType = attType2;
-            pendingAttValue = attValue2;
-            pendingAttLocation = attLoc2;
-            pendingAttProp = attProp2;
+            pendingAttCode = Arrays.copyOf(pendingAttCode, pendingAttListSize * 2);
+            pendingAttType = Arrays.copyOf(pendingAttType, pendingAttListSize * 2);
+            pendingAttValue = Arrays.copyOf(pendingAttValue, pendingAttListSize * 2);
+            pendingAttLocation = Arrays.copyOf(pendingAttLocation, pendingAttListSize * 2);
+            pendingAttProp = Arrays.copyOf(pendingAttProp, pendingAttListSize * 2);
         }
 
         pendingAttCode[pendingAttListSize] = attName;
@@ -477,7 +501,9 @@ public final class ComplexContentOutputter extends SequenceReceiver {
      * if not.
      *
      * @param nodeName the proposed name, including proposed prefix
-     * @param seq      sequence number, used for generating a substitute prefix when necessary
+     * @param seq      sequence number, used for generating a substitute prefix when necessary.
+     *                 The value 0 is used for element names; values greater than 0 are used
+     *                 for attribute names.
      * @return a nameCode to use in place of the proposed nameCode (or the original nameCode
      *         if no change is needed)
      * @throws net.sf.saxon.trans.XPathException
@@ -504,6 +530,14 @@ public final class ComplexContentOutputter extends SequenceReceiver {
             }
         }
         // no declaration of this prefix: declare it now
+        if (seq > 0 && nsprefix.isEmpty()) {
+            // This is an attribute and the prefix is "" - need to invent a prefix
+            // See bug 3068 and unit test ParserTest/testXercesSchemaDefaultedAttributes
+            String prefix = getSubstitutePrefix(binding, seq);
+            NodeName newName = new FingerprintedQName(prefix, nodeName.getURI(), nodeName.getLocalPart());
+            namespace(newName.getNamespaceBinding(), 0);
+            return newName;
+        }
         namespace(binding, 0);
         return nodeName;
     }
@@ -520,6 +554,9 @@ public final class ComplexContentOutputter extends SequenceReceiver {
      */
 
     private String getSubstitutePrefix(NamespaceBinding nscode, int seq) {
+        if (nscode.getURI().equals(NamespaceConstant.XML)) {
+            return "xml";
+        }
         return nscode.getPrefix() + '_' + seq;
     }
 
@@ -612,6 +649,10 @@ public final class ComplexContentOutputter extends SequenceReceiver {
                 }
                 endDocument();
                 previousAtomic = false;
+            } else if (item instanceof Orphan && ((Orphan) item).isDisableOutputEscaping()) {
+                // see test case doe-0185 - needed for output buffered within try/catch
+                characters(item.getStringValueCS(), locationId, ReceiverOptions.DISABLE_ESCAPING);
+                previousAtomic = false;
             } else {
                 int copyOptions = CopyOptions.TYPE_ANNOTATIONS;
                 if (copyNamespaces == NodeInfo.LOCAL_NAMESPACES) {
@@ -696,5 +737,17 @@ public final class ComplexContentOutputter extends SequenceReceiver {
     public boolean usesTypeAnnotations() {
         return nextReceiver.usesTypeAnnotations();
     }
+
+    public void beforeBulkCopy() throws XPathException {
+        level++;
+        if (pendingStartTagDepth >= 0) {
+            startContent();
+        }
+    }
+
+    public void afterBulkCopy() {
+        level--;
+        previousAtomic = false;
+    }
 }
 
