@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -17,6 +17,7 @@ import net.sf.saxon.lib.SaxonOutputKeys;
 import net.sf.saxon.lib.TraceListener;
 import net.sf.saxon.om.Item;
 import net.sf.saxon.om.NamespaceBinding;
+import net.sf.saxon.om.NamespaceBindingSet;
 import net.sf.saxon.om.NodeName;
 import net.sf.saxon.trace.InstructionInfo;
 import net.sf.saxon.trans.SaxonErrorCode;
@@ -32,6 +33,7 @@ import org.xml.sax.SAXException;
 import org.xml.sax.ext.LexicalHandler;
 
 import javax.xml.transform.Result;
+import javax.xml.transform.sax.TransformerHandler;
 import java.util.Properties;
 import java.util.Stack;
 
@@ -183,7 +185,7 @@ public class ContentHandlerProxy implements Receiver {
     }
 
     /**
-     * Notify an unparsed entity URI. This implementation does nothing: the event is ignored.
+     * Notify an unparsed entity URI.
      *
      * @param name     The name of the unparsed entity
      * @param systemID The system identifier of the unparsed entity
@@ -191,7 +193,15 @@ public class ContentHandlerProxy implements Receiver {
      */
 
     public void setUnparsedEntity(String name, String systemID, String publicID) throws XPathException {
-        // no-op
+        // Avoid passing unparsed entities to a general-purpose DTDHandler because it's probably
+        // expecting them in the context of the rest of the DTD, which we don't make available.
+        if (handler instanceof TransformerHandler) {
+            try {
+                ((TransformerHandler)handler).unparsedEntityDecl(name, publicID, systemID, "unknown");
+            } catch (SAXException e) {
+                throw new XPathException(e);
+            }
+        }
     }
 
     /**
@@ -326,21 +336,23 @@ public class ContentHandlerProxy implements Receiver {
      * any children for the element.
      */
 
-    public void namespace(NamespaceBinding namespaceBinding, int properties) throws XPathException {
-        String prefix = namespaceBinding.getPrefix();
-        if (prefix.equals("xml")) {
-            return;
-        }
-        String uri = namespaceBinding.getURI();
-        if (!undeclareNamespaces && uri.isEmpty() && prefix.length() != 0) {
-            // This is a namespace undeclaration, but the ContentHandler doesn't want to know about undeclarations
-            return;
-        }
-        try {
-            handler.startPrefixMapping(prefix, uri);
-            namespaceStack.push(prefix);
-        } catch (SAXException err) {
-            handleSAXException(err);
+    public void namespace(NamespaceBindingSet namespaceBindings, int properties) throws XPathException {
+        for (NamespaceBinding ns : namespaceBindings) {
+            String prefix = ns.getPrefix();
+            if (prefix.equals("xml")) {
+                return;
+            }
+            String uri = ns.getURI();
+            if (!undeclareNamespaces && uri.isEmpty() && prefix.length() != 0) {
+                // This is a namespace undeclaration, but the ContentHandler doesn't want to know about undeclarations
+                return;
+            }
+            try {
+                handler.startPrefixMapping(prefix, uri);
+                namespaceStack.push(prefix);
+            } catch (SAXException err) {
+                handleSAXException(err);
+            }
         }
     }
 
