@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -10,9 +10,6 @@ package net.sf.saxon.event;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.Version;
-import net.sf.saxon.evpull.EventIterator;
-import net.sf.saxon.evpull.EventIteratorToReceiver;
-import net.sf.saxon.evpull.PullEventSource;
 import net.sf.saxon.expr.number.Numberer_en;
 import net.sf.saxon.expr.parser.ExplicitLocation;
 import net.sf.saxon.expr.parser.Location;
@@ -25,14 +22,13 @@ import net.sf.saxon.pull.StaxBridge;
 import net.sf.saxon.trans.SaxonErrorCode;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.Type;
-import net.sf.saxon.value.Whitespace;
 import org.xml.sax.*;
 
 import javax.xml.stream.XMLStreamReader;
 import javax.xml.transform.Source;
 import javax.xml.transform.sax.SAXSource;
+import javax.xml.transform.stax.StAXSource;
 import javax.xml.transform.stream.StreamSource;
-import java.lang.reflect.Method;
 import java.util.List;
 import java.util.Map;
 
@@ -45,15 +41,6 @@ public abstract class Sender {
     // Converted to an abstract static class in Saxon 9.3
 
     private static Class[] EMPTY_CLASS_ARRAY = new Class[0];
-    private static Class staxSourceClass;
-
-    static {
-        try {
-            staxSourceClass = Class.forName("javax.xml.transform.stax.StAXSource");
-        } catch (Exception err) {
-            // no action; if StAXSource isn't available then we don't use it.
-        }
-    }
 
     private Sender() {
     }
@@ -106,16 +93,18 @@ public abstract class Sender {
             }
         }
 
-        if (options.getSpaceStrippingRule() != null) {
-            next = new Stripper(options.getSpaceStrippingRule(), next);
-        } else if (options.getStripSpace() == Whitespace.ALL) {
-            next = new Stripper(AllElementsSpaceStrippingRule.getInstance(), next);
-        } else if (options.getStripSpace() == Whitespace.XSLT) {
-            Controller controller = pipe.getController();
-            if (controller != null) {
-                next = controller.makeStripper(next);
-            }
+        SpaceStrippingRule strippingRule = options.getSpaceStrippingRule();
+        if (strippingRule != null && !(strippingRule instanceof NoElementsSpaceStrippingRule)) {
+            next = strippingRule.makeStripper(next);
         }
+//        else if (options.getStripSpace() == Whitespace.ALL) {
+//            next = new Stripper(AllElementsSpaceStrippingRule.getInstance(), next);
+//        } else if (options.getStripSpace() == Whitespace.XSLT) {
+//            Controller controller = pipe.getController();
+//            if (controller != null) {
+//                next = controller.makeStripper(next);
+//            }
+//        }
 
         if (source instanceof TreeInfo) {
             source = ((TreeInfo)source).getRootNode();
@@ -125,7 +114,7 @@ public abstract class Sender {
             NodeInfo ns = (NodeInfo) source;
             String baseURI = ns.getBaseURI();
             if (schemaValidation != Validation.PRESERVE) {
-                next = config.getDocumentValidator(next, baseURI, options);
+                next = config.getDocumentValidator(next, baseURI, options, null);
             }
 
             int kind = ns.getNodeKind();
@@ -141,10 +130,6 @@ public abstract class Sender {
             sendPullSource((PullSource) source, next, options);
             return;
 
-        } else if (source instanceof PullEventSource) {
-            sendPullEventSource((PullEventSource) source, next, options);
-            return;
-
         } else if (source instanceof EventSource) {
             ((EventSource) source).send(next);
             return;
@@ -187,17 +172,8 @@ public abstract class Sender {
                 send(ps, next, options);
             }
             return;
-        } else if (staxSourceClass != null && staxSourceClass.isAssignableFrom(source.getClass())) {
-            // Test for a StAXSource
-            // Use reflection to avoid problems if JAXP 1.4 not installed
-            XMLStreamReader reader = null;
-            try {
-                Method getReaderMethod = staxSourceClass.getMethod("getXMLStreamReader", EMPTY_CLASS_ARRAY);
-                reader = (XMLStreamReader) getReaderMethod.invoke(source);
-            } catch (Exception e) {
-                // no action
-            }
-            //XMLStreamReader reader = ((StAXSource)source).getXMLStreamReader();
+        } else if (source instanceof StAXSource) {
+            XMLStreamReader reader = ((StAXSource)source).getXMLStreamReader();
             if (reader == null) {
                 throw new XPathException("Saxon can only handle a StAXSource that wraps an XMLStreamReader");
             }
@@ -516,7 +492,7 @@ public abstract class Sender {
 //#if EE==true
             // Add a document validator to the pipeline
             if (!(receiver instanceof DocumentValidator)) {
-                receiver = config.getDocumentValidator(receiver, systemId, options);
+                receiver = config.getDocumentValidator(receiver, systemId, options, null);
             }
 //#endif
         }
@@ -555,27 +531,6 @@ public abstract class Sender {
         }
     }
 
-    private static void sendPullEventSource(PullEventSource source, Receiver receiver, ParseOptions options)
-            throws XPathException {
-        PipelineConfiguration pipe = receiver.getPipelineConfiguration();
-        boolean xInclude = options.isXIncludeAware();
-        if (xInclude) {
-            throw new XPathException("XInclude processing is not supported with a pull parser");
-        }
-        receiver = makeValidator(receiver, source.getSystemId(), options);
-
-        receiver.open();
-        EventIterator provider = source.getEventIterator();
-
-        receiver.setPipelineConfiguration(pipe);
-        SequenceReceiver out = SequenceReceiver.isTrueSequenceReceiver(receiver)
-                ? (SequenceReceiver) receiver
-                : new TreeReceiver(receiver);
-        EventIteratorToReceiver.copy(provider, out);
-        receiver.close();
-    }
-
-
     /**
      * Configure a SAX parser to ensure it has the correct namesapce properties set
      *
