@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -18,6 +18,7 @@ import net.sf.saxon.type.SimpleType;
 import net.sf.saxon.type.Type;
 
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.Iterator;
 import java.util.List;
 
@@ -130,9 +131,7 @@ public class StartTagBuffer extends ProxyReceiver implements NamespaceResolver {
 
         countStack[depth] = 0;
         if (++depth >= countStack.length) {
-            int[] newstack = new int[depth * 2];
-            System.arraycopy(countStack, 0, newstack, 0, depth);
-            countStack = newstack;
+            countStack = Arrays.copyOf(countStack, depth * 2);
         }
 
         // Ensure that the element namespace is output, unless this is done
@@ -147,23 +146,25 @@ public class StartTagBuffer extends ProxyReceiver implements NamespaceResolver {
         attCount = 0;
     }
 
-    public void namespace(NamespaceBinding namespaceBinding, int properties) throws XPathException {
+    public void namespace(NamespaceBindingSet namespaceBindings, int properties) throws XPathException {
 
-        if (!acceptAttributes) {
-            throw NoOpenStartTagException.makeNoOpenStartTagException(
-                    Type.NAMESPACE, namespaceBinding.getPrefix(),
-                    getPipelineConfiguration().getHostLanguage(),
-                    inDocument, false, ExplicitLocation.UNKNOWN_LOCATION);
-        }
-
-        // avoid duplicates
-        for (int n = 0; n < countStack[depth - 1]; n++) {
-            if (namespaces[namespacesSize - 1 - n].equals(namespaceBinding)) {
-                return;
+        for (NamespaceBinding ns : namespaceBindings) {
+            if (!acceptAttributes) {
+                throw NoOpenStartTagException.makeNoOpenStartTagException(
+                        Type.NAMESPACE, ns.getPrefix(),
+                        getPipelineConfiguration().getHostLanguage(),
+                        inDocument, false, ExplicitLocation.UNKNOWN_LOCATION);
             }
+
+            // avoid duplicates
+            for (int n = 0; n < countStack[depth - 1]; n++) {
+                if (namespaces[namespacesSize - 1 - n].equals(namespaceBindings)) {
+                    return;
+                }
+            }
+            addToStack(ns);
+            countStack[depth - 1]++;
         }
-        addToStack(namespaceBinding);
-        countStack[depth - 1]++;
     }
 
     /**
@@ -212,9 +213,7 @@ public class StartTagBuffer extends ProxyReceiver implements NamespaceResolver {
     private void addToStack(NamespaceBinding binding) {
         // expand the stack if necessary
         if (namespacesSize + 1 >= namespaces.length) {
-            NamespaceBinding[] newlist = new NamespaceBinding[namespacesSize * 2];
-            System.arraycopy(namespaces, 0, newlist, 0, namespacesSize);
-            namespaces = newlist;
+            namespaces = Arrays.copyOf(namespaces, namespacesSize * 2);
         }
         namespaces[namespacesSize++] = binding;
     }
