@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,16 +7,14 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.AtomizerCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.AtomizerAdjunct;
-import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.ma.arrays.ArrayItemType;
 import net.sf.saxon.event.ReceiverOptions;
 import net.sf.saxon.expr.instruct.Block;
 import net.sf.saxon.expr.instruct.Choose;
 import net.sf.saxon.expr.instruct.ValueOf;
+import net.sf.saxon.expr.parser.*;
+import net.sf.saxon.ma.arrays.ArrayItemType;
 import net.sf.saxon.om.*;
+import net.sf.saxon.pattern.NameTest;
 import net.sf.saxon.pattern.NodeKindTest;
 import net.sf.saxon.pattern.NodeTest;
 import net.sf.saxon.trans.XPathException;
@@ -25,8 +23,6 @@ import net.sf.saxon.type.*;
 import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.Cardinality;
 
-import java.util.Map;
-
 
 /**
  * An Atomizer is an expression corresponding essentially to the fn:data() function: it
@@ -107,13 +103,8 @@ public final class Atomizer extends UnaryExpression {
                 return operand;
             }
             SequenceIterator iter = val.iterate();
-            while (true) {
-                // if all items in the sequence are atomic (they generally will be, since this is
-                // done at compile time), then return the sequence
-                Item i = iter.next();
-                if (i == null) {
-                    return operand;
-                }
+            Item i;
+            while ((i = iter.next()) != null) {
                 if (i instanceof NodeInfo) {
                     return this;
                 }
@@ -127,6 +118,9 @@ public final class Atomizer extends UnaryExpression {
                     }
                 }
             }
+            // if all items in the sequence are atomic (they generally will be, since this is
+            // done at compile time), then return the sequence
+            return operand;
         } else if (operand instanceof ValueOf && (((ValueOf) operand).getOptions() & ReceiverOptions.DISABLE_ESCAPING) == 0) {
             // XSLT users tend to use ValueOf unnecessarily
             return ((ValueOf) operand).convertToCastAsString();
@@ -145,7 +139,7 @@ public final class Atomizer extends UnaryExpression {
         untyped = untyped | !visitor.getStaticContext().getPackageData().isSchemaAware();
 
         // If the configuration allows typed data, check whether the content type of these particular nodes is untyped
-        final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        final TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
         computeSingleValued(th);
         resetLocalStaticProperties();
         ItemType operandType = getOperandItemType();
@@ -212,7 +206,7 @@ public final class Atomizer extends UnaryExpression {
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         Expression exp = super.optimize(visitor, contextInfo);
         if (exp == this) {
-            final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+            final TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
             Expression operand = getBaseExpression();
             if (th.isSubType(operand.getItemType(), BuiltInAtomicType.ANY_ATOMIC)) {
                 return operand;
@@ -244,6 +238,23 @@ public final class Atomizer extends UnaryExpression {
                 Block newBlock = new Block(atomizedChildren);
                 return newBlock.typeCheck(visitor, contextInfo).optimize(visitor, contextInfo);
             }
+            if (untyped && operand instanceof AxisExpression &&
+                    ((AxisExpression)operand).getAxis() == AxisInfo.ATTRIBUTE &&
+                    ((AxisExpression) operand).getNodeTest() instanceof NameTest) {
+                StructuredQName name = ((AxisExpression) operand).getNodeTest().getMatchingNodeName();
+                FingerprintedQName qName = new FingerprintedQName(name, visitor.getConfiguration().getNamePool());
+                AttributeGetter ag = new AttributeGetter(qName);
+                int checks = 0;
+                if (((AxisExpression) operand).isContextPossiblyUndefined()) {
+                    checks |= AttributeGetter.CHECK_CONTEXT_ITEM_PRESENT;
+                }
+                if (!(((AxisExpression) operand).getContextItemType() instanceof NodeTest)) {
+                    checks |= AttributeGetter.CHECK_CONTEXT_ITEM_IS_NODE;
+                }
+                ag.setRequiredChecks(checks);
+                ExpressionTool.copyLocationInfo(this, ag);
+                return ag;
+            }
         }
         return exp;
     }
@@ -270,6 +281,16 @@ public final class Atomizer extends UnaryExpression {
         return p | StaticProperty.NON_CREATIVE;
     }
 
+    /**
+     * Reset the static properties of the expression to -1, so that they have to be recomputed
+     * next time they are used.
+     */
+    @Override
+    public void resetLocalStaticProperties() {
+        super.resetLocalStaticProperties();
+        operandItemType = null;
+    }
+
     /**
      * Copy an expression. This makes a deep copy.
      *
@@ -286,20 +307,11 @@ public final class Atomizer extends UnaryExpression {
         return copy;
     }
 
-    //#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
     @Override
-    public AtomizerAdjunct getStreamingAdjunct() {
-        return new AtomizerAdjunct();
+    public String getStreamerName() {
+        return "Atomizer";
     }
 
-//#endif
-
     /**
      * Iterate over the sequence of values
      */
@@ -307,7 +319,7 @@ public final class Atomizer extends UnaryExpression {
     /*@NotNull*/
     public SequenceIterator iterate(XPathContext context) throws XPathException {
         SequenceIterator base = getBaseExpression().iterate(context);
-        return getAtomizingIterator(base, untyped);
+        return getAtomizingIterator(base, untyped && operandItemType instanceof NodeTest);
     }
 
     /**
@@ -376,7 +388,8 @@ public final class Atomizer extends UnaryExpression {
         } else if (in instanceof JavaExternalObjectType) {
             return in.getAtomizedItemType();
         } else if (in instanceof ArrayItemType) {
-            return ((ArrayItemType)in).getMemberType().getPrimaryType().getAtomizedItemType();
+            PlainType ait = ((ArrayItemType)in).getMemberType().getPrimaryType().getAtomizedItemType();
+            return ait == null ? ErrorType.getInstance() : ait;
         } else if (in instanceof FunctionItemType) {
             return ErrorType.getInstance();
         }
@@ -468,14 +481,15 @@ public final class Atomizer extends UnaryExpression {
      * iterator
      *
      * @param base    the supplied iterator, the input to atomization
-     * @param untyped this can safely be set to true if it is known that all nodes in the base sequence will
-     *                be untyped; but it is always OK to set it to false.
+     * @param oneToOne this can safely be set to true if it is known that all nodes in the base sequence will
+     *                be untyped and that there will be no arrays in the base sequence; but it is always OK
+     *                to set it to false.
      * @return an iterator that returns atomic values, the result of the atomization
      * @throws net.sf.saxon.trans.XPathException
      *          if a dynamic evaluation error occurs
      */
 
-    public static SequenceIterator getAtomizingIterator(SequenceIterator base, boolean untyped) throws XPathException {
+    public static SequenceIterator getAtomizingIterator(SequenceIterator base, boolean oneToOne) throws XPathException {
         int properties = base.getProperties();
         if ((properties & SequenceIterator.LAST_POSITION_FINDER) != 0) {
             int count = ((LastPositionFinder) base).getLength();
@@ -488,27 +502,13 @@ public final class Atomizer extends UnaryExpression {
         } else if ((properties & SequenceIterator.ATOMIZING) != 0) {
             return new AxisAtomizingIterator((AtomizedValueIterator)base);
         }
-        if (untyped) {
+        if (oneToOne) {
             return new UntypedAtomizingIterator(base);
         } else {
             return new AtomizingIterator(base);
         }
     }
 
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Atomizer expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new AtomizerCompiler();
-    }
-    //#endif
-
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -537,7 +537,7 @@ public final class Atomizer extends UnaryExpression {
      * Implement the mapping function. This is stateless, so there is a singleton instance.
      */
 
-    public static class AtomizingFunction implements MappingFunction<Item, AtomicValue> {
+    public static class AtomizingFunction implements MappingFunction {
 
         /**
          * Private constructor, ensuring that everyone uses the singleton instance
