@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,10 +7,6 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.AxisExpressionCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.AxisExpressionAdjunct;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.om.*;
@@ -24,7 +20,9 @@ import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.z.IntIterator;
 import net.sf.saxon.z.IntSet;
 
-import java.util.*;
+import java.util.HashSet;
+import java.util.Iterator;
+import java.util.Set;
 
 
 /**
@@ -89,7 +87,6 @@ public final class AxisExpression extends Expression {
 
     /**
      * Simplify an expression
-     *
      */
 
     /*@NotNull*/
@@ -116,7 +113,7 @@ public final class AxisExpression extends Expression {
         doneTypeCheck = true;
         if (contextInfo.getItemType() == ErrorType.getInstance()) {
             XPathException err = new XPathException("Axis step " + toString() +
-                    " cannot be used here: the context item is absent");
+                                                            " cannot be used here: the context item is absent");
             err.setIsTypeError(true);
             err.setErrorCode("XPDY0002");
             err.setLocation(getLocation());
@@ -124,13 +121,13 @@ public final class AxisExpression extends Expression {
         } else {
             staticInfo = contextInfo;
         }
-        Configuration config = getConfiguration();
+        Configuration config = visitor.getConfiguration();
         TypeHierarchy th = config.getTypeHierarchy();
         int relation = th.relationship(contextInfo.getItemType(), AnyNodeTest.getInstance());
 
         if (relation == TypeHierarchy.DISJOINT) {
             XPathException err = new XPathException("Axis step " + toString() +
-                    " cannot be used here: the context item is not a node");
+                                                            " cannot be used here: the context item is not a node");
             err.setIsTypeError(true);
             err.setErrorCode("XPTY0020");
             err.setLocation(getLocation());
@@ -170,16 +167,16 @@ public final class AxisExpression extends Expression {
         if (test != null && !AxisInfo.getTargetUType(UType.ANY_NODE, axis).overlaps(test.getUType())) {
             if (warnings) {
                 visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select " +
-                                test.getUType().toStringWithIndefiniteArticle(),
-                        getLocation());
+                                             test.getUType().toStringWithIndefiniteArticle(),
+                                     getLocation());
             }
             return Literal.makeEmptySequence();
         }
 
-        if (test instanceof NameTest && axis == AxisInfo.NAMESPACE && !((NameTest)test).getNamespaceURI().isEmpty()) {
+        if (test instanceof NameTest && axis == AxisInfo.NAMESPACE && !((NameTest) test).getNamespaceURI().isEmpty()) {
             if (warnings) {
                 visitor.issueWarning("The names of namespace nodes are never prefixed, so this axis step will never select anything",
-                    getLocation());
+                                     getLocation());
             }
             return Literal.makeEmptySequence();
         }
@@ -191,8 +188,8 @@ public final class AxisExpression extends Expression {
         if (targetUType.equals(UType.VOID)) {
             if (warnings) {
                 visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis starting at " +
-                                originUType.toStringWithIndefiniteArticle() + " node will never select anything",
-                        getLocation());
+                                             originUType.toStringWithIndefiniteArticle() + " will never select anything",
+                                     getLocation());
             }
             return Literal.makeEmptySequence();
         }
@@ -201,9 +198,9 @@ public final class AxisExpression extends Expression {
         if (!targetUType.overlaps(testUType)) {
             if (warnings) {
                 visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis starting at " +
-                                originUType.toStringWithIndefiniteArticle() + " will never select " +
-                                test.getUType().toStringWithIndefiniteArticle(),
-                        getLocation());
+                                             originUType.toStringWithIndefiniteArticle() + " will never select " +
+                                             test.getUType().toStringWithIndefiniteArticle(),
+                                     getLocation());
             }
             return Literal.makeEmptySequence();
         }
@@ -246,8 +243,9 @@ public final class AxisExpression extends Expression {
 
                             if (selectedElementNames.intersect(outermostElementNames).isEmpty()) {
                                 if (warnings) {
-                                    visitor.issueWarning("Starting at a document node, the step is selecting an element whose name " +
-                                            "is not among the names of child elements permitted for this document node type", getLocation());
+                                    visitor.issueWarning(
+                                            "Starting at a document node, the step is selecting an element whose name " +
+                                                    "is not among the names of child elements permitted for this document node type", getLocation());
                                 }
 
                                 return Literal.makeEmptySequence();
@@ -261,8 +259,8 @@ public final class AxisExpression extends Expression {
                                 SchemaDeclaration decl = config.getElementDeclaration(outermostElementName);
                                 if (decl == null) {
                                     if (warnings) {
-                                        visitor.issueWarning("Element " + config.getNamePool().getDisplayName(outermostElementName) +
-                                                " is not declared in the schema", getLocation());
+                                        visitor.issueWarning("Element " + config.getNamePool().getEQName(outermostElementName) +
+                                                                     " is not declared in the schema", getLocation());
                                     }
                                     itemType = elementTest;
                                 } else {
@@ -302,10 +300,11 @@ public final class AxisExpression extends Expression {
                 if (!(ct == AnyType.getInstance() || ct == Untyped.getInstance() || ct == AnySimpleType.getInstance() ||
                         ct == BuiltInAtomicType.ANY_ATOMIC || ct == BuiltInAtomicType.UNTYPED_ATOMIC ||
                         ct == BuiltInAtomicType.STRING)) {
-                    // TODO: this test could be more precise, e.g. string is not possible for elements and attribute nodes
                     if (warnings) {
-                        visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select any typed nodes, " +
-                                "because the expression is being compiled in an environment that is not schema-aware", getLocation());
+                        visitor.issueWarning(
+                                "The " + AxisInfo.axisName[axis] + " axis will never select any typed nodes, " +
+                                        "because the expression is being compiled in an environment that is not schema-aware",
+                                getLocation());
                     }
                     return Literal.makeEmptySequence();
                 }
@@ -318,16 +317,18 @@ public final class AxisExpression extends Expression {
                 if (warnings) {
                     if ((axis == AxisInfo.CHILD || axis == AxisInfo.DESCENDANT || axis == AxisInfo.DESCENDANT_OR_SELF) &&
                             UType.PARENT_NODE_KINDS.union(UType.ATTRIBUTE).subsumes(kind)) {
-                        visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select any " +
-                                        kind +
-                                        " nodes when starting at " +
+                        visitor.issueWarning(
+                                "The " + AxisInfo.axisName[axis] + " axis will never select any " +
+                                        kind + " nodes when starting at " +
                                         (origin == Type.ATTRIBUTE ? "an attribute node" : getStartingNodeDescription(contentType)),
                                 getLocation());
                     } else if (axis == AxisInfo.CHILD && kind.equals(UType.TEXT) &&
                             (getParentExpression() instanceof Atomizer)) {
-                        visitor.issueWarning("Selecting the text nodes of an element with simple content may give the " +
-                                "wrong answer in the presence of comments or processing instructions. It is usually " +
-                                "better to omit the '/text()' step", getLocation());
+                        visitor.issueWarning(
+                                "Selecting the text nodes of an element with simple content may give the " +
+                                        "wrong answer in the presence of comments or processing instructions. It is usually " +
+                                        "better to omit the '/text()' step",
+                                getLocation());
                     } else if (axis == AxisInfo.ATTRIBUTE) {
                         Iterator extensions = config.getExtensionsOfType(contentType);
                         boolean found = false;
@@ -353,11 +354,12 @@ public final class AxisExpression extends Expression {
                             }
                         }
                         if (!found) {
-                            visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select " +
-                                    (targetName == null ?
-                                            "any attribute nodes" :
-                                            "an attribute node named " + getDiagnosticName(targetName, env)) +
-                                    " when starting at " + getStartingNodeDescription(contentType), getLocation());
+                            visitor.issueWarning(
+                                    "The " + AxisInfo.axisName[axis] + " axis will never select " +
+                                            (targetName == null ?
+                                                     "any attribute nodes" :
+                                                     "an attribute node named " + getDiagnosticName(targetName, env)) +
+                                            " when starting at " + getStartingNodeDescription(contentType), getLocation());
                             // Despite the warning, leave the expression unchanged. This is because
                             // we don't necessarily know about all extended types at compile time:
                             // in particular, we don't seal the XML Schema namespace to block extensions
@@ -372,10 +374,10 @@ public final class AxisExpression extends Expression {
                 // can never be defined as an extension of a type with simple content
                 if (warnings) {
                     visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select any " +
-                            kind.toString() +
-                            " nodes when starting at " +
-                            getStartingNodeDescription(contentType) +
-                            ", as this type requires simple content", getLocation());
+                                                 kind.toString() +
+                                                 " nodes when starting at " +
+                                                 getStartingNodeDescription(contentType) +
+                                                 ", as this type requires simple content", getLocation());
                 }
                 return Literal.makeEmptySequence();
             } else if (((ComplexType) contentType).isEmptyContent() &&
@@ -388,17 +390,19 @@ public final class AxisExpression extends Expression {
                 }
                 if (warnings) {
                     visitor.issueWarning("The " + AxisInfo.axisName[axis] + " axis will never select any" +
-                            " nodes when starting at " +
-                            getStartingNodeDescription(contentType) +
-                            ", as this type requires empty content", getLocation());
+                                                 " nodes when starting at " +
+                                                 getStartingNodeDescription(contentType) +
+                                                 ", as this type requires empty content", getLocation());
                 }
                 return Literal.makeEmptySequence();
             } else if (axis == AxisInfo.ATTRIBUTE) {
                 if (targetfp == -1) {
                     if (warnings) {
                         if (!((ComplexType) contentType).allowsAttributes()) {
-                            visitor.issueWarning("The complex type " + contentType.getDescription() +
-                                    " allows no attributes other than the standard attributes in the xsi namespace", getLocation());
+                            visitor.issueWarning(
+                                    "The complex type " + contentType.getDescription() +
+                                            " allows no attributes other than the standard attributes in the xsi namespace",
+                                    getLocation());
                         }
                     }
                 } else {
@@ -417,8 +421,10 @@ public final class AxisExpression extends Expression {
                         }
                         if (schemaType == null) {
                             if (warnings) {
-                                visitor.issueWarning("The complex type " + contentType.getDescription() +
-                                        " does not allow an attribute named " + getDiagnosticName(targetName, env), getLocation());
+                                visitor.issueWarning(
+                                        "The complex type " + contentType.getDescription() +
+                                                " does not allow an attribute named " + getDiagnosticName(targetName, env),
+                                        getLocation());
                                 return Literal.makeEmptySequence();
                             }
                         } else {
@@ -443,8 +449,10 @@ public final class AxisExpression extends Expression {
                         ((ComplexType) contentType).gatherAllPermittedChildren(children, false);
                         if (children.isEmpty()) {
                             if (warnings) {
-                                visitor.issueWarning("The complex type " + contentType.getDescription() +
-                                        " does not allow children", getLocation());
+                                visitor.issueWarning(
+                                        "The complex type " + contentType.getDescription() +
+                                                " does not allow children",
+                                        getLocation());
                             }
                             return Literal.makeEmptySequence();
                         }
@@ -464,9 +472,9 @@ public final class AxisExpression extends Expression {
                     if (schemaType == null) {
                         if (warnings) {
                             String message = "The complex type " + contentType.getDescription() +
-                                " does not allow a child element named " + getDiagnosticName(childElement, env);
+                                    " does not allow a child element named " + getDiagnosticName(childElement, env);
                             HashSet<StructuredQName> permitted = new HashSet<StructuredQName>();
-                            ((ComplexType)contentType).gatherAllPermittedChildren(permitted, false);
+                            ((ComplexType) contentType).gatherAllPermittedChildren(permitted, false);
                             for (StructuredQName sq : permitted) {
                                 if (sq.getLocalPart().equals(childElement.getLocalPart()) && !sq.equals(childElement)) {
                                     message += ". Perhaps the namespace is " +
@@ -482,17 +490,22 @@ public final class AxisExpression extends Expression {
                         itemType = new CombinedNodeTest(
                                 test,
                                 Token.INTERSECT,
-                                new ContentTypeTest(Type.ELEMENT, schemaType, getConfiguration(), true));
+                                new ContentTypeTest(Type.ELEMENT, schemaType, config, true));
                         computedCardinality = ((ComplexType) contentType).getElementParticleCardinality(childElement, true);
                         ExpressionTool.resetStaticProperties(this);
                         if (computedCardinality == StaticProperty.ALLOWS_ZERO) {
                             // this shouldn't happen, because we've already checked for this a different way.
                             // but it's worth being safe (there was a bug involving an incorrect inference here)
-                            visitor.issueWarning("The complex type " + contentType.getDescription() +
-                                    " appears not to allow a child element named " + getDiagnosticName(childElement, env), getLocation());
+                            visitor.issueWarning(
+                                    "The complex type " + contentType.getDescription() +
+                                            " appears not to allow a child element named " +
+                                            getDiagnosticName(childElement, env),
+                                    getLocation());
                             return Literal.makeEmptySequence();
                         }
-                        if (!Cardinality.allowsMany(computedCardinality)) {
+                        if (!Cardinality.allowsMany(computedCardinality) &&
+                                !(getParentExpression() instanceof FirstItemExpression) &&
+                                !visitor.isOptimizeForPatternMatching()) {
                             // if there can be at most one child of this name, create a FirstItemExpression
                             // to stop the search after the first one is found
                             return FirstItemExpression.makeFirstItemExpression(this);
@@ -545,7 +558,7 @@ public final class AxisExpression extends Expression {
                             //return this;
                         }
                         if (usefulChildren.size() < children.size()) {
-                            NodeTest childTest = makeUnionNodeTest(usefulChildren, getConfiguration().getNamePool());
+                            NodeTest childTest = makeUnionNodeTest(usefulChildren, config.getNamePool());
                             AxisExpression first = new AxisExpression(AxisInfo.CHILD, childTest);
                             ExpressionTool.copyLocationInfo(this, first);
                             byte nextAxis;
@@ -562,8 +575,11 @@ public final class AxisExpression extends Expression {
                         }
                     } else {
                         if (warnings) {
-                            visitor.issueWarning("The complex type " + contentType.getDescription() +
-                                    " does not allow a descendant element named " + getDiagnosticName(targetName, env), getLocation());
+                            visitor.issueWarning(
+                                    "The complex type " + contentType.getDescription() +
+                                            " does not allow a descendant element named " +
+                                            getDiagnosticName(targetName, env),
+                                    getLocation());
                         }
                     }
                 } catch (SchemaException e) {
@@ -662,8 +678,8 @@ public final class AxisExpression extends Expression {
 
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) {
         doneOptimize = true; // This ensures no more warnings about empty axes, because (a) we've probably output the
-                             // warning already, and (b) we're now looking at a different expression from what the user
-                             // wrote. In particular, prevent spurious warnings after function inlining.
+        // warning already, and (b) we're now looking at a different expression from what the user
+        // wrote. In particular, prevent spurious warnings after function inlining.
         staticInfo = contextInfo;
         return this;
     }
@@ -739,8 +755,10 @@ public final class AxisExpression extends Expression {
     /**
      * Copy an expression. This makes a deep copy.
      *
+     * @param rebindings a mutable list of (old binding, new binding) pairs
+     *                   that is used to update the bindings held in any
+     *                   local variable references that are copied.
      * @return the copy of the original expression
-     * @param rebindings
      */
 
     /*@NotNull*/
@@ -821,6 +839,26 @@ public final class AxisExpression extends Expression {
         }
     }
 
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @param contextItemType the static type of the context item for the expression evaluation
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        // See W3C bug 30032
+        UType reachable = AxisInfo.getTargetUType(contextItemType, axis);
+        if (test == null) {
+            return reachable;
+        } else {
+            return reachable.intersection(test.getUType());
+        }
+    }
+
+
     /**
      * Determine the intrinsic dependencies of an expression, that is, those which are not derived
      * from the dependencies of its subexpressions. For example, position() has an intrinsic dependency
@@ -952,6 +990,9 @@ public final class AxisExpression extends Expression {
         return staticInfo.isPossiblyAbsent();
     }
 
+    public ContextItemStaticInfo getContextItemStaticInfo() {
+        return staticInfo;
+    }
 
     /**
      * Convert this expression to an equivalent XSLT pattern
@@ -1026,7 +1067,7 @@ public final class AxisExpression extends Expression {
         if (item == null) {
             // Might as well do the test anyway, whether or not contextMaybeUndefined is set
             XPathException err = new XPathException("The context item for axis step " +
-                    toString() + " is absent");
+                                                            toString() + " is absent");
             err.setErrorCode("XPDY0002");
             err.setXPathContext(context);
             err.setLocation(getLocation());
@@ -1041,7 +1082,7 @@ public final class AxisExpression extends Expression {
             }
         } catch (ClassCastException cce) {
             XPathException err = new XPathException("The context item for axis step " +
-                    toString() + " is not a node");
+                                                            toString() + " is not a node");
             err.setErrorCode("XPTY0020");
             err.setXPathContext(context);
             err.setLocation(getLocation());
@@ -1076,19 +1117,6 @@ public final class AxisExpression extends Expression {
         }
     }
 
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Axis expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new AxisExpressionCompiler();
-    }
-    //#endif
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -1105,8 +1133,9 @@ public final class AxisExpression extends Expression {
             } else if (axis == AxisInfo.NAMESPACE) {
                 known = NodeKindTest.makeNodeKindTest(Type.NAMESPACE);
             }
+            int targetVersion = destination.getIntOption("targetVersion", 1);
             try {
-                destination.emitAttribute("jsTest", test == null ? "return true;" : test.generateJavaScriptItemTypeTest(known));
+                destination.emitAttribute("jsTest", test == null ? "return true;" : test.generateJavaScriptItemTypeTest(known, targetVersion));
             } catch (XPathException e) {
                 e.maybeSetLocation(getLocation());
                 throw e;
@@ -1156,20 +1185,11 @@ public final class AxisExpression extends Expression {
         return fsb.toString();
     }
 
-    //#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
     @Override
-    protected StreamingAdjunct getStreamingAdjunct() {
-        return new AxisExpressionAdjunct();
+    public String getStreamerName() {
+        return "AxisExpression";
     }
 
-    //#endif
-
     /**
      * Find any necessary preconditions for the satisfaction of this expression
      * as a set of boolean expressions to be evaluated on the context node
