@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,10 +7,7 @@
 
 package net.sf.saxon.expr;
 
-import net.sf.saxon.expr.parser.ContextItemStaticInfo;
-import net.sf.saxon.expr.parser.ExpressionTool;
-import net.sf.saxon.expr.parser.ExpressionVisitor;
-import net.sf.saxon.expr.parser.Token;
+import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.om.GroundedValue;
 import net.sf.saxon.om.Item;
 import net.sf.saxon.om.SequenceTool;
@@ -86,7 +83,7 @@ public abstract class BinaryExpression extends Expression {
      * @return the left-hand operand child expression
      */
 
-    public Expression getLhsExpression() {
+    public final Expression getLhsExpression() {
         return lhs.getChildExpression();
     }
 
@@ -113,7 +110,7 @@ public abstract class BinaryExpression extends Expression {
      * @return the right-hand operand expression
      */
 
-    public Expression getRhsExpression() {
+    public final Expression getRhsExpression() {
         return rhs.getChildExpression();
     }
 
@@ -427,7 +424,16 @@ public abstract class BinaryExpression extends Expression {
 
     @Override
     public String toShortString() {
-        return getLhsExpression().toShortString() + " " + displayOperator() + " " + getRhsExpression().toShortString();
+        return parenthesize(getLhsExpression()) + " " + displayOperator() + " " + parenthesize(getRhsExpression());
+    }
+
+    private String parenthesize(Expression operand) {
+        String operandStr = operand.toShortString();
+        if (operand instanceof BinaryExpression &&
+                XPathParser.operatorPrecedence(((BinaryExpression) operand).operator) < XPathParser.operatorPrecedence(operator)) {
+            operandStr = "(" + operandStr + ")";
+        }
+        return operandStr;
     }
 
     /**
