@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -714,8 +714,8 @@ public abstract class Calculator {
             if (Float.isInfinite(quotient)) {
                 // bug 29171, test cbcl-numeric-idivide-008
                 return new DecimalIdivDecimal().compute(
-                    new DecimalValue(((NumericValue) a).getDecimalValue()),
-                    new DecimalValue(((NumericValue) b).getDecimalValue()),
+                    new BigDecimalValue(((NumericValue) a).getDecimalValue()),
+                    new BigDecimalValue(((NumericValue) b).getDecimalValue()),
                     c);
             } else {
                 return Converter.FLOAT_TO_INTEGER.convert(new FloatValue(quotient)).asAtomic();
@@ -736,7 +736,7 @@ public abstract class Calculator {
             if (a instanceof IntegerValue && b instanceof IntegerValue) {
                 return ((IntegerValue) a).plus((IntegerValue) b);
             } else {
-                return new DecimalValue(
+                return new BigDecimalValue(
                         ((NumericValue) a).getDecimalValue().add(((NumericValue) b).getDecimalValue()));
             }
         }
@@ -755,7 +755,7 @@ public abstract class Calculator {
             if (a instanceof IntegerValue && b instanceof IntegerValue) {
                 return ((IntegerValue) a).minus((IntegerValue) b);
             } else {
-                return new DecimalValue(
+                return new BigDecimalValue(
                         ((NumericValue) a).getDecimalValue().subtract(((NumericValue) b).getDecimalValue()));
             }
         }
@@ -774,7 +774,7 @@ public abstract class Calculator {
             if (a instanceof IntegerValue && b instanceof IntegerValue) {
                 return ((IntegerValue) a).times((IntegerValue) b);
             } else {
-                return new DecimalValue(
+                return new BigDecimalValue(
                         ((NumericValue) a).getDecimalValue().multiply(((NumericValue) b).getDecimalValue()));
             }
         }
@@ -798,16 +798,16 @@ public abstract class Calculator {
         }
     }
 
-    public static DecimalValue decimalDivide(NumericValue a, NumericValue b) throws XPathException {
+    public static BigDecimalValue decimalDivide(NumericValue a, NumericValue b) throws XPathException {
         final BigDecimal A = a.getDecimalValue();
         final BigDecimal B = b.getDecimalValue();
 //        int scale = Math.max(DecimalValue.DIVIDE_PRECISION,
 //                Math.max(A.scale(), B.scale()));
-        int scale = Math.max(DecimalValue.DIVIDE_PRECISION, A.scale() - B.scale() + DecimalValue.DIVIDE_PRECISION);
+        int scale = Math.max(BigDecimalValue.DIVIDE_PRECISION, A.scale() - B.scale() + BigDecimalValue.DIVIDE_PRECISION);
 
         try {
             BigDecimal result = A.divide(B, scale, BigDecimal.ROUND_HALF_DOWN);
-            return new DecimalValue(result);
+            return new BigDecimalValue(result);
         } catch (ArithmeticException err) {
             if (b.compareTo(0) == 0) {
                 throw new XPathException("Decimal divide by zero", "FOAR0001");
@@ -832,7 +832,7 @@ public abstract class Calculator {
             try {
                 //BigDecimal quotient = A.divide(B, 0, BigDecimal.ROUND_DOWN);
                 //BigDecimal remainder = A.subtract(quotient.multiply(B));
-                return new DecimalValue(A.remainder(B));
+                return new BigDecimalValue(A.remainder(B));
             } catch (ArithmeticException err) {
                 if (((NumericValue) b).compareTo(0) == 0) {
                     throw new XPathException("Decimal modulo zero", "FOAR0001", c);
