@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,9 +7,6 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.CardinalityCheckerCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.CardinalityCheckerAdjunct;
 import net.sf.saxon.event.SequenceReceiver;
 import net.sf.saxon.event.TypeCheckingFilter;
 import net.sf.saxon.expr.parser.*;
@@ -20,6 +17,7 @@ import net.sf.saxon.pattern.DocumentNodeTest;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.XPathException;
+import net.sf.saxon.tree.iter.ArrayIterator;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.Type;
@@ -65,7 +63,7 @@ public final class CardinalityChecker extends UnaryExpression {
 
     public static Expression makeCardinalityChecker(Expression sequence, int cardinality, RoleDiagnostic role) {
         Expression result;
-        if (sequence instanceof Literal && Cardinality.subsumes(cardinality, SequenceTool.getCardinality(((Literal)sequence).getValue()))) {
+        if (sequence instanceof Literal && Cardinality.subsumes(cardinality, SequenceTool.getCardinality(((Literal) sequence).getValue()))) {
             return sequence;
         }
         if (sequence instanceof Atomizer && !Cardinality.allowsMany(cardinality)) {
@@ -228,7 +226,7 @@ public final class CardinalityChecker extends UnaryExpression {
                         role.getMessage() + " is an empty sequence", role.getErrorCode(), context);
             } else if (count > 1 && !Cardinality.allowsMany(requiredCardinality)) {
                 typeError("A sequence of more than one item is not allowed as the " +
-                        role.getMessage() + depictSequenceStart(base.getAnother(), 2),
+                        role.getMessage() + depictSequenceStart(base, 2),
                         role.getErrorCode(), context);
             }
             return base;
@@ -253,12 +251,8 @@ public final class CardinalityChecker extends UnaryExpression {
             FastStringBuffer sb = new FastStringBuffer(FastStringBuffer.C64);
             int count = 0;
             sb.append(" (");
-            while (true) {
-                Item next = seq.next();
-                if (next == null) {
-                    sb.append(") ");
-                    return sb.toString();
-                }
+            Item next;
+            while ((next = seq.next()) != null) {
                 if (count++ > 0) {
                     sb.append(", ");
                 }
@@ -269,6 +263,8 @@ public final class CardinalityChecker extends UnaryExpression {
 
                 sb.append(Err.depict(next));
             }
+            sb.append(") ");
+            return sb.toString();
         } catch (XPathException e) {
             return "";
         }
@@ -297,8 +293,9 @@ public final class CardinalityChecker extends UnaryExpression {
             }
             Item second = iter.next();
             if (second != null) {
+                Item[] leaders = new Item[]{first, second};
                 typeError("A sequence of more than one item is not allowed as the " +
-                    role.getMessage() + depictSequenceStart(iter.getAnother(), 2), role.getErrorCode(), context);
+                    role.getMessage() + depictSequenceStart(new ArrayIterator(leaders), 2), role.getErrorCode(), context);
                 return null;
             }
         }
@@ -456,33 +453,11 @@ public final class CardinalityChecker extends UnaryExpression {
         return getBaseExpression().toShortString();
     }
 
-//#ifdefined BYTECODE
 
-    /**
-     * Return the compiler of the CardinalityCheck expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new CardinalityCheckerCompiler();
+    public String getStreamerName() {
+        return "CardinalityChecker";
     }
-//#endif
-
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public CardinalityCheckerAdjunct getStreamingAdjunct() {
-        return new CardinalityCheckerAdjunct();
-    }
-
-//#endif
-
 
     @Override
     public void setLocation(Location id) {
