@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,28 +8,24 @@
 package net.sf.saxon.expr;
 
 
-import com.saxonica.ee.bytecode.CastExpressionCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
+import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.functions.String_1;
 import net.sf.saxon.lib.ConversionRules;
 import net.sf.saxon.om.GroundedValue;
 import net.sf.saxon.om.Sequence;
-import net.sf.saxon.om.StandardNames;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
 
-import java.util.Map;
-
 
 /**
  * Cast Expression: implements "cast as data-type ( expression )". It also allows an internal
  * cast, which has the same semantics as a user-requested cast, but maps an empty sequence to
  * an empty sequence.
  * <p/>
- * This expression class does not handle casting to a union type.
+ * This expression class does not handle casting to a list or union type.
  */
 
 public class CastExpression extends CastingExpression implements Callable {
@@ -56,11 +52,13 @@ public class CastExpression extends CastingExpression implements Callable {
         typeCheckChildren(visitor, contextInfo);
         SequenceType atomicType = SequenceType.makeSequenceType(BuiltInAtomicType.ANY_ATOMIC, getCardinality());
 
-        final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        Configuration config = visitor.getConfiguration();
+        final TypeHierarchy th = config.getTypeHierarchy();
         RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.TYPE_OP, "cast as", 0);
         ItemType sourceItemType = null;
 
-        Expression operand = TypeChecker.staticTypeCheck(getBaseExpression(), atomicType, false, role, visitor);
+        TypeChecker tc = config.getTypeChecker(false);
+        Expression operand = tc.staticTypeCheck(getBaseExpression(), atomicType, role, visitor);
         setBaseExpression(operand);
         sourceItemType = operand.getItemType();
 
@@ -76,7 +74,6 @@ public class CastExpression extends CastingExpression implements Callable {
                 throw err;
             }
         }
-        boolean isXPath30 = visitor.getStaticContext().getXPathVersion() >= 30;
 
         PlainType sourceType = (PlainType) sourceItemType;
         int r = th.relationship(sourceType, getTargetType());
@@ -88,7 +85,7 @@ public class CastExpression extends CastingExpression implements Callable {
             converter = new Converter.UpCastingConverter(getTargetType());
         } else {
 
-            ConversionRules rules = getConfiguration().getConversionRules();
+            ConversionRules rules = visitor.getConfiguration().getConversionRules();
 
             if (sourceType.isAtomicType() && sourceType != BuiltInAtomicType.ANY_ATOMIC) {
                 converter = rules.getConverter((AtomicType)sourceType, getTargetType());
@@ -104,20 +101,12 @@ public class CastExpression extends CastingExpression implements Callable {
                         converter = converter.setNamespaceResolver(getRetainedStaticContext());
                     }
                 }
-                if (converter.isXPath30Conversion() && !isXPath30
-                        && !(operand instanceof Literal && getTargetType().getPrimitiveType() == StandardNames.XS_QNAME)) {
-                    XPathException err = new XPathException("Casting from " + sourceType + " to " + getTargetType() +
-                            " requires XPath 3.0 functionality to be enabled");
-                    err.setErrorCode("XPTY0004");
-                    err.setLocation(getLocation());
-                    err.setIsTypeError(true);
-                    throw err;
-                }
+
             }
         }
 
         if (operand instanceof Literal) {
-            return preEvaluate(isXPath30);
+            return preEvaluate();
         }
 
         return this;
@@ -142,7 +131,7 @@ public class CastExpression extends CastingExpression implements Callable {
 
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
-        final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        final TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
         Expression e2 = super.optimize(visitor, contextInfo);
         if (e2 != this) {
             return e2;
@@ -218,7 +207,7 @@ public class CastExpression extends CastingExpression implements Callable {
         }
 
         if (operand instanceof Literal) {
-            return preEvaluate(visitor.getStaticContext().getXPathVersion() >= 30);
+            return preEvaluate();
         }
         return this;
     }
@@ -227,16 +216,9 @@ public class CastExpression extends CastingExpression implements Callable {
      * Perform early (compile-time) evaluation
      */
 
-    protected Expression preEvaluate(boolean isXPath30) throws XPathException {
+    protected Expression preEvaluate() throws XPathException {
         GroundedValue literalOperand = ((Literal) getBaseExpression()).getValue();
         if (literalOperand instanceof AtomicValue && converter != null) {
-            if (converter.isXPath30Conversion() && !isXPath30 &&
-                    !(converter instanceof StringConverter.StringToQName && isOperandIsStringLiteral())) {
-                XPathException e = new XPathException("Casting from " + ((AtomicValue) literalOperand).getPrimitiveType() + " to " + getTargetType() + " requires XPath 3.0 to be enabled");
-                e.setLocation(getLocation());
-                e.setErrorCode("XPTY0004");
-                throw e;
-            }
             ConversionResult result = converter.convert((AtomicValue) literalOperand);
             if (result instanceof ValidationFailure) {
                 ValidationFailure err = (ValidationFailure) result;
@@ -246,7 +228,9 @@ public class CastExpression extends CastingExpression implements Callable {
                 }
                 throw new XPathException(err.getMessage(), code, this.getLocation());
             } else {
-                return Literal.makeLiteral((AtomicValue) result);
+                Expression lit = Literal.makeLiteral((AtomicValue) result);
+                ExpressionTool.copyLocationInfo(this, lit);
+                return lit;
             }
         }
         if (literalOperand.getLength() == 0) {
@@ -280,6 +264,32 @@ public class CastExpression extends CastingExpression implements Callable {
         return getTargetType();
     }
 
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        return getTargetType().getUType();
+    }
+
+    /**
+     * Determine the special properties of this expression
+     * @return the expression properties
+     */
+    @Override
+    public int computeSpecialProperties() {
+        int p = super.computeSpecialProperties();
+        if (getTargetType() == BuiltInAtomicType.UNTYPED_ATOMIC) {
+            p = p &~ StaticProperty.NOT_UNTYPED_ATOMIC;
+        }
+        return p;
+    }
+
     /**
      * For an expression that returns an integer or a sequence of integers, get
      * a lower and upper bound on the values of the integers that may be returned, from
@@ -362,14 +372,6 @@ public class CastExpression extends CastingExpression implements Callable {
                 e.setErrorCode("XPTY0004");
                 throw e;
             }
-            if (converter.isXPath30Conversion() && getRetainedStaticContext().getXPathVersion() < 30 &&
-                    !(converter instanceof StringConverter.StringToQName && isOperandIsStringLiteral())) {
-                XPathException e = new XPathException("Casting from " + value.getPrimitiveType() + " to " + getTargetType() + " requires XPath 3.0 to be enabled");
-                e.setXPathContext(context);
-                e.setLocation(getLocation());
-                e.setErrorCode("XPTY0004");
-                throw e;
-            }
             if (getTargetType().isNamespaceSensitive()) {
                 converter = converter.setNamespaceResolver(getRetainedStaticContext());
             }
@@ -450,19 +452,18 @@ public class CastExpression extends CastingExpression implements Callable {
         out.endElement();
     }
 
-
-//#ifdefined BYTECODE
-
     /**
-     * Return the compiler of the Cast expression
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
      *
-     * @return the relevant ExpressionCompiler
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in export() output displaying the expression.
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new CastExpressionCompiler();
+    public String getExpressionName() {
+        return "cast";
     }
-//#endif
+
 
 
 }
