@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,8 +7,6 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.LetExpressionCompiler;
 import net.sf.saxon.expr.parser.ContextItemStaticInfo;
 import net.sf.saxon.expr.parser.ExpressionTool;
 import net.sf.saxon.expr.parser.ExpressionVisitor;
@@ -33,27 +31,6 @@ public class EagerLetExpression extends LetExpression {
         return e;
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the EagerLet expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new LetExpressionCompiler();
-    }
-//#endif
-
-    /**
-     * Evaluate the variable.
-     */
-
-//    protected ValueRepresentation eval(XPathContext context) throws XPathException {
-//        return ExpressionTool.eagerEvaluate(sequence, context);
-//    }
-
 }
 
 
