@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,23 +7,12 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.CallableExpressionCompiler;
-import com.saxonica.ee.bytecode.CompiledExpression;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.IsWholeNumberCompiler;
-import com.saxonica.ee.bytecode.util.CannotCompileException;
-import com.saxonica.ee.stream.Posture;
-import com.saxonica.ee.stream.PostureAndSweep;
-import com.saxonica.ee.stream.Sweep;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.event.SequenceReceiver;
-import net.sf.saxon.evpull.EmptyEventIterator;
-import net.sf.saxon.evpull.EventIterator;
-import net.sf.saxon.evpull.EventIteratorOverSequence;
-import net.sf.saxon.evpull.SingletonEventIterator;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.functions.*;
+import net.sf.saxon.functions.KeyFn;
+import net.sf.saxon.functions.SuperId;
+import net.sf.saxon.functions.SystemFunction;
 import net.sf.saxon.lib.Logger;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.NodeKindTest;
@@ -38,6 +27,7 @@ import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.AtomicType;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.SchemaType;
+import net.sf.saxon.type.UType;
 import net.sf.saxon.value.*;
 import net.sf.saxon.z.IntHashSet;
 import net.sf.saxon.z.IntIterator;
@@ -66,7 +56,7 @@ import java.util.*;
  * no subexpressions.</p>
  */
 
-public abstract class Expression implements /*InstructionInfo,*/ IdentityComparable {
+public abstract class Expression implements IdentityComparable {
 
     public static final int EVALUATE_METHOD = 1;
     public static final int ITERATE_METHOD = 2;
@@ -83,6 +73,8 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     private RetainedStaticContext retainedStaticContext;
     private int[] slotsUsed;
     private int evaluationMethod;
+    private Map<String, Object> extraProperties;
+    private int cost = -1;
 
 //    public int serial;  // used to identify expressions for internal diagnostics
 //    private static int nextSerial = 0;
@@ -122,8 +114,10 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     }
 
     /**
-     * Get the interpreted form of the expression. This normally returns the expression itself,
-     * except for a CompiledExpression (bytecode-generated) in which case it returns the original
+     * Get the interpreted form of the expression. Normally this returns the expression unchanged;
+     * but on a CompiledExpression (the result of bytecode generation) it returns the original
+     * interpreted form.
+     * @return the interpreted form of this expression
      */
 
     public Expression getInterpretedExpression() {
@@ -141,11 +135,25 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         Iterable<Operand> ops = operands();
         for (Operand o : ops) {
             Expression child = o.getChildExpression();
-            if (child.getParentExpression() != this) {
-                System.err.println("*** Bad parent pointer found in " + child.toShortString() +
-                    " at " + child.getLocation().getLineNumber() + " ***");
+            boolean badOperand = o.getParentExpression() != this;
+            boolean badExpression = child.getParentExpression() != this;
+            if (badOperand || badExpression) {
+                String message = "*** Bad parent pointer found in " +
+                        (badOperand ? "operand " : "expression ") +
+                        child.toShortString() +
+                        " at " + child.getLocation().getSystemId() + "#" + child.getLocation().getLineNumber() + " ***";
+                try {
+                    Configuration config = getConfiguration();
+                    Logger logger = config == null ? null : config.getLogger();
+                    if (logger != null) {
+                        logger.warning(message);
+                    } else {
+                        throw new IllegalStateException(message);
+                    }
+                } catch (Exception err) {
+                    throw new IllegalStateException(message);
+                }
                 child.setParentExpression(Expression.this);
-                //throw new IllegalStateException("Bad parent pointer");
             }
             if (child.getRetainedStaticContext() == null) {
                 child.setRetainedStaticContext(getRetainedStaticContext());
@@ -190,10 +198,6 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return parentExpression;
     }
 
-    public String getSystemId() {
-        return getLocation().getSystemId();
-    }
-
     /**
      * Set the parent expression of this expression in the expression tree.
      * @param parent the parent expression
@@ -211,17 +215,35 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
      * @throws IllegalStateException if invalid parent pointers are found
      */
 
-//    public Expression verifyParentPointers() throws IllegalStateException {
-//        for (Operand o : operands()) {
-//            Expression parent = o.getChildExpression().getParentExpression();
-//            if (parent != this) {
-//                throw new IllegalStateException("Invalid parent pointer in " +
-//                    o.getChildExpression().getClass() + o.getChildExpression().serial + " -> " + parent.getClass() + parent.serial);
-//            }
-//            o.getChildExpression().verifyParentPointers();
-//        }
-//        return this;
-//    }
+    public Expression verifyParentPointers() throws IllegalStateException {
+        for (Operand o : operands()) {
+            Expression parent = o.getChildExpression().getParentExpression();
+            if (parent != this) {
+                throw new IllegalStateException("Invalid parent pointer in " + parent.toShortString() +
+                                                        " subexpression " + o.getChildExpression().toShortString());
+            }
+            if (o.getParentExpression() != this) {
+                throw new IllegalStateException("Invalid parent pointer in operand object " + parent.toShortString() +
+                                                        " subexpression " + o.getChildExpression().toShortString());
+            }
+            if (ExpressionTool.findOperand(parent, o.getChildExpression()) == null) {
+                throw new IllegalStateException("Incorrect parent pointer in " + parent.toShortString() +
+                                                        " subexpression " + o.getChildExpression().toShortString());
+            }
+            o.getChildExpression().verifyParentPointers();
+        }
+        return this;
+    }
+
+    public void verifyParentPointers(String message) {
+        //System.err.println("Verify: " + message);
+        try {
+            verifyParentPointers();
+        } catch (IllegalStateException err) {
+            System.err.println("Verify failed at: " + message + "(" + err.getMessage() + ")");
+            //throw err;
+        }
+    }
 
     /**
      * Restore parent pointers for the subtree rooted at this expression
@@ -245,13 +267,6 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
      */
 
     public abstract int getImplementationMethod();
-//    {
-//        if (Cardinality.allowsMany(getCardinality())) {
-//            return ITERATE_METHOD;
-//        } else {
-//            return EVALUATE_METHOD;
-//        }
-//    }
 
     /**
      * Determine whether this expression implements its own method for static type checking
@@ -276,6 +291,65 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return false;
     }
 
+    /**
+     * Ask whether the expression can be lifted out of a loop, assuming it has no dependencies
+     * on the controlling variable/focus of the loop
+     * @param forStreaming true if the expression is being optimized for streaming
+     */
+
+    public boolean isLiftable(boolean forStreaming) {
+        int p = getSpecialProperties();
+        return (p & StaticProperty.NON_CREATIVE) != 0 && (p & StaticProperty.HAS_SIDE_EFFECTS) == 0
+                && ((getDependencies() & StaticProperty.DEPENDS_ON_ASSIGNABLE_GLOBALS) == 0);
+    }
+
+    /**
+     * Get the innermost scoping expression of this expression, for expressions that directly
+     * depend on something in the dynamic context. For example, in the case of a local variable
+     * reference this is the expression that causes the relevant variable to be bound; for a
+     * context item expression it is the innermost focus-setting container. For expressions
+     * that have no intrinsic dependency on the dynamic context, the value returned is null;
+     * the scoping container for such an expression is the innermost scoping container of its
+     * operands.
+     * @return the innermost scoping container of this expression
+     */
+
+    public Expression getScopingExpression() {
+        int d = getIntrinsicDependencies() & StaticProperty.DEPENDS_ON_FOCUS;
+        if (d != 0) {
+            if (d == StaticProperty.DEPENDS_ON_CONTEXT_DOCUMENT) {
+                return ExpressionTool.getContextDocumentSettingContainer(this);
+            } else {
+                return ExpressionTool.getFocusSettingContainer(this);
+            }
+        } else {
+            return null;
+        }
+    }
+
+    /**
+     * Ask whether the expression is multithreaded (that is, whether its operands are
+     * evaluated in parallel)
+     * @param config the Saxon configuration
+     * @return true if execution will be multithreaded
+     */
+
+    public boolean isMultiThreaded(Configuration config) {
+        return false;
+    }
+
+    /**
+     * Ask whether common subexpressions found in the operands of this expression can
+     * be extracted and evaluated outside the expression itself. The result is irrelevant
+     * in the case of operands evaluated with a different focus, which will never be
+     * extracted in this way, even if they have no focus dependency.
+     * @return true by default, unless overridden in a subclass
+     */
+
+    public boolean allowExtractingCommonSubexpressions() {
+        return true;
+    }
+
     /**
      * Simplify an expression. This performs any static optimization (by rewriting the expression
      * as a different expression). The default implementation simplifies its operands.
@@ -292,9 +366,8 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     }
 
     /**
-     * Simplify an expression. This performs any static optimization (by rewriting the expression
+     * Simplify this expression. This performs any static optimization (by rewriting the expression
      * as a different expression). The default implementation simplifies its operands.
-     * @return the simplified expression (or the original if unchanged, or if modified in-situ)
      * @throws net.sf.saxon.trans.XPathException
      *          if an error is discovered during expression
      *          rewriting
@@ -331,13 +404,48 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         }
     }
 
+    /**
+     * Set the retained static context on this expression and on all subexpressions in the expression tree.
+     * The supplied retained static context is applied to all subexpressions, unless they have a closer
+     * ancestor with an existing retained static context, in which case that is used instead,
+     *
+     * @param rsc the static context to be retained
+     */
+
+    public void setRetainedStaticContextThoroughly(RetainedStaticContext rsc) {
+//        if (rsc != null) {
+//            System.err.println("setRSC " + toShortString() + " - " + rsc);
+//        }
+        retainedStaticContext = rsc;
+        for (Operand o : operands()) {
+            if (o != null) {
+                Expression child = o.getChildExpression();
+                if (child != null) {
+                    if (child.getLocalRetainedStaticContext() == null) {
+                        child.setRetainedStaticContextThoroughly(rsc);
+                    } else {
+                        rsc = child.getLocalRetainedStaticContext();
+                        for (Operand p : child.operands()) {
+                            Expression grandchild = p.getChildExpression();
+                            if (grandchild != null) {
+                                grandchild.setRetainedStaticContextThoroughly(rsc);
+                            }
+                        }
+                    }
+                }
+            }
+        }
+    }
+
     /**
      * Set the parts of the static context that might be needed by the function, without
      * passing them on to subexpressions. Used for dynamic function calls only.
      */
 
     public void setRetainedStaticContextLocally(RetainedStaticContext rsc) {
-        retainedStaticContext = rsc;
+        if (rsc != null) {
+            retainedStaticContext = rsc;
+        }
     }
 
     /**
@@ -348,11 +456,16 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     public final RetainedStaticContext getRetainedStaticContext() {
         if (retainedStaticContext == null) {
             Expression parent = getParentExpression();
-            retainedStaticContext = parent == null ? null : parent.getRetainedStaticContext();
+            return parent == null ? null : parent.getRetainedStaticContext();
+            //setRetainedStaticContextLocally(parent == null ? null : parent.getRetainedStaticContext());
         }
         return retainedStaticContext;
     }
 
+    public RetainedStaticContext getLocalRetainedStaticContext() {
+        return retainedStaticContext;
+    }
+
     /**
      * Get the saved static base URI as a string
      *
@@ -382,6 +495,11 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return false;
     }
 
+    /**
+     * Ask whether this expression is a call on a particular function
+     * @param localName the name of the function in question, in the standard fn namespace
+     */
+
     public boolean isCallOnSystemFunction(String localName) {
         return false;
     }
@@ -481,7 +599,10 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor,
                                /*@Nullable*/ ContextItemStaticInfo contextInfo) throws XPathException {
-        optimizeChildren(visitor, contextInfo);
+        if (visitor.incrementAndTestDepth()) {  // protect against infinite recursion
+            optimizeChildren(visitor, contextInfo);
+            visitor.decrementDepth();
+        }
         return Expression.this;
     }
 
@@ -518,11 +639,14 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
      */
 
     public int getCost() {
-        int i = getNetCost();
-        for (Operand o : operands()) {
-            i += o.getChildExpression().getCost();
+        if (cost < 0) {
+            int i = getNetCost();
+            for (Operand o : operands()) {
+                i += o.getChildExpression().getCost();
+            }
+            cost = i;
         }
-        return i;
+        return cost;
     }
 
     /**
@@ -533,85 +657,10 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
      * the operands
      */
 
-    protected int getNetCost() {
+    public int getNetCost() {
         return 1;
     }
 
-    /**
-     * Offer promotion for this subexpression. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is accepted. This method is always called at compile time.
-     *
-     *
-     * @param offer  details of the offer, for example the offer to move
-     *               expressions that don't depend on the context to an outer level in
-     *               the containing expression
-     * @return if the offer is not accepted, return this expression unchanged.
-     *         Otherwise return the result of rewriting the expression to promote
-     *         this subexpression
-     * @throws XPathException if any error is detected
-     */
-
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        Expression exp = offer.accept(Expression.this);
-        if (exp != null) {
-            return exp;
-        } else {
-            promoteChildren(offer);
-            return Expression.this;
-        }
-    }
-
-    /**
-     * Helper method for implementations of 
-     * {@link #promote(net.sf.saxon.expr.parser.PromotionOffer)} 
-     * to apply the requested promotion to all subexpressions
-     * @param offer the promotion offer
-     * @throws XPathException if any failure occurs
-     */
-    
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        if (offer.action == PromotionOffer.FOCUS_INDEPENDENT) {
-            for (Operand o : operands()) {
-                if (o.hasSameFocus()) {
-                    o.setChildExpression(doPromotion(o.getChildExpression(), offer));
-                }
-            }
-        } else {
-            for (Operand o : operands()) {
-                o.setChildExpression(doPromotion(o.getChildExpression(), offer));
-            }
-        }
-    }
-    
-    /**
-      * Promote a subexpression if possible, and if the expression was changed, carry out housekeeping
-      * to reset the static properties and correct the parent pointers in the tree
-      *
-      * @param subexpression the subexpression that is a candidate for promotion
-      * @param offer         details of the promotion being considered @return the result of the promotion. This will be the current expression if no promotion
-      *                      actions have taken place
-      * @return the expression that results from doing the promotion as requested. Returns null if and only
-      *         if the supplied subexpression is null.
-      * @throws net.sf.saxon.trans.XPathException
-      *          if an error occurs
-      */
-
-     public final Expression doPromotion(Expression subexpression, PromotionOffer offer)
-             throws XPathException {
-         if (subexpression == null) {
-             return null;
-         }
-         Expression e = subexpression.promote(offer);
-         if (e != subexpression) {
-             //e.restoreParentPointers();
-             adoptChildExpression(e);
-         } else if (offer.accepted) {
-             ExpressionTool.resetStaticProperties(e);
-         }
-         return e;
-     }
-    
 
     /**
      * Replace this expression by a simpler expression that delivers the results without regard
@@ -663,7 +712,7 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     }
 
     /**
-     * Determine the data type of the expression, if possible. All expression return
+     * Determine the static item type of the expression, as precisely possible. All expression return
      * sequences, in general; this method determines the type of the items within the
      * sequence, assuming that (a) this is known in advance, and (b) it is the same for
      * all items in the sequence.
@@ -678,6 +727,25 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     /*@NotNull*/
     public abstract ItemType getItemType();
 
+    /**
+     * Get the static type of the expression as a SequenceType
+     */
+
+    public SequenceType getStaticType() {
+        return SequenceType.makeSequenceType(getItemType(), getCardinality());
+    }
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType the static type of the context item
+     */
+
+    public UType getStaticUType(UType contextItemType) {
+        return UType.ANY;
+    }
+
     /**
      * Determine which aspects of the context the expression depends on. The result is
      * a bitwise-or'ed value composed from constants such as XPathContext.VARIABLES and
@@ -793,43 +861,6 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return SingletonIterator.makeIterator(value);
     }
 
-    /**
-     * Deliver the result of the expression as a sequence of events.
-     * <p/>
-     * <p>The events (of class {@link net.sf.saxon.evpull.PullEvent}) are either complete
-     * items, or one of startElement, endElement, startDocument, or endDocument, known
-     * as semi-nodes. The stream of events may also include a nested EventIterator.
-     * If a start-end pair exists in the sequence, then the events between
-     * this pair represent the content of the document or element. The content sequence will
-     * have been processed to the extent that any attribute and namespace nodes in the
-     * content sequence will have been merged into the startElement event. Namespace fixup
-     * will have been performed: that is, unique prefixes will have been allocated to element
-     * and attribute nodes, and all namespaces will be declared by means of a namespace node
-     * in the startElement event or in an outer startElement forming part of the sequence.
-     * However, duplicate namespaces may appear in the sequence.</p>
-     * <p>The content of an element or document may include adjacent or zero-length text nodes,
-     * atomic values, and nodes represented as nodes rather than broken down into events.</p>
-     *
-     * @param context The dynamic evaluation context
-     * @return the result of the expression as an iterator over a sequence of PullEvent objects
-     * @throws XPathException if a dynamic error occurs during expression evaluation
-     */
-
-    /*@Nullable*/
-    public EventIterator iterateEvents(XPathContext context) throws XPathException {
-        int m = getImplementationMethod();
-        if ((m & EVALUATE_METHOD) != 0) {
-            Item item = evaluateItem(context);
-            if (item == null) {
-                return EmptyEventIterator.getInstance();
-            } else {
-                return new SingletonEventIterator(item);
-            }
-        } else {
-            return new EventIteratorOverSequence(iterate(context));
-        }
-    }
-
     /**
      * Get the effective boolean value of the expression. This returns false if the value
      * is the empty sequence, a zero-length string, a number equal to zero, or the boolean
@@ -1120,6 +1151,20 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         }
     }
 
+    /**
+     * Ask whether this expression is an instruction. In XSLT streamability analysis this
+     * is used to distinguish constructs corresponding to XSLT instructions from other constructs,
+     * typically XPath expressions (but also things like attribute constructors on a literal result element,
+     * references to attribute sets, etc. However, an XPath expression within a text value template
+     * in a text node of an XSLT stylesheet is treated as an instruction.
+     * In a non-XSLT environment (such as XQuery) this property has no meaning and its setting is undefined.
+     * @return true if this construct originates as an XSLT instruction
+     */
+
+    public boolean isInstruction() {
+        return false;
+    }
+
     /**
      * Compute the static properties. This should only be done once for each
      * expression.
@@ -1141,6 +1186,15 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         staticProperties = -1;
     }
 
+    /**
+     * Ask whether the static properties of the expression have been computed
+     * @return true if the static properties have been computed
+     */
+
+    public boolean isStaticPropertiesKnown() {
+        return staticProperties != -1;
+    }
+
     /**
      * Compute the static cardinality of this expression
      *
@@ -1266,8 +1320,9 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
      * Copy an expression. This makes a deep copy.
      *
      * @return the copy of the original expression
-     * @param rebindings if a local variable is encountered whose binding is present in this
-     *                   map, then the binding will be replaced with a new binding.
+     * @param rebindings a mutable list of (old binding, new binding) pairs
+     *                   that is used to update the bindings held in any
+     *                   local variable references that are copied.
      */
 
     /*@NotNull*/
@@ -1319,24 +1374,6 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         throw new XPathException("Cannot convert the expression {" + toString() + "} to a pattern");
     }
 
-//#ifdefined STREAM
-
-    /**
-     * Convert this expression to a streaming pattern (a pattern used internally to match nodes during
-     * push processing of an event stream)
-     *
-     *
-     * @param config           the Saxon configuration
-     * @return the equivalent pattern if conversion succeeds; otherwise null
-     */
-
-    public Pattern toStreamingPattern(Configuration config) {
-        StreamingAdjunct adj = makeStreamingAdjunct(config, Expression.this);
-        return adj.toStreamingPattern(config);
-    }
-
-//#endif
-
     /**
      * Get the local variables (identified by their slot numbers) on which this expression depends.
      * Should only be called if the caller has established that there is a dependency on local variables.
@@ -1363,11 +1400,7 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
     }
 
     private static void gatherSlotsUsed(Expression exp, IntHashSet slots) {
-//#ifdefined BYTECODE
-        if (exp instanceof CompiledExpression) {
-            exp = ((CompiledExpression)exp).getInterpretedExpression();
-        }
-//#endif
+        exp = exp.getInterpretedExpression();
         if (exp instanceof LocalVariableReference) {
             slots.add(((LocalVariableReference)exp).getSlotNumber());
         } else if (exp instanceof SuppliedParameterReference) {
@@ -1446,16 +1479,6 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return new MonoIterator<String>("expression");
     }
 
-    /**
-     * Get the host language (XSLT, XQuery, XPath) used to implement the code in this container
-     *
-     * @return typically {@link net.sf.saxon.Configuration#XSLT} or {@link net.sf.saxon.Configuration#XQUERY}
-     */
-
-    public int getHostLanguage() {
-        return getPackageData().getHostLanguage();
-    }
-
     /**
      * Add a representation of this expression to a PathMap. The PathMap captures a map of the nodes visited
      * by an expression in a source tree.
@@ -1569,122 +1592,49 @@ public abstract class Expression implements /*InstructionInfo,*/ IdentityCompara
         return System.identityHashCode(Expression.this.getLocation());
     }
 
-    //#ifdefined BYTECODE
-
     /**
-     * Return the compiler relating to a  particular expression
-     *
-     * @return the relevant ExpressionCompiler
-     * @throws CannotCompileException if it is not possible to generate bytecode for this expression
+     * Set an extra property on the expression. (Used for Saxon-EE properties, e.g. properties
+     * relating to streaming).
+     * @param name the name of the property
+     * @param value the value of the property, or null to remove the property
      */
-    public ExpressionCompiler getExpressionCompiler() throws CannotCompileException {
-        if (this instanceof Callable) {
-            if (this instanceof IntegratedFunctionCall) {
-                IntegratedFunctionCall fc = (IntegratedFunctionCall) Expression.this;
-                if (fc.getFunction().getDefinition() instanceof IsWholeNumber) {
-                    return new IsWholeNumberCompiler();
-                }
+
+    public void setExtraProperty(String name, Object value) {
+        if (extraProperties == null) {
+            if (value == null) {
+                return;
             }
-            return new CallableExpressionCompiler();
+            extraProperties = new HashMap<String, Object>(4);
         }
-        throw new CannotCompileException(Expression.this);
-    }
-
-//#endif
-
-//#ifdefined STREAM
-
-    /**
-     * Make a streaming adjunct for a particular expression.
-     *
-     * @param config the Saxon Configuration
-     * @param exp    the expression to be compiled
-     * @return the compiler for this kind of expression
-     */
-
-    public static StreamingAdjunct makeStreamingAdjunct(Configuration config, final Expression exp) {
-        StreamingAdjunct ec = exp.getStreamingAdjunct();
-        ec.setConfiguration(config);
-        ec.setExpression(exp);
-        return ec;
-    }
-
-    /**
-     * Get an object that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-
-    protected StreamingAdjunct getStreamingAdjunct() {
-        return new StreamingAdjunct();
-    }
-
-    /**
-     * Get the posture and sweep of this expression as defined in the W3C streamability specifications.
-     * This provides an assessment of stylesheet code against the W3C criteria for guaranteed
-     * streamability, and is implemented to allow these criteria to be tested. It is not the
-     * case that all expression that emerge as streamable from this analysis are currently
-     * capable of being streamed by Saxon
-     *
-     * @param allowExtensions  if false, the definition of "guaranteed streamability" in the
-     *                         W3C specification is used. If true, Saxon extensions are permitted, which make some
-     * @param contextInfo      Information about the context item type and posture
-     * @param reasons          the caller may supply a list, in which case the implementation may add to this
-     *                         list a message explaining why the construct is not streamable, suitable for inclusion in an
-     * @return the posture and sweep of the expression
-     */
-
-    public PostureAndSweep getStreamability(boolean allowExtensions, ContextItemStaticInfo contextInfo, List<String> reasons) {
-        if (postureAndSweep == null) {
-            postureAndSweep = makeStreamingAdjunct(getConfiguration(), Expression.this)
-                    .computeStreamability(allowExtensions, contextInfo, reasons);
-        }
-        return postureAndSweep;
-    }
-
-    private PostureAndSweep postureAndSweep = null;
-
-    public Posture getPosture() {
-        if (postureAndSweep != null) {
-            return postureAndSweep.getPosture();
+        if (value == null) {
+            extraProperties.remove(name);
         } else {
-            throw new IllegalStateException("getPosture() called when streamability has not yet been computed");
-        }
-    }
-
-    public Sweep getSweep() {
-        if (postureAndSweep != null) {
-            return postureAndSweep.getSweep();
-        } else {
-            throw new IllegalStateException(getClass().getName() + ".getSweep() called when streamability has not yet been computed");
+            extraProperties.put(name, value);
         }
     }
 
     /**
-     * Get the posture and sweep of the expression if these have already been computed
-     * @return the posture and sweep if known, or null if the streamability has not yet been assessed.
+     * Get the value of an extra property on the expression
+     * @param name  the name of the property
+     * @return the value of the property if set, or null otherwise
      */
 
-    public PostureAndSweep getPostureAndSweepIfKnown() {
-        return postureAndSweep;
-    }
-
-    public void clearStreamabilityData() {
-        postureAndSweep = null;
-        for (Operand o : operands()) {
-            o.getChildExpression().clearStreamabilityData();
+    public Object getExtraProperty(String name) {
+        if (extraProperties == null) {
+            return null;
+        } else {
+            return extraProperties.get(name);
         }
     }
 
-    public void setPostureAndSweep(PostureAndSweep ps) {
-        postureAndSweep = ps;
-    }
+    /**
+     * Get the (partial) name of a class that supports streaming of this kind of expression
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
+     */
 
-//#endif
-
-    //project:preconditions
-    public Expression optimizeForType(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType) throws XPathException {
-        return Expression.this;
+    public String getStreamerName() {
+        return null;
     }
 
 
