@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,15 +8,12 @@
 package net.sf.saxon.expr;
 
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.FilterExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.FilterExpressionAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.instruct.Choose;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.functions.PositionAndLast;
 import net.sf.saxon.functions.SystemFunction;
-import net.sf.saxon.functions.VendorFunctionLibrary;
+import net.sf.saxon.functions.registry.VendorFunctionSetHE;
 import net.sf.saxon.lib.FeatureKeys;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.om.*;
@@ -26,6 +23,7 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.iter.EmptyIterator;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
+import net.sf.saxon.value.StringValue;
 
 import java.math.BigInteger;
 
@@ -116,6 +114,20 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         return getBase().getItemType();
     }
 
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @param contextItemType
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        return getBase().getStaticUType(contextItemType);
+    }
+
+
     /**
      * Get the base expression
      *
@@ -235,14 +247,17 @@ public final class FilterExpression extends BinaryExpression implements ContextS
 
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
-        final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        Configuration config = visitor.getConfiguration();
+        final TypeHierarchy th = config.getTypeHierarchy();
         getLhs().typeCheck(visitor, contextInfo);
         getBase().setFiltered(true);
         if (Literal.isEmptySequence(getBase())) {
             return getBase();
         }
 
-        getRhs().typeCheck(visitor, new ContextItemStaticInfo(getBase().getItemType(), false, getBase()));
+        ContextItemStaticInfo baseItemType = config.makeContextItemStaticInfo(getSelectExpression().getItemType(), false);
+        baseItemType.setContextSettingExpression(getBase());
+        getRhs().typeCheck(visitor, baseItemType);
 
         // The filter expression usually doesn't need to be sorted
 
@@ -293,21 +308,16 @@ public final class FilterExpression extends BinaryExpression implements ContextS
 
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType) throws XPathException {
-        final Configuration config = getConfiguration();
+        final Configuration config = visitor.getConfiguration();
         final Optimizer opt = config.obtainOptimizer();
-        final boolean debug = config.getBooleanProperty(FeatureKeys.TRACE_OPTIMIZER_DECISIONS);
+        final boolean tracing = config.getBooleanProperty(FeatureKeys.TRACE_OPTIMIZER_DECISIONS);
         final TypeHierarchy th = config.getTypeHierarchy();
 
         getLhs().optimize(visitor, contextItemType);
         getBase().setFiltered(true);
-
-        Expression originalFilter;
-        try {
-            originalFilter = getFilter().copy(new RebindingMap());
-        } catch (UnsupportedOperationException err) {
-            originalFilter = null;
-        }
-        ContextItemStaticInfo baseItemType = new ContextItemStaticInfo(getBase().getItemType(), false, getBase());
+        
+        ContextItemStaticInfo baseItemType = config.makeContextItemStaticInfo(getSelectExpression().getItemType(), false);
+        baseItemType.setContextSettingExpression(getBase());
         getRhs().optimize(visitor, baseItemType);
 
         // The filter expression usually doesn't need to be sorted
@@ -320,12 +330,15 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         // Rewrite child::X[last()] as child::X[empty(following-sibling::X)] - especially useful for patterns
 
         if (getFilter() instanceof IsLastExpression &&
-            ((IsLastExpression) getFilter()).getCondition() &&
-            getBase() instanceof AxisExpression &&
-            ((AxisExpression) getBase()).getAxis() == AxisInfo.CHILD) {
+                ((IsLastExpression) getFilter()).getCondition() &&
+                getBase() instanceof AxisExpression &&
+                ((AxisExpression) getBase()).getAxis() == AxisInfo.CHILD) {
             NodeTest test = ((AxisExpression) getBase()).getNodeTest();
             AxisExpression fs = new AxisExpression(AxisInfo.FOLLOWING_SIBLING, test);
-            setFilter(SystemFunction.makeCall("empty", getRetainedStaticContext(), new Expression[]{fs}));
+            setFilter(SystemFunction.makeCall("empty", getRetainedStaticContext(), fs));
+            if (tracing) {
+                opt.trace(config, "Replaced [last()] predicate by test for following-sibling", this);
+            }
         }
 
         // if the result of evaluating the filter cannot include numeric values, then we can use
@@ -341,14 +354,14 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         // the filter expression may have been reduced to a constant boolean by previous optimizations
         if (getFilter() instanceof Literal && ((Literal) getFilter()).getValue() instanceof BooleanValue) {
             if (((BooleanValue) ((Literal) getFilter()).getValue()).getBooleanValue()) {
-                if (debug) {
+                if (tracing) {
                     opt.trace("Redundant filter removed", getBase());
                 }
                 return getBase();
             } else {
                 Expression result = Literal.makeEmptySequence();
                 ExpressionTool.copyLocationInfo(this, result);
-                if (debug) {
+                if (tracing) {
                     opt.trace("Filter expression eliminated because predicate is always false", result);
                 }
                 return result;
@@ -364,17 +377,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         // determine whether the filter is indexable
         if (!filterIsPositional && !visitor.isOptimizeForStreaming()) {
             int isIndexable = opt.isIndexableFilter(getFilter());
-            if (isIndexable == 0 && getFilter() != originalFilter && originalFilter != null) {
-                // perhaps the original filter was indexable; if so, revert to the original
-                // this happens when [@a = 1] is rewritten as [some $x in @a satisfies $x eq 1]
-                // TODO: this rollback mechanism is very unsatisfactory. Better: make the some expression indexable!
-                int origIndexable = opt.isIndexableFilter(originalFilter);
-                if (origIndexable != 0) {
-                    isIndexable = origIndexable;
-                    setFilter(originalFilter);
-                    adoptChildExpression(originalFilter);
-                }
-            }
+
             // If the filter is indexable consider creating a key, or an indexed filter expression
             // (This happens in Saxon-EE only)
             if (isIndexable != 0) {
@@ -402,7 +405,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 ExpressionTool.copyLocationInfo(this, f1);
                 FilterExpression f2 = new FilterExpression(f1, p1);
                 ExpressionTool.copyLocationInfo(this, f2);
-                if (debug) {
+                if (tracing) {
                     opt.trace("Composite filter replaced by nested filter expressions", f2);
                 }
                 return f2.optimize(visitor, contextItemType);
@@ -415,7 +418,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 ExpressionTool.copyLocationInfo(this, f1);
                 FilterExpression f2 = new FilterExpression(f1, p0);
                 ExpressionTool.copyLocationInfo(this, f2);
-                if (debug) {
+                if (tracing) {
                     opt.trace("Composite filter replaced by nested filter expressions", f2);
                 }
                 return f2.optimize(visitor, contextItemType);
@@ -432,9 +435,9 @@ public final class FilterExpression extends BinaryExpression implements ContextS
             }
         }
 
-        Expression subsequence = tryToRewritePositionalFilter(visitor);
+        Expression subsequence = tryToRewritePositionalFilter(visitor, tracing);
         if (subsequence != null) {
-            if (debug) {
+            if (tracing) {
                 subsequence.setRetainedStaticContext(getRetainedStaticContext());  // Avoids errors in debug explain
                 opt.trace("Rewrote Filter Expression as:", subsequence);
             }
@@ -448,11 +451,11 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         // of evaluation, so we evaluate the cheapest predicates first
 
         if (!filterIsPositional && getLhsExpression() instanceof FilterExpression &&
-                !((FilterExpression)getLhsExpression()).isFilterIsPositional()) {
+                !((FilterExpression) getLhsExpression()).isFilterIsPositional()) {
             Expression base = ((FilterExpression) getLhsExpression()).getLhsExpression();
             Expression predicate1 = ((FilterExpression) getLhsExpression()).getRhsExpression();
             Expression predicate2 = getRhsExpression();
-            if (predicate1.getCost() > 2*predicate2.getCost()) {
+            if (predicate1.getCost() > 2 * predicate2.getCost()) {
                 FilterExpression fe1 = new FilterExpression(base, predicate2);
                 FilterExpression fe2 = new FilterExpression(fe1, predicate1);
                 ExpressionTool.copyLocationInfo(this, fe2);
@@ -461,39 +464,14 @@ public final class FilterExpression extends BinaryExpression implements ContextS
 
         }
 
-        // If any subexpressions within the filter are not dependent on the focus,
-        // promote them: this causes them to be evaluated once, outside the filter
-        // expression. Note: we do this even if the filter is numeric, because it ensures that
-        // the subscript is pre-evaluated, allowing direct indexing into the sequence.
-
-        PromotionOffer offer = new PromotionOffer(opt);
-        offer.action = PromotionOffer.FOCUS_INDEPENDENT;
-        offer.promoteDocumentDependent = (getBase().getSpecialProperties() & StaticProperty.CONTEXT_DOCUMENT_NODESET) != 0;
-        offer.containingExpression = this;
-        filter2 = doPromotion(getFilter(), offer);
-        if (filter2 != getFilter()) {
-            setFilter(filter2);
-            adoptChildExpression(filter2);
-            ExpressionTool.resetStaticProperties(filter2);
+        final Sequence sequence = tryEarlyEvaluation(visitor);
+        if (sequence != null) {
+            GroundedValue value = SequenceTool.toGroundedValue(sequence);
+            return Literal.makeLiteral(value);
         }
 
-        if (offer.containingExpression instanceof LetExpression) {
-            if (debug) {
-                opt.trace("Subexpression extracted from filter because independent of context", offer.containingExpression);
-            }
-            ExpressionTool.resetStaticProperties(filter2);
-            offer.containingExpression = offer.containingExpression.optimize(visitor, contextItemType);
-        }
-        Expression result = offer.containingExpression;
+        return this;
 
-        if (result instanceof FilterExpression) {
-            final Sequence sequence = ((FilterExpression) result).tryEarlyEvaluation(visitor);
-            if (sequence != null) {
-                GroundedValue value = SequenceTool.toGroundedValue(sequence);
-                return Literal.makeLiteral(value);
-            }
-        }
-        return result;
     }
 
     /**
@@ -506,7 +484,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
      */
     @Override
     public int getCost() {
-        return getLhsExpression().getCost() + 5*getRhsExpression().getCost();
+        return getLhsExpression().getCost() + 5 * getRhsExpression().getCost();
     }
 
     /**
@@ -552,7 +530,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 XPathContext context = visitor.getStaticContext().makeEarlyEvaluationContext();
                 return SequenceExtent.makeSequenceExtent(iterate(context));
             }
-        } catch (Exception e) {
+        } catch (XPathException e) {
             // can happen for a variety of reasons, for example the filter references a global parameter,
             // references the doc() function, etc.
             return null;
@@ -600,40 +578,46 @@ public final class FilterExpression extends BinaryExpression implements ContextS
      * @throws XPathException if an error occurs
      */
 
-    private Expression tryToRewritePositionalFilter(ExpressionVisitor visitor) throws XPathException {
-        if (visitor.isOptimizeForStreaming()) {
-            // TODO: we're suppressing these optimizations because they generate expressions that
-            // can't currently be streamed. But in principle the optimizations are still worth doing.
-            return null;
-        }
-        TypeHierarchy th = getConfiguration().getTypeHierarchy();
+    private Expression tryToRewritePositionalFilter(ExpressionVisitor visitor, boolean tracing) throws XPathException {
+        Configuration config = visitor.getConfiguration();
+        TypeHierarchy th = config.getTypeHierarchy();
         if (getFilter() instanceof Literal) {
             GroundedValue val = ((Literal) getFilter()).getValue();
             if (val instanceof NumericValue) {
-                if (((NumericValue) val).isWholeNumber()) {
-                    long lvalue = ((NumericValue) val).longValue();
-                    if (lvalue <= 0 || lvalue >= Integer.MAX_VALUE) {
-                        return Literal.makeEmptySequence();
-                    } else if (lvalue == 1) {
-                        return FirstItemExpression.makeFirstItemExpression(getBase());
+                Expression result;
+                int lvalue = ((NumericValue)val).asSubscript();
+                if (lvalue != -1) {
+                    if (lvalue == 1) {
+                        result = FirstItemExpression.makeFirstItemExpression(getBase());
                     } else {
-                        return new SubscriptExpression(getBase(), getFilter());
+                        result = new SubscriptExpression(getBase(), getFilter());
                     }
                 } else {
-                    return Literal.makeEmptySequence();
+                    result = Literal.makeEmptySequence();
                 }
+                if (tracing) {
+                    Optimizer.trace(config, "Rewriting numeric filter expression with constant subscript", result);
+                }
+                return result;
             } else {
-                return ExpressionTool.effectiveBooleanValue(val.iterate()) ? getBase() : Literal.makeEmptySequence();
+                Expression result = ExpressionTool.effectiveBooleanValue(val.iterate()) ? getBase() : Literal.makeEmptySequence();
+                if (tracing) {
+                    Optimizer.trace(config, "Rewriting boolean filter expression with constant subscript", result);
+                }
+                return result;
             }
         }
         if (th.isSubType(getFilter().getItemType(), NumericType.getInstance()) &&
                 !Cardinality.allowsMany(getFilter().getCardinality()) &&
                 (getFilter().getDependencies() & StaticProperty.DEPENDS_ON_FOCUS) == 0) {
-            return new SubscriptExpression(getBase(), getFilter());
+            Expression result = new SubscriptExpression(getBase(), getFilter());
+            if (tracing) {
+                Optimizer.trace(config, "Rewriting numeric filter expression with focus-independent subscript", result);
+            }
+            return result;
         }
         if (getFilter() instanceof ComparisonExpression) {
-            VendorFunctionLibrary lib = getConfiguration().getVendorFunctionLibrary();
-            StaticContext env = visitor.getStaticContext();
+            //VendorFunctionLibrary lib = getConfiguration().getVendorFunctionLibrary();
             Expression lhs = ((ComparisonExpression) getFilter()).getLhsExpression();
             Expression rhs = ((ComparisonExpression) getFilter()).getRhsExpression();
             int operator = ((ComparisonExpression) getFilter()).getSingletonOperator();
@@ -668,7 +652,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 comparand = new LocalVariableReference(let);
                 LocalVariableReference existsArg = new LocalVariableReference(let);
                 Expression exists = SystemFunction.makeCall("exists", getRetainedStaticContext(), existsArg);
-                Expression rewrite = tryToRewritePositionalFilterSupport(getBase(), comparand, operator, th, lib, env);
+                Expression rewrite = tryToRewritePositionalFilterSupport(getBase(), comparand, operator, th);
                 if (rewrite == null) {
                     return this;
                 }
@@ -676,7 +660,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 let.setAction(choice);
                 return let;
             } else {
-                return tryToRewritePositionalFilterSupport(getBase(), comparand, operator, th, lib, env);
+                return tryToRewritePositionalFilterSupport(getBase(), comparand, operator, th);
             }
         } else if (getFilter() instanceof IntegerRangeTest) {
             // rewrite SEQ[position() = N to M]
@@ -694,7 +678,11 @@ public final class FilterExpression extends BinaryExpression implements ContextS
             }
             if (ExpressionTool.dependsOnFocus(max)) {
                 if (max.isCallOn(PositionAndLast.Last.class)) {
-                    return SystemFunction.makeCall("subsequence", getRetainedStaticContext(), getBase(), min);
+                    Expression result = SystemFunction.makeCall("subsequence", getRetainedStaticContext(), getBase(), min);
+                    if (tracing) {
+                        Optimizer.trace(config, "Rewriting numeric range filter expression using subsequence()", result);
+                    }
+                    return result;
                 } else {
                     return null;
                 }
@@ -711,6 +699,9 @@ public final class FilterExpression extends BinaryExpression implements ContextS
             Expression length = new ArithmeticExpression(max, Token.MINUS, minMinusOne);
             Expression subs = SystemFunction.makeCall("subsequence", getRetainedStaticContext(), getBase(), min, length);
             let.setAction(subs);
+            if (tracing) {
+                Optimizer.trace(config, "Rewriting numeric range filter expression using subsequence()", subs);
+            }
             return let;
 
         } else {
@@ -720,7 +711,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
 
     private static Expression tryToRewritePositionalFilterSupport(
             Expression start, Expression comparand, int operator,
-            TypeHierarchy th, VendorFunctionLibrary lib, StaticContext env)
+            TypeHierarchy th)
             throws XPathException {
         if (th.isSubType(comparand.getItemType(), BuiltInAtomicType.INTEGER)) {
             switch (operator) {
@@ -745,7 +736,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                         ArithmeticExpression decrement = new ArithmeticExpression(
                                 comparand, Token.MINUS, Literal.makeLiteral(Int64Value.makeIntegerValue(1)));
                         decrement.setCalculator(Calculator.getCalculator(      // bug 2704
-                                StandardNames.XS_INTEGER, StandardNames.XS_INTEGER, Calculator.MINUS, true));
+                                                                               StandardNames.XS_INTEGER, StandardNames.XS_INTEGER, Calculator.MINUS, true));
                         args[2] = decrement;
                     }
                     return SystemFunction.makeCall("subsequence", start.getRetainedStaticContext(), args);
@@ -796,14 +787,13 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                     LocalVariableReference isWholeArg = new LocalVariableReference(let);
                     LocalVariableReference arithArg = new LocalVariableReference(let);
                     LocalVariableReference floorArg = new LocalVariableReference(let);
-                    Expression isWhole = lib.makeSaxonFunction(
-                            "is-whole-number", new Expression[]{isWholeArg}, env);
+                    Expression isWhole = VendorFunctionSetHE.getInstance().makeFunction("is-whole-number", 1).makeFunctionCall(isWholeArg);
                     Expression minusOne = new ArithmeticExpression(
                             arithArg, Token.MINUS, Literal.makeLiteral(Int64Value.makeIntegerValue(1)));
                     Expression floor = SystemFunction.makeCall("floor", start.getRetainedStaticContext(), floorArg);
                     Expression choice = Choose.makeConditional(isWhole, minusOne, floor);
                     Expression subs = SystemFunction.makeCall(
-                        "subsequence", start.getRetainedStaticContext(), start, Literal.makeLiteral(Int64Value.makeIntegerValue(1)), choice);
+                            "subsequence", start.getRetainedStaticContext(), start, Literal.makeLiteral(Int64Value.makeIntegerValue(1)), choice);
                     let.setAction(subs);
                     //decl.fixupReferences(let);
                     return let;
@@ -811,7 +801,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                 case Token.FLE: {
                     Expression floor = SystemFunction.makeCall("floor", start.getRetainedStaticContext(), comparand);
                     return SystemFunction.makeCall(
-                        "subsequence", start.getRetainedStaticContext(), start, Literal.makeLiteral(Int64Value.makeIntegerValue(1)), floor);
+                            "subsequence", start.getRetainedStaticContext(), start, Literal.makeLiteral(Int64Value.makeIntegerValue(1)), floor);
                 }
                 case Token.FNE: {
                     // rewrite SEQ[position() ne V] as
@@ -824,8 +814,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                     let.setSequence(comparand);
                     LocalVariableReference isWholeArg = new LocalVariableReference(let);
                     LocalVariableReference castArg = new LocalVariableReference(let);
-                    Expression isWhole = lib.makeSaxonFunction(
-                            "is-whole-number", new Expression[]{isWholeArg}, env);
+                    Expression isWhole = VendorFunctionSetHE.getInstance().makeFunction("is-whole-number", 1).makeFunctionCall(isWholeArg);
                     ExpressionTool.copyLocationInfo(start, isWhole);
                     Expression cast = new CastExpression(castArg, BuiltInAtomicType.INTEGER, false);
                     ExpressionTool.copyLocationInfo(start, cast);
@@ -846,8 +835,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                     LocalVariableReference isWholeArg = new LocalVariableReference(let);
                     LocalVariableReference arithArg = new LocalVariableReference(let);
                     LocalVariableReference ceilingArg = new LocalVariableReference(let);
-                    Expression isWhole = lib.makeSaxonFunction(
-                            "is-whole-number", new Expression[]{isWholeArg}, env);
+                    Expression isWhole = VendorFunctionSetHE.getInstance().makeFunction("is-whole-number", 1).makeFunctionCall(isWholeArg);
                     Expression plusOne = new ArithmeticExpression(
                             arithArg, Token.PLUS, Literal.makeLiteral(Int64Value.makeIntegerValue(1)));
                     Expression ceiling = SystemFunction.makeCall("ceiling", start.getRetainedStaticContext(), ceilingArg);
@@ -867,35 +855,6 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         }
     }
 
-    /**
-     * Promote this expression if possible
-     *
-     * @param offer details of the promotion that is possible
-     * @return the promoted expression (or the original expression, unchanged)
-     */
-
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        Expression exp = offer.accept(this);
-        if (exp != null) {
-            return exp;
-        } else {
-            if ((offer.action == PromotionOffer.RANGE_INDEPENDENT) && (getBase() instanceof FilterExpression)) {
-                // Try to handle the situation where we have (EXP[PRED1])[PRED2], and
-                // EXP[PRED2] does not depend on the range variable, but PRED1 does
-                TypeHierarchy th = offer.getOptimizer().getConfiguration().getTypeHierarchy();
-                FilterExpression newfe = promoteIndependentPredicates(offer.bindingList, offer.getOptimizer(), th);
-                if (newfe != this) {
-                    return newfe.promote(offer);
-                }
-            }
-            setBase(doPromotion(getBase(), offer));
-            // Don't pass on requests to the filter. We could pass them on, but only after augmenting
-            // them to say we are interested in subexpressions that don't depend on either the
-            // outer context or the inner context.
-            return this;
-        }
-    }
-
     /**
      * Replace this expression by a simpler expression that delivers the results without regard
      * to order.
@@ -1008,7 +967,7 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                     filterType == BuiltInAtomicType.DECIMAL || filterType == BuiltInAtomicType.FLOAT) {
                 return StaticProperty.ALLOWS_ZERO_OR_ONE;
             }
-            if (getFilter() instanceof ArithmeticExpression || getFilter() instanceof ArithmeticExpression10) {
+            if (getFilter() instanceof ArithmeticExpression) {
                 return StaticProperty.ALLOWS_ZERO_OR_ONE;
             }
         }
@@ -1075,15 +1034,15 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         TypeHierarchy th = config.getTypeHierarchy();
         Pattern basePattern = base.toPattern(config, is30);
         if (!isPositional(th)) {
-            return new PatternWithPredicate(basePattern, filter);
+            return new BasePatternWithPredicate(basePattern, filter);
         } else if (basePattern instanceof NodeTestPattern &&
                 basePattern.getItemType() instanceof NodeTest &&
                 filterIsPositional &&
                 base instanceof AxisExpression &&
-                ((AxisExpression)base).getAxis() == AxisInfo.CHILD &&
+                ((AxisExpression) base).getAxis() == AxisInfo.CHILD &&
                 (filter.getDependencies() & StaticProperty.DEPENDS_ON_LAST) == 0) {
-            if (filter instanceof Literal && ((Literal)filter).getValue() instanceof IntegerValue) {
-                return new SimplePositionalPattern((NodeTest) basePattern.getItemType(), (int)((IntegerValue) ((Literal) filter).getValue()).longValue());
+            if (filter instanceof Literal && ((Literal) filter).getValue() instanceof IntegerValue) {
+                return new SimplePositionalPattern((NodeTest) basePattern.getItemType(), (int) ((IntegerValue) ((Literal) filter).getValue()).longValue());
             } else {
                 return new GeneralPositionalPattern((NodeTest) basePattern.getItemType(), filter);
             }
@@ -1120,9 +1079,8 @@ public final class FilterExpression extends BinaryExpression implements ContextS
                         ExpressionTool.ebvError("sequence of two or more items starting with a numeric value");
                     } else {
                         // Filter is a constant number
-                        NumericValue subscript = (NumericValue)first;
-                        if (subscript.isWholeNumber() && subscript.signum() > 0 && subscript.longValue() < Integer.MAX_VALUE) {
-                            int pos = (int) subscript.longValue();
+                        int pos = ((NumericValue)first).asSubscript();
+                        if (pos != -1) {
                             if (getBase() instanceof VariableReference) {
                                 Sequence baseVal = ((VariableReference) getBase()).evaluateVariable(context);
                                 if (baseVal instanceof MemoClosure) {
@@ -1191,28 +1149,13 @@ public final class FilterExpression extends BinaryExpression implements ContextS
 
     }
 
-    /**
-     * Determine which aspects of the context the expression depends on. The result is
-     * a bitwise-or'ed value composed from constants such as XPathContext.VARIABLES and
-     * XPathContext.CURRENT_NODE
-     *
-     * @return the dependencies
-     */
-
-//    public int computeDependencies() {
-//        // not all dependencies in the filter expression matter, because the context node,
-//        // position, and size are not dependent on the outer context.
-//        return getBase().getDependencies() |
-//                (getFilter().getDependencies() & (StaticProperty.DEPENDS_ON_XSLT_CONTEXT |
-//                        StaticProperty.DEPENDS_ON_LOCAL_VARIABLES |
-//                        StaticProperty.DEPENDS_ON_USER_FUNCTIONS));
-//    }
-
     /**
      * Copy an expression. This makes a deep copy.
      *
+     * @param rebindings a mutable list of (old binding, new binding) pairs
+     *                   that is used to update the bindings held in any
+     *                   local variable references that are copied.
      * @return the copy of the original expression
-     * @param rebindings
      */
 
     /*@NotNull*/
@@ -1225,33 +1168,17 @@ public final class FilterExpression extends BinaryExpression implements ContextS
         return fe;
     }
 
-//#ifdefined BYTECODE
 
     /**
-     * Return the compiler of the Filter expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new FilterExpressionCompiler();
+    public String getStreamerName() {
+        return "FilterExpression";
     }
-//#endif
-
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public FilterExpressionAdjunct getStreamingAdjunct() {
-        return new FilterExpressionAdjunct();
-    }
-
-
-//#endif
 
     /**
      * The toString() method for an expression attempts to give a representation of the expression
