@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -208,6 +208,7 @@ public abstract class FunctionCall extends Expression {
                                   ExpressionVisitor visitor) throws XPathException {
 //        boolean isXSLT = target instanceof UserFunction &&
 //            ((UserFunction) target).getPackageData().getHostLanguage() == Configuration.XSLT;
+        TypeChecker tc = visitor.getConfiguration().getTypeChecker(visitor.getStaticContext().isInBackwardsCompatibleMode());
         SequenceType[] argTypes = target.getFunctionItemType().getArgumentTypes();
         int n = target.getArity();
         for (int i = 0; i < n; i++) {
@@ -218,10 +219,9 @@ public abstract class FunctionCall extends Expression {
 //            if (isXSLT) {
 //                role.setErrorCode("XTTE0790");
 //            }
-            setArg(i, TypeChecker.staticTypeCheck(
+            setArg(i, tc.staticTypeCheck(
                 getArg(i),
                 argTypes[i],
-                visitor.getStaticContext().isInBackwardsCompatibleMode(),
                 role, visitor));
         }
     }
@@ -282,13 +282,14 @@ public abstract class FunctionCall extends Expression {
      */
 
     public Expression preEvaluate(ExpressionVisitor visitor) throws XPathException {
-        if (getIntrinsicDependencies() != 0) {
+        if ((getIntrinsicDependencies() & ~StaticProperty.DEPENDS_ON_STATIC_CONTEXT) != 0) {
             return this;
         }
         try {
             Literal lit = Literal.makeLiteral(SequenceExtent.makeSequenceExtent(
                     iterate(visitor.getStaticContext().makeEarlyEvaluationContext())));
             ExpressionTool.copyLocationInfo(this, lit);
+            Optimizer.trace(visitor.getConfiguration(), "Pre-evaluated function call " + toShortString(), lit);
             return lit;
         } catch (NoDynamicContextException e) {
             // early evaluation failed, usually because implicit timezone required
@@ -329,11 +330,11 @@ public abstract class FunctionCall extends Expression {
         int numArgs = getArity();
         String msg = null;
         if (min == max && numArgs != min) {
-            msg = "Function " + getDisplayName() + " must have " + min + pluralArguments(min);
+            msg = "Function " + getDisplayName() + " must have " + pluralArguments(min);
         } else if (numArgs < min) {
-            msg = "Function " + getDisplayName() + " must have at least " + min + pluralArguments(min);
+            msg = "Function " + getDisplayName() + " must have at least " + pluralArguments(min);
         } else if (numArgs > max) {
-            msg = "Function " + getDisplayName() + " must have no more than " + max + pluralArguments(max);
+            msg = "Function " + getDisplayName() + " must have no more than " + pluralArguments(max);
         }
         if (msg != null) {
             XPathException err = new XPathException(msg, "XPST0017");
@@ -365,7 +366,7 @@ public abstract class FunctionCall extends Expression {
      */
 
     public static String pluralArguments(int num) {
-        return num == 1 ? " argument" : " arguments";
+        return num == 1 ? "one argument" : (num + " arguments");
     }
 
 
@@ -411,8 +412,8 @@ public abstract class FunctionCall extends Expression {
      *         in explain() output displaying the expression.
      */
 
-    public final String getExpressionName() {
-        return getDisplayName();
+    public String getExpressionName() {
+        return "functionCall";
     }
 
     /**
