@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,14 +7,12 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.GeneralComparisonCompiler;
-import com.saxonica.ee.stream.adjunct.GeneralComparisonAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.expr.sort.AtomicComparer;
 import net.sf.saxon.expr.sort.CodepointCollator;
 import net.sf.saxon.expr.sort.GenericAtomicComparer;
+import net.sf.saxon.expr.sort.UntypedNumericComparer;
 import net.sf.saxon.functions.Minimax;
 import net.sf.saxon.functions.SystemFunction;
 import net.sf.saxon.lib.ConversionRules;
@@ -24,6 +22,7 @@ import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
+import net.sf.saxon.value.StringValue;
 
 import java.util.ArrayList;
 import java.util.List;
@@ -35,7 +34,7 @@ import java.util.List;
  * =, !=, <, >, etc. This implementation is not used when in backwards-compatible mode
  */
 
-public class GeneralComparison extends BinaryExpression implements ComparisonExpression, Callable {
+public abstract class GeneralComparison extends BinaryExpression implements ComparisonExpression, Callable {
 
     public static final int ONE_TO_ONE = 0;
     public static final int MANY_TO_ONE = 1;
@@ -116,25 +115,25 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
 
     /*@NotNull*/
     @Override
-    public Expression simplify() throws XPathException {
-        Expression e = super.simplify();
-        if (e != this) {
-            ExpressionTool.copyLocationInfo(this, e);
-            return e;
-        } else if (getRetainedStaticContext().isBackwardsCompatibility()) {
-            GeneralComparison10 gc10 = new GeneralComparison10(getLhsExpression(), getOperator(), getRhsExpression());
-            gc10.setRetainedStaticContext(getRetainedStaticContext());
-            gc10.setAtomicComparer(getAtomicComparer());
-            ExpressionTool.copyLocationInfo(this, gc10);
-            return gc10;
-        } else {
-            GeneralComparison20 gc20 = new GeneralComparison20(getLhsExpression(), getOperator(), getRhsExpression());
-            gc20.setRetainedStaticContext(getRetainedStaticContext());
-            gc20.setAtomicComparer(getAtomicComparer());
-            ExpressionTool.copyLocationInfo(this, gc20);
-            return gc20;
-        }
-    }
+//    public Expression simplify() throws XPathException {
+//        Expression e = super.simplify();
+//        if (e != this) {
+//            ExpressionTool.copyLocationInfo(this, e);
+//            return e;
+//        } else if (getRetainedStaticContext().isBackwardsCompatibility()) {
+//            GeneralComparison10 gc10 = new GeneralComparison10(getLhsExpression(), getOperator(), getRhsExpression());
+//            gc10.setRetainedStaticContext(getRetainedStaticContext());
+//            gc10.setAtomicComparer(getAtomicComparer());
+//            ExpressionTool.copyLocationInfo(this, gc10);
+//            return gc10;
+//        } else {
+//            GeneralComparison20 gc20 = new GeneralComparison20(getLhsExpression(), getOperator(), getRhsExpression());
+//            gc20.setRetainedStaticContext(getRetainedStaticContext());
+//            gc20.setAtomicComparer(getAtomicComparer());
+//            ExpressionTool.copyLocationInfo(this, gc20);
+//            return gc20;
+//        }
+//    }
 
     /**
      * Get a name identifying the kind of expression, in terms meaningful to a user.
@@ -202,7 +201,7 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
 
-        final Configuration config = getConfiguration();
+        final Configuration config = visitor.getConfiguration();
         final TypeHierarchy th = config.getTypeHierarchy();
 
         Expression oldOp0 = getLhsExpression();
@@ -223,13 +222,14 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
 
         SequenceType atomicType = SequenceType.ATOMIC_SEQUENCE;
 
+        TypeChecker tc = config.getTypeChecker(false);
         RoleDiagnostic role0 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 0);
         //role0.setSourceLocator(this);
-        setLhsExpression(TypeChecker.staticTypeCheck(getLhsExpression(), atomicType, false, role0, visitor));
+        setLhsExpression(tc.staticTypeCheck(getLhsExpression(), atomicType, role0, visitor));
 
         RoleDiagnostic role1 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 1);
         //role1.setSourceLocator(this);
-        setRhsExpression(TypeChecker.staticTypeCheck(getRhsExpression(), atomicType, false, role1, visitor));
+        setRhsExpression(tc.staticTypeCheck(getRhsExpression(), atomicType, role1, visitor));
 
         if (getLhsExpression() != oldOp0) {
             adoptChildExpression(getLhsExpression());
@@ -269,7 +269,7 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
             // then no static type checking is possible
         } else {
             if (!Type.isPossiblyComparable(pt0, pt1, Token.isOrderedOperator(singletonOperator))) {
-                String message = "Cannot compare " + t0.toString() + " to " + t1.toString();
+                String message = "In {" + toShortString() + "}: cannot compare " + t0.toString() + " to " + t1.toString();
                 if (Cardinality.allowsZero(c0) || Cardinality.allowsZero(c1)) {
                     if (!doneWarnings) { // avoid duplicate warnings
                         doneWarnings = true;
@@ -306,7 +306,7 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
 
             Expression e0 = getLhsExpression();
             Expression e1 = getRhsExpression();
-
+            
             if (t0.equals(BuiltInAtomicType.UNTYPED_ATOMIC)) {
                 if (t1.equals(BuiltInAtomicType.UNTYPED_ATOMIC)) {
                     e0 = new CastExpression(getLhsExpression(), BuiltInAtomicType.STRING, false);
@@ -314,16 +314,20 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
                     e1 = new CastExpression(getRhsExpression(), BuiltInAtomicType.STRING, false);
                     adoptChildExpression(e1);
                 } else if (th.isSubType(t1, NumericType.getInstance())) {
-                    e0 = new CastExpression(getLhsExpression(), BuiltInAtomicType.DOUBLE, false);
-                    adoptChildExpression(e0);
+//                    e0 = new CastExpression(getLhsExpression(), BuiltInAtomicType.DOUBLE, false);
+//                    adoptChildExpression(e0);
+                    Expression vun = makeCompareUntypedToNumeric(getLhsExpression(), getRhsExpression(), singletonOperator);
+                    return vun.typeCheck(visitor, contextInfo);
                 } else {
                     e0 = new CastExpression(getLhsExpression(), pt1, false);
                     adoptChildExpression(e0);
                 }
             } else if (t1.equals(BuiltInAtomicType.UNTYPED_ATOMIC)) {
                 if (th.isSubType(t0, NumericType.getInstance())) {
-                    e1 = new CastExpression(getRhsExpression(), BuiltInAtomicType.DOUBLE, false);
-                    adoptChildExpression(e1);
+//                    e1 = new CastExpression(getRhsExpression(), BuiltInAtomicType.DOUBLE, false);
+//                    adoptChildExpression(e1);
+                    Expression vun = makeCompareUntypedToNumeric(getRhsExpression(), getLhsExpression(), Token.inverse(singletonOperator));
+                    return vun.typeCheck(visitor, contextInfo);
                 } else {
                     e1 = new CastExpression(getRhsExpression(), pt0, false);
                     adoptChildExpression(e1);
@@ -333,6 +337,7 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
             ValueComparison vc = new ValueComparison(e0, singletonOperator, e1);
             vc.setAtomicComparer(comparer);
             ExpressionTool.copyLocationInfo(this, vc);
+            Optimizer.trace(config, "Replaced general comparison by value comparison", vc);
             return vc.typeCheck(visitor, contextInfo);
         }
 
@@ -353,14 +358,26 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
         if ((getLhsExpression() instanceof Literal) && (getRhsExpression() instanceof Literal)) {
             return Literal.makeLiteral(evaluateItem(env.makeEarlyEvaluationContext()));
         }
-
         return this;
     }
 
+    private Expression makeCompareUntypedToNumeric(Expression lhs, Expression rhs, int operator) {
+        ValueComparison vc = new ValueComparison(lhs, operator, rhs);
+        vc.setAtomicComparer(new UntypedNumericComparer());
+        ExpressionTool.copyLocationInfo(this, vc);
+        Optimizer.trace(getConfiguration(), "Replaced general comparison by value comparison", vc);
+        return vc;
+    }
+
+
     private static Expression makeMinOrMax(Expression exp, String function) throws XPathException {
-        Expression fn = SystemFunction.makeCall(function, exp.getRetainedStaticContext(), exp);
-        ((Minimax) ((SystemFunctionCall)fn).getTargetFunction()).setIgnoreNaN(true);
-        return fn;
+        if (Cardinality.allowsMany(exp.getCardinality())) {
+            Expression fn = SystemFunction.makeCall(function, exp.getRetainedStaticContext(), exp);
+            ((Minimax) ((SystemFunctionCall) fn).getTargetFunction()).setIgnoreNaN(true);
+            return fn;
+        } else {
+            return exp;
+        }
     }
 
     /**
@@ -372,7 +389,7 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
 
-        final TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        final TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
         final StaticContext env = visitor.getStaticContext();
 
         optimizeChildren(visitor, contextInfo);
@@ -482,10 +499,13 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
         // comparison isn't known in advance. For example [(1, U1) < ("fred", U2)]
         // involves both string and numeric comparisons.
 
-        // Only do this optimization for a many-to-many comparison, because it prevents
-        // early exit on a many-to-one comparison.
+        // Generally, do this optimization for a many-to-many comparison, because it prevents
+        // early exit on a many-to-one comparison. But with a many-to-one comparison, do it
+        // if the "many" branch can be lifted up the expression tree.
 
-        if (operator != Token.EQUALS && operator != Token.NE && comparisonCardinality == MANY_TO_MANY &&
+        if (operator != Token.EQUALS &&
+                operator != Token.NE &&
+                (comparisonCardinality == MANY_TO_MANY || comparisonCardinality == MANY_TO_ONE && manyOperandIsLiftable()) &&
                 (th.isSubType(t0, NumericType.getInstance()) || th.isSubType(t1, NumericType.getInstance()))) {
 
             // System.err.println("** using minimax optimization **");
@@ -523,10 +543,24 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
         }
 
         // Finally, convert to use the GeneralComparisonEE algorithm if in Saxon-EE
-        return getConfiguration().obtainOptimizer()
+        return visitor.getConfiguration().obtainOptimizer()
                 .optimizeGeneralComparison(visitor, this, false, contextInfo);
     }
 
+    private boolean manyOperandIsLiftable() {
+        if (getParentExpression() instanceof ContextSwitchingExpression &&
+                ((ContextSwitchingExpression)getParentExpression()).getActionExpression() == this) {
+            for (Operand o : operands()) {
+                if (Cardinality.allowsMany(o.getChildExpression().getCardinality())) {
+                    if (ExpressionTool.dependsOnFocus(o.getChildExpression())) {
+                        return false;
+                    }
+                }
+            }
+            return true;
+        };
+        return false;
+    }
 
     /**
      * Copy an expression. This makes a deep copy.
@@ -536,16 +570,16 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
      */
 
     /*@NotNull*/
-    public Expression copy(RebindingMap rebindings) {
-        GeneralComparison gc = new GeneralComparison(getLhsExpression().copy(rebindings), operator, getRhsExpression().copy(rebindings));
-        ExpressionTool.copyLocationInfo(this, gc);
-        gc.setRetainedStaticContext(getRetainedStaticContext());
-        gc.comparer = comparer;
-        gc.singletonOperator = singletonOperator;
-        gc.needsRuntimeCheck = needsRuntimeCheck;
-        gc.comparisonCardinality = comparisonCardinality;
-        return gc;
-    }
+//    public abstract Expression copy(RebindingMap rebindings) {
+//        GeneralComparison gc = new GeneralComparison(getLhsExpression().copy(rebindings), operator, getRhsExpression().copy(rebindings));
+//        ExpressionTool.copyLocationInfo(this, gc);
+//        gc.setRetainedStaticContext(getRetainedStaticContext());
+//        gc.comparer = comparer;
+//        gc.singletonOperator = singletonOperator;
+//        gc.needsRuntimeCheck = needsRuntimeCheck;
+//        gc.comparisonCardinality = comparisonCardinality;
+//        return gc;
+//    }
 
     /**
      * Evaluate the expression in a given context
@@ -796,24 +830,28 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
             if (u0) {
                 // a0 is untyped atomic
                 if (a1 instanceof NumericValue) {
-                    return quickCompare((UntypedAtomicValue)a0, (NumericValue)a1, operator, rules);
+                    return UntypedNumericComparer.quickCompare((UntypedAtomicValue) a0, (NumericValue) a1, operator, rules);
+                } else if (a1 instanceof StringValue) {
+                    // no conversion needed
                 } else {
                     StringConverter sc = a1.getItemType().getPrimitiveItemType().getStringConverter(rules);
                     if (a1 instanceof QualifiedNameValue) {
                         sc = (StringConverter)sc.setNamespaceResolver(nsResolver);
                     }
-                    a0 = sc.convertString(a0.getStringValueCS()).asAtomic();
+                    a0 = ((UntypedAtomicValue)a0).obtainConversionResult(a1.getItemType().getPrimitiveType(), sc).asAtomic();
                 }
             } else {
                 // a1 is untyped atomic
                 if (a0 instanceof NumericValue) {
-                    return quickCompare((UntypedAtomicValue)a1, (NumericValue)a0, Token.inverse(operator), rules);
+                    return UntypedNumericComparer.quickCompare((UntypedAtomicValue) a1, (NumericValue) a0, Token.inverse(operator), rules);
+                } else if (a0 instanceof StringValue) {
+                    // no conversion needed
                 } else {
                     StringConverter sc = a0.getItemType().getPrimitiveItemType().getStringConverter(rules);
                     if (a0 instanceof QualifiedNameValue) {
                         sc = (StringConverter)sc.setNamespaceResolver(nsResolver);
                     }
-                    a1 = sc.convertString(a1.getStringValueCS()).asAtomic();
+                    a1 = ((UntypedAtomicValue) a1).obtainConversionResult(a0.getItemType().getPrimitiveType(), sc).asAtomic();
                 }
             }
             checkTypes = false; // No further checking needed if conversion succeeded
@@ -833,6 +871,18 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
         return BuiltInAtomicType.BOOLEAN;
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        return UType.BOOLEAN;
+    }
+
     /**
      * Return the singleton form of the comparison operator, e.g. FEQ for EQUALS, FGT for GT
      *
@@ -860,34 +910,22 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
     }
 
     protected GeneralComparison getInverseComparison() {
-        GeneralComparison gc2 = new GeneralComparison(getRhsExpression(), Token.inverse(operator), getLhsExpression());
+        GeneralComparison20 gc2 = new GeneralComparison20(getRhsExpression(), Token.inverse(operator), getLhsExpression());
         gc2.setRetainedStaticContext(getRetainedStaticContext());
         return gc2;
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the GeneralComparison expression
+  /**
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new GeneralComparisonCompiler();
+    public String getStreamerName() {
+        return "GeneralComparison";
     }
 
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public GeneralComparisonAdjunct getStreamingAdjunct() {
-        return new GeneralComparisonAdjunct();
-    }
-
-    //#endif
 
 //    protected String displayOperator() {
 //        return "many-to-many " + super.displayOperator();
@@ -920,160 +958,6 @@ public class GeneralComparison extends BinaryExpression implements ComparisonExp
         out.emitAttribute("comp", comparer.save());
     }
 
-    /**
-     * Optimized routine to compare an untyped atomic value with a numeric value.
-     * This attempts to deliver a quick answer if the comparison if obviously false,
-     * without performing the full string-to-double conversion
-     * @param a0 the untypedAtomic comparand
-     * @param a1 the numeric comparand
-     * @param operator the comparison operator: a singleton operator such as Token.FEQ
-     * @param rules the conversion rules
-     * @return the result of the comparison
-     * @throws XPathException if the first operand is not convertible to a double
-     */
-
-    public static boolean quickCompare(
-            UntypedAtomicValue a0, NumericValue a1, int operator, ConversionRules rules)
-            throws XPathException {
-        double d1 = a1.getDoubleValue();
-        double d0;
-        int comp;
-        DoubleValue dv0 = a0.getDoubleValueIfKnown();
-        if (dv0 != null) {
-            d0 = dv0.getDoubleValue();
-            comp = Double.compare(d0, d1);
-        } else {
-            CharSequence cs = Whitespace.trimWhitespace(a0.getStringValueCS());
-
-            boolean simple = true;
-            int wholePartLength = 0;
-            int firstDigit = -1;
-            int decimalPoints = 0;
-            char sign = '?';
-            for (int i=0; i<cs.length(); i++) {
-                char c = cs.charAt(i);
-                if (c >= '0' && c <= '9') {
-                    if (firstDigit < 0) {
-                        firstDigit = c - '0';
-                    }
-                    if (decimalPoints == 0) {
-                        wholePartLength++;
-                    }
-                } else if (c == '-') {
-                    if (sign != '?' || wholePartLength > 0 || decimalPoints > 0) {
-                        simple = false;
-                        break;
-                    }
-                    sign = c;
-                } else if (c == '.') {
-                    if (decimalPoints > 0) {
-                        simple = false;
-                        break;
-                    }
-                    decimalPoints = 1;
-                } else {
-                    simple = false;
-                    break;
-                }
-            }
-            if (firstDigit < 0) {
-                simple = false;
-            }
-            if (simple && wholePartLength > 0 && wholePartLength <= 10) {
-                double lowerBound = bounds[firstDigit][wholePartLength];
-                double upperBound = bounds[firstDigit+1][wholePartLength];
-                if (sign == '-') {
-                    double temp = lowerBound;
-                    lowerBound = -upperBound;
-                    upperBound = -temp;
-                }
-                switch (operator) {
-                    case Token.FEQ:
-                        if (lowerBound > d1 || upperBound < d1) {
-                            return false;
-                        }
-                        break;
-                    case Token.FLE:
-                        if (lowerBound > d1) {
-                            return false;
-                        }
-                        if (upperBound <= d1) {
-                            return true;
-                        }
-                        break;
-                    case Token.FLT:
-                        if (lowerBound >= d1) {
-                            return false;
-                        }
-                        if (upperBound < d1) {
-                            return true;
-                        }
-                        break;
-                    case Token.FGE:
-                        if (upperBound < d1) {
-                            return false;
-                        }
-                        if (lowerBound >= d1) {
-                            return true;
-                        }
-                        break;
-                    case Token.FGT:
-                        if (upperBound <= d1) {
-                            return false;
-                        }
-                        if (lowerBound > d1) {
-                            return true;
-                        }
-                        break;
-                    case Token.FNE:
-                        if (upperBound < d1 || lowerBound > d1) {
-                            return true;
-                        }
-                        break;
-                    default:
-                        break;
-                }
-            }
-            if (simple && decimalPoints == 0 && wholePartLength <= 15 && a1 instanceof Int64Value) {
-                long l0 = Long.parseLong(cs.toString());
-                comp = new Long(l0).compareTo(a1.longValue());
-            } else {
-                d0 = ((DoubleValue)BuiltInAtomicType.DOUBLE.getStringConverter(rules).convertString(cs).asAtomic()).getDoubleValue();
-                a0.setDoubleValue(new DoubleValue(d0));
-                comp = Double.compare(d0, d1);
-            }
-        }
-        switch (operator) {
-            case Token.FEQ:
-                return comp == 0;
-            case Token.FLE:
-                return comp <= 0;
-            case Token.FLT:
-                return comp < 0;
-            case Token.FGE:
-                return comp >= 0;
-            case Token.FGT:
-                return comp > 0;
-            case Token.FNE:
-            default:
-                return comp != 0;
-        }
-    }
-
-    private static double[][] bounds = {
-            {1, 0e0, 0e1, 0e2, 0e3, 0e4, 0e5, 0e6, 0e7, 0e8, 0e9, 0e10},
-            {1, 1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9, 1e10},
-            {1, 2e0, 2e1, 2e2, 2e3, 2e4, 2e5, 2e6, 2e7, 2e8, 2e9, 2e10},
-            {1, 3e0, 3e1, 3e2, 3e3, 3e4, 3e5, 3e6, 3e7, 3e8, 3e9, 3e10},
-            {1, 4e0, 4e1, 4e2, 4e3, 4e4, 4e5, 4e6, 4e7, 4e8, 4e9, 4e10},
-            {1, 5e0, 5e1, 5e2, 5e3, 5e4, 5e5, 5e6, 5e7, 5e8, 5e9, 5e10},
-            {1, 6e0, 6e1, 6e2, 6e3, 6e4, 6e5, 6e6, 6e7, 6e8, 6e9, 6e10},
-            {1, 7e0, 7e1, 7e2, 7e3, 7e4, 7e5, 7e6, 7e7, 7e8, 7e9, 7e10},
-            {1, 8e0, 8e1, 8e2, 8e3, 8e4, 8e5, 8e6, 8e7, 8e8, 8e9, 8e10},
-            {1, 9e0, 9e1, 9e2, 9e3, 9e4, 9e5, 9e6, 9e7, 9e8, 9e9, 9e10},
-            {1, 10e0, 10e1, 10e2, 10e3, 10e4, 10e5, 10e6, 10e7, 10e8, 10e9, 10e10}
-    };
-
     /**
      * Find any necessary preconditions for the satisfaction of this expression
      * as a set of boolean expressions to be evaluated on the context node
