@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,7 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.LetExpressionCompiler;
-import com.saxonica.ee.stream.Sweep;
-import com.saxonica.ee.stream.adjunct.LetExpressionAdjunct;
-import com.saxonica.ee.stream.adjunct.LetExpressionAdjunctB;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
-import net.sf.saxon.evpull.EventIterator;
 import net.sf.saxon.expr.instruct.DocumentInstr;
-import net.sf.saxon.expr.instruct.GlobalVariable;
 import net.sf.saxon.expr.instruct.TailCall;
 import net.sf.saxon.expr.instruct.TailCallReturner;
 import net.sf.saxon.expr.parser.*;
@@ -24,10 +16,12 @@ import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.SchemaType;
-import net.sf.saxon.value.Cardinality;
+import net.sf.saxon.type.UType;
 import net.sf.saxon.value.IntegerValue;
 import net.sf.saxon.value.SequenceType;
 
+import java.util.ArrayList;
+
 
 /**
  * A LetExpression represents the XQuery construct let $x := expr return expr. It is used
@@ -38,6 +32,8 @@ public class LetExpression extends Assignation implements TailCallReturner {
 
     private int evaluationMode = ExpressionTool.UNDECIDED;
     private boolean needsEagerEvaluation = false;
+    private boolean needsLazyEvaluation = false;
+    private boolean isInstruction;
 
     /**
      * Create a LetExpression
@@ -47,6 +43,30 @@ public class LetExpression extends Assignation implements TailCallReturner {
         //System.err.println("let");
     }
 
+
+    /**
+     * Say whether this expression originates as an XSLT instruction
+     *
+     * @param inst true if this is an xsl:variable instruction
+     */
+
+    public void setInstruction(boolean inst) {
+        isInstruction = inst;
+    }
+
+    /**
+     * Ask whether this expression is an instruction. In XSLT streamability analysis this
+     * is used to distinguish constructs corresponding to XSLT instructions from other constructs.
+     *
+     * @return true if this construct originates as an XSLT instruction
+     */
+
+    @Override
+    public boolean isInstruction() {
+        return isInstruction;
+    }
+
+
     /**
      * Get a name identifying the kind of expression, in terms meaningful to a user.
      *
@@ -59,58 +79,46 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return "let";
     }
 
+    /**
+     * Say that this variable must be evaluated eagerly. This is a requirement for example
+     * for variables appearing within a try/catch, or for variables referenced from within
+     * a construct that uses multi-threading, for example xsl:result-document or a multi-threaded
+     * xsl:for-each.
+     * @param req set to true if eager evaluation is required
+     */
+
     public void setNeedsEagerEvaluation(boolean req) {
+        if (req && needsLazyEvaluation) {
+            // No action - see bug #2903, Lazy evaluation wins
+        }
         this.needsEagerEvaluation = req;
     }
 
-//    private void recomputeRefCount() {
-//        if (refCount != FilterExpression.FILTERED) {
-//            setRefCount(countReferences(this, this, false));
-//        }
-//    }
-//
-//    private static int countReferences(LetExpression let, Expression child, boolean inLoop) {
-//        List<Integer> visits = new ArrayList<Integer>();
-//        visits.add(0);
-//        return countReferencesInternal(let, child, inLoop, visits);
-//    }
-//
-//    private static int countReferencesInternal(LetExpression let, Expression child, boolean inLoop, List<Integer> visits) {
-//        int total = 0;
-//        for (Operand o : child.operands()) {
-//            Expression g = o.getChildExpression();
-//            boolean repeated = inLoop || o.isEvaluatedRepeatedly();
-//            if (!repeated && !o.hasSameFocus() && ExpressionTool.dependsOnFocus(let.getSequence())) {
-//                repeated = true;
-//            }
-//            if (g instanceof LocalVariableReference && ((LocalVariableReference) g).getBinding() == let) {
-//                total += repeated ? 10 : 1;
-//            } else {
-//                int v = visits.get(0) + 1;
-//                if (v > 100) {
-//                    return 17; // enough; no point wasting more time on this
-//                }
-//                visits.set(0, v);
-//                total += countReferencesInternal(let, g, repeated, visits);
-//            }
-//            if (total >= 3) {
-//                break;
-//            }
-//        }
-//        return total;
-//    }
-//
-//    public void setRefCount(int refCount) {
-//        this.refCount = refCount;
-//    }
+    public void setNeedsLazyEvaluation(boolean req) {
+        if (req && needsEagerEvaluation) {
+            this.needsEagerEvaluation = false; // Bug 2903: lazy evaluation wins
+        }
+        this.needsLazyEvaluation = req;
+    }
+
+    public boolean isNeedsLazyEvaluation() {
+        return needsLazyEvaluation;
+    }
+
+    /**
+     * Ask whether the expression can be lifted out of a loop, assuming it has no dependencies
+     * on the controlling variable/focus of the loop
+     * @param forStreaming
+     */
+    @Override
+    public boolean isLiftable(boolean forStreaming) {
+        return super.isLiftable(forStreaming) && !needsEagerEvaluation;
+    }
 
     @Override
     public void resetLocalStaticProperties() {
         super.resetLocalStaticProperties();
-        references = null;
-//        if (refCount != FilterExpression.FILTERED) {
-//            setRefCount(-1);
-//        }
+        references = new ArrayList<VariableReference>(); // bug 3233
     }
 
     /**
@@ -138,12 +146,7 @@ public class LetExpression extends Assignation implements TailCallReturner {
                               getSequence() instanceof Literal ? ((Literal) getSequence()).getValue() : null,
                               getSequence().getSpecialProperties(), this);
 
-//        boolean indexed = refCount == FilterExpression.FILTERED;
-//        refCount = 0;
         getActionOp().typeCheck(visitor, contextInfo);
-//        if (indexed) {
-//            refCount = FilterExpression.FILTERED;
-//        }
         return this;
     }
 
@@ -178,7 +181,8 @@ public class LetExpression extends Assignation implements TailCallReturner {
                                       boolean backwardsCompatible,
                                       RoleDiagnostic role, ExpressionVisitor visitor)
             throws XPathException {
-        setAction(TypeChecker.staticTypeCheck(getAction(), req, backwardsCompatible, role, visitor));
+        TypeChecker tc = visitor.getConfiguration().getTypeChecker(backwardsCompatible);
+        setAction(tc.staticTypeCheck(getAction(), req, role, visitor));
         return this;
     }
 
@@ -214,10 +218,11 @@ public class LetExpression extends Assignation implements TailCallReturner {
             return getSequence();
         }
 
-        if (getSequence() instanceof Literal) {
+        if (getSequence() instanceof Literal && opt.isOptionSet(OptimizerOptions.INLINE_VARIABLES)) {
             // inline the variable: replace all references by the constant value
             // This relies on the fact that optimizing the action part will cause any references to be inlined
             opt.trace("Inlined constant variable " + getVariableName(), getSequence());
+            replaceVariable(getSequence());
             return getAction().optimize(visitor, contextItemType);
         }
 
@@ -276,19 +281,41 @@ public class LetExpression extends Assignation implements TailCallReturner {
         // can be evaluated in streaming mode, but an arbitrary expression using copy() inline can't (e.g.
         // if it appears in a path expression or as an operand of an arithmetic expression)
 
-        if (considerRemoval && references.size() == 1 && ExpressionTool.dependsOnFocus(getSequence())) {
+        if (considerRemoval && references.size() == 1 && ExpressionTool.dependsOnFocus(getSequence()) ) {
             if (visitor.isOptimizeForStreaming()) {
                 considerRemoval = false;
             }
+            // Disallow inlining if the focus of the variable reference differs from the containing binding.
+            Expression child = references.get(0);
+            Expression parent = child.getParentExpression();
+            while (parent != null && parent != this) {
+                if (!ExpressionTool.findOperand(parent, child).hasSameFocus()) {
+                    considerRemoval = false;
+                    break;
+                }
+                child = parent;
+                parent = child.getParentExpression();
+            }
+        }
+
+
+        if (considerRemoval && references.size() == 1) {
+            if (ExpressionTool.changesXsltContext(getSequence())) {
+                // Don't inline variables whose initializer might contain a call to xsl:result-document
+                considerRemoval = false;
+            } else if ((getSequence().getDependencies() & StaticProperty.DEPENDS_ON_CURRENT_GROUP) != 0) {
+                // Don't inline variables that depend on current-group() or current-grouping-key()
+                considerRemoval = false;
+            } else if (references.get(0).isInLoop()) {
+                considerRemoval = false;
+            }
         }
 
-        // Don't inline variables whose initializer might contain a call to xsl:result-document
-        if (considerRemoval && references.size() == 1 && ExpressionTool.changesXsltContext(getSequence())) {
-            considerRemoval = false;
-        }
-
-        if (considerRemoval && (references.size() == 1 || getSequence() instanceof Literal || getSequence() instanceof VariableReference)) {
+        if (considerRemoval && (references.size() == 1 || getSequence() instanceof Literal ||
+                                        getSequence() instanceof VariableReference) &&
+                opt.isOptionSet(OptimizerOptions.INLINE_VARIABLES)) {
             inlineReferences();
+            opt.trace("Inlined references to $" + getVariableName(), getAction());
             references = null;
             return getAction().optimize(visitor, contextItemType);
         }
@@ -334,7 +361,7 @@ public class LetExpression extends Assignation implements TailCallReturner {
             setEvaluationMode(ExpressionTool.eagerEvaluationMode(getSequence()));
         } else if (isIndexedVariable()) {
             setEvaluationMode(ExpressionTool.MAKE_CLOSURE);
-        } else {
+        } else if (evaluationMode == ExpressionTool.UNDECIDED) {
             setEvaluationMode(ExpressionTool.lazyEvaluationMode(getSequence()));
         }
         return this;
@@ -347,11 +374,9 @@ public class LetExpression extends Assignation implements TailCallReturner {
         for (VariableReference ref : references) {
             Expression parent = ref.getParentExpression();
             if (parent != null) {
-                for (Operand o : parent.operands()) {
-                    if (o.getChildExpression() == ref) {
-                        o.setChildExpression(getSequence().copy(new RebindingMap()));
-                        break;
-                    }
+                Operand o = ExpressionTool.findOperand(parent, ref);
+                if (o != null) {
+                    o.setChildExpression(getSequence().copy(new RebindingMap()));
                 }
                 ExpressionTool.resetStaticProperties(parent);
             }
@@ -371,7 +396,6 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return getSequence().getCost() + getAction().getCost();
     }
 
-
     /**
      * Determine whether all references to this variable are using the value either
      * (a) by atomizing it, or (b) by taking its string value. (This excludes usages
@@ -434,24 +458,6 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return getAction().getIntegerBounds();
     }
 
-    //#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public StreamingAdjunct getStreamingAdjunct() {
-        if (getAction().getPostureAndSweepIfKnown() != null && getAction().getSweep() == Sweep.CONSUMING) {
-            return new LetExpressionAdjunctB();
-        } else {
-            return new LetExpressionAdjunct();
-        }
-    }
-
- //#endif
-
     /**
      * An implementation of Expression must provide at least one of the methods evaluateItem(), iterate(), or process().
      * This method indicates which of these methods is provided directly. The other methods will always be available
@@ -487,25 +493,6 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return let.getAction().iterate(context);
     }
 
-    /**
-     * Iterate over the result of the expression to return a sequence of events
-     */
-
-    public EventIterator iterateEvents(XPathContext context) throws XPathException {
-        // minimize stack consumption by evaluating nested LET expressions iteratively
-        LetExpression let = this;
-        while (true) {
-            Sequence val = let.eval(context);
-            context.setLocalVariable(let.getLocalSlotNumber(), val);
-            if (let.getAction() instanceof LetExpression) {
-                let = (LetExpression) let.getAction();
-            } else {
-                break;
-            }
-        }
-        return let.getAction().iterateEvents(context);
-    }
-
 
     /**
      * Evaluate the variable.
@@ -518,11 +505,21 @@ public class LetExpression extends Assignation implements TailCallReturner {
         if (evaluationMode == ExpressionTool.UNDECIDED) {
             setEvaluationMode(ExpressionTool.lazyEvaluationMode(getSequence()));
         }
-        int savedOutputState = context.getTemporaryOutputState();
-        context.setTemporaryOutputState(StandardNames.XSL_VARIABLE);
-        Sequence result = ExpressionTool.evaluate(getSequence(), evaluationMode, context, getNominalReferenceCount());
-        context.setTemporaryOutputState(savedOutputState);
-        return result;
+        try {
+            int savedOutputState = context.getTemporaryOutputState();
+            context.setTemporaryOutputState(StandardNames.XSL_VARIABLE);
+            Sequence result = ExpressionTool.evaluate(getSequence(), evaluationMode, context, getNominalReferenceCount());
+            context.setTemporaryOutputState(savedOutputState);
+            return result;
+        } catch (ClassCastException e) {
+            // Probably the evaluation mode is wrong, as a result of an expression rewrite. Try again.
+            setEvaluationMode(ExpressionTool.lazyEvaluationMode(getSequence()));
+            int savedOutputState = context.getTemporaryOutputState();
+            context.setTemporaryOutputState(StandardNames.XSL_VARIABLE);
+            Sequence result = ExpressionTool.evaluate(getSequence(), evaluationMode, context, getNominalReferenceCount());
+            context.setTemporaryOutputState(savedOutputState);
+            return result;
+        }
     }
 
     /**
@@ -604,6 +601,21 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return getAction().getItemType();
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (isInstruction()) {
+            return UType.ANY;
+        }
+        return getAction().getStaticUType(contextItemType);
+    }
+
     /**
      * Determine the static cardinality of the expression
      */
@@ -635,53 +647,6 @@ public class LetExpression extends Assignation implements TailCallReturner {
         return ExpressionTool.markTailFunctionCalls(getAction(), qName, arity);
     }
 
-    /**
-     * Promote this expression if possible
-     */
-
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        Expression exp = offer.accept(this);
-        if (exp != null) {
-            return exp;
-        } else {
-            // pass the offer on to the sequence expression
-            Expression seq2 = doPromotion(getSequence(), offer);
-            if (seq2 != getSequence()) {
-                // if we've extracted a global variable, it may need to be marked indexable
-                if (seq2 instanceof VariableReference) {
-                    Binding b = ((VariableReference) seq2).getBinding();
-                    if (b instanceof GlobalVariable && isIndexedVariable) {
-                        ((GlobalVariable) b).setIndexedVariable();
-                    }
-                }
-                setSequence(seq2);
-            }
-            if (offer.action == PromotionOffer.EXTRACT_GLOBAL_VARIABLES) {
-                setAction(doPromotion(getAction(), offer));
-            } else if (offer.action == PromotionOffer.RANGE_INDEPENDENT ||
-                    offer.action == PromotionOffer.FOCUS_INDEPENDENT) {
-                // Pass the offer to the action expression after adding the variable bound by this let expression,
-                // so that a subexpression must depend on neither variable if it is to be promoted
-                Binding[] savedBindingList = offer.bindingList;
-                offer.bindingList = extendBindingList(offer.bindingList);
-                setAction(doPromotion(getAction(), offer));
-                offer.bindingList = savedBindingList;
-            }
-            // if this results in the expression (let $x := $y return Z), replace all references to
-            // to $x by references to $y in the Z part, and eliminate this LetExpression by
-            // returning the action part.
-            if (getSequence() instanceof VariableReference && !needsEagerEvaluation) {
-                rebuildReferenceList(true);
-                if (references != null) {
-                    inlineReferences();
-                    return getAction();
-                }
-            }
-            references = null;
-            return this;
-        }
-    }
-
 
     /**
      * Copy an expression. This makes a deep copy.
@@ -697,9 +662,11 @@ public class LetExpression extends Assignation implements TailCallReturner {
         let.isIndexedVariable = isIndexedVariable;
         let.hasLoopingReference = hasLoopingReference;
         let.setNeedsEagerEvaluation(needsEagerEvaluation);
+        let.setNeedsLazyEvaluation(needsLazyEvaluation);
         let.setVariableQName(variableName);
         let.setRequiredType(requiredType);
         let.setSequence(getSequence().copy(rebindings));
+        let.setInstruction(isInstruction());
         Expression newAction = getAction().copy(rebindings);
         let.setAction(newAction);
         ExpressionTool.rebindVariableReferences(newAction, this, let);
@@ -762,18 +729,6 @@ public class LetExpression extends Assignation implements TailCallReturner {
         let.getAction().evaluatePendingUpdates(context, pul);
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Let expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new LetExpressionCompiler();
-    }
-//#endif
 
     /**
      * The toString() method for an expression attempts to give a representation of the expression
@@ -806,8 +761,7 @@ public class LetExpression extends Assignation implements TailCallReturner {
     public void export(ExpressionPresenter out) throws XPathException {
         out.startElement("let", this);
         out.emitAttribute("var", variableName);
-        out.emitAttribute("as", getSequence().getItemType().toString() +
-                Cardinality.getOccurrenceIndicator(getSequence().getCardinality()));
+        out.emitAttribute("as", getSequence().getStaticType().toExportString());
         if (isIndexedVariable()) {
             out.emitAttribute("indexable", "true");
         }
