@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,10 +8,6 @@
 package net.sf.saxon.expr;
 
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.LiteralCompiler;
-import com.saxonica.ee.stream.adjunct.LiteralAdjunct;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import com.saxonica.functions.hof.FunctionLiteral;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.event.SequenceReceiver;
@@ -27,11 +23,9 @@ import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
-import net.sf.saxon.value.StringValue;
 
 import java.util.ArrayList;
 import java.util.List;
-import java.util.Map;
 
 
 /**
@@ -47,7 +41,7 @@ public class Literal extends Expression {
     /**
      * Create a literal as a wrapper around a Value
      *
-     * @param value     the value of this literal
+     * @param value the value of this literal
      */
 
     protected Literal(GroundedValue value) {
@@ -109,8 +103,9 @@ public class Literal extends Expression {
      * example, to reorder the predicates in a filter expression so cheaper predicates are
      * evaluated first.
      */
+
     @Override
-    public int getCost() {
+    public int getNetCost() {
         return 0;
     }
 
@@ -133,6 +128,28 @@ public class Literal extends Expression {
         }
     }
 
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @param contextItemType the static context item type
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (value.getLength() == 0) {
+            return UType.VOID;
+        } else if (value instanceof AtomicValue) {
+            return ((AtomicValue) value).getUType();
+        } else if (value instanceof Function) {
+            return UType.FUNCTION;
+        } else {
+            return super.getStaticUType(contextItemType);
+        }
+    }
+
+
     /**
      * Determine the cardinality
      */
@@ -188,7 +205,7 @@ public class Literal extends Expression {
      * are used to indicate values limited by the size of a string or the size of a sequence.
      *
      * @return the lower and upper bounds of integer values in the result, or null to indicate
-     *         unknown or not applicable.
+     * unknown or not applicable.
      */
     /*@Nullable*/
     @Override
@@ -214,33 +231,13 @@ public class Literal extends Expression {
         return value.getLength() == 0;
     }
 
-    /**
-     * Offer promotion for this subexpression. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is accepted. This method is always called at compile time.
-     *
-     *
-     * @param offer  details of the offer, for example the offer to move
-     *               expressions that don't depend on the context to an outer level in
-     *               the containing expression
-     * @return if the offer is not accepted, return this expression unchanged.
-     *         Otherwise return the result of rewriting the expression to promote
-     *         this subexpression
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any error is detected
-     */
-    @Override
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        return this;
-    }
-
     /**
      * Copy an expression. This makes a deep copy.
      *
-     * @return a copy of the original literal. Note that the
-     *         underlying value is not copied; the code relies on the caller
-     *         treating the underlying value as immutable.
      * @param rebindings
+     * @return a copy of the original literal. Note that the
+     * underlying value is not copied; the code relies on the caller
+     * treating the underlying value as immutable.
      */
 
     /*@NotNull*/
@@ -256,8 +253,7 @@ public class Literal extends Expression {
      * @param config the Saxon configuration
      * @param is30   true if this is XSLT 3.0
      * @return the equivalent pattern
-     * @throws net.sf.saxon.trans.XPathException
-     *          if conversion is not possible
+     * @throws net.sf.saxon.trans.XPathException if conversion is not possible
      */
     @Override
     public Pattern toPattern(Configuration config, boolean is30) throws XPathException {
@@ -275,7 +271,7 @@ public class Literal extends Expression {
      * @param pathMap        the PathMap to which the expression should be added
      * @param pathMapNodeSet the set of nodes within the path map
      * @return the pathMapNode representing the focus established by this expression, in the case where this
-     *         expression is the first operand of a path expression or filter expression
+     * expression is the first operand of a path expression or filter expression
      */
 
     public PathMap.PathMapNodeSet addToPathMap(PathMap pathMap, PathMap.PathMapNodeSet pathMapNodeSet) {
@@ -304,10 +300,9 @@ public class Literal extends Expression {
      *
      * @param context supplies the context for evaluation
      * @return a SequenceIterator that can be used to iterate over the result
-     *         of the expression
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any dynamic error occurs evaluating the
-     *          expression
+     * of the expression
+     * @throws net.sf.saxon.trans.XPathException if any dynamic error occurs evaluating the
+     *                                           expression
      */
 
     /*@NotNull*/
@@ -323,10 +318,9 @@ public class Literal extends Expression {
      * provide its own implementation.
      *
      * @return a SequenceIterator that can be used to iterate over the result
-     *         of the expression
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any dynamic error occurs evaluating the
-     *          expression
+     * of the expression
+     * @throws net.sf.saxon.trans.XPathException if any dynamic error occurs evaluating the
+     *                                           expression
      */
 
     /*@NotNull*/
@@ -355,7 +349,7 @@ public class Literal extends Expression {
     public void process(XPathContext context) throws XPathException {
         SequenceReceiver out = context.getReceiver();
         if (value instanceof Item) {
-            out.append((Item)value);
+            out.append((Item) value, getLocation(), NodeInfo.ALL_NAMESPACES);
         } else {
             SequenceIterator iter = value.iterate();
             Item it;
@@ -403,9 +397,8 @@ public class Literal extends Expression {
      *
      * @param context The context in which the expression is to be evaluated
      * @return the effective boolean value
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any dynamic error occurs evaluating the
-     *          expression
+     * @throws net.sf.saxon.trans.XPathException if any dynamic error occurs evaluating the
+     *                                           expression
      */
 
     public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
@@ -436,9 +429,9 @@ public class Literal extends Expression {
      *
      * @param obj the other expression
      * @return true if the two literals are equal. The test here requires (a) identity in the
-     *         sense defined by XML Schema (same value in the same value space), and (b) identical type
-     *         annotations. For example the literal xs:int(3) is not equal (as an expression) to xs:short(3),
-     *         because the two expressions are not interchangeable.
+     * sense defined by XML Schema (same value in the same value space), and (b) identical type
+     * annotations. For example the literal xs:int(3) is not equal (as an expression) to xs:short(3),
+     * because the two expressions are not interchangeable.
      */
 
     public boolean equals(Object obj) {
@@ -459,20 +452,36 @@ public class Literal extends Expression {
                 if (m0 == null || m1 == null) {
                     return false;
                 }
+                if (m0 == m1) {
+                    continue;
+                }
                 boolean n0 = m0 instanceof NodeInfo;
                 boolean n1 = m1 instanceof NodeInfo;
                 if (n0 != n1) {
                     return false;
                 }
-                if (n0 && n1 && !((NodeInfo) m0).isSameNodeInfo((NodeInfo) m1)) {
-                    return false;
-                }
-                if (m0 instanceof AtomicValue && m1 instanceof AtomicValue) {
-                    if (!((AtomicValue) m0).isIdentical((AtomicValue) m1) ||
-                            ((AtomicValue) m0).getItemType() != ((AtomicValue) m1).getItemType()) {
+                if (n0) {
+                    if (((NodeInfo) m0).isSameNodeInfo((NodeInfo) m1)) {
+                        continue;
+                    } else {
                         return false;
                     }
                 }
+                boolean a0 = m0 instanceof AtomicValue;
+                boolean a1 = m1 instanceof AtomicValue;
+                if (a0 != a1) {
+                    return false;
+                }
+                if (a0) {
+                    if (((AtomicValue) m0).isIdentical((AtomicValue) m1) &&
+                            ((AtomicValue) m0).getItemType() == ((AtomicValue) m1).getItemType()) {
+                        continue;
+                    } else {
+                        return false;
+                    }
+                }
+                // don't attempt to compare functions, maps, and arrays
+                return false;
             }
         } catch (XPathException err) {
             return false;
@@ -510,58 +519,69 @@ public class Literal extends Expression {
         exportValue(value, out);
     }
 
-    public static void exportValue(Sequence value, ExpressionPresenter out) {
-        try {
-            if (value.head() == null) {
-                out.startElement("empty");
-                out.endElement();
-            } else if (value instanceof AtomicValue) {
-                exportAtomicValue((AtomicValue)value, out);
-            } else if (value instanceof IntegerRange) {
-                out.startElement("range");
-                out.emitAttribute("from", "" + ((IntegerRange) value).getStart());
-                out.emitAttribute("to", "" + ((IntegerRange) value).getEnd());
-                out.endElement();
-            } else if (value instanceof NodeInfo) {
-                out.startElement("node");
-                out.emitAttribute("kind", ((NodeInfo) value).getNodeKind() + "");
-                out.emitAttribute("baseUri", ((NodeInfo) value).getBaseURI());
-                String ser = QueryResult.serialize((NodeInfo) value);
-                out.emitAttribute("content", ser);
-                out.endElement();
-            } else if (value instanceof MapItem) {
-                out.startElement("map");
-                out.emitAttribute("size", "" + ((MapItem) value).size());
-                for (KeyValuePair kvp : (MapItem)value) {
-                    exportAtomicValue(kvp.key, out);
-                    exportValue(kvp.value, out);
-                }
-                out.endElement();
-            } else if (value instanceof Function) {
-                ((Function) value).export(out);
-            } else if (value instanceof ObjectValue) {
-                out.startElement("externalObject");
-                out.emitAttribute("class", ((ObjectValue) value).getObject().getClass().getName());
-                out.endElement();
-            } else {
-                out.startElement("literal");
-                if (value instanceof GroundedValue) {
-                    out.emitAttribute("count", ((GroundedValue) value).getLength() + "");
-                }
-                SequenceIterator iter = value.iterate();
-                Item it;
-                while ((it = iter.next()) != null) {
-                    exportValue(it, out);
-                }
-                out.endElement();
-
+    public static void exportValue(Sequence value, ExpressionPresenter out) throws XPathException {
+        if (value.head() == null) {
+            out.startElement("empty");
+            out.endElement();
+        } else if (value instanceof AtomicValue) {
+            exportAtomicValue((AtomicValue) value, out);
+        } else if (value instanceof IntegerRange) {
+            out.startElement("range");
+            out.emitAttribute("from", "" + ((IntegerRange) value).getStart());
+            out.emitAttribute("to", "" + ((IntegerRange) value).getEnd());
+            out.endElement();
+        } else if (value instanceof NodeInfo) {
+            out.startElement("node");
+            out.emitAttribute("kind", ((NodeInfo) value).getNodeKind() + "");
+            out.emitAttribute("baseUri", ((NodeInfo) value).getBaseURI());
+            String ser = QueryResult.serialize((NodeInfo) value);
+            out.emitAttribute("content", ser);
+            out.endElement();
+        } else if (value instanceof MapItem) {
+            out.startElement("map");
+            out.emitAttribute("size", "" + ((MapItem) value).size());
+            for (KeyValuePair kvp : (MapItem) value) {
+                exportAtomicValue(kvp.key, out);
+                exportValue(kvp.value, out);
             }
-        } catch (XPathException e) {
-            throw new AssertionError();
+            out.endElement();
+        } else if (value instanceof Function) {
+            ((Function) value).export(out);
+        } else if (value instanceof ObjectValue) {
+            out.startElement("externalObject");
+            out.emitAttribute("class", ((ObjectValue) value).getObject().getClass().getName());
+            out.endElement();
+        } else {
+            out.startElement("literal");
+            if (value instanceof GroundedValue) {
+                out.emitAttribute("count", ((GroundedValue) value).getLength() + "");
+            }
+            SequenceIterator iter = value.iterate();
+            Item it;
+            while ((it = iter.next()) != null) {
+                exportValue(it, out);
+            }
+            out.endElement();
+
         }
     }
 
-    public static void exportAtomicValue(AtomicValue value, ExpressionPresenter out) {
+    /**
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
+     *
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in export() output displaying the expression.
+     */
+    @Override
+    public String getExpressionName() {
+        return "literal";
+    }
+
+    public static void exportAtomicValue(AtomicValue value, ExpressionPresenter out) throws XPathException {
+        if ("JS".equals(out.getOption("target"))) {
+            value.checkValidInJavascript();
+        }
         AtomicType type = value.getItemType();
         String val = value.getStringValue();
         if (type.equals(BuiltInAtomicType.STRING)) {
@@ -581,11 +601,11 @@ public class Literal extends Expression {
             out.emitAttribute("val", val);
             out.endElement();
         } else if (type.equals(BuiltInAtomicType.BOOLEAN)) {
-            out.startElement(((BooleanValue)value).effectiveBooleanValue() ? "true" : "false");
+            out.startElement(((BooleanValue) value).effectiveBooleanValue() ? "true" : "false");
             out.endElement();
         } else if (value instanceof QualifiedNameValue) {
             out.startElement("qName");
-            out.emitAttribute("pre", ((QualifiedNameValue)value).getPrefix());
+            out.emitAttribute("pre", ((QualifiedNameValue) value).getPrefix());
             out.emitAttribute("uri", ((QualifiedNameValue) value).getNamespaceURI());
             out.emitAttribute("loc", ((QualifiedNameValue) value).getLocalName());
             if (!type.equals(BuiltInAtomicType.QNAME)) {
@@ -647,7 +667,7 @@ public class Literal extends Expression {
      * @param exp   an expression
      * @param value true or false
      * @return true if the expression is a literal and the literal represents the boolean value given in the
-     *         second argument
+     * second argument
      */
 
     public static boolean isConstantBoolean(Expression exp, boolean value) {
@@ -658,6 +678,26 @@ public class Literal extends Expression {
         return false;
     }
 
+    /**
+     * Test if a literal has the effective boolean value true (or false)
+     *
+     * @param exp   an expression
+     * @param value true or false
+     * @return true if the expression is a literal and the literal represents the boolean value given in the
+     * second argument
+     */
+
+    public static boolean hasEffectiveBooleanValue(Expression exp, boolean value) {
+        if (exp instanceof Literal) {
+            try {
+                return value == ((Literal) exp).getValue().effectiveBooleanValue();
+            } catch (XPathException err) {
+                return false;
+            }
+        }
+        return false;
+    }
+
     /**
      * Test if a literal represents the integer value 1
      *
@@ -678,8 +718,8 @@ public class Literal extends Expression {
      * document outside the subtree rooted at the context node.
      *
      * @return true if the expression has no dependencies on the context node, or if the only dependencies
-     *         on the context node are downward selections using the self, child, descendant, attribute, and namespace
-     *         axes.
+     * on the context node are downward selections using the self, child, descendant, attribute, and namespace
+     * axes.
      */
 
     public boolean isSubtreeExpression() {
@@ -696,33 +736,11 @@ public class Literal extends Expression {
         return new Literal(EmptySequence.getInstance());
     }
 
-//    /**
-//     * Factory method to create a literal as a wrapper around a Value (factory method)
-//     *
-//     * @param container the container in the expression tree. Should not be null.
-//     *                  Saxon does not reject a null value for the container, but it may cause
-//     *                  subsequent failures if a container is not supplied at some stage.
-//     * @param value     the value of this literal
-//     * @return the Literal
-//     */
-//
-//    public static Literal makeLiteral(GroundedValue value) {
-//        value = value.reduce();
-//        if (value instanceof StringValue) {
-//            return new StringLiteral((StringValue) value, container);
-////#ifdefined HOF
-//        } else if (value instanceof Function && !(value instanceof MapItem || value instanceof ArrayItem)) {
-//            return new FunctionLiteral((Function) value, container);
-////#endif
-//        } else {
-//            return new Literal(value, container);
-//        }
-//    }
 
     /**
      * Factory method to create a literal as a wrapper around a Value (factory method)
      *
-     * @param value     the value of this literal
+     * @param value the value of this literal
      * @return the Literal
      */
 
@@ -739,26 +757,14 @@ public class Literal extends Expression {
         }
     }
 
-//#ifdefined STREAM
-
-    @Override
-    protected StreamingAdjunct getStreamingAdjunct() {
-        return new LiteralAdjunct();
-    }
-
-    //#endif
-
-//#ifdefined BYTECODE
-
     /**
-     * Return the compiler of the Literal expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new LiteralCompiler();
+    public String getStreamerName() {
+        return "Literal";
     }
-//#endif
-
 }
\ No newline at end of file
