@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,15 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.util.CannotCompileException;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.Configuration;
+import net.sf.saxon.expr.instruct.AnalyzeString;
 import net.sf.saxon.expr.oper.OperandArray;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.functions.*;
 import net.sf.saxon.functions.Error;
-import net.sf.saxon.om.LazySequence;
+import net.sf.saxon.functions.registry.BuiltInFunctionSet;
+import net.sf.saxon.lib.NamespaceConstant;
+import net.sf.saxon.ma.map.MapFunctionSet;
 import net.sf.saxon.om.Sequence;
 import net.sf.saxon.om.SequenceTool;
 import net.sf.saxon.pattern.NodeSetPattern;
@@ -27,7 +27,7 @@ import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.TypeHierarchy;
 import net.sf.saxon.value.IntegerValue;
 
-import java.util.Map;
+import java.util.Arrays;
 
 /**
  * A call to a system-defined function (specifically, a function implemented as an instance
@@ -35,8 +35,12 @@ import java.util.Map;
  */
 public class SystemFunctionCall extends StaticFunctionCall implements Negatable {
 
+    public Evaluator[] argumentEvaluators;
+
     public SystemFunctionCall(SystemFunction target, Expression[] arguments) {
         super(target, arguments);
+        argumentEvaluators = new Evaluator[arguments.length];
+        Arrays.fill(argumentEvaluators, Evaluator.LAZY_SEQUENCE);
     }
 
     /**
@@ -63,7 +67,7 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
     @Override
     public Expression preEvaluate(ExpressionVisitor visitor) throws XPathException {
         SystemFunction target = getTargetFunction();
-        if ((target.getDetails().properties & StandardFunction.LATE) == 0) {
+        if ((target.getDetails().properties & BuiltInFunctionSet.LATE) == 0) {
             return super.preEvaluate(visitor);
         } else {
             // Early evaluation of this function is suppressed
@@ -85,15 +89,36 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         checkFunctionCall(getTargetFunction(), visitor);
         // Give the function an opportunity to use the type information now available
         getTargetFunction().supplyTypeInformation(visitor, contextInfo, getArguments());
-        if ((getTargetFunction().getDetails().properties & StandardFunction.LATE) == 0) {
+        if ((getTargetFunction().getDetails().properties & BuiltInFunctionSet.LATE) == 0) {
             return preEvaluateIfConstant(visitor);
         }
+        allocateArgumentEvaluators(getArguments());
         return this;
     }
 
+    private void allocateArgumentEvaluators(Expression[] arguments) {
+        for (int i = 0; i < arguments.length; i++) {
+            Expression arg = arguments[i];
+            int cardinality = isCallOn(Concat.class) ?
+                    StaticProperty.ALLOWS_ZERO_OR_ONE :
+                    getTargetFunction().getDetails().argumentTypes[i].getCardinality();
+            if (arg instanceof Literal) {
+                argumentEvaluators[i] = Evaluator.LITERAL;
+            } else if (arg instanceof VariableReference) {
+                argumentEvaluators[i] = Evaluator.VARIABLE;
+            } else if (cardinality == StaticProperty.EXACTLY_ONE) {
+                argumentEvaluators[i] = Evaluator.SINGLE_ITEM;
+            } else if (cardinality == StaticProperty.ALLOWS_ZERO_OR_ONE) {
+                argumentEvaluators[i] = Evaluator.OPTIONAL_ITEM;
+            } else {
+                argumentEvaluators[i] = Evaluator.LAZY_SEQUENCE;
+            }
+        }
+    }
+
     @Override
     public SystemFunction getTargetFunction() {
-        return (SystemFunction)super.getTargetFunction();
+        return (SystemFunction) super.getTargetFunction();
     }
 
     /**
@@ -108,18 +133,25 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
     @Override
     public int getIntrinsicDependencies() {
         int properties = getTargetFunction().getDetails().properties;
-        if ((properties & StandardFunction.FOCUS) != 0) {
+        if ((properties & BuiltInFunctionSet.FOCUS) != 0) {
             int dep = 0;
-            if ((properties & StandardFunction.CITEM) != 0) {
+            if ((properties & BuiltInFunctionSet.CDOC) != 0) {
+                dep |= StaticProperty.DEPENDS_ON_CONTEXT_DOCUMENT;
+            }
+            if ((properties & BuiltInFunctionSet.CITEM) != 0) {
                 dep |= StaticProperty.DEPENDS_ON_CONTEXT_ITEM;
             }
-            if ((properties & StandardFunction.POSN) != 0) {
+            if ((properties & BuiltInFunctionSet.POSN) != 0) {
                 dep |= StaticProperty.DEPENDS_ON_POSITION;
             }
-            if ((properties & StandardFunction.LAST) != 0) {
+            if ((properties & BuiltInFunctionSet.LAST) != 0) {
                 dep |= StaticProperty.DEPENDS_ON_LAST;
             }
             return dep;
+        } else if ((properties & BuiltInFunctionSet.BASE) != 0) {
+            return StaticProperty.DEPENDS_ON_STATIC_CONTEXT;
+        } else if ((properties & BuiltInFunctionSet.DCOLL) != 0) {
+            return StaticProperty.DEPENDS_ON_STATIC_CONTEXT;
         } else if (isCallOn(RegexGroup.class)) {
             return StaticProperty.DEPENDS_ON_CURRENT_GROUP;
         } else if (isCallOnSystemFunction("current-merge-group")) {
@@ -156,6 +188,48 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         return getTargetFunction().getSpecialProperties(getArguments());
     }
 
+    /**
+     * Return the estimated cost of evaluating an expression. This is a very crude measure based
+     * on the syntactic form of the expression (we have no knowledge of data values). We take
+     * the cost of evaluating a simple scalar comparison or arithmetic expression as 1 (one),
+     * and we assume that a sequence has length 5. The resulting estimates may be used, for
+     * example, to reorder the predicates in a filter expression so cheaper predicates are
+     * evaluated first.
+     */
+    @Override
+    public int getNetCost() {
+        return getTargetFunction().getNetCost();
+    }
+
+    @Override
+    public Expression getScopingExpression() {
+        if (isCallOn(RegexGroup.class)) {
+            Expression parent = getParentExpression();
+            while (parent != null) {
+                if (parent instanceof AnalyzeString) {
+                    return parent;
+                }
+                parent = parent.getParentExpression();
+            }
+            return null;
+        } else {
+            return super.getScopingExpression();
+        }
+    }
+
+    /**
+     * Ask whether the expression can be lifted out of a loop, assuming it has no dependencies
+     * on the controlling variable/focus of the loop
+     *
+     * @param forStreaming true if the expression is being optimized for streaming
+     */
+    @Override
+    public boolean isLiftable(boolean forStreaming) {
+        return super.isLiftable(forStreaming) &&
+                !isCallOn(CurrentMergeGroup.class) && !isCallOn(CurrentMergeKey.class) &&
+                (!forStreaming || !isCallOn(MapFunctionSet.MapEntry.class));
+    }
+
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         Expression sf = super.optimize(visitor, contextInfo);
         if (sf == this) {
@@ -165,13 +239,16 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
             if (sfo != null) {
                 sfo.setParentExpression(getParentExpression());
                 ExpressionTool.copyLocationInfo(this, sfo);
+                if (sfo instanceof SystemFunctionCall) {
+                    ((SystemFunctionCall) sfo).allocateArgumentEvaluators(((SystemFunctionCall) sfo).getArguments());
+                }
                 return sfo;
             }
         }
         if (sf instanceof SystemFunctionCall) {
             // If any arguments are known to be empty, pre-evaluate the result
-            StandardFunction.Entry details = ((SystemFunctionCall) sf).getTargetFunction().getDetails();
-            if ((details.properties & StandardFunction.UO) != 0) {
+            BuiltInFunctionSet.Entry details = ((SystemFunctionCall) sf).getTargetFunction().getDetails();
+            if ((details.properties & BuiltInFunctionSet.UO) != 0) {
                 // First argument does not need to be in any particular order
                 setArg(0, getArg(0).unordered(true, visitor.isOptimizeForStreaming()));
             }
@@ -183,6 +260,7 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
                     }
                 }
             }
+            ((SystemFunctionCall) sf).allocateArgumentEvaluators(((SystemFunctionCall) sf).getArguments());
         }
         return sf;
     }
@@ -212,8 +290,8 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
     /**
      * Copy an expression. This makes a deep copy.
      *
-     * @return the copy of the original expression
      * @param rebindings
+     * @return the copy of the original expression
      */
     @Override
     public Expression copy(RebindingMap rebindings) {
@@ -221,7 +299,11 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         for (int i = 0; i < args.length; i++) {
             args[i] = getArg(i).copy(rebindings);
         }
-        return new SystemFunctionCall(getTargetFunction(), args);
+        SystemFunction target = getTargetFunction();
+        if (target instanceof StatefulSystemFunction) {
+            target = ((StatefulSystemFunction)target).copy();
+        }
+        return target.makeFunctionCall(args);
     }
 
     /**
@@ -240,7 +322,7 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
     //@Override
     public IntegerValue[] getIntegerBounds() {
         SystemFunction fn = getTargetFunction();
-        if ((fn.getDetails().properties & StandardFunction.FILTER) != 0) {
+        if ((fn.getDetails().properties & BuiltInFunctionSet.FILTER) != 0) {
             return getArg(0).getIntegerBounds();
         }
         return fn.getIntegerBounds();
@@ -328,6 +410,8 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         if (isCallOn(Doc.class) || isCallOn(DocumentFn.class) || isCallOn(CollectionFn.class)) {
             getArg(0).addToPathMap(pathMap, pathMapNodeSet);
             return new PathMap.PathMapNodeSet(pathMap.makeNewRoot(this));
+        } else if (isCallOn(KeyFn.class)) {
+            return ((KeyFn) getTargetFunction()).addToPathMap(pathMap, pathMapNodeSet);
         } else {
             return super.addToPathMap(pathMap, pathMapNodeSet);
         }
@@ -346,9 +430,9 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         SystemFunction fn = getTargetFunction();
         if (fn instanceof Root_1) {
             if (is30 &&
-                (getArg(0) instanceof ContextItemExpression ||
-                    (getArg(0) instanceof ItemChecker &&
-                        ((ItemChecker) getArg(0)).getBaseExpression() instanceof ContextItemExpression))) {
+                    (getArg(0) instanceof ContextItemExpression ||
+                             (getArg(0) instanceof ItemChecker &&
+                                      ((ItemChecker) getArg(0)).getBaseExpression() instanceof ContextItemExpression))) {
                 return new NodeSetPattern(this);
             }
         }
@@ -360,11 +444,32 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
         int numArgs = operanda.getNumberOfOperands();
         Sequence[] actualArgs = new Sequence[numArgs];
         for (int i = 0; i < numArgs; i++) {
-            actualArgs[i] = new LazySequence(operanda.getOperandExpression(i).iterate(context));
+            Expression exp = operanda.getOperandExpression(i);
+            actualArgs[i] = argumentEvaluators[i].evaluate(exp, context);
         }
         return actualArgs;
     }
 
+    @Override
+    public void resetLocalStaticProperties() {
+        super.resetLocalStaticProperties();
+        if (argumentEvaluators != null) {
+            allocateArgumentEvaluators(getArguments());
+        } 
+    }
+
+    /**
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
+     *
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in explain() output displaying the expression.
+     */
+    @Override
+    public String getExpressionName() {
+        return "sysFuncCall";
+    }
+
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -373,55 +478,44 @@ public class SystemFunctionCall extends StaticFunctionCall implements Negatable
      */
     @Override
     public void export(ExpressionPresenter out) throws XPathException {
-        out.startElement("fn", this);
-        out.emitAttribute("name", getFunctionName().getDisplayName());
-        getTargetFunction().exportAttributes(out);
-        for (Operand o : operands()) {
-            o.getChildExpression().export(out);
+        if (getFunctionName().hasURI(NamespaceConstant.FN)) {
+            out.startElement("fn", this);
+            out.emitAttribute("name", getFunctionName().getLocalPart());
+            getTargetFunction().exportAttributes(out);
+            for (Operand o : operands()) {
+                o.getChildExpression().export(out);
+            }
+            getTargetFunction().exportAdditionalArguments(this, out);
+            out.endElement();
+        } else {
+            // Function was implemented as an IntegratedFunctionCall in 9.7 and we retain the same export format
+            out.startElement("ifCall", this);
+            out.emitAttribute("name", getFunctionName().getEQName());
+            out.emitAttribute("type", getTargetFunction().getFunctionItemType().getResultType().toExportString());
+            getTargetFunction().exportAttributes(out);
+            for (Operand o : operands()) {
+                o.getChildExpression().export(out);
+            }
+            getTargetFunction().exportAdditionalArguments(this, out);
+            out.endElement();
         }
-        out.endElement();
     }
 
-    //#ifdefined BYTECODE
-
     /**
-     * Return the bytecode compiler for this system function call.
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() throws CannotCompileException {
-        ExpressionCompiler compiler = getTargetFunction().getExpressionCompiler();
-        if (compiler == null) {
-            if (getTargetFunction() instanceof ContextItemAccessorFunction) {
-                Expression exp = ((ContextItemAccessorFunction)getTargetFunction()).makeContextItemExplicit();
-                return exp.getExpressionCompiler();
-            }
-            return super.getExpressionCompiler();
-        }
-        return compiler;
-    }
-    //#endif
-
-    //#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * Subclass representing a system function call that has been optimized; this overrides the
+     * optimize() method to do nothing, thus ensuring that optimization converges.
      */
 
-    public StreamingAdjunct getStreamingAdjunct() {
-        StreamingAdjunct sa = getTargetFunction().getStreamingAdjunct();
-        if (sa == null) {
-            if (getTargetFunction() instanceof ContextItemAccessorFunction) {
-                Expression exp = ((ContextItemAccessorFunction) getTargetFunction()).makeContextItemExplicit();
-                return exp.getStreamingAdjunct();
-            }
-            return new StreamingAdjunct();
+    public abstract static class Optimized extends SystemFunctionCall {
+
+        public Optimized(SystemFunction target, Expression[] arguments) {
+            super(target, arguments);
+        }
+        @Override
+        public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
+            return this; // prevent infinite optimization
         }
-        return sa;
     }
-//#endif
+
 
 }
