@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,13 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.UserFunctionCallCompiler;
-import com.saxonica.ee.stream.adjunct.UserFunctionCallAdjunct;
-import net.sf.saxon.evpull.EmptyEventIterator;
-import net.sf.saxon.evpull.EventIterator;
 import net.sf.saxon.expr.instruct.Block;
 import net.sf.saxon.expr.instruct.UserFunction;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.om.*;
+import net.sf.saxon.om.Item;
+import net.sf.saxon.om.Sequence;
+import net.sf.saxon.om.SequenceIterator;
+import net.sf.saxon.om.StructuredQName;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trace.LocationKind;
 import net.sf.saxon.trans.SaxonErrorCode;
@@ -25,6 +23,7 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.AnyItemType;
 import net.sf.saxon.type.ItemType;
+import net.sf.saxon.type.UType;
 import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.EmptySequence;
 import net.sf.saxon.value.SequenceType;
@@ -46,6 +45,16 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
     private int tailCall = NOT_TAIL_CALL;
     private StructuredQName name;
 
+    public boolean isBeingInlined() {
+        return beingInlined;
+    }
+
+    public void setBeingInlined(boolean beingInlined) {
+        this.beingInlined = beingInlined;
+    }
+
+    private boolean beingInlined = false;
+
 
     /*@Nullable*/ private int[] argumentEvaluationModes = null;
 
@@ -72,7 +81,6 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
     }
 
 
-
     /**
      * Set the static type
      *
@@ -108,18 +116,16 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
     }
 
     /**
-      * Get the binding slot to be used. This is the offset within the binding vector of the containing
-      * component where the actual target template is to be found.
-      *
-      * @return the offset in the binding vector of the containing package where the target template
-      *         can be found.
-      */
-
-     public int getBindingSlot() {
-         return bindingSlot;
-     }
-
+     * Get the binding slot to be used. This is the offset within the binding vector of the containing
+     * component where the actual target template is to be found.
+     *
+     * @return the offset in the binding vector of the containing package where the target template
+     * can be found.
+     */
 
+    public int getBindingSlot() {
+        return bindingSlot;
+    }
 
 
     /**
@@ -171,7 +177,7 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
     }
 
     public SymbolicName getSymbolicName() {
-        return new SymbolicName(StandardNames.XSL_FUNCTION, getFunctionName(), getArity());
+        return new SymbolicName.F(getFunctionName(), getArity());
     }
 
     public Component getTarget() {
@@ -190,6 +196,7 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
 
     /**
      * Set the argument evaluation modes
+     *
      * @param evalModes the argument evaluation modes to be used
      */
 
@@ -224,6 +231,23 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         }
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @param contextItemType the static type of the context item
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        UserFunction f = getFunction();
+        if (f == null) {
+            // Happens when called during parsing
+            return UType.ANY;
+        }
+        return f.getResultType().getPrimaryType().getUType();
+    }
+
     public int getIntrinsicDependencies() {
         return StaticProperty.DEPENDS_ON_USER_FUNCTIONS;
     }
@@ -251,14 +275,22 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         // Inherit the properties of the function being called if possible. But we have to prevent
         // looping when the function is recursive. For safety, we only consider the properties of the
         // function body if it contains no further function calls.
-        if (function == null || function.getBody() == null) {
+        // TODO: this approach is not safe if the function can be overridden in another package
+        if (function == null) {
             return super.computeSpecialProperties();
-        } else {
+//        } else if (function.getDeclaredStreamability() != FunctionStreamability.UNCLASSIFIED) {
+//            // TODO: this is a first approximation. Need to consider grounded nodes in the result...
+//            int p = StaticProperty.ORDERED_NODESET;
+//            if (function.getDeclaredStreamability() == FunctionStreamability.SHALLOW_DESCENT) {
+//                p |= StaticProperty.PEER_NODESET;
+//            }
+//            return p;
+        } else if (function.getBody() != null) {
             int props;
             List<UserFunction> calledFunctions = new ArrayList<UserFunction>();
             ExpressionTool.gatherCalledFunctions(function.getBody(), calledFunctions);
             if (calledFunctions.isEmpty()) {
-                props = function.getBody().computeSpecialProperties();
+                props = function.getBody().getSpecialProperties();
             } else {
                 props = super.computeSpecialProperties();
             }
@@ -266,14 +298,16 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
                 props |= StaticProperty.NON_CREATIVE;
             }
             return props;
+        } else {
+            return super.computeSpecialProperties();
         }
     }
 
     /**
      * Copy an expression. This makes a deep copy.
      *
-     * @return the copy of the original expression
      * @param rebindings
+     * @return the copy of the original expression
      */
 
     /*@NotNull*/
@@ -291,11 +325,6 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
             a2[i] = getArg(i).copy(rebindings);
         }
         ufc.setArguments(a2);
-        if (argumentEvaluationModes != null) {
-            int[] am2 = new int[argumentEvaluationModes.length];
-            System.arraycopy(argumentEvaluationModes, 0, am2, 0, am2.length);
-            ufc.argumentEvaluationModes = am2;
-        }
         ExpressionTool.copyLocationInfo(this, ufc);
         return ufc;
     }
@@ -321,17 +350,7 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         }
         if (function != null) {
             checkFunctionCall(function, visitor);
-            if (function.getFunctionName() == null) {
-                // This is an inline function item; add type-checking for the argument types and the result type
-                throw new IllegalStateException(); //bug 2879
-//                Expression body = function.getBody();
-//                body = body.typeCheck(visitor, ContextItemStaticInfo.ABSENT);
-//                RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.FUNCTION_RESULT, "", 0);
-//                body = TypeChecker.staticTypeCheck(body, function.getResultType(), visitor.getStaticContext().isInBackwardsCompatibleMode(), role, visitor);
-//                function.setBody(body);
-            }
-            computeArgumentEvaluationModes();
-            if (staticType == SequenceType.ANY_SEQUENCE) {
+            if (staticType == null || staticType == SequenceType.ANY_SEQUENCE) {
                 // try to get a better type
                 staticType = function.getResultType();
             }
@@ -341,10 +360,10 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
 
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType) throws XPathException {
+        //function.setOnFirstCall(new DeferredAction(visitor, contextItemType));
         Expression e = super.optimize(visitor, contextItemType);
         if (e == this && function != null) {
-            computeArgumentEvaluationModes();
-            Expression e2 = getConfiguration().obtainOptimizer().tryInlineFunctionCall(
+            Expression e2 = visitor.getConfiguration().obtainOptimizer().tryInlineFunctionCall(
                     this, visitor, contextItemType);
             if (e2 != this) {
                 return e2.optimize(visitor, contextItemType);
@@ -368,11 +387,13 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
      * Compute the evaluation mode of each argument
      */
 
-    public void computeArgumentEvaluationModes() {
+    public synchronized void computeArgumentEvaluationModes() {
         int numArgs = getArity();
         argumentEvaluationModes = new int[numArgs];
         for (int i = 0; i < numArgs; i++) {
-            if (function.getParameterDefinitions()[i].isIndexedVariable()) {
+            if (i == 0 && function.getDeclaredStreamability().isConsuming()) {
+                argumentEvaluationModes[i] = ExpressionTool.STREAMING_ARGUMENT;
+            } else if (function.getParameterDefinitions()[i].isIndexedVariable()) {
                 argumentEvaluationModes[i] = ExpressionTool.MAKE_INDEXED_VARIABLE;
             } else {
                 Expression arg = getArg(i);
@@ -417,9 +438,9 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
      * @param pathMap        the PathMap to which the expression should be added
      * @param pathMapNodeSet the PathMapNodeSet to which the paths embodied in this expression should be added
      * @return the pathMapNode representing the focus established by this expression, in the case where this
-     *         expression is the first operand of a path expression or filter expression. For an expression that does
-     *         navigation, it represents the end of the arc in the path map that describes the navigation route. For other
-     *         expressions, it is the same as the input pathMapNode.
+     * expression is the first operand of a path expression or filter expression. For an expression that does
+     * navigation, it represents the end of the arc in the path map that describes the navigation route. For other
+     * expressions, it is the same as the input pathMapNode.
      */
 
     public PathMap.PathMapNodeSet addToPathMap(PathMap pathMap, PathMap.PathMapNodeSet pathMapNodeSet) {
@@ -494,7 +515,6 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         UserFunction targetFunction;
         Sequence[] actualArgs = evaluateArguments(context);
         XPathContextMajor c2;
-
         if (isTailCall()) {
             requestTailCall(context, actualArgs);
             return EmptySequence.getInstance();
@@ -502,23 +522,27 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
 
         if (bindingSlot >= 0) {
             Component target = getTargetComponent(context);
-            targetFunction = (UserFunction)target.getCode();
-            Component<UserFunction> targetComponent = (Component<UserFunction>)target;
-            if (targetComponent.getVisibility() == Visibility.ABSENT) {
-                throw new XPathException("Cannot call a function defined with visibility=absent", "XTDE3052");
+            if (target.isHiddenAbstractComponent()) {
+                throw new XPathException("Cannot call an abstract function (" +
+                                                 name.getDisplayName() +
+                                                 ") with no implementation", "XTDE3052");
             }
+            targetFunction = (UserFunction) target.getActor();
             c2 = targetFunction.makeNewContext(context);
-            c2.setCurrentComponent(targetComponent);
+            c2.setCurrentComponent(target);
+            c2.setOrigin(this);
         } else {
             targetFunction = function;
             c2 = targetFunction.makeNewContext(context);
+            c2.setOrigin(this);
         }
 
         try {
-            return targetFunction.call(c2, actualArgs);
+            Sequence s = targetFunction.call(c2, actualArgs);
+            return s;
         } catch (StackOverflowError err) {
             throw new XPathException("Too many nested function calls. May be due to infinite recursion",
-                SaxonErrorCode.SXLM0001, getLocation());
+                                     SaxonErrorCode.SXLM0001, getLocation());
         }
     }
 
@@ -527,7 +551,7 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
             TailCallLoop.TailCallComponent info = new TailCallLoop.TailCallComponent();
             Component target = getTargetComponent(context);
             info.component = target;
-            info.function = (UserFunction) target.getCode();
+            info.function = (UserFunction) target.getActor();
             ((XPathContextMajor) context).requestTailCall(info, actualArgs);
 
         } else {
@@ -556,9 +580,9 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
 
         if (bindingSlot >= 0) {
             Component target = getTargetComponent(context);
-            UserFunction targetFunction = (UserFunction) target.getCode();
-            if (target.getVisibility() == Visibility.ABSENT) {
-                throw new XPathException("Cannot call a function defined with visibility=absent", "XTDE3052");
+            UserFunction targetFunction = (UserFunction) target.getActor();
+            if (target.getVisibility() == Visibility.ABSTRACT) {
+                throw new XPathException("Cannot call a function defined with visibility=abstract", "XTDE3052");
             }
             XPathContextMajor c2 = targetFunction.makeNewContext(context);
             c2.setCurrentComponent(target);
@@ -569,7 +593,7 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         }
     }
 
-    private Component getTargetComponent(XPathContext context) {
+    public Component getTargetComponent(XPathContext context) {
         if (bindingSlot == -1) {
             // fallback for non-package code
             return function.getDeclaringComponent();
@@ -579,49 +603,31 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
     }
 
     public UserFunction getTargetFunction(XPathContext context) {
-        return (UserFunction)getTargetComponent(context).getCode();
+        return (UserFunction) getTargetComponent(context).getActor();
     }
 
-    /**
-     * Process the function call in pull mode
-     *
-     * @param context the XPath dynamic context
-     * @throws XPathException
-     */
-
-    public EventIterator iterateEvents(XPathContext context) throws XPathException {
-
-        Sequence[] actualArgs = evaluateArguments(context);
-
-        if (isTailCall()) {
-            requestTailCall(context, actualArgs);
-            return EmptyEventIterator.getInstance();
-        }
-
-        Component target = getTargetComponent(context);
-        UserFunction targetFunction = (UserFunction) target.getCode();
-        Component<UserFunction> targetComponent = (Component<UserFunction>) target;
-        if (targetComponent.getVisibility() == Visibility.ABSENT) {
-            throw new XPathException("Cannot call a function defined with visibility=absent", "XTDE3052");
-        }
-        XPathContextMajor c2 = targetFunction.makeNewContext(context);
-        c2.setCurrentComponent(targetComponent);
-        return targetFunction.iterateEvents(actualArgs, c2);
-
-    }
-
-
     public Sequence[] evaluateArguments(XPathContext c) throws XPathException {
+        return evaluateArguments(c, false);
+    }
+
+
+    public Sequence[] evaluateArguments(XPathContext c, boolean streamed) throws XPathException {
         int numArgs = getArity();
         Sequence[] actualArgs = new Sequence[numArgs];
-        if (argumentEvaluationModes == null) {
-            // should have been done at compile time
-            computeArgumentEvaluationModes();
+        synchronized(this) {
+            if (argumentEvaluationModes == null) {
+                // should have been done at compile time
+                computeArgumentEvaluationModes();
+            }
         }
         for (int i = 0; i < numArgs; i++) {
 
             int refs = 10; // was function.getParameterDefinitions()[i].getReferenceCount();
-            actualArgs[i] = ExpressionTool.evaluate(getArg(i), argumentEvaluationModes[i], c, refs);
+            int evalMode = argumentEvaluationModes[i];
+            if (evalMode == ExpressionTool.STREAMING_ARGUMENT && !streamed) {
+                evalMode = ExpressionTool.ITERATE_AND_MATERIALIZE;
+            }
+            actualArgs[i] = ExpressionTool.evaluate(getArg(i), evalMode, c, refs);
 
             if (actualArgs[i] == null) {
                 actualArgs[i] = EmptySequence.getInstance();
@@ -645,14 +651,14 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         if (getFunctionName() != null) {
             out.emitAttribute("name", getFunctionName().getEQName());
             out.emitAttribute("tailCall",
-                    tailCall == NOT_TAIL_CALL ? "false" : tailCall == SELF_TAIL_CALL ? "self" : "foreign");
+                              tailCall == NOT_TAIL_CALL ? "false" : tailCall == SELF_TAIL_CALL ? "self" : "foreign");
         }
         out.emitAttribute("bSlot", "" + getBindingSlot());
         if (getArgumentEvaluationModes() != null && getArity() > 0) {
             FastStringBuffer fsb = new FastStringBuffer(FastStringBuffer.C64);
             int[] e = getArgumentEvaluationModes();
             for (int i : e) {
-                fsb.append(i+" ");
+                fsb.append(i + " ");
             }
             out.emitAttribute("eval", Whitespace.trim(fsb));
         }
@@ -671,6 +677,18 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         return LocationKind.FUNCTION_CALL;
     }
 
+    /**
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
+     *
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in explain() output displaying the expression.
+     */
+    @Override
+    public String getExpressionName() {
+        return "userFunctionCall";
+    }
+
     public Object getProperty(String name) {
         if (name.equals("target")) {
             return function;
@@ -682,24 +700,5 @@ public class UserFunctionCall extends FunctionCall implements UserFunctionResolv
         return getFunctionName();
     }
 
-//#ifdefined STREAM
-    @Override
-    public UserFunctionCallAdjunct getStreamingAdjunct() {
-        return new UserFunctionCallAdjunct();
-    }
-//#endif
 
-
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the UserFunctionCall expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new UserFunctionCallCompiler();
-    }
-//#endif
 }
\ No newline at end of file
