@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,14 +7,9 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.ValueComparisonCompiler;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.expr.sort.AtomicComparer;
-import net.sf.saxon.expr.sort.CodepointCollatingComparer;
-import net.sf.saxon.expr.sort.CodepointCollator;
-import net.sf.saxon.expr.sort.GenericAtomicComparer;
+import net.sf.saxon.expr.sort.*;
 import net.sf.saxon.functions.*;
 import net.sf.saxon.lib.StringCollator;
 import net.sf.saxon.om.Sequence;
@@ -25,8 +20,6 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
 
-import java.util.Map;
-
 /**
  * ValueComparison: a boolean expression that compares two atomic values
  * for equals, not-equals, greater-than or less-than. Implements the operators
@@ -143,7 +136,7 @@ public final class ValueComparison extends BinaryExpression implements Compariso
         resetLocalStaticProperties();
         typeCheckChildren(visitor, contextInfo);
 
-        Configuration config = getConfiguration();
+        Configuration config = visitor.getConfiguration();
         StaticContext env = visitor.getStaticContext();
 
         if (Literal.isEmptySequence(getLhsExpression())) {
@@ -154,13 +147,18 @@ public final class ValueComparison extends BinaryExpression implements Compariso
             return resultWhenEmpty == null ? getRhsExpression() : Literal.makeLiteral(resultWhenEmpty);
         }
 
+        if (comparer instanceof UntypedNumericComparer) {
+            return this; // we've already done all that needs to be done
+        }
+
         final SequenceType optionalAtomic = SequenceType.OPTIONAL_ATOMIC;
+        TypeChecker tc = config.getTypeChecker(false);
 
         RoleDiagnostic role0 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 0);
-        setLhsExpression(TypeChecker.staticTypeCheck(getLhsExpression(), optionalAtomic, false, role0, visitor));
+        setLhsExpression(tc.staticTypeCheck(getLhsExpression(), optionalAtomic, role0, visitor));
 
         RoleDiagnostic role1 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 1);
-        setRhsExpression(TypeChecker.staticTypeCheck(getRhsExpression(), optionalAtomic, false, role1, visitor));
+        setRhsExpression(tc.staticTypeCheck(getRhsExpression(), optionalAtomic, role1, visitor));
 
         PlainType t0 = getLhsExpression().getItemType().getAtomizedItemType();
         PlainType t1 = getRhsExpression().getItemType().getAtomizedItemType();
@@ -209,8 +207,9 @@ public final class ValueComparison extends BinaryExpression implements Compariso
                         (opt1 ? "?" : "") + " will fail unless " + which + " empty", getLocation());
                 needsRuntimeCheck = true;
             } else {
-                XPathException err = new XPathException("Cannot compare " + t0.toString() +
-                        " to " + t1.toString());
+                String message = "In {" + toShortString() + "}: cannot compare " +
+                        t0.toString() + " to " + t1.toString();
+                XPathException err = new XPathException(message);
                 err.setIsTypeError(true);
                 err.setErrorCode("XPTY0004");
                 err.setLocation(getLocation());
@@ -266,7 +265,7 @@ public final class ValueComparison extends BinaryExpression implements Compariso
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
 
-        TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
 
         optimizeChildren(visitor, contextInfo);
 
@@ -298,12 +297,12 @@ public final class ValueComparison extends BinaryExpression implements Compariso
         // TODO: extend this to cases where N is not a literal
 
         if (getLhsExpression().isCallOn(Count.class) && Literal.isAtomic(getRhsExpression())) {
-            Expression e2 = optimizeCount(visitor, false);
+            Expression e2 = optimizeCount(false);
             if (e2 != null) {
                 return e2.optimize(visitor, contextInfo);
             }
         } else if (getRhsExpression().isCallOn(Count.class) && Literal.isAtomic(getLhsExpression())) {
-            Expression e2 = optimizeCount(visitor, true);
+            Expression e2 = optimizeCount(true);
             if (e2 != null) {
                 return e2.optimize(visitor, contextInfo);
             }
@@ -502,14 +501,13 @@ public final class ValueComparison extends BinaryExpression implements Compariso
      * until the result of the comparison is deducible; for example to evaluate (count(X)>2) we can stop at the third
      * item in the sequence.
      *
-     * @param visitor  the expression visitor
      * @param inverted true if the call to count(X) is the right-hand operand
      * @return the optimized expression, or null if no optimization is possible
      * @throws XPathException if an error occurs
      */
 
     /*@Nullable*/
-    private Expression optimizeCount(ExpressionVisitor visitor, boolean inverted) throws XPathException {
+    private Expression optimizeCount(boolean inverted) throws XPathException {
         SystemFunctionCall countFn = (SystemFunctionCall) (inverted ? getRhsExpression() : getLhsExpression());
         Expression sequence = countFn.getArg(0);
         sequence = sequence.unordered(false, false);
@@ -716,7 +714,7 @@ public final class ValueComparison extends BinaryExpression implements Compariso
      * @throws XPathException if the values are not comparable
      */
 
-    static boolean compare(AtomicValue v0, int op, AtomicValue v1, AtomicComparer comparer, boolean checkTypes)
+    public static boolean compare(AtomicValue v0, int op, AtomicValue v1, AtomicComparer comparer, boolean checkTypes)
             throws XPathException {
         if (checkTypes &&
                 !Type.isGuaranteedComparable(v0.getPrimitiveType(), v1.getPrimitiveType(), Token.isOrderedOperator(op))) {
@@ -817,6 +815,18 @@ public final class ValueComparison extends BinaryExpression implements Compariso
         return BuiltInAtomicType.BOOLEAN;
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        return UType.BOOLEAN;
+    }
+
     /**
      * Determine the static cardinality.
      */
@@ -842,34 +852,7 @@ public final class ValueComparison extends BinaryExpression implements Compariso
         out.emitAttribute("comp", comparer.save());
     }
 
-//#ifdefined BYTECODE
 
-    /**
-     * Return the compiler of the ValueComparison expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ValueComparisonCompiler();
-    }
-//#endif
-
-
-  /*  *//**
-     * Find any necessary preconditions for the satisfaction of this expression
-     * as a set of boolean expressions to be evaluated on the context node
-     *
-     * @return A set of conditions, or null if none have been computed
-     *//*
-    public Set<Expression> getPreconditions() {
-        HashSet<Expression> pre = new HashSet<Expression>();
-        Expression r = RealPreconditions.reducedFunction(this);
-        if(r != null) {
-            pre.add(r);
-        }
-        return pre;
-    }*/
 
 }
 
