@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,22 +7,19 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.VariableReferenceCompiler;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
-import com.saxonica.ee.stream.adjunct.VariableReferenceAdjunct;
+import com.saxonica.ee.stream.Streamability;
 import net.sf.saxon.event.SequenceReceiver;
+import net.sf.saxon.expr.flwor.LocalVariableBinding;
 import net.sf.saxon.expr.instruct.GlobalParam;
+import net.sf.saxon.expr.instruct.LocalParam;
+import net.sf.saxon.expr.instruct.LocalParamBlock;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.lib.StandardErrorListener;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.NodeTest;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
-import net.sf.saxon.type.AnyItemType;
-import net.sf.saxon.type.AtomicType;
-import net.sf.saxon.type.ItemType;
-import net.sf.saxon.type.TypeHierarchy;
+import net.sf.saxon.type.*;
 import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.IntegerValue;
 import net.sf.saxon.value.SequenceType;
@@ -87,21 +84,17 @@ public abstract class VariableReference extends Expression implements BindingRef
      * Create a clone copy of this VariableReference
      *
      * @return the cloned copy
-     * @param rebindings
+     * @param rebindings variables that need to switch to new bindings
      */
 
     /*@NotNull*/
     public abstract Expression copy(RebindingMap rebindings);
-//    {
-//        if (binding == null) {
-//            //System.err.println("copy unbound variable " + this);
-//            throw new UnsupportedOperationException("Cannot copy a variable reference whose binding is unknown");
-//        }
-//        VariableReference ref = new VariableReference();
-//        ref.copyFrom(this);
-//        binding.addReference(inLoop);
-//        return ref;
-//    }
+
+
+    @Override
+    public int getNetCost() {
+        return 0;
+    }
 
     protected void copyFrom(VariableReference ref) {
         binding = ref.binding;
@@ -112,7 +105,7 @@ public abstract class VariableReference extends Expression implements BindingRef
         inLoop = ref.inLoop;
         filtered = ref.filtered;
         //#ifdefined STREAM
-        setPostureAndSweep(ref.getPostureAndSweepIfKnown());
+        Streamability.setPostureAndSweep(this, Streamability.getPostureAndSweepIfKnown(ref));
         //#endif
         ExpressionTool.copyLocationInfo(ref, this);
     }
@@ -138,7 +131,8 @@ public abstract class VariableReference extends Expression implements BindingRef
         // Although the variable may be a context document node-set at the point it is defined,
         // the context at the point of use may be different, so this property cannot be transferred.
         int dependencies = getDependencies();
-        staticProperties = (properties & ~StaticProperty.CONTEXT_DOCUMENT_NODESET) |
+        staticProperties = (properties & ~StaticProperty.CONTEXT_DOCUMENT_NODESET
+                        & ~StaticProperty.ALL_NODES_NEWLY_CREATED) |
                 StaticProperty.NON_CREATIVE |
                 type.getCardinality() |
                 dependencies;
@@ -215,6 +209,7 @@ public abstract class VariableReference extends Expression implements BindingRef
 //        }
 
 
+
 //  following code removed because it causes error181 to blow the stack - need to check for circularities well
 //            if (binding instanceof GlobalVariable) {
 //                ((GlobalVariable)binding).typeCheck(visitor, AnyItemType.getInstance());
@@ -233,7 +228,7 @@ public abstract class VariableReference extends Expression implements BindingRef
     }
 
     /**
-     * Type-check the expression. At this stage details of the static type must be known.
+     * Optimize the expression. At this stage details of the static type must be known.
      * If the variable has a compile-time value, this is substituted for the variable reference
      */
 
@@ -243,12 +238,16 @@ public abstract class VariableReference extends Expression implements BindingRef
                 ((LetExpression)binding).getSequence() instanceof Literal &&
                 !((LetExpression) binding).isIndexedVariable) {
             Expression val = ((LetExpression) binding).getSequence();
+            Optimizer.trace(visitor.getConfiguration(), "Replaced variable " + getDisplayName() + " by its value", val);
             binding = null;
             return val.copy(new RebindingMap());
         }
         if (constantValue != null) {
             binding = null;
-            return Literal.makeLiteral(constantValue);
+            Expression result = Literal.makeLiteral(constantValue);
+            ExpressionTool.copyLocationInfo(this, result);
+            Optimizer.trace(visitor.getConfiguration(), "Replaced variable " + getDisplayName() + " by its value", result);
+            return result;
         }
         return this;
     }
@@ -259,8 +258,12 @@ public abstract class VariableReference extends Expression implements BindingRef
      * to be located at run-time.
      */
 
-    public void fixup(Binding binding) {
-        this.binding = binding;
+    public void fixup(Binding newBinding) {
+        boolean indexed = binding instanceof LocalBinding && ((LocalBinding)binding).isIndexedVariable();
+        this.binding = newBinding;
+        if (indexed && newBinding instanceof LocalBinding) {
+            ((LocalBinding)newBinding).setIndexedVariable();
+        }
         resetLocalStaticProperties();
     }
 
@@ -316,6 +319,33 @@ public abstract class VariableReference extends Expression implements BindingRef
         }
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType the static context item type
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (binding != null) {
+            if (binding.isGlobal() ||
+                    binding instanceof LocalParam ||
+                    (binding instanceof LetExpression && ((LetExpression)binding).isInstruction()) ||
+                    binding instanceof LocalVariableBinding) {
+                SequenceType st = binding.getRequiredType();
+                if (st != null) {
+                    return st.getPrimaryType().getUType();
+                } else {
+                    return UType.ANY;
+                }
+            } else if (binding instanceof Assignation) {
+                return ((Assignation) binding).getSequence().getStaticUType(contextItemType);
+            }
+        }
+        return UType.ANY;
+    }
+
     /**
      * For an expression that returns an integer or a sequence of integers, get
      * a lower and upper bound on the values of the integers that may be returned, from
@@ -385,7 +415,7 @@ public abstract class VariableReference extends Expression implements BindingRef
                 staticType.getPrimaryType() instanceof NodeTest) {
             p |= StaticProperty.SINGLE_DOCUMENT_NODESET;
         }
-        return p;
+        return p &~ StaticProperty.ALL_NODES_NEWLY_CREATED;
     }
 
     /**
@@ -429,14 +459,6 @@ public abstract class VariableReference extends Expression implements BindingRef
         return d;
     }
 
-    /**
-     * Promote this expression if possible
-     */
-
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        return this;
-    }
-
     /**
      * An implementation of Expression must provide at least one of the methods evaluateItem(), iterate(), or process().
      * This method indicates which of these methods is provided. This implementation provides both all three methods
@@ -448,6 +470,36 @@ public abstract class VariableReference extends Expression implements BindingRef
                 | ITERATE_METHOD | PROCESS_METHOD;
     }
 
+    /**
+     * Get the innermost scoping expression of this expression, for expressions that directly
+     * depend on something in the dynamic context. For example, in the case of a local variable
+     * reference this is the expression that causes the relevant variable to be bound; for a
+     * context item expression it is the innermost focus-setting container. For expressions
+     * that have no intrinsic dependency on the dynamic context, the value returned is null;
+     * the scoping container for such an expression is the innermost scoping container of its
+     * operands.
+     *
+     * @return the innermost scoping container of this expression
+     */
+    @Override
+    public Expression getScopingExpression() {
+        if (binding instanceof Expression) {
+            if (binding instanceof LocalParam && ((LocalParam)binding).getParentExpression() instanceof LocalParamBlock) {
+                LocalParamBlock block = (LocalParamBlock)((LocalParam) binding).getParentExpression();
+                return block.getParentExpression();
+            } else {
+                return (Expression) binding;
+            }
+        }
+        Expression parent = getParentExpression();
+        while (parent != null) {
+            if (parent.hasVariableBinding(binding)) {
+                return parent;
+            }
+            parent = parent.getParentExpression();
+        }
+        return null;
+    }
 
     /**
      * Add a representation of this expression to a PathMap. The PathMap captures a map of the nodes visited
@@ -486,19 +538,21 @@ public abstract class VariableReference extends Expression implements BindingRef
     public SequenceIterator iterate(XPathContext c) throws XPathException {
         try {
             Sequence actual = evaluateVariable(c);
+            assert actual != null;
             return actual.iterate();
         } catch (XPathException err) {
             err.maybeSetLocation(getLocation());
             throw err;
         } catch (NullPointerException err) {
+            err.printStackTrace();
             String msg = "Internal error: no value for variable $" + getDisplayName() +
-                    " at line " + getLocation().getLineNumber() + (getSystemId() == null ? "" : " of " + getSystemId());
+                    " at line " + getLocation().getLineNumber() + (getLocation().getSystemId() == null ? "" : " of " + getLocation().getSystemId());
             StandardErrorListener.printStackTrace(c.getConfiguration().getLogger(), c);
             throw new AssertionError(msg);
         } catch (AssertionError err) {
             err.printStackTrace();
             String msg = err.getMessage() + ". Variable reference $" + getDisplayName() +
-                    " at line " + getLocation().getLineNumber() + (getSystemId() == null ? "" : " of " + getSystemId());
+                    " at line " + getLocation().getLineNumber() + (getLocation().getSystemId() == null ? "" : " of " + getLocation().getSystemId());
             StandardErrorListener.printStackTrace(c.getConfiguration().getLogger(), c);
             throw new AssertionError(msg);
         }
@@ -507,6 +561,7 @@ public abstract class VariableReference extends Expression implements BindingRef
     public Item evaluateItem(XPathContext c) throws XPathException {
         try {
             Sequence actual = evaluateVariable(c);
+            assert actual != null;
             return actual.head();
         } catch (XPathException err) {
             err.maybeSetLocation(getLocation());
@@ -519,12 +574,9 @@ public abstract class VariableReference extends Expression implements BindingRef
             SequenceIterator iter = evaluateVariable(c).iterate();
             SequenceReceiver out = c.getReceiver();
             Location loc = getLocation();
-            while (true) {
-                Item it = iter.next();
-                if (it == null) {
-                    break;
-                }
-                out.append(it, loc, NodeInfo.ALL_NAMESPACES);
+            Item item;
+            while ((item = iter.next()) != null) {
+                out.append(item, loc, NodeInfo.ALL_NAMESPACES);
             }
         } catch (XPathException err) {
             err.maybeSetLocation(getLocation());
@@ -598,17 +650,6 @@ public abstract class VariableReference extends Expression implements BindingRef
         }
     }
 
-    public boolean isCompatible(VariableReference other) {
-        if (!getEQName().equals(other.getEQName())) {
-            return false;
-        }
-        if (getItemType().getPrimitiveType() != other.getItemType().getPrimitiveType()) {
-            //TODO - this might be wrong. Need to check
-            return false;
-        }
-        return false;
-    }
-
     /**
      * The toString() method for an expression attempts to give a representation of the expression
      * in an XPath-like form, but there is no guarantee that the syntax will actually be true XPath.
@@ -631,6 +672,18 @@ public abstract class VariableReference extends Expression implements BindingRef
         return "$" + getDisplayName();
     }
 
+    /**
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
+     *
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in export() output displaying the expression.
+     */
+    @Override
+    public String getExpressionName() {
+        return "varRef";
+    }
+
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -639,9 +692,6 @@ public abstract class VariableReference extends Expression implements BindingRef
     public void export(ExpressionPresenter destination) throws XPathException {
         destination.startElement("varRef", this);
         destination.emitAttribute("name", variableName);
-//        if (this instanceof LocalVariableReference) {
-//            destination.emitAttribute("local-slot", "" + ((LocalVariableReference) this).getSlotNumber());
-//        }
         if (binding instanceof LocalBinding) {
             destination.emitAttribute("slot", "" + ((LocalBinding) binding).getLocalSlotNumber());
         }
@@ -649,32 +699,15 @@ public abstract class VariableReference extends Expression implements BindingRef
         destination.endElement();
     }
 
-    //#ifdefined STREAM
-
     /**
-     * Get a class that supports streamed evaluation of this expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    protected StreamingAdjunct getStreamingAdjunct() {
-        return new VariableReferenceAdjunct();
+    public String getStreamerName() {
+        return "VariableReference";
     }
-
-    //#endif
-
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the VariableReference expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new VariableReferenceCompiler();
-    }
-//#endif
-
 }
 
