@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,12 @@
 
 package net.sf.saxon.expr;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.VennExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
-import com.saxonica.ee.stream.adjunct.VennExpressionAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.instruct.Block;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.expr.sort.DocumentSorter;
 import net.sf.saxon.expr.sort.GlobalOrderComparer;
+import net.sf.saxon.functions.CurrentGroupCall;
 import net.sf.saxon.functions.SystemFunction;
 import net.sf.saxon.om.AxisInfo;
 import net.sf.saxon.om.SequenceIterator;
@@ -24,11 +21,11 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.Type;
 import net.sf.saxon.type.TypeHierarchy;
+import net.sf.saxon.type.UType;
 import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.SequenceType;
 
 import java.util.HashSet;
-import java.util.Map;
 import java.util.Set;
 
 
@@ -112,6 +109,27 @@ public class VennExpression extends BinaryExpression {
         }
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        switch (operator) {
+            case Token.UNION:
+                return getLhsExpression().getStaticUType(contextItemType).union(getRhsExpression().getStaticUType(contextItemType));
+            case Token.INTERSECT:
+                return getLhsExpression().getStaticUType(contextItemType).intersection(getRhsExpression().getStaticUType(contextItemType));
+            case Token.EXCEPT:
+            default:
+                return getLhsExpression().getStaticUType(contextItemType);
+        }
+    }
+
+
     /**
      * Determine the static cardinality of the expression
      */
@@ -257,21 +275,23 @@ public class VennExpression extends BinaryExpression {
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, final ContextItemStaticInfo contextInfo) throws XPathException {
 
+        Configuration config = visitor.getConfiguration();
+        TypeChecker tc = config.getTypeChecker(false);
         typeCheckChildren(visitor, contextInfo);
 
         if (!(getLhsExpression() instanceof Pattern)) {
             final RoleDiagnostic role0 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 0);
-            setLhsExpression(TypeChecker.staticTypeCheck(getLhsExpression(), SequenceType.NODE_SEQUENCE, false, role0, visitor));
+            setLhsExpression(tc.staticTypeCheck(getLhsExpression(), SequenceType.NODE_SEQUENCE, role0, visitor));
         }
 
         if (!(getRhsExpression() instanceof Pattern)) {
             final RoleDiagnostic role1 = new RoleDiagnostic(RoleDiagnostic.BINARY_EXPR, Token.tokens[operator], 1);
-            setRhsExpression(TypeChecker.staticTypeCheck(getRhsExpression(), SequenceType.NODE_SEQUENCE, false, role1, visitor));
+            setRhsExpression(tc.staticTypeCheck(getRhsExpression(), SequenceType.NODE_SEQUENCE, role1, visitor));
         }
 
         // For the intersect and except operators, if the types are disjoint then we can simplify
         if (operator != Token.UNION) {
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
+            TypeHierarchy th = config.getTypeHierarchy();
             ItemType t0 = getLhsExpression().getItemType();
             ItemType t1 = getRhsExpression().getItemType();
             if (th.relationship(t0, t1) == TypeHierarchy.DISJOINT) {
@@ -315,39 +335,56 @@ public class VennExpression extends BinaryExpression {
             return e;
         }
 
-        final Configuration config = getConfiguration();
+        final Configuration config = visitor.getConfiguration();
         final TypeHierarchy th = config.getTypeHierarchy();
 
         // If either operand is an empty sequence, simplify the expression. This can happen
         // after reduction with constructs of the form //a[condition] | //b[not(condition)],
         // common in XPath 1.0 because there were no conditional expressions.
 
+        Expression lhs = getLhsExpression();
+        Expression rhs = getRhsExpression();
         switch (operator) {
             case Token.UNION:
-                if (Literal.isEmptySequence(getLhsExpression()) &&
-                        (getRhsExpression().getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
-                    return getRhsExpression();
+                if (Literal.isEmptySequence(lhs) &&
+                        (rhs.getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
+                    return rhs;
                 }
-                if (Literal.isEmptySequence(getRhsExpression()) &&
-                        (getLhsExpression().getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
-                    return getLhsExpression();
+                if (Literal.isEmptySequence(rhs) &&
+                        (lhs.getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
+                    return lhs;
+                }
+                if (lhs instanceof CurrentGroupCall && rhs instanceof ContextItemExpression) {
+                    return lhs;
                 }
                 break;
             case Token.INTERSECT:
-                if (Literal.isEmptySequence(getLhsExpression())) {
-                    return getLhsExpression();
+                if (Literal.isEmptySequence(lhs)) {
+                    return lhs;
                 }
-                if (Literal.isEmptySequence(getRhsExpression())) {
-                    return getRhsExpression();
+                if (Literal.isEmptySequence(rhs)) {
+                    return rhs;
+                }
+                if (lhs instanceof CurrentGroupCall && rhs instanceof ContextItemExpression) {
+                    return rhs;
                 }
                 break;
             case Token.EXCEPT:
-                if (Literal.isEmptySequence(getLhsExpression())) {
-                    return getLhsExpression();
+                if (Literal.isEmptySequence(lhs)) {
+                    return lhs;
                 }
-                if (Literal.isEmptySequence(getRhsExpression()) &&
-                        (getLhsExpression().getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
-                    return getLhsExpression();
+                if (Literal.isEmptySequence(rhs) &&
+                        (lhs.getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
+                    return lhs;
+                }
+                if (lhs instanceof CurrentGroupCall && rhs instanceof ContextItemExpression) {
+                    // Test case si-group-055.
+                    // The streaming code has problems with (current-group() except .) so we
+                    // optimize it away. This is a bit of a hack, because the difficulty may
+                    // affect other expressions as well. The problem arises because part of the
+                    // pattern needs to be evaluated with each node in the group as anchor node,
+                    // and part with only the first node in the group as anchor node.
+                    return new TailExpression(lhs, 2);
                 }
                 break;
         }
@@ -355,9 +392,9 @@ public class VennExpression extends BinaryExpression {
         // If both are axis expressions on the same axis, merge them
         // ie. rewrite (axis::test1 | axis::test2) as axis::(test1 | test2)
 
-        if (getLhsExpression() instanceof AxisExpression && getRhsExpression() instanceof AxisExpression) {
-            final AxisExpression a1 = (AxisExpression) getLhsExpression();
-            final AxisExpression a2 = (AxisExpression) getRhsExpression();
+        if (lhs instanceof AxisExpression && rhs instanceof AxisExpression) {
+            final AxisExpression a1 = (AxisExpression) lhs;
+            final AxisExpression a2 = (AxisExpression) rhs;
             if (a1.getAxis() == a2.getAxis()) {
                 AxisExpression ax = new AxisExpression(a1.getAxis(),
                         new CombinedNodeTest(a1.getNodeTest(),
@@ -379,9 +416,9 @@ public class VennExpression extends BinaryExpression {
 
         // TODO: generalize this code to handle all distributive operators
 
-        if (getLhsExpression() instanceof SlashExpression && getRhsExpression() instanceof SlashExpression && operator == Token.UNION) {
-            final SlashExpression path1 = (SlashExpression) getLhsExpression();
-            final SlashExpression path2 = (SlashExpression) getRhsExpression();
+        if (lhs instanceof SlashExpression && rhs instanceof SlashExpression && operator == Token.UNION) {
+            final SlashExpression path1 = (SlashExpression) lhs;
+            final SlashExpression path2 = (SlashExpression) rhs;
 
             if (path1.getFirstStep().equals(path2.getFirstStep())) {
                 final VennExpression venn = new VennExpression(
@@ -398,9 +435,9 @@ public class VennExpression extends BinaryExpression {
         // Try merging two non-positional filter expressions:
         // A[exp0] | A[exp1] becomes A[exp0 or exp1]
 
-        if (getLhsExpression() instanceof FilterExpression && getRhsExpression() instanceof FilterExpression) {
-            final FilterExpression exp0 = (FilterExpression) getLhsExpression();
-            final FilterExpression exp1 = (FilterExpression) getRhsExpression();
+        if (lhs instanceof FilterExpression && rhs instanceof FilterExpression) {
+            final FilterExpression exp0 = (FilterExpression) lhs;
+            final FilterExpression exp1 = (FilterExpression) rhs;
 
             if (!exp0.isPositional(th) &&
                     !exp1.isPositional(th) &&
@@ -432,22 +469,22 @@ public class VennExpression extends BinaryExpression {
         // Convert @*|node() into @*,node() to eliminate the sorted merge operation
         // Avoid doing this when streaming because xsl:value-of select="@*,node()" is not currently streamable
         if (!visitor.isOptimizeForStreaming() && operator == Token.UNION &&
-                getLhsExpression() instanceof AxisExpression && getRhsExpression() instanceof AxisExpression) {
-            AxisExpression a0 = (AxisExpression) getLhsExpression();
-            AxisExpression a1 = (AxisExpression) getRhsExpression();
+                lhs instanceof AxisExpression && rhs instanceof AxisExpression) {
+            AxisExpression a0 = (AxisExpression) lhs;
+            AxisExpression a1 = (AxisExpression) rhs;
             if (a0.getAxis() == AxisInfo.ATTRIBUTE && a1.getAxis() == AxisInfo.CHILD) {
-                return new Block(new Expression[]{getLhsExpression(), getRhsExpression()});
+                return new Block(new Expression[]{lhs, rhs});
             } else if (a1.getAxis() == AxisInfo.ATTRIBUTE && a0.getAxis() == AxisInfo.CHILD) {
-                return new Block(new Expression[]{getRhsExpression(), getLhsExpression()});
+                return new Block(new Expression[]{rhs, lhs});
             }
         }
 
         // Convert (A intersect B) to use a serial search where one operand is a singleton
-        if (operator == Token.INTERSECT && !Cardinality.allowsMany(getLhsExpression().getCardinality())) {
-            return new SingletonIntersectExpression(getLhsExpression(), operator, getRhsExpression().unordered(false, false));
+        if (operator == Token.INTERSECT && !Cardinality.allowsMany(lhs.getCardinality())) {
+            return new SingletonIntersectExpression(lhs, operator, rhs.unordered(false, false));
         }
-        if (operator == Token.INTERSECT && !Cardinality.allowsMany(getRhsExpression().getCardinality())) {
-            return new SingletonIntersectExpression(getRhsExpression(), operator, getLhsExpression().unordered(false, false));
+        if (operator == Token.INTERSECT && !Cardinality.allowsMany(rhs.getCardinality())) {
+            return new SingletonIntersectExpression(rhs, operator, lhs.unordered(false, false));
         }
 
         // If the types of the operands are disjoint, simplify "intersect" and "except"
@@ -455,10 +492,10 @@ public class VennExpression extends BinaryExpression {
             if (operator == Token.INTERSECT) {
                 return Literal.makeEmptySequence();
             } else if (operator == Token.EXCEPT) {
-                if ((getLhsExpression().getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
-                    return getLhsExpression();
+                if ((lhs.getSpecialProperties() & StaticProperty.ORDERED_NODESET) != 0) {
+                    return lhs;
                 } else {
-                    return new DocumentSorter(getLhsExpression());
+                    return new DocumentSorter(lhs);
                 }
             }
         }
@@ -653,33 +690,15 @@ public class VennExpression extends BinaryExpression {
         }
     }
 
-//#ifdefined STREAM
-
-
     /**
-     * Get a class that supports streamed evaluation of this expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    protected StreamingAdjunct getStreamingAdjunct() {
-        return new VennExpressionAdjunct();
+    public String getStreamerName() {
+        return "VennExpression";
     }
-    //#endif
-
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Venn expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new VennExpressionCompiler();
-    }
-//#endif
-
-
 }
 
