@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -19,6 +19,8 @@ import net.sf.saxon.om.Sequence;
 import net.sf.saxon.om.StructuredQName;
 import net.sf.saxon.regex.RegexIterator;
 import net.sf.saxon.trans.*;
+import net.sf.saxon.trans.rules.Rule;
+import net.sf.saxon.trans.rules.RuleManager;
 import net.sf.saxon.tree.iter.SingletonIterator;
 import net.sf.saxon.tree.iter.UnfailingIterator;
 
@@ -37,7 +39,7 @@ public class XPathContextMajor extends XPathContextMinor {
     private ParameterSet localParameters;
     private ParameterSet tunnelParameters;
     /*@Nullable*/ private TailCallLoop.TailCallInfo tailCallInfo;
-    private Component<? extends Mode> currentMode;
+    private Component.M currentMode;
     /*@Nullable*/ private Rule currentTemplate;
     private GroupIterator currentGroupIterator;
     private GroupIterator currentMergeGroupIterator;
@@ -47,7 +49,7 @@ public class XPathContextMajor extends XPathContextMinor {
 
     private URIResolver uriResolver;
     private UnfailingErrorListener errorListener;
-    private Component<? extends ComponentCode> currentComponent;
+    private Component currentComponent;
     XPathException currentException;
 
 
@@ -152,7 +154,7 @@ public class XPathContextMajor extends XPathContextMinor {
         c.currentTemplate = p.getCurrentTemplateRule();
         c.currentRegexIterator = p.getCurrentRegexIterator();
         c.currentGroupIterator = p.getCurrentGroupIterator();
-        c.currentMergeGroupIterator = p.getCurrentMergeGroupIterator();
+        c.currentMergeGroupIterator = null;
         c.caller = prev;
         c.tailCallInfo = null;
         c.threadManager = ((XPathContextMajor) p).threadManager;
@@ -371,9 +373,7 @@ public class XPathContextMajor extends XPathContextMinor {
     public void requestTailCall(TailCallLoop.TailCallInfo targetFn, Sequence[] variables) {
         if (variables != null) {
             if (variables.length > stackFrame.slots.length) {
-                Sequence[] v2 = new Sequence[variables.length];
-                System.arraycopy(variables, 0, v2, 0, variables.length);
-                stackFrame.slots = v2;
+                stackFrame.slots = Arrays.copyOf(variables, variables.length);
             } else {
                 System.arraycopy(variables, 0, stackFrame.slots, 0, variables.length);
             }
@@ -428,7 +428,7 @@ public class XPathContextMajor extends XPathContextMinor {
      * @param mode the new current mode
      */
 
-    public void setCurrentMode(Component<? extends Mode> mode) {
+    public void setCurrentMode(Component.M mode) {
         this.currentMode = mode;
     }
 
@@ -438,8 +438,8 @@ public class XPathContextMajor extends XPathContextMinor {
      * @return the current mode. May return null if the current mode is the default mode.
      */
 
-    public Component<? extends Mode> getCurrentMode() {
-        Component<? extends Mode> m = currentMode;
+    public Component.M getCurrentMode() {
+        Component.M m = currentMode;
         if (m == null) {
             RuleManager rm = getController().getRuleManager();
             if (rm != null) {
