@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,8 +7,7 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.AnalyzeStringCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
+import net.sf.saxon.Configuration;
 import net.sf.saxon.Version;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
@@ -149,7 +148,20 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
      * @return if allow XSLT 3.0 features
      */
     public boolean isAllow30features() {
-        return getRetainedStaticContext().getXPathVersion() >= 30;
+        return true;
+    }
+
+    /**
+     * Ask whether common subexpressions found in the operands of this expression can
+     * be extracted and evaluated outside the expression itself. The result is irrelevant
+     * in the case of operands evaluated with a different focus, which will never be
+     * extracted in this way, even if they have no focus dependency.
+     *
+     * @return false for this kind of expression
+     */
+    @Override
+    public boolean allowExtractingCommonSubexpressions() {
+        return false;
     }
 
     /**
@@ -176,29 +188,30 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
 
+        Configuration config = visitor.getConfiguration();
         selectOp.typeCheck(visitor, contextInfo);
         regexOp.typeCheck(visitor, contextInfo);
         flagsOp.typeCheck(visitor, contextInfo);
 
         if (matchingOp != null) {
-            matchingOp.typeCheck(visitor, new ContextItemStaticInfo(BuiltInAtomicType.STRING, false));
+            matchingOp.typeCheck(visitor, config.makeContextItemStaticInfo(BuiltInAtomicType.STRING, false));
         }
         if (nonMatchingOp != null) {
-            nonMatchingOp.typeCheck(visitor, new ContextItemStaticInfo(BuiltInAtomicType.STRING, false));
+            nonMatchingOp.typeCheck(visitor, config.makeContextItemStaticInfo(BuiltInAtomicType.STRING, false));
         }
 
+        TypeChecker tc = visitor.getConfiguration().getTypeChecker(false);
 
         RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "analyze-string/select", 0);
-        SequenceType required = visitor.getStaticContext().getXPathVersion() >= 30 ?
-            SequenceType.OPTIONAL_STRING : SequenceType.SINGLE_STRING;
+        SequenceType required = SequenceType.OPTIONAL_STRING;
         // see bug 7976
-        setSelect(TypeChecker.staticTypeCheck(getSelect(), required, false, role, visitor));
+        setSelect(tc.staticTypeCheck(getSelect(), required, role, visitor));
 
         role = new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "analyze-string/regex", 0);
-        setRegex(TypeChecker.staticTypeCheck(getRegex(), SequenceType.SINGLE_STRING, false, role, visitor));
+        setRegex(tc.staticTypeCheck(getRegex(), SequenceType.SINGLE_STRING, role, visitor));
 
         role = new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "analyze-string/flags", 0);
-        setFlags(TypeChecker.staticTypeCheck(getFlags(), SequenceType.SINGLE_STRING, false, role, visitor));
+        setFlags(tc.staticTypeCheck(getFlags(), SequenceType.SINGLE_STRING, role, visitor));
 
         return this;
     }
@@ -207,15 +220,16 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
 
+        Configuration config = visitor.getConfiguration();
         selectOp.optimize(visitor, contextInfo);
         regexOp.optimize(visitor, contextInfo);
         flagsOp.optimize(visitor, contextInfo);
 
         if (matchingOp != null) {
-            matchingOp.optimize(visitor, new ContextItemStaticInfo(BuiltInAtomicType.STRING, false));
+            matchingOp.optimize(visitor, config.makeContextItemStaticInfo(BuiltInAtomicType.STRING, false));
         }
         if (nonMatchingOp != null) {
-            nonMatchingOp.optimize(visitor, new ContextItemStaticInfo(BuiltInAtomicType.STRING, false));
+            nonMatchingOp.optimize(visitor, config.makeContextItemStaticInfo(BuiltInAtomicType.STRING, false));
         }
 
         if (pattern == null && getRegex() instanceof StringLiteral && getFlags() instanceof StringLiteral) {
@@ -223,16 +237,16 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
                 final CharSequence regex = ((StringLiteral) this.getRegex()).getStringValue();
                 final CharSequence flagstr = ((StringLiteral) getFlags()).getStringValue();
 
-                String hostLang = visitor.getStaticContext().getXPathVersion() >= 30 ? "XP30" : "XP20";
+                String hostLang = "XP30";
                 List<String> warnings = new ArrayList<String>();
-                pattern = Version.platform.compileRegularExpression(regex, flagstr.toString(), hostLang, warnings);
+                pattern = Version.platform.compileRegularExpression(config, regex, flagstr.toString(), hostLang, warnings);
                 for (String w : warnings) {
                     visitor.getStaticContext().issueWarning(w, getLocation());
                 }
 
                 if ((!isAllow30features() || !useXsltErrorCodes) && pattern.matches("")) {
                     // prevent it being reported more than once
-                    pattern = Version.platform.compileRegularExpression("x", "", "XP20", warnings);
+                    pattern = Version.platform.compileRegularExpression(config, "x", "", "XP20", warnings);
                     invalidRegex("The regular expression must not be one that matches a zero-length string",
                             useXsltErrorCodes ? "XTDE1150" : "FORX0003");
                 }
@@ -261,18 +275,13 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
      * Copy an expression. This makes a deep copy.
      *
      * @return the copy of the original expression
-     * @param rebindings
+     * @param rm the rebinding map
      */
 
     /*@NotNull*/
-    public Expression copy(RebindingMap rebindings) {
+    public Expression copy(RebindingMap rm) {
         AnalyzeString a2 = new AnalyzeString(
-                copy(getSelect(), rebindings),
-                copy(getRegex(), rebindings),
-                copy(getFlags(), rebindings),
-                copy(getMatching(), rebindings),
-                copy(getNonMatching(), rebindings),
-                pattern);
+                copy(getSelect(), rm), copy(getRegex(), rm), copy(getFlags(), rm), copy(getMatching(), rm), copy(getNonMatching(), rm), pattern);
         ExpressionTool.copyLocationInfo(this, a2);
         return a2;
     }
@@ -351,25 +360,6 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
         return dependencies;
     }
 
-    /**
-     * Handle promotion offers, that is, non-local tree rewrites.
-     *
-     * @param offer The type of rewrite being offered
-     * @throws XPathException
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setSelect(doPromotion(getSelect(), offer));
-        setRegex(doPromotion(getRegex(), offer));
-        setFlags(doPromotion(getFlags(), offer));
-        if (getMatching() != null) {
-            setMatching(doPromotion(getMatching(), offer));
-        }
-        if (getNonMatching() != null) {
-            setNonMatching(doPromotion(getNonMatching(), offer));
-        }
-    }
-
 
     /**
      * ProcessLeavingTail: called to do the real work of this instruction. This method
@@ -422,9 +412,9 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
         RegularExpression re = pattern;
         if (re == null) {
             CharSequence flagstr = getFlags().evaluateAsString(context);
-            String dialect = getRetainedStaticContext().getXPathVersion() >= 30 ? "XP30" : "XP20";
+            String dialect = "XP30";
             re = Version.platform.compileRegularExpression(
-                getRegex().evaluateAsString(context), flagstr.toString(), dialect, null);
+                    context.getConfiguration(), getRegex().evaluateAsString(context), flagstr.toString(), dialect, null);
 
             if ((dialect.equals("XP20") || !useXsltErrorCodes) && re.matches("")) {
                 dynamicError("The regular expression must not be one that matches a zero-length string",
@@ -463,18 +453,17 @@ public class AnalyzeString extends Instruction implements ContextOriginator {
         return new ContextMappingIterator(fn, c2);
     }
 
-    //#ifdefined BYTECODE
-
     /**
-     * Return the compiler of the AnalyzeString expression
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
      *
-     * @return the relevant ExpressionCompiler
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in explain() output displaying the expression.
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new AnalyzeStringCompiler();
+    public String getExpressionName() {
+        return "analyzeString";
     }
-    //#endif
 
     /**
      * Diagnostic print of expression structure. The abstract expression tree
