@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,22 +7,16 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ApplyTemplatesCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.ApplyTemplatesAdjunct;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.om.FocusTrackingIterator;
-import net.sf.saxon.om.Sequence;
-import net.sf.saxon.om.SequenceIterator;
-import net.sf.saxon.om.StandardNames;
+import net.sf.saxon.om.*;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.*;
+import net.sf.saxon.trans.rules.RuleManager;
 import net.sf.saxon.tree.iter.EmptyIterator;
 
 import java.util.ArrayList;
 import java.util.List;
-import java.util.Map;
 
 
 /**
@@ -41,7 +35,7 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
     protected boolean implicitSelect;
     protected boolean inStreamableConstruct = false;
     protected RuleManager ruleManager;
-    private int bindingSlot; // for binding the mode
+    private int bindingSlot = -1; // for binding the mode
 
     protected ApplyTemplates() {
     }
@@ -246,7 +240,7 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
      * doing any further analysis to find out more precisely).
      */
 
-    public final boolean createsNewNodes() {
+    public final boolean mayCreateNewNodes() {
         return true;
     }
 
@@ -260,8 +254,8 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
 
     protected TailCall apply(XPathContext context, boolean returnTailCall) throws XPathException {
 
-        Component<? extends Mode> targetMode = getTargetMode(context);
-        Mode thisMode = targetMode.getCode();
+        Component.M targetMode = getTargetMode(context);
+        Mode thisMode = targetMode.getActor();
 
         // handle parameters if any
 
@@ -321,13 +315,13 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
      * @return the mode to be used, as a Component
      */
 
-    public Component<? extends Mode> getTargetMode(XPathContext context) {
-        Component<? extends Mode> targetMode;
+    public Component.M getTargetMode(XPathContext context) {
+        Component.M targetMode;
         if (useCurrentMode) {
             targetMode = context.getCurrentMode();
         } else {
             if (bindingSlot >= 0) {
-                targetMode = (Component<? extends Mode>)context.getTargetComponent(bindingSlot);
+                targetMode = (Component.M)context.getTargetComponent(bindingSlot);
                 if (targetMode.getVisibility() == Visibility.ABSTRACT) {
                     throw new AssertionError("Modes cannot be abstract");
                 }
@@ -409,19 +403,6 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
         return mode==null ? null : mode.getSymbolicName();
     }
 
-    /**
-     * Handle promotion offers, that is, non-local tree rewrites.
-     *
-     * @param offer The type of rewrite being offered
-     * @throws XPathException
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setSelect(doPromotion(getSelect(), offer));
-        WithParam.promoteParams(actualParams, offer);
-        WithParam.promoteParams(tunnelParams, offer);
-    }
-
     /**
      * Add a representation of this expression to a PathMap. The PathMap captures a map of the nodes visited
      * by an expression in a source tree.
@@ -545,14 +526,14 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
     protected static class ApplyTemplatesPackage implements TailCall {
 
         private Sequence selectedItems;
-        private Component<? extends Mode> targetMode;
+        private Component.M targetMode;
         private ParameterSet params;
         private ParameterSet tunnelParams;
         private XPathContextMajor evaluationContext;
         private Location locationId;
 
         ApplyTemplatesPackage(Sequence selectedItems,
-                              Component<? extends Mode> targetMode,
+                              Component.M targetMode,
                               ParameterSet params,
                               ParameterSet tunnelParams,
                               XPathContextMajor context,
@@ -569,29 +550,14 @@ public class ApplyTemplates extends Instruction implements ITemplateCall, Compon
             evaluationContext.setCurrentIterator(new FocusTrackingIterator(selectedItems.iterate()));
             evaluationContext.setCurrentMode(targetMode);
             evaluationContext.setCurrentComponent(targetMode);
-            return targetMode.getCode().applyTemplates(params, tunnelParams, evaluationContext, locationId);
+            return targetMode.getActor().applyTemplates(params, tunnelParams, evaluationContext, locationId);
         }
     }
 
-//#ifdefined BYTECODE
-    /**
-     * Return the compiler of the ApplyTemplates expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ApplyTemplatesCompiler();
+    public String getStreamerName() {
+        return "ApplyTemplates";
     }
-//#endif
-
-//#ifdefined STREAM
-    @Override
-    public ApplyTemplatesAdjunct getStreamingAdjunct() {
-        return new ApplyTemplatesAdjunct();
-    }
-//#endif
-
 
 }
 
