@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,14 +7,11 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.BlockCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.BlockAdjunct;
-import net.sf.saxon.evpull.BlockEventIterator;
-import net.sf.saxon.evpull.EmptyEventIterator;
-import net.sf.saxon.evpull.EventIterator;
 import net.sf.saxon.expr.*;
-import net.sf.saxon.expr.parser.*;
+import net.sf.saxon.expr.parser.ContextItemStaticInfo;
+import net.sf.saxon.expr.parser.ExpressionTool;
+import net.sf.saxon.expr.parser.ExpressionVisitor;
+import net.sf.saxon.expr.parser.RebindingMap;
 import net.sf.saxon.om.AxisInfo;
 import net.sf.saxon.om.Item;
 import net.sf.saxon.om.SequenceIterator;
@@ -31,7 +28,6 @@ import net.sf.saxon.value.SequenceExtent;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.List;
-import java.util.Map;
 
 
 /**
@@ -61,6 +57,11 @@ public class Block extends Instruction {
         }
     }
 
+    @Override
+    public boolean isInstruction() {
+        return false;
+    }
+
     /**
      * Get the n'th child expression
      * @param n the position of the child expression required (zero-based)
@@ -95,6 +96,26 @@ public class Block extends Instruction {
         return Arrays.asList(operanda);
     }
 
+    /**
+     * Ask whether this expression is, or contains, the binding of a given variable
+     *
+     * @param binding the variable binding
+     * @return true if this expression is the variable binding (for example a ForExpression
+     * or LetExpression) or if it is a FLWOR expression that binds the variable in one of its
+     * clauses.
+     */
+    @Override
+    public boolean hasVariableBinding(Binding binding) {
+        if (binding instanceof LocalParam) {
+            for (Operand o : operanda) {
+                if (o.getChildExpression() == binding) {
+                    return true;
+                }
+            }
+        }
+        return false;
+    }
+
     /**
      * Static factory method to create a block. If one of the arguments is already a block,
      * the contents will be merged into a new composite block
@@ -230,13 +251,13 @@ public class Block extends Instruction {
 
     /**
      * Determine whether the block includes any instructions that might return nodes with a type annotation
-     *
+     * @param insn the instruction (for example this block)
      * @param th the type hierarchy cache
      * @return true if any expression in the block can return type-annotated nodes
      */
 
-    private boolean mayReturnTypedNodes(TypeHierarchy th) {
-        for (Operand o : operands()) {
+    public static boolean mayReturnTypedNodes(Instruction insn, TypeHierarchy th) {
+        for (Operand o : insn.operands()) {
             Expression exp = o.getChildExpression();
             if ((exp.getSpecialProperties() & StaticProperty.ALL_NODES_UNTYPED) == 0) {
                 ItemType it = exp.getItemType();
@@ -328,15 +349,45 @@ public class Block extends Instruction {
         if (size() == 0) {
             return ErrorType.getInstance();
         }
-        ItemType t1 = child(0).getItemType();
+        ItemType t1 = null;
         TypeHierarchy th = getConfiguration().getTypeHierarchy();
-        for (int i = 1; i < size(); i++) {
-            t1 = Type.getCommonSuperType(t1, child(i).getItemType(), th);
-            if (t1 instanceof AnyItemType) {
-                return t1;  // no point going any further
+        for (int i = 0; i < size(); i++) {
+            Expression child = child(i);
+            if (!(child instanceof Message)) {
+                ItemType t = child.getItemType();
+                t1 = (t1 == null ? t : Type.getCommonSuperType(t1, t, th));
+                if (t1 instanceof AnyItemType) {
+                    return t1;  // no point going any further
+                }
             }
         }
-        return t1;
+        return t1 == null ? ErrorType.getInstance() : t1;
+    }
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (isInstruction()) {
+            return super.getStaticUType(contextItemType);
+        } else {
+            if (size() == 0) {
+                return UType.VOID;
+            }
+            UType t1 = child(0).getStaticUType(contextItemType);
+            for (int i = 1; i < size(); i++) {
+                t1 = t1.union(child(i).getStaticUType(contextItemType));
+                if (t1 == UType.ANY) {
+                    return t1;  // no point going any further
+                }
+            }
+            return t1;
+        }
     }
 
     /**
@@ -363,15 +414,8 @@ public class Block extends Instruction {
      * returns true.
      */
 
-    public final boolean createsNewNodes() {
-        for (Operand o : operands()) {
-            Expression child = o.getChildExpression();
-            int props = child.getSpecialProperties();
-            if ((props & StaticProperty.NON_CREATIVE) == 0) {
-                return true;
-            }
-        }
-        return false;
+    public final boolean mayCreateNewNodes() {
+        return someOperandCreatesNewNodes();
     }
 
 
@@ -554,7 +598,7 @@ public class Block extends Instruction {
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         typeCheckChildren(visitor, contextInfo);
-        if (!mayReturnTypedNodes(getConfiguration().getTypeHierarchy())) {
+        if (!mayReturnTypedNodes(this, visitor.getConfiguration().getTypeHierarchy())) {
             resetLocalStaticProperties();
             allNodesUntyped = true;
         }
@@ -600,19 +644,6 @@ public class Block extends Instruction {
     }
 
 
-    /**
-     * Handle promotion offers, that is, non-local tree rewrites.
-     *
-     * @param offer The type of rewrite being offered
-     * @throws XPathException
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        for (int c = 0; c < size(); c++) {
-            setChild(c, doPromotion(child(c), offer));
-        }
-    }
-
     /**
      * Check that any elements and attributes constructed or returned by this expression are acceptable
      * in the content model of a given complex type. It's always OK to say yes, since the check will be
@@ -692,23 +723,6 @@ public class Block extends Instruction {
         }
     }
 
-    /**
-     * Get an EventIterator over the results of all the child expressions
-     *
-     * @param context the XPath dynamic context
-     * @return an EventIterator
-     */
-
-    public EventIterator iterateEvents(XPathContext context) throws XPathException {
-        if (size() == 0) {
-            return EmptyEventIterator.getInstance();
-        } else if (size() == 1) {
-            return child(0).iterateEvents(context);
-        } else {
-            return new BlockEventIterator(operanda, context);
-        }
-    }
-
 
     /**
      * Evaluate an updating expression, adding the results to a Pending Update List.
@@ -726,29 +740,9 @@ public class Block extends Instruction {
         }
     }
 
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Block expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new BlockCompiler();
-    }
-    //#endif
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public BlockAdjunct getStreamingAdjunct() {
-        return new BlockAdjunct();
+    public String getStreamerName() {
+        return "Block";
     }
 
-    //#endif
 }
\ No newline at end of file
