@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,11 +7,6 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ChooseCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.ChooseAdjunct;
-import net.sf.saxon.evpull.EmptyEventIterator;
-import net.sf.saxon.evpull.EventIterator;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.functions.BooleanFn;
@@ -32,7 +27,6 @@ import net.sf.saxon.value.SequenceType;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.List;
-import java.util.Map;
 
 
 /**
@@ -44,6 +38,7 @@ public class Choose extends Instruction {
 
     private Operand[] conditionOps;
     private Operand[] actionOps;
+    private boolean isInstruction;
 
 
     // The class implements both xsl:choose and xsl:if. There is a list of boolean
@@ -112,6 +107,29 @@ public class Choose extends Instruction {
         return new Choose(conditions, actions);
     }
 
+    /**
+     * Say whether this choose expression originates as an XSLT instruction
+     *
+     * @param inst true if this is an xsl:choose or xsl:if instruction
+     */
+
+    public void setInstruction(boolean inst) {
+        isInstruction = inst;
+    }
+
+    /**
+     * Ask whether this expression is an instruction. In XSLT streamability analysis this
+     * is used to distinguish constructs corresponding to XSLT instructions from other constructs.
+     *
+     * @return true if this construct originates as an XSLT instruction
+     */
+
+    @Override
+    public boolean isInstruction() {
+        return isInstruction;
+    }
+
+
     public int size() {
         return conditionOps.length;
     }
@@ -184,6 +202,19 @@ public class Choose extends Instruction {
         return operanda;
     }
 
+    /**
+     * Ask whether common subexpressions found in the operands of this expression can
+     * be extracted and evaluated outside the expression itself. The result is irrelevant
+     * in the case of operands evaluated with a different focus, which will never be
+     * extracted in this way, even if they have no focus dependency.
+     *
+     * @return false for this kind of expression
+     */
+    @Override
+    public boolean allowExtractingCommonSubexpressions() {
+        return false;
+    }
+
     /**
      * Atomize all the action expressions
      */
@@ -235,28 +266,36 @@ public class Choose extends Instruction {
     }
 
     private Expression removeRedundantBranches(ExpressionVisitor visitor) {
+        Expression result = removeRedundantBranches0(visitor);
+        if (result != this) {
+            ExpressionTool.copyLocationInfo(this, result);
+        }
+        return result;
+    }
+
+    private Expression removeRedundantBranches0(ExpressionVisitor visitor) {
         // Eliminate a redundant if (false)
 
         boolean compress = false;
         for (int i = 0; i < size(); i++) {
             Expression condition = getCondition(i);
-            if (Literal.isConstantBoolean(condition, false) ||
-                    (i < size()-1 && Literal.isConstantBoolean(condition, true))) {
+            if (condition instanceof Literal) {
                 compress = true;
+                break;
             }
         }
-
         int size = size();
+        boolean changed = false;
         if (compress) {
             List<Expression> conditions = new ArrayList<Expression>(size);
             List<Expression> actions = new ArrayList<Expression>(size);
             for (int i = 0; i < size; i++) {
                 Expression condition = getCondition(i);
-                if (!Literal.isConstantBoolean(condition, false)) {
+                if (!Literal.hasEffectiveBooleanValue(condition, false)) {
                     conditions.add(condition);
                     actions.add(getAction(i));
                 }
-                if (Literal.isConstantBoolean(condition, true)) {
+                if (Literal.hasEffectiveBooleanValue(condition, true)) {
                     break;
                 }
             }
@@ -264,9 +303,9 @@ public class Choose extends Instruction {
                 Literal lit = Literal.makeEmptySequence();
                 ExpressionTool.copyLocationInfo(this, lit);
                 return lit;
-            } else if (conditions.size() == 1 && Literal.isConstantBoolean(conditions.get(0), true)) {
+            } else if (conditions.size() == 1 && Literal.hasEffectiveBooleanValue(conditions.get(0), true)) {
                 return actions.get(0);
-            } else {
+            } else if (conditions.size() != size) {
                 Expression[] c = conditions.toArray(new Expression[conditions.size()]);
                 Expression[] a = actions.toArray(new Expression[actions.size()]);
                 Choose result = new Choose(c, a);
@@ -277,7 +316,7 @@ public class Choose extends Instruction {
 
         // See if only condition left is: if (true) then x else ()
 
-        if (size() == 1 && Literal.isConstantBoolean(getCondition(0), true)) {
+        if (size() == 1 && Literal.hasEffectiveBooleanValue(getCondition(0), true)) {
             return getAction(0);
         }
 
@@ -285,7 +324,9 @@ public class Choose extends Instruction {
 
         if (Literal.isEmptySequence(getAction(size() - 1))) {
             if (size() == 1) {
-                return Literal.makeEmptySequence();
+                Literal lit = Literal.makeEmptySequence();
+                ExpressionTool.copyLocationInfo(this, lit);
+                return lit;
             } else {
                 Expression[] conditions = new Expression[size-1];
                 Expression[] actions = new Expression[size-1];
@@ -299,7 +340,7 @@ public class Choose extends Instruction {
 
         // Flatten an "else if"
 
-        if (Literal.isConstantBoolean(getCondition(size - 1), true) &&
+        if (Literal.hasEffectiveBooleanValue(getCondition(size - 1), true) &&
                 getAction(size - 1) instanceof Choose) {
             Choose choose2 = (Choose) getAction(size - 1);
             int newLen = size + choose2.size() - 1;
@@ -321,8 +362,8 @@ public class Choose extends Instruction {
         if (size == 2 &&
                 Literal.isConstantBoolean(getAction(0), true) &&
                 Literal.isConstantBoolean(getAction(1), false) &&
-                Literal.isConstantBoolean(getCondition(1), true)) {
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
+                Literal.hasEffectiveBooleanValue(getCondition(1), true)) {
+            TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
             if (th.isSubType(getCondition(0).getItemType(), BuiltInAtomicType.BOOLEAN) &&
                     getCondition(0).getCardinality() == StaticProperty.EXACTLY_ONE) {
                 return getCondition(0);
@@ -335,9 +376,10 @@ public class Choose extends Instruction {
 
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
+        TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
         for (int i = 0; i < size(); i++) {
             conditionOps[i].typeCheck(visitor, contextInfo);
-            XPathException err = TypeChecker.ebvError(getCondition(i), getConfiguration().getTypeHierarchy());
+            XPathException err = TypeChecker.ebvError(getCondition(i), th);
             if (err != null) {
                 err.setLocator(getCondition(i).getLocation());
                 throw err;
@@ -360,7 +402,7 @@ public class Choose extends Instruction {
                     // be satisfied.
                     // Also, don't throw a type error if the branch will never be executed; this can happen with
                     // a typeswitch where the purpose of the condition is to test the type.
-                    if (Literal.isEmptySequence(getAction(i)) || Literal.isConstantBoolean(getCondition(i), false)) {
+                    if (Literal.isEmptySequence(getAction(i)) || Literal.hasEffectiveBooleanValue(getCondition(i), false)) {
                         setAction(i, new ErrorExpression(err));
                     } else {
                         throw err;
@@ -407,9 +449,10 @@ public class Choose extends Instruction {
                                       RoleDiagnostic role, ExpressionVisitor visitor)
             throws XPathException {
         int size = size();
+        TypeChecker tc = getConfiguration().getTypeChecker(backwardsCompatible);
         for (int i = 0; i < size; i++) {
             try {
-                setAction(i, TypeChecker.staticTypeCheck(getAction(i), req, backwardsCompatible, role, visitor));
+                setAction(i, tc.staticTypeCheck(getAction(i), req, role, visitor));
             } catch (XPathException err) {
                 if (err.isStaticError()) {
                     throw err;
@@ -424,7 +467,7 @@ public class Choose extends Instruction {
         }
         // If the last condition isn't true(), then we need to consider the fall-through case, which returns
         // an empty sequence
-        if (!Literal.isConstantBoolean(getCondition(size - 1), true) &&
+        if (!Literal.hasEffectiveBooleanValue(getCondition(size - 1), true) &&
                 !Cardinality.allowsZero(req.getCardinality())) {
             Expression[] c = new Expression[size + 1];
             Expression[] a = new Expression[size + 1];
@@ -478,12 +521,13 @@ public class Choose extends Instruction {
                 }
             }
         }
+
         if (size == 0) {
             return Literal.makeEmptySequence();
         }
         Expression e = removeRedundantBranches(visitor);
         if (e instanceof Choose) {
-            return getConfiguration().obtainOptimizer().trySwitch((Choose) e);
+            return visitor.getConfiguration().obtainOptimizer().trySwitch((Choose) e, visitor);
         } else {
             return e;
         }
@@ -507,6 +551,7 @@ public class Choose extends Instruction {
         }
         Choose ch2 = new Choose(c2, a2);
         ExpressionTool.copyLocationInfo(this, ch2);
+        ch2.setInstruction(isInstruction());
         return ch2;
     }
 
@@ -633,6 +678,26 @@ public class Choose extends Instruction {
         return type;
     }
 
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (isInstruction()) {
+            return super.getStaticUType(contextItemType);
+        } else {
+            UType type = getAction(0).getStaticUType(contextItemType);
+            for (int i = 1; i < size(); i++) {
+                type = type.union(getAction(i).getStaticUType(contextItemType));
+            }
+            return type;
+        }
+    }
+
     /**
      * Compute the cardinality of the sequence returned by evaluating this instruction
      *
@@ -644,7 +709,7 @@ public class Choose extends Instruction {
         boolean includesTrue = false;
         for (int i = 0; i < size(); i++) {
             card = Cardinality.union(card, getAction(i).getCardinality());
-            if (Literal.isConstantBoolean(getCondition(i), true)) {
+            if (Literal.hasEffectiveBooleanValue(getCondition(i), true)) {
                 includesTrue = true;
             }
         }
@@ -678,7 +743,7 @@ public class Choose extends Instruction {
      * (Nodes created by the conditions can't contribute to the result).
      */
 
-    public final boolean createsNewNodes() {
+    public final boolean mayCreateNewNodes() {
         for (Operand action : actions()) {
             int props = action.getChildExpression().getSpecialProperties();
             if ((props & StaticProperty.NON_CREATIVE) == 0) {
@@ -688,30 +753,6 @@ public class Choose extends Instruction {
         return false;
     }
 
-    /**
-     * Handle promotion offers, that is, non-local tree rewrites.
-     *
-     * @param offer The type of rewrite being offered
-     * @throws XPathException
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        // xsl:when acts as a guard: expressions inside the when mustn't be evaluated if the when is false,
-        // and conditions after the first mustn't be evaluated if a previous condition is true. So we
-        // don't pass all promotion offers on
-        if (offer.action == PromotionOffer.EXTRACT_GLOBAL_VARIABLES) {
-            for (int i = 0; i < size(); i++) {
-                setCondition(i, doPromotion(getCondition(i), offer));
-            }
-            for (int i = 0; i < size(); i++) {
-                setAction(i, doPromotion(getAction(i), offer));
-            }
-        } else {
-            // in other cases, only the first xsl:when condition is promoted
-            setCondition(0, doPromotion(getCondition(0), offer));
-        }
-    }
-
     /**
      * Replace this expression by a simpler expression that delivers the results without regard
      * to order.
@@ -913,35 +954,6 @@ public class Choose extends Instruction {
     }
 
 
-    /**
-     * Deliver the result of the expression as a sequence of events.
-     * <p/>
-     * <p>The events (of class {@link net.sf.saxon.evpull.PullEvent}) are either complete
-     * items, or one of startElement, endElement, startDocument, or endDocument, known
-     * as semi-nodes. The stream of events may also include a nested EventIterator.
-     * If a start-end pair exists in the sequence, then the events between
-     * this pair represent the content of the document or element. The content sequence will
-     * have been processed to the extent that any attribute and namespace nodes in the
-     * content sequence will have been merged into the startElement event. Namespace fixup
-     * will have been performed: that is, unique prefixes will have been allocated to element
-     * and attribute nodes, and all namespaces will be declared by means of a namespace node
-     * in the startElement event or in an outer startElement forming part of the sequence.
-     * However, duplicate namespaces may appear in the sequence.</p>
-     * <p>The content of an element or document may include adjacent or zero-length text nodes,
-     * atomic values, and nodes represented as nodes rather than broken down into events.</p>
-     *
-     * @param context The dynamic evaluation context
-     * @return the result of the expression as an iterator over a sequence of PullEvent objects
-     * @throws net.sf.saxon.trans.XPathException
-     *          if a dynamic error occurs during expression evaluation
-     */
-
-    public EventIterator iterateEvents(XPathContext context) throws XPathException {
-        int i = choose(context);
-        return i < 0 ? EmptyEventIterator.getInstance() : getAction(i).iterateEvents(context);
-    }
-
-
     /**
      * Evaluate an updating expression, adding the results to a Pending Update List.
      * The default implementation of this method, which is used for non-updating expressions,
@@ -958,34 +970,21 @@ public class Choose extends Instruction {
         }
     }
 
-//#ifdefined BYTECODE
-
     /**
-     * Return the compiler of the Choose expression
+     * Get a name identifying the kind of expression, in terms meaningful to a user.
      *
-     * @return the relevant ExpressionCompiler
+     * @return a name identifying the kind of expression, in terms meaningful to a user.
+     * The name will always be in the form of a lexical XML QName, and should match the name used
+     * in explain() output displaying the expression.
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ChooseCompiler();
+    public String getExpressionName() {
+        return "choose";
     }
-//#endif
 
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
     @Override
-    public ChooseAdjunct getStreamingAdjunct() {
-        return new ChooseAdjunct();
+    public String getStreamerName() {
+        return "Choose";
     }
-
-
-//#endif
-
-
 }
 
