@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,8 +7,7 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ComputedAttributeCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
+import net.sf.saxon.Configuration;
 import net.sf.saxon.event.ReceiverOptions;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
@@ -25,8 +24,6 @@ import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
 import net.sf.saxon.value.StringValue;
 
-import java.util.Map;
-
 /**
  * An instruction derived from an xsl:attribute element in stylesheet, or from
  * an attribute constructor in XQuery, in cases where the attribute name is not known
@@ -183,12 +180,13 @@ public final class ComputedAttribute extends AttributeCreator {
         nameOp.typeCheck(visitor, contextItemType);
 
         RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "attribute/name", 0);
+        Configuration config = visitor.getConfiguration();
+        TypeHierarchy th = config.getTypeHierarchy();
 
         if (allowNameAsQName) {
             // Can only happen in XQuery
-            setNameExp(TypeChecker.staticTypeCheck(getNameExp(),
-                    SequenceType.SINGLE_ATOMIC, false, role, visitor));
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
+            setNameExp(config.getTypeChecker(false).staticTypeCheck(getNameExp(),
+                                                                                        SequenceType.SINGLE_ATOMIC, role, visitor));
             ItemType nameItemType = getNameExp().getItemType();
             boolean maybeString = th.relationship(nameItemType, BuiltInAtomicType.STRING) != TypeHierarchy.DISJOINT ||
                     th.relationship(nameItemType, BuiltInAtomicType.UNTYPED_ATOMIC) != TypeHierarchy.DISJOINT;
@@ -202,7 +200,6 @@ public final class ComputedAttribute extends AttributeCreator {
                 throw err;
             }
         } else {
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
             if (!th.isSubType(getNameExp().getItemType(), BuiltInAtomicType.STRING)) {
                 setNameExp(SystemFunction.makeCall("string", getRetainedStaticContext(), getNameExp()));
             }
@@ -213,7 +210,7 @@ public final class ComputedAttribute extends AttributeCreator {
 
 //            Deleted MHK Aug 2014 because the namespace expression is an AVT and therefore necessarily a string
 //            role = new RoleLocator(RoleLocator.INSTRUCTION, "attribute/namespace", 0);
-//            namespace = TypeChecker.staticTypeCheck(
+//            namespace = config.getTypeChecker().staticTypeCheck(
 //                    namespace, SequenceType.SINGLE_STRING, false, role, visitor);
         }
 
@@ -224,7 +221,7 @@ public final class ComputedAttribute extends AttributeCreator {
                 if (val instanceof StringValue) {
                     String[] parts = NameChecker.checkQNameParts(val.getStringValueCS());
                     if (getNamespaceExp() == null) {
-                        String uri = getNamespaceResolver().getURIForPrefix(parts[0], true);
+                        String uri = getNamespaceResolver().getURIForPrefix(parts[0], false);
                         if (uri == null) {
                             XPathException se = new XPathException("Prefix " + parts[0] + " has not been declared");
                             if (isXSLT()) {
@@ -284,31 +281,11 @@ public final class ComputedAttribute extends AttributeCreator {
                 getRetainedStaticContext(), getValidationAction(), getSchemaType(), allowNameAsQName);
         ExpressionTool.copyLocationInfo(this, exp);
         exp.setSelect(getSelect().copy(rebindings));
+        exp.setInstruction(isInstruction());
         return exp;
     }
 
 
-    /**
-     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is not accepted - this is appropriate in the case of simple expressions
-     * such as constant values and variable references where promotion would give no performance
-     * advantage. This method is always called at compile time.
-     *
-     * @param offer details of the offer, for example the offer to move
-     *              expressions that don't depend on the context to an outer level in
-     *              the containing expression
-     * @throws XPathException if any error is detected
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setNameExp(doPromotion(getNameExp(), offer));
-        if (getNamespaceExp() != null) {
-            setNamespace(doPromotion(getNamespaceExp(), offer));
-        }
-        super.promoteChildren(offer);
-    }
-
     /**
      * Check that any elements and attributes constructed or returned by this expression are acceptable
      * in the content model of a given complex type. It's always OK to say yes, since the check will be
@@ -484,19 +461,6 @@ public final class ComputedAttribute extends AttributeCreator {
         return node;
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the ComputedAttribute expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ComputedAttributeCompiler();
-    }
-//#endif
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -511,6 +475,13 @@ public final class ComputedAttribute extends AttributeCreator {
         if (type != null) {
             out.emitAttribute("type", type.getStructuredQName().getEQName());
         }
+        String flags = "";
+        if (isLocal()) {
+            flags += "l";
+        }
+        if (!flags.isEmpty()) {
+            out.emitAttribute("flags", flags);
+        }
         out.setChildRole("name");
         getNameExp().export(out);
         if (getNamespaceExp() != null) {
