@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,12 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ComputedElementCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.event.Receiver;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.functions.SystemFunction;
-import net.sf.saxon.lib.FeatureKeys;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.lib.StandardURIChecker;
 import net.sf.saxon.lib.Validation;
@@ -26,9 +23,6 @@ import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
-import net.sf.saxon.value.StringValue;
-
-import java.util.Map;
 
 
 /**
@@ -131,7 +125,6 @@ public class ComputedElement extends ElementCreator {
             setNamespaceExp(getNamespaceExp().simplify());
         }
         Configuration config = getConfiguration();
-        setLazyConstruction(config.getBooleanProperty(FeatureKeys.LAZY_CONSTRUCTION_MODE));
         boolean schemaAware = getPackageData().isSchemaAware();
         preservingTypes |= !schemaAware;
 
@@ -152,12 +145,14 @@ public class ComputedElement extends ElementCreator {
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         super.typeCheck(visitor, contextInfo);
+        Configuration config = visitor.getConfiguration();
+        TypeHierarchy th = config.getTypeHierarchy();
+
         RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "element/name", 0);
         if (allowNameAsQName) {
             // Can only happen in XQuery
-            setNameExp(TypeChecker.staticTypeCheck(getNameExp(),
-                    SequenceType.SINGLE_ATOMIC, false, role, visitor));
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
+            setNameExp(config.getTypeChecker(false).staticTypeCheck(getNameExp(),
+                                                                                SequenceType.SINGLE_ATOMIC, role, visitor));
             ItemType supplied = getNameExp().getItemType();
             if (th.relationship(supplied, BuiltInAtomicType.STRING) == TypeHierarchy.DISJOINT &&
                     th.relationship(supplied, BuiltInAtomicType.UNTYPED_ATOMIC) == TypeHierarchy.DISJOINT &&
@@ -169,7 +164,6 @@ public class ComputedElement extends ElementCreator {
                 throw de;
             }
         } else {
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
             if (!th.isSubType(getNameExp().getItemType(), BuiltInAtomicType.STRING)) {
                 setNameExp(SystemFunction.makeCall("string", getRetainedStaticContext(), getNameExp()));
             }
@@ -240,28 +234,6 @@ public class ComputedElement extends ElementCreator {
     }
 
 
-    /**
-     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is not accepted - this is appropriate in the case of simple expressions
-     * such as constant values and variable references where promotion would give no performance
-     * advantage. This method is always called at compile time.
-     *
-     * @param offer details of the offer, for example the offer to move
-     *              expressions that don't depend on the context to an outer level in
-     *              the containing expression
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any error is detected
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setNameExp(doPromotion(getNameExp(), offer));
-        if (getNamespaceExp() != null) {
-            setNamespaceExp(doPromotion(getNamespaceExp(), offer));
-        }
-        super.promoteChildren(offer);
-    }
-
     /**
      * Check that any elements and attributes constructed or returned by this expression are acceptable
      * in the content model of a given complex type. It's always OK to say yes, since the check will be
@@ -440,19 +412,6 @@ public class ComputedElement extends ElementCreator {
         return StandardNames.XSL_ELEMENT;
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the ComputedElement expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ComputedElementCompiler();
-    }
-//#endif
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -461,6 +420,9 @@ public class ComputedElement extends ElementCreator {
     public void export(ExpressionPresenter out) throws XPathException {
         out.startElement("compElem", this);
         String flags = getInheritanceFlags();
+        if (isLocal()) {
+            flags += "l";
+        }
         if (!flags.isEmpty()) {
             out.emitAttribute("flags", flags);
         }
