@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,14 +7,14 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.CopyCompiler;
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.CopyAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.event.*;
 import net.sf.saxon.expr.*;
-import net.sf.saxon.expr.parser.*;
+import net.sf.saxon.expr.parser.ContextItemStaticInfo;
+import net.sf.saxon.expr.parser.ExpressionTool;
+import net.sf.saxon.expr.parser.ExpressionVisitor;
+import net.sf.saxon.expr.parser.RebindingMap;
 import net.sf.saxon.lib.ParseOptions;
 import net.sf.saxon.lib.Validation;
 import net.sf.saxon.om.*;
@@ -25,7 +25,6 @@ import net.sf.saxon.tree.util.NamespaceIterator;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.BooleanValue;
 import net.sf.saxon.value.SequenceType;
-import net.sf.saxon.value.Whitespace;
 
 import java.util.Iterator;
 
@@ -61,11 +60,6 @@ public class Copy extends ElementCreator {
         this.inheritNamespacesToChildren = inheritNamespaces;
         setValidationAction(validation, schemaType);
         preservingTypes = schemaType == null && validation == Validation.PRESERVE;
-        if (copyNamespaces) {
-            setLazyConstruction(false);
-            // can't do lazy construction at present in cases where namespaces need to be copied from the
-            // source document.
-        }
     }
 
     /**
@@ -138,7 +132,6 @@ public class Copy extends ElementCreator {
             this.resultItemType = selectItemType;
         }
 
-        verifyLazyConstruction();
         checkContentSequence(visitor.getStaticContext());
         return this;
     }
@@ -198,15 +191,6 @@ public class Copy extends ElementCreator {
         return contentOp;
     }
 
-    private final static OperandRole FOCUS_CONTROLLING_SELECT =
-            new OperandRole(OperandRole.SETS_NEW_FOCUS, OperandUsage.INSPECTION, SequenceType.OPTIONAL_ITEM);
-    public final static OperandRole FOCUS_CONTROLLED_CONTENT =
-            new OperandRole(OperandRole.USES_NEW_FOCUS, OperandUsage.ABSORPTION, SequenceType.ANY_SEQUENCE);
-    public final static OperandRole SAME_FOCUS_CONTENT =
-            new OperandRole(0, OperandUsage.ABSORPTION, SequenceType.ANY_SEQUENCE);
-
-
-
 
     /**
      * Get the item type of the result of this instruction.
@@ -231,7 +215,7 @@ public class Copy extends ElementCreator {
             return selectItemType;
         }
         // The rest of the code handles the complications of schema-awareness
-        Configuration config = getConfiguration();
+        Configuration config = th.getConfiguration();
         if (getSchemaType() != null) {
             int e = th.relationship(selectItemType, NodeKindTest.ELEMENT);
             if (e == TypeHierarchy.SAME_TYPE || e == TypeHierarchy.SUBSUMED_BY) {
@@ -323,7 +307,7 @@ public class Copy extends ElementCreator {
         Expression exp = super.optimize(visitor, contextItemType);
         if (exp == this) {
             if (resultItemType == null) {
-                resultItemType = computeItemType(getConfiguration().getTypeHierarchy());
+                resultItemType = computeItemType(visitor.getConfiguration().getTypeHierarchy());
             }
             if (visitor.isOptimizeForStreaming()) {
                 UType type = contextItemType.getItemType().getUType();
@@ -346,10 +330,6 @@ public class Copy extends ElementCreator {
         return exp;
     }
 
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setContentExpression(doPromotion(getContentExpression(), offer));
-    }
-
     /**
      * Callback from ElementCreator when constructing an element
      *
@@ -477,10 +457,10 @@ public class Copy extends ElementCreator {
                 boolean pop = false;
                 if (!preservingTypes) {
                     ParseOptions options = new ParseOptions(getValidationOptions());
-                    options.setStripSpace(Whitespace.NONE);
+                    options.setSpaceStrippingRule(NoElementsSpaceStrippingRule.getInstance());
                     controller.getConfiguration().prepareValidationReporting(context, options);
                     Receiver val = controller.getConfiguration().
-                        getDocumentValidator(out, source.getBaseURI(), options);
+                        getDocumentValidator(out, source.getBaseURI(), options, getLocation());
                     if (val != out) {
                         SequenceReceiver sr = new TreeReceiver(val);
                         sr.setPipelineConfiguration(out.getPipelineConfiguration());
@@ -524,7 +504,7 @@ public class Copy extends ElementCreator {
         Controller controller = context.getController();
         SequenceReceiver saved = context.getReceiver();
         SequenceOutputter seq = controller.allocateSequenceOutputter(1);
-        seq.getPipelineConfiguration().setHostLanguage(getHostLanguage());
+        seq.getPipelineConfiguration().setHostLanguage(getPackageData().getHostLanguage());
         seq.getPipelineConfiguration().setLocationIsCodeLocation(true);
         context.setReceiver(seq);
         process(context);
@@ -535,29 +515,17 @@ public class Copy extends ElementCreator {
         return item;
     }
 
-//#ifdefined STREAM
-
     /**
-     * Get a class that supports streamed evaluation of this expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public CopyAdjunct getStreamingAdjunct() {
-        return new CopyAdjunct();
+    public String getStreamerName() {
+        return "Copy";
     }
 
-    /**
-     * Return the compiler of the Copy expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new CopyCompiler();
-    }
-//#endif
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -576,6 +544,9 @@ public class Copy extends ElementCreator {
         if (inheritNamespacesFromParent) {
             flags += "n";
         }
+        if (isLocal()) {
+            flags += "l";
+        }
         out.emitAttribute("flags", flags);
         out.setChildRole("content");
         getContentExpression().export(out);
