@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,15 +7,11 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.InterpretedExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.CopyOfAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.event.*;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.expr.sort.DocumentSorter;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.lib.ParseOptions;
 import net.sf.saxon.lib.Validation;
@@ -26,16 +22,16 @@ import net.sf.saxon.pattern.NodeKindTest;
 import net.sf.saxon.pattern.NodeTest;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
+import net.sf.saxon.tree.tiny.TinyBuilder;
+import net.sf.saxon.tree.tiny.TinyNodeImpl;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.tree.wrapper.VirtualCopy;
 import net.sf.saxon.tree.wrapper.VirtualUntypedCopy;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.SequenceType;
-import net.sf.saxon.value.Whitespace;
 
 import java.net.URI;
 import java.net.URISyntaxException;
-import java.util.Map;
 
 
 /**
@@ -51,7 +47,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
     private SchemaType schemaType;
     private boolean requireDocumentOrElement = false;
     private boolean rejectDuplicateAttributes;
-    private boolean readOnce = false;
+    //private boolean readOnce = false;
     private boolean validating;
     private boolean copyLineNumbers = true;
     private boolean copyForUpdate = false;
@@ -140,16 +136,6 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         this.isSchemaAware = schemaAware;
     }
 
-    /**
-     * Set the "saxon:read-once" optimization mode
-     *
-     * @param b true to enable the optimization
-     */
-
-    public void setReadOnce(boolean b) {
-        readOnce = b;
-    }
-
     /**
      * Set whether line numbers are to be copied from the source to the result.
      * Default is false.
@@ -166,7 +152,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
      * The result depends on the type of the select expression.
      */
 
-    public final boolean createsNewNodes() {
+    public final boolean mayCreateNewNodes() {
         return !getSelect().getItemType().isPlainType();
     }
 
@@ -191,7 +177,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
      * Test whether this expression requires a document or element node
      *
      * @return true if this expression requires the value of the argument to be a document or element node,
-     *         false if there is no such requirement
+     * false if there is no such requirement
      */
 
     public boolean isDocumentOrElementRequired() {
@@ -240,6 +226,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
 
     /**
      * Say whether accumulator values should be copied from the source document
+     *
      * @param copy true if values should be copied
      */
 
@@ -249,6 +236,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
 
     /**
      * Ask whether accumulator values should be copied from the source document
+     *
      * @return true if values should be copied
      */
 
@@ -260,8 +248,8 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
     /**
      * Copy an expression. This makes a deep copy.
      *
+     * @param rebindings information about variables whose bindings need to be replaced
      * @return the copy of the original expression
-     * @param rebindings
      */
 
     /*@NotNull*/
@@ -269,7 +257,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         CopyOf c = new CopyOf(getSelect().copy(rebindings), copyNamespaces, validation, schemaType, rejectDuplicateAttributes);
         c.setCopyForUpdate(copyForUpdate);
         c.setCopyLineNumbers(copyLineNumbers);
-        c.setReadOnce(readOnce);
+        //c.setReadOnce(readOnce);
         c.isSchemaAware = isSchemaAware;
         c.setCopyAccumulators(copyAccumulators);
         return c;
@@ -308,7 +296,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
                     }
                     if (e != TypeHierarchy.DISJOINT || a != TypeHierarchy.DISJOINT) {
                         // it might be an element or attribute
-                        return AnyNodeTest.getInstance();
+                        return in instanceof NodeTest ? AnyNodeTest.getInstance() : AnyItemType.getInstance();
                     } else {
                         // it can't be an element or attribute, so stripping type annotations can't affect it
                         return in;
@@ -377,10 +365,6 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         return getSelect().getDependencies();
     }
 
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setSelect(doPromotion(getSelect(), offer));
-    }
-
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         typeCheckChildren(visitor, contextInfo);
@@ -388,9 +372,11 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
             // this implies the expression is actually an XQuery validate{} expression, hence the error messages
             RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.TYPE_OP, "validate", 0);
             role.setErrorCode("XQTY0030");
-            setSelect(TypeChecker.staticTypeCheck(getSelect(), SequenceType.SINGLE_NODE, false, role, visitor));
+            Configuration config = visitor.getConfiguration();
+            setSelect(config.getTypeChecker(false).staticTypeCheck(
+                    getSelect(), SequenceType.SINGLE_NODE, role, visitor));
 
-            TypeHierarchy th = getConfiguration().getTypeHierarchy();
+            TypeHierarchy th = config.getTypeHierarchy();
             ItemType t = getSelect().getItemType();
             if (th.isSubType(t, NodeKindTest.ATTRIBUTE)) {
                 throw new XPathException("validate{} expression cannot be applied to an attribute", "XQTY0030");
@@ -413,26 +399,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
 
     /*@NotNull*/
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType) throws XPathException {
-        if (readOnce) {
-            Expression optcopy;
-            if (!(getSelect() instanceof DocumentSorter)) {
-                optcopy = getConfiguration().obtainOptimizer().optimizeCopy(visitor, contextItemType, getSelect());
-                if (optcopy != null) {
-                    optcopy = optcopy.typeCheck(visitor, contextItemType);
-                    ExpressionTool.copyLocationInfo(this, optcopy);
-                    return optcopy;
-                }
-            }
-            selectOp.optimize(visitor, contextItemType);
-            optcopy = getConfiguration().obtainOptimizer().optimizeCopy(visitor, contextItemType, getSelect());
-            if (optcopy != null) {
-                ExpressionTool.copyLocationInfo(this, optcopy);
-                return optcopy;
-            }
-        } else {
-            selectOp.optimize(visitor, contextItemType);
-        }
-
+        selectOp.optimize(visitor, contextItemType);
         if (Literal.isEmptySequence(getSelect())) {
             return getSelect();
         }
@@ -464,9 +431,6 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         if (rejectDuplicateAttributes) {
             fsb.append('a');
         }
-        if (readOnce) {
-            fsb.append('o');
-        }
         if (validating) {
             fsb.append('v');
         }
@@ -485,27 +449,24 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         if (copyAccumulators) {
             fsb.append('m');
         }
-        if (fsb.length() > 0) {
+        if (!fsb.isEmpty()) {
             out.emitAttribute("flags", fsb.toString());
         }
         getSelect().export(out);
         out.endElement();
     }
 
-    //#ifdefined BYTECODE
-
     /**
-     * Get a class that supports streamed evaluation of this expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public CopyOfAdjunct getStreamingAdjunct() {
-        return new CopyOfAdjunct();
+    public String getStreamerName() {
+        return "CopyOf";
     }
 
-    //#endif
-
     /**
      * Add a representation of this expression to a PathMap. The PathMap captures a map of the nodes visited
      * by an expression in a source tree.
@@ -522,9 +483,9 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
      * @param pathMap        the PathMap to which the expression should be added
      * @param pathMapNodeSet the PathMapNodeSet to which the paths embodied in this expression should be added
      * @return the pathMapNodeSet representing the points in the source document that are both reachable by this
-     *         expression, and that represent possible results of this expression. For an expression that does
-     *         navigation, it represents the end of the arc in the path map that describes the navigation route. For other
-     *         expressions, it is the same as the input pathMapNode.
+     * expression, and that represent possible results of this expression. For an expression that does
+     * navigation, it represents the end of the arc in the path map that describes the navigation route. For other
+     * expressions, it is the same as the input pathMapNode.
      */
 
     public PathMap.PathMapNodeSet addToPathMap(PathMap pathMap, PathMap.PathMapNodeSet pathMapNodeSet) {
@@ -548,155 +509,215 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
 
     public TailCall processLeavingTail(XPathContext context) throws XPathException {
 
-        Controller controller = context.getController();
         SequenceReceiver out = context.getReceiver();
-        boolean copyBaseURI = out.getSystemId() == null;
-        // if the copy is being attached to an existing parent, it inherits the base URI of the parent
 
-        int copyOptions = CopyOptions.TYPE_ANNOTATIONS;
-        if (copyNamespaces) {
-            copyOptions |= CopyOptions.ALL_NAMESPACES;
-        }
         if (copyAccumulators) {
-            copyOptions |= CopyOptions.ACCUMULATORS;
-        }
-        if (copyForUpdate) {
-            copyOptions |= CopyOptions.FOR_UPDATE;
-        }
-        //int whichNamespaces = (copyNamespaces ? NodeInfo.ALL_NAMESPACES : NodeInfo.NO_NAMESPACES);
-
-        SequenceIterator iter = getSelect().iterate(context);
-        Item item;
-        while ((item = iter.next()) != null) {
-            if (item instanceof NodeInfo) {
-                NodeInfo source = (NodeInfo) item;
-                int kind = source.getNodeKind();
-                if (requireDocumentOrElement &&
-                        !(kind == Type.ELEMENT || kind == Type.DOCUMENT)) {
-                    XPathException e = new XPathException("Operand of validate expression must be a document or element node");
-                    e.setXPathContext(context);
-                    e.setErrorCode("XQTY0030");
-                    throw e;
-                }
-                final Configuration config = controller.getConfiguration();
-                switch (kind) {
-
-                    case Type.ELEMENT: {
-                        Receiver eval = out;
-                        if (validating) {
-                            ParseOptions options = new ParseOptions();
-                            options.setSchemaValidationMode(validation);
-                            options.setTopLevelType(schemaType);
-                            options.setTopLevelElement(NameOfNode.makeName(source).getStructuredQName());
-                            config.prepareValidationReporting(context, options);
-                            eval = config.getElementValidator(out, options, getLocation());
-                        }
-                        if (copyBaseURI) {
-                            eval.setSystemId(computeNewBaseUri(source, getStaticBaseURIString()));
-                        }
-
-                        Receiver savedReceiver = null;
-                        PipelineConfiguration savedPipe = null;
-                        if (copyLineNumbers) {
-                            savedReceiver = eval;
-                            PipelineConfiguration pipe = eval.getPipelineConfiguration();
-                            savedPipe = new PipelineConfiguration(pipe);
-                            LocationCopier copier = new LocationCopier(false);
-                            pipe.setComponent(CopyInformee.class.getName(), copier);
-                        }
-                        source.copy(eval, copyOptions, getLocation());
-                        if (copyLineNumbers) {
-                            eval = savedReceiver;
-                            eval.setPipelineConfiguration(savedPipe);
-                        }
-                        break;
+            // Try to create a virtual copy if we can. This is cheaper
+            if (mustPush()) {
+                // This typically happens with the combination copy-accumulators=yes, validation=strict
+                // Test case accumulators-070
+                // We have to create a physical copy because of the validation requirement, but this makes
+                // it difficult to copy the accumulator values.
+                SequenceIterator iter = getSelect().iterate(context);
+                Item item;
+                while ((item = iter.next()) != null) {
+                    if (item instanceof NodeInfo) {
+                        TinyBuilder builder = new TinyBuilder(out.getPipelineConfiguration());
+                        builder.open();
+                        copyOneNode(context, new TreeReceiver(builder), (NodeInfo)item, CopyOptions.ALL_NAMESPACES);
+                        builder.close();
+                        TinyNodeImpl copy = (TinyNodeImpl)builder.getCurrentRoot();
+                        copy.getTree().setCopiedFrom((NodeInfo)item);
+                        out.append(copy);
+                    } else {
+                        out.append(item);
                     }
-                    case Type.ATTRIBUTE:
-                        if (schemaType != null && schemaType.isComplexType()) {
-                            XPathException e = new XPathException("When copying an attribute with schema validation, the requested type must not be a complex type");
-                            e.setLocation(getLocation());
-                            e.setXPathContext(context);
-                            e.setErrorCode("XTTE1535");
-                            throw dynamicError(getLocation(), e, context);
-                        }
-                        try {
-                            copyAttribute(source, (SimpleType) schemaType, validation, this, context, rejectDuplicateAttributes);
-                        } catch (NoOpenStartTagException err) {
-                            XPathException e = new XPathException(err.getMessage());
-                            e.setLocation(getLocation());
-                            e.setXPathContext(context);
-                            e.setErrorCodeQName(err.getErrorCodeQName());
-                            throw dynamicError(getLocation(), e, context);
-                        }
-                        break;
-                    case Type.TEXT:
-                        out.characters(source.getStringValueCS(), getLocation(), 0);
-                        break;
-
-                    case Type.PROCESSING_INSTRUCTION:
-                        if (copyBaseURI) {
-                            out.setSystemId(source.getBaseURI());
-                        }
-                        out.processingInstruction(source.getDisplayName(), source.getStringValueCS(), getLocation(), 0);
-                        break;
-
-                    case Type.COMMENT:
-                        out.comment(source.getStringValueCS(), getLocation(), 0);
-                        break;
-
-                    case Type.NAMESPACE:
-                        try {
-                            source.copy(out, 0, getLocation());
-                        } catch (NoOpenStartTagException err) {
-                            XPathException e = new XPathException(err.getMessage());
-                            e.setXPathContext(context);
-                            e.setErrorCodeQName(err.getErrorCodeQName());
-                            throw dynamicError(getLocation(), e, context);
-                        }
-                        break;
-
-                    case Type.DOCUMENT: {
-                        ParseOptions options = new ParseOptions();
-                        options.setSchemaValidationMode(validation);
-                        options.setStripSpace(Whitespace.NONE);
-                        options.setTopLevelType(schemaType);
-                        config.prepareValidationReporting(context, options);
-                        Receiver val = config.getDocumentValidator(out, source.getBaseURI(), options);
-                        //val.setPipelineConfiguration(out.getPipelineConfiguration());
-                        if (copyBaseURI) {
-                            val.setSystemId(source.getBaseURI());
-                        }
-                        Receiver savedReceiver = null;
-                        PipelineConfiguration savedPipe = null;
-                        if (copyLineNumbers) {
-                            savedReceiver = val;
-                            savedPipe = new PipelineConfiguration(val.getPipelineConfiguration());
-                            LocationCopier copier = new LocationCopier(true);
-                            val.getPipelineConfiguration().setComponent(CopyInformee.class.getName(), copier);
-
-                        }
-                        source.copy(val, copyOptions, getLocation());
-                        if (copyLineNumbers) {
-                            val = savedReceiver;
-                            val.setPipelineConfiguration(savedPipe);
-                        }
-//                        if (val != out) {
-//                            See bug 2403
-//                            val.close(); // needed to flush out unresolved IDREF values when validating: test copy-5021
-//                        }
-                        break;
-                    }
-                    default:
-                        throw new IllegalArgumentException("Unknown node kind " + source.getNodeKind());
                 }
 
             } else {
-                out.append(item, getLocation(), NodeInfo.ALL_NAMESPACES);
+                // Use the iterate() method to create a virtual copy.
+                SequenceIterator iter = iterate(context);
+                Item item;
+                while ((item = iter.next()) != null) {
+                    out.append(item);
+                }
+            }
+        } else {
+
+            int copyOptions = CopyOptions.TYPE_ANNOTATIONS;
+            if (copyNamespaces) {
+                copyOptions |= CopyOptions.ALL_NAMESPACES;
+            }
+
+            if (copyForUpdate) {
+                copyOptions |= CopyOptions.FOR_UPDATE;
+            }
+            //int whichNamespaces = (copyNamespaces ? NodeInfo.ALL_NAMESPACES : NodeInfo.NO_NAMESPACES);
+
+            SequenceIterator iter = getSelect().iterate(context);
+            Item item;
+            while ((item = iter.next()) != null) {
+                if (item instanceof NodeInfo) {
+                    copyOneNode(context, out, (NodeInfo) item, copyOptions);
+
+                } else {
+                    out.append(item, getLocation(), NodeInfo.ALL_NAMESPACES);
+                }
             }
         }
         return null;
     }
 
+    private void copyOneNode(XPathContext context, SequenceReceiver out, NodeInfo item, int copyOptions) throws XPathException {
+        Controller controller = context.getController();
+        boolean copyBaseURI = out.getSystemId() == null;
+        int kind = item.getNodeKind();
+        if (requireDocumentOrElement &&
+                !(kind == Type.ELEMENT || kind == Type.DOCUMENT)) {
+            XPathException e = new XPathException("Operand of validate expression must be a document or element node");
+            e.setXPathContext(context);
+            e.setErrorCode("XQTY0030");
+            throw e;
+        }
+        final Configuration config = controller.getConfiguration();
+        switch (kind) {
+
+            case Type.ELEMENT: {
+                Receiver eval = out;
+                if (validating) {
+                    ParseOptions options = new ParseOptions();
+                    options.setSchemaValidationMode(validation);
+                    SchemaType type = schemaType;
+                    if (type == null && (validation == Validation.STRICT || validation == Validation.LAX)) {
+                        // Bug 3062
+                        String xsitype = item.getAttributeValue(NamespaceConstant.SCHEMA_INSTANCE, "type");
+                        if (xsitype != null) {
+                            StructuredQName typeName;
+                            try {
+                                typeName = StructuredQName.fromLexicalQName(
+                                        xsitype,
+                                        true,
+                                        false,
+                                        new InscopeNamespaceResolver(item));
+                            } catch (XPathException e) {
+                                throw new XPathException("Invalid QName in xsi:type attribute of element being validated: "
+                                                                 + xsitype + ". " + e.getMessage(), "XTTE1510");
+                            }
+                            type = config.getSchemaType(typeName);
+                            if (type == null) {
+                                throw new XPathException("Unknown xsi:type in element being validated: " + xsitype, "XTTE1510");
+                            }
+                        }
+                    }
+                    options.setTopLevelType(type);
+                    options.setTopLevelElement(NameOfNode.makeName(item).getStructuredQName());
+                    options.setErrorListener(context.getErrorListener());
+                    config.prepareValidationReporting(context, options);
+                    eval = config.getElementValidator(out, options, getLocation());
+                }
+                if (copyBaseURI) {
+                    eval.setSystemId(computeNewBaseUri(item, getStaticBaseURIString()));
+                }
+
+                Receiver savedReceiver = null;
+                PipelineConfiguration savedPipe = null;
+                if (copyLineNumbers) {
+                    savedReceiver = eval;
+                    PipelineConfiguration pipe = eval.getPipelineConfiguration();
+                    savedPipe = new PipelineConfiguration(pipe);
+                    LocationCopier copier = new LocationCopier(false);
+                    pipe.setComponent(CopyInformee.class.getName(), copier);
+                }
+                item.copy(eval, copyOptions, getLocation());
+                if (copyLineNumbers) {
+                    eval = savedReceiver;
+                    assert eval != null;
+                    eval.setPipelineConfiguration(savedPipe);
+                }
+                break;
+            }
+            case Type.ATTRIBUTE:
+                if (schemaType != null && schemaType.isComplexType()) {
+                    XPathException e = new XPathException("When copying an attribute with schema validation, the requested type must not be a complex type");
+                    e.setLocation(getLocation());
+                    e.setXPathContext(context);
+                    e.setErrorCode("XTTE1535");
+                    throw dynamicError(getLocation(), e, context);
+                }
+                try {
+                    copyAttribute(item, (SimpleType) schemaType, validation, this, context, rejectDuplicateAttributes);
+                } catch (NoOpenStartTagException err) {
+                    XPathException e = new XPathException(err.getMessage());
+                    e.setLocation(getLocation());
+                    e.setXPathContext(context);
+                    e.setErrorCodeQName(err.getErrorCodeQName());
+                    throw dynamicError(getLocation(), e, context);
+                }
+                break;
+            case Type.TEXT:
+                out.characters(item.getStringValueCS(), getLocation(), 0);
+                break;
+
+            case Type.PROCESSING_INSTRUCTION:
+                if (copyBaseURI) {
+                    out.setSystemId(item.getBaseURI());
+                }
+                out.processingInstruction(item.getDisplayName(), item.getStringValueCS(), getLocation(), 0);
+                break;
+
+            case Type.COMMENT:
+                out.comment(item.getStringValueCS(), getLocation(), 0);
+                break;
+
+            case Type.NAMESPACE:
+                try {
+                    item.copy(out, 0, getLocation());
+                } catch (NoOpenStartTagException err) {
+                    XPathException e = new XPathException(err.getMessage());
+                    e.setXPathContext(context);
+                    e.setErrorCodeQName(err.getErrorCodeQName());
+                    throw dynamicError(getLocation(), e, context);
+                }
+                break;
+
+            case Type.DOCUMENT: {
+                ParseOptions options = new ParseOptions();
+                options.setSchemaValidationMode(validation);
+                options.setSpaceStrippingRule(NoElementsSpaceStrippingRule.getInstance());
+                options.setTopLevelType(schemaType);
+                options.setErrorListener(context.getErrorListener());
+                config.prepareValidationReporting(context, options);
+                Receiver val = config.getDocumentValidator(out, item.getBaseURI(), options, getLocation());
+                //val.setPipelineConfiguration(out.getPipelineConfiguration());
+                if (copyBaseURI) {
+                    val.setSystemId(item.getBaseURI());
+                }
+                Receiver savedReceiver = null;
+                PipelineConfiguration savedPipe = null;
+                if (copyLineNumbers) {
+                    savedReceiver = val;
+                    savedPipe = new PipelineConfiguration(val.getPipelineConfiguration());
+                    LocationCopier copier = new LocationCopier(true);
+                    val.getPipelineConfiguration().setComponent(CopyInformee.class.getName(), copier);
+
+                }
+                item.copy(val, copyOptions, getLocation());
+                if (copyLineNumbers) {
+                    val = savedReceiver;
+                    assert val != null;
+                    val.setPipelineConfiguration(savedPipe);
+                }
+                //                        if (val != out) {
+                //                            See bug 2403
+                //                            val.close(); // needed to flush out unresolved IDREF values when validating: test copy-5021
+                //                        }
+                break;
+            }
+            default:
+                throw new IllegalArgumentException("Unknown node kind " + item.getNodeKind());
+        }
+    }
+
     public static String computeNewBaseUri(NodeInfo source, String staticBaseURI) {
         // These rules are the rules for xsl:copy-of instruction in XSLT. The same code is used to support the
         // validate{} expression in XQuery. XQuery says nothing about the base URI of a node that results
@@ -754,7 +775,7 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         } catch (XPathException e) {
             e.maybeSetContext(context);
             e.maybeSetLocation(instruction.getLocation());
-            if (instruction.getHostLanguage() == Configuration.XQUERY && e.getErrorCodeLocalPart().equals("XTTE0950")) {
+            if (instruction.getPackageData().getHostLanguage() == Configuration.XQUERY && e.getErrorCodeLocalPart().equals("XTTE0950")) {
                 e.setErrorCode("XQTY0086");
             }
             throw e;
@@ -764,12 +785,12 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
     /**
      * Validate an attribute node and return the type annotation to be used
      *
-     * @param source           the node to be copied
-     * @param schemaType       the simple type against which the value is to be validated, if any
-     * @param validation       one of preserve, strip, strict, lax
-     * @param context          the dynamic context
-     * @throws XPathException if the attribute is not valid
+     * @param source     the node to be copied
+     * @param schemaType the simple type against which the value is to be validated, if any
+     * @param validation one of preserve, strip, strict, lax
+     * @param context    the dynamic context
      * @return the type annotation to be used for the attribute
+     * @throws XPathException if the attribute is not valid
      */
 
 
@@ -780,17 +801,16 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         if (schemaType != null) {
             if (schemaType.isNamespaceSensitive()) {
                 XPathException err = new XPathException("Cannot create a parentless attribute whose " +
-                        "type is namespace-sensitive (such as xs:QName)");
+                                                                "type is namespace-sensitive (such as xs:QName)");
                 err.setErrorCode("XTTE1545");
                 throw err;
             }
             ValidationFailure err = schemaType.validateContent(
                     value, DummyNamespaceResolver.getInstance(), context.getConfiguration().getConversionRules());
             if (err != null) {
-                ValidationException ve = new ValidationException("Attribute being copied does not match the required type. " +
-                        err.getMessage());
-                ve.setErrorCodeQName(err.getErrorCodeQName());
-                throw ve;
+                err.setMessage("Attribute being copied does not match the required type. " +
+                                       err.getMessage());
+                throw err.makeException();
             }
             annotation = schemaType;
         } else if (validation == Validation.STRICT || validation == Validation.LAX) {
@@ -815,17 +835,34 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
         return annotation;
     }
 
+    private boolean mustPush() {
+        return schemaType != null || validation == Validation.LAX || validation == Validation.STRICT ||
+                /*!copyNamespaces ||*/ copyForUpdate;
+    }
+
     /*@NotNull*/
     public SequenceIterator iterate(XPathContext context) throws XPathException {
         final Controller controller = context.getController();
         assert controller != null;
-        if (schemaType == null && copyNamespaces && !copyForUpdate) {
+        if (schemaType == null /*&& copyNamespaces*/ && !copyForUpdate) {
             if (validation == Validation.PRESERVE) {
                 // create a virtual copy of the underlying nodes
-                ItemMappingFunction<Item, Item> copier = new ItemMappingFunction<Item, Item>() {
-                    public Item mapItem(Item item) {
+                ItemMappingFunction copier = new ItemMappingFunction() {
+                    public Item mapItem(Item item) throws XPathException {
                         if (item instanceof NodeInfo) {
+                            if (((NodeInfo) item).getTreeInfo().isTyped()) {
+                                if (!copyNamespaces && ((NodeInfo) item).getNodeKind() == Type.ELEMENT) {
+                                    // A lot of extra work here just to check for error XTTE0950, but the conditions are rare
+                                    Sink sink = new Sink(controller.makePipelineConfiguration());
+                                    ((NodeInfo) item).copy(sink, CopyOptions.TYPE_ANNOTATIONS, getLocation());
+                                }
+                                if (((NodeInfo) item).getNodeKind() == Type.ATTRIBUTE &&
+                                        ((SimpleType) ((NodeInfo) item).getSchemaType()).isNamespaceSensitive()) {
+                                    throw new XPathException("Cannot copy an attribute with namespace-sensitive content except as part of its containing element", "XTTE0950");
+                                }
+                            }
                             VirtualCopy vc = VirtualCopy.makeVirtualCopy((NodeInfo) item);
+                            vc.setDropNamespaces(!copyNamespaces);
                             vc.getTreeInfo().setCopyAccumulators(copyAccumulators);
                             if (((NodeInfo) item).getNodeKind() == Type.ELEMENT) {
                                 vc.setSystemId(computeNewBaseUri((NodeInfo) item, getStaticBaseURIString()));
@@ -836,59 +873,43 @@ public class CopyOf extends Instruction implements ValidatingInstruction {
                         }
                     }
                 };
-                return new ItemMappingIterator<Item, Item>(getSelect().iterate(context), copier, true);
+                return new ItemMappingIterator(getSelect().iterate(context), copier, true);
             } else if (validation == Validation.STRIP) {
                 // create a virtual copy of the underlying nodes
-                ItemMappingFunction<Item, Item> copier = new ItemMappingFunction<Item, Item>() {
+                ItemMappingFunction copier = new ItemMappingFunction() {
                     public Item mapItem(Item item) {
                         if (!(item instanceof NodeInfo)) {
                             return item;
                         }
                         VirtualCopy vc = VirtualUntypedCopy.makeVirtualUntypedTree((NodeInfo) item, (NodeInfo) item);
                         vc.getTreeInfo().setCopyAccumulators(copyAccumulators);
+                        vc.setDropNamespaces(!copyNamespaces);
                         if (((NodeInfo) item).getNodeKind() == Type.ELEMENT) {
                             vc.setSystemId(computeNewBaseUri((NodeInfo) item, getStaticBaseURIString()));
                         }
                         return vc;
                     }
                 };
-                return new ItemMappingIterator<Item, Item>(getSelect().iterate(context), copier, true);
+                return new ItemMappingIterator(getSelect().iterate(context), copier, true);
             }
         }
         SequenceReceiver saved = context.getReceiver();
         PipelineConfiguration pipe = controller.makePipelineConfiguration();
         pipe.setLocationIsCodeLocation(true);
         SequenceOutputter out = new SequenceOutputter(pipe);
-        pipe.setHostLanguage(getHostLanguage());
+        pipe.setHostLanguage(getPackageData().getHostLanguage());
         context.setReceiver(out);
         try {
             process(context);
         } catch (XPathException err) {
-            if (err instanceof ValidationException) {
-                ((ValidationException) err).setLocation(getLocation());
-                ((ValidationException) err).setSystemId(getSystemId());
-            }
             err.maybeSetLocation(getLocation());
+            err.maybeSetContext(context);
             throw err;
         }
         context.setReceiver(saved);
         return out.getSequence().iterate();
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the CopyOf expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        //return new CopyOfCompiler();
-        return new InterpretedExpressionCompiler();
-    }
-//#endif
-
 
 }
 
