@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,13 +7,10 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.FixedElementCompiler;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.event.Receiver;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.lib.FeatureKeys;
 import net.sf.saxon.lib.Validation;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.CombinedNodeTest;
@@ -25,8 +22,6 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.*;
 
-import java.util.Map;
-
 
 /**
  * An instruction that creates an element node whose name is known statically.
@@ -83,7 +78,6 @@ public class FixedElement extends ElementCreator {
     /*@NotNull*/
     public Expression simplify() throws XPathException {
         final Configuration config = getConfiguration();
-        setLazyConstruction(config.getBooleanProperty(FeatureKeys.LAZY_CONSTRUCTION_MODE));
         preservingTypes |= !getPackageData().isSchemaAware();
         return super.simplify();
     }
@@ -156,7 +150,7 @@ public class FixedElement extends ElementCreator {
         if (namespaceBindings.length == 0) {
             return;
         }
-        TypeHierarchy th = getConfiguration().getTypeHierarchy();
+        TypeHierarchy th = visitor.getConfiguration().getTypeHierarchy();
         ItemType contentType = getContentExpression().getItemType();
         boolean ok = th.relationship(contentType, NodeKindTest.ATTRIBUTE) == TypeHierarchy.DISJOINT;
         if (!ok) {
@@ -166,8 +160,7 @@ public class FixedElement extends ElementCreator {
                 for (Operand o : getContentExpression().operands()) {
                     Expression exp = o.getChildExpression();
                     if (exp instanceof FixedAttribute) {
-                        int attNameCode = ((FixedAttribute) exp).getAttributeNameCode();
-                        if (NamePool.isPrefixed(attNameCode)) {
+                        if (!((FixedAttribute)exp).getAttributeName().hasURI("")) {
                             ok = false;
                             break;
                         }
@@ -248,7 +241,7 @@ public class FixedElement extends ElementCreator {
                                                  NodeName elementName, Expression content) throws XPathException {
         final Configuration config = env.getConfiguration();
         ItemType itemType;
-        int fp = elementName.allocateNameCode(config.getNamePool()) & NamePool.FP_MASK;   // Bug #2563 - namespaced alias in unoptimized export
+        int fp = elementName.obtainFingerprint(config.getNamePool());   // Bug #2563 - namespaced alias in unoptimized export
         if (schemaType == null) {
             if (validation == Validation.STRICT) {
                 SchemaDeclaration decl = config.getElementDeclaration(fp);
@@ -295,12 +288,12 @@ public class FixedElement extends ElementCreator {
                         try {
                             config.checkTypeDerivationIsOK(xsiType, declaredType, 0);
                         } catch (SchemaException e) {
-                            ValidationException ve = new ValidationException("The specified xsi:type " + xsiType.getDescription() +
+                            ValidationFailure ve = new ValidationFailure("The specified xsi:type " + xsiType.getDescription() +
                                 " is not validly derived from the required type " + schemaType.getDescription());
                             ve.setConstraintReference(1, "cvc-elt", "4.3");
                             ve.setErrorCode(instr.isXSLT() ? "XTTE1515" : "XQDY0027");
-                            ve.setLocation(instr.getLocation());
-                            throw ve;
+                            ve.setLocator(instr.getLocation());
+                            throw ve.makeException();
                         }
                     }
                 }
@@ -432,7 +425,7 @@ public class FixedElement extends ElementCreator {
      */
 
     private SchemaType testForXSIType(FixedAttribute fat, StaticContext env) throws XPathException {
-        int att = fat.getAttributeNameCode() & NamePool.FP_MASK;
+        int att = fat.getAttributeFingerprint();
         if (att == StandardNames.XSI_TYPE) {
             Expression attValue = fat.getSelect();
             if (attValue instanceof StringLiteral) {
@@ -549,19 +542,6 @@ public class FixedElement extends ElementCreator {
         return namespaceBindings;
     }
 
-//#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the FixedElement expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new FixedElementCompiler();
-    }
-//#endif
-
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -571,25 +551,29 @@ public class FixedElement extends ElementCreator {
         out.startElement("elem", this);
         if (elementName.getPrefix().isEmpty()) {
             out.emitAttribute("name", elementName.getLocalPart());
-            if (!elementName.hasURI("")) {
-                out.emitAttribute("nsuri", elementName.getURI());
-            }
         } else {
             out.emitAttribute("name", elementName.getStructuredQName());
         }
+        out.emitAttribute("nsuri", elementName.getURI());
         String flags = getInheritanceFlags();
         if (!elementName.getURI().isEmpty() && elementName.getPrefix().isEmpty()) {
             flags += "d";  // "d" to indicate default namespace
         }
+        if (isLocal()) {
+            flags += "l";
+        }
         if (!flags.isEmpty()) {
             out.emitAttribute("flags", flags);
         }
         FastStringBuffer fsb = new FastStringBuffer(FastStringBuffer.C256);
         if (namespaceBindings.length > 0) {
             for (NamespaceBinding ns : namespaceBindings) {
-                fsb.append(ns.getPrefix());
-                fsb.append('=');
-                fsb.append(ns.getURI());
+                String prefix = ns.getPrefix();
+                fsb.append(prefix.isEmpty() ? "#" : prefix);
+                if (!ns.getURI().equals(getRetainedStaticContext().getURIForPrefix(prefix, true))) {
+                    fsb.append('=');
+                    fsb.append(ns.getURI());
+                }
                 fsb.append(' ');
             }
             fsb.setLength(fsb.length() - 1);
@@ -619,7 +603,7 @@ public class FixedElement extends ElementCreator {
      */
     @Override
     public String getExpressionName() {
-        return "element{" + elementName.getDisplayName() + "}";
+        return "element";
     }
 }
 
