@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,9 +7,6 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.SlashExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.ForEachAdjunct;
 import net.sf.saxon.Controller;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
@@ -20,8 +17,8 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.ErrorType;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.SchemaType;
-
-import java.util.Map;
+import net.sf.saxon.type.UType;
+import net.sf.saxon.value.Cardinality;
 
 
 /**
@@ -36,6 +33,7 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
     protected Operand selectOp;
     protected Operand actionOp;
     protected Operand threadsOp;
+    protected boolean isInstruction;
 
     /**
      * Create an xsl:for-each instruction
@@ -66,6 +64,29 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
         this.containsTailCall = containsTailCall && action instanceof TailCallReturner;
     }
 
+    /**
+     * Say whether this ForEach expression originates as an XSLT instruction
+     *
+     * @param inst true if this is an xsl:for-each instruction; false if it is the XPath "!" operator
+     */
+
+    public void setInstruction(boolean inst) {
+        isInstruction = inst;
+    }
+
+    /**
+     * Ask whether this expression is an instruction. In XSLT streamability analysis this
+     * is used to distinguish constructs corresponding to XSLT instructions from other constructs.
+     *
+     * @return true if this construct originates as an XSLT instruction
+     */
+
+    @Override
+    public boolean isInstruction() {
+        return isInstruction;
+    }
+
+
     /**
      * Get the select expression
      *
@@ -213,13 +234,31 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
         return getAction().getItemType();
     }
 
+
+    /**
+     * Get the static type of the expression as a UType, following precisely the type
+     * inference rules defined in the XSLT 3.0 specification.
+     *
+     * @return the static item type of the expression according to the XSLT 3.0 defined rules
+     * @param contextItemType
+     */
+    @Override
+    public UType getStaticUType(UType contextItemType) {
+        if (isInstruction()) {
+            return super.getStaticUType(contextItemType);
+        } else {
+            return getAction().getStaticUType(getSelect().getStaticUType(contextItemType));
+        }
+    }
+
+
     /**
      * Determine whether this instruction creates new nodes.
      * This implementation returns true if the "action" creates new nodes.
      * (Nodes created by the condition can't contribute to the result).
      */
 
-    public final boolean createsNewNodes() {
+    public final boolean mayCreateNewNodes() {
         int props = getAction().getSpecialProperties();
         return (props & StaticProperty.NON_CREATIVE) == 0;
     }
@@ -234,9 +273,14 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
             return Literal.makeEmptySequence();
         }
 
-        ContextItemStaticInfo cit = new ContextItemStaticInfo(getSelect().getItemType(), false, getSelect());
+        ContextItemStaticInfo cit = visitor.getConfiguration().makeContextItemStaticInfo(getSelect().getItemType(), false);
+        cit.setContextSettingExpression(getSelect());
         actionOp.typeCheck(visitor, cit);
 
+        if (!Cardinality.allowsMany(getSelect().getCardinality())) {
+            actionOp.setOperandRole(actionOp.getOperandRole().modifyProperty(OperandRole.SINGLETON, true));
+        }
+
         return this;
     }
 
@@ -244,7 +288,8 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
     public Expression optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         selectOp.optimize(visitor, contextInfo);
 
-        ContextItemStaticInfo cit = new ContextItemStaticInfo(getSelect().getItemType(), false, getSelect());
+        ContextItemStaticInfo cit = visitor.getConfiguration().makeContextItemStaticInfo(getSelect().getItemType(), false);
+        cit.setContextSettingExpression(getSelect());
         actionOp.optimize(visitor, cit);
 
         if (!visitor.isOptimizeForStreaming()) {
@@ -257,27 +302,8 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
             }
         }
 
-        // If any subexpressions within the body of the for-each are not dependent on the focus,
-        // promote them: this causes them to be evaluated once, outside the for-each loop
-
-        PromotionOffer offer = new PromotionOffer(getConfiguration().obtainOptimizer());
-        offer.action = PromotionOffer.FOCUS_INDEPENDENT;
-        offer.promoteDocumentDependent = (getSelect().getSpecialProperties() & StaticProperty.CONTEXT_DOCUMENT_NODESET) != 0;
-        offer.promoteXSLTFunctions = false;
-        offer.containingExpression = this;
-        offer.bindingList = new Binding[0];
-        setAction(doPromotion(getAction(), offer));
-
-        if (offer.containingExpression instanceof LetExpression) {
-            ((LetExpression) offer.containingExpression).setEvaluationMode(ExpressionTool.MAKE_MEMO_CLOSURE);
-            offer.containingExpression = offer.containingExpression.optimize(visitor, contextInfo);
-        }
-        Expression e2 = offer.containingExpression;
-        if (e2 != this) {
-            return e2;
-        }
         if (threadsOp != null && !Literal.isEmptySequence(getThreads())) {
-            return getConfiguration().obtainOptimizer().generateMultithreadedInstruction(this);
+            return visitor.getConfiguration().obtainOptimizer().generateMultithreadedInstruction(this);
         }
         return this;
     }
@@ -334,6 +360,7 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
     public Expression copy(RebindingMap rebindings) {
         ForEach f2 = new ForEach(getSelect().copy(rebindings), getAction().copy(rebindings), containsTailCall, getThreads());
         ExpressionTool.copyLocationInfo(this, f2);
+        f2.setInstruction(isInstruction());
         return f2;
     }
 
@@ -374,22 +401,11 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
         return p;
     }
 
-    /**
-     * Handle promotion offers, that is, non-local tree rewrites.
-     *
-     * @param offer The type of rewrite being offered
-     * @throws XPathException
-     */
-
-    protected void promoteChildren(PromotionOffer offer) throws XPathException {
-        setSelect(doPromotion(getSelect(), offer));
-        if (offer.action == PromotionOffer.EXTRACT_GLOBAL_VARIABLES) {
-            // Don't pass on other requests
-            setAction(doPromotion(getAction(), offer));
-        }
+    @Override
+    public boolean alwaysCreatesNewNodes() {
+        return (getAction() instanceof Instruction) && ((Instruction)getAction()).alwaysCreatesNewNodes();
     }
 
-
     /**
      * An implementation of Expression must provide at least one of the methods evaluateItem(), iterate(), or process().
      * This method indicates which of these methods is provided. This implementation provides both iterate() and
@@ -539,32 +555,15 @@ public class ForEach extends Instruction implements ContextMappingFunction, Cont
         return "forEach";
     }
 
-//#ifdefined BYTECODE
-
     /**
-     * Return the compiler of the ForEach expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new SlashExpressionCompiler();
+    public String getStreamerName() {
+        return "ForEach";
     }
-//#endif
-
-//#ifdefined STREAM
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public ForEachAdjunct getStreamingAdjunct() {
-        return new ForEachAdjunct();
-    }
-
-//#endif
-
-
 }
 
