@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,9 +7,6 @@
 
 package net.sf.saxon.expr.instruct;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.ResultDocumentCompiler;
-import com.saxonica.ee.stream.adjunct.ResultDocumentAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.event.ComplexContentOutputter;
@@ -27,7 +24,6 @@ import net.sf.saxon.style.StylesheetPackage;
 import net.sf.saxon.style.XSLResultDocument;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.Err;
-import net.sf.saxon.trans.SaxonErrorCode;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.ErrorType;
 import net.sf.saxon.type.ItemType;
@@ -233,6 +229,12 @@ public class ResultDocument extends Instruction
         return async;
     }
 
+    @Override
+    public boolean isMultiThreaded(Configuration config) {
+        return isAsynchronous() && config.isLicensedFeature(Configuration.LicenseFeature.SCHEMA_VALIDATION)
+                && config.getBooleanProperty(FeatureKeys.ALLOW_MULTITHREADING);
+    }
+
     /*@NotNull*/
     public Expression typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
         typeCheckChildren(visitor, contextInfo);
@@ -376,8 +378,8 @@ public class ResultDocument extends Instruction
         if (context.getTemporaryOutputState() != 0) {
             // Check this before starting a new thread (because in temporary output state the context is sometimes
             // not sufficiently shared to allow the error to be passed back
-            XPathException err = new XPathException("Cannot execute xsl:result-document while evaluating " +
-                                                            context.getNamePool().getDisplayName(context.getTemporaryOutputState()));
+            XPathException err = new XPathException("Cannot execute xsl:result-document while evaluating xsl:" +
+                                                            context.getNamePool().getLocalName(context.getTemporaryOutputState()));
             err.setErrorCode("XTDE1480");
             err.setLocation(getLocation());
             throw err;
@@ -400,14 +402,17 @@ public class ResultDocument extends Instruction
         SequenceReceiver saved = context.getReceiver();
         String savedOutputUri = context.getCurrentOutputUri();
         if (context.getTemporaryOutputState() != 0) {
-            XPathException err = new XPathException("Cannot execute xsl:result-document while evaluating " +
-                                                            context.getNamePool().getDisplayName(context.getTemporaryOutputState()));
+            XPathException err = new XPathException("Cannot execute xsl:result-document while evaluating xsl:" +
+                                                            context.getNamePool().getLocalName(context.getTemporaryOutputState()));
             err.setErrorCode("XTDE1480");
             err.setLocation(getLocation());
             throw err;
         }
 
         Properties computedLocalProps = gatherOutputProperties(context);
+        if (getStaticBaseURIString() != null) {
+            computedLocalProps.setProperty(SaxonOutputKeys.PARAMETER_DOCUMENT_BASE_URI, getStaticBaseURIString());
+        }
         String nextInChain = computedLocalProps.getProperty(SaxonOutputKeys.NEXT_IN_CHAIN);
         //TODO: reinstate this code
 //        if (nextInChain != null && nextInChain.length() > 0) {
@@ -420,6 +425,7 @@ public class ResultDocument extends Instruction
 
         Receiver out;
         OutputURIResolver resolver = null;
+        CloseableResult closeableResult = null;
         Result result = null;
         boolean buildTree = SaxonOutputKeys.isBuildTree(computedLocalProps);
         if (getHref() == null) {
@@ -448,7 +454,8 @@ public class ResultDocument extends Instruction
             resolver = controller.getOutputURIResolver().newInstance();
 
             try {
-                result = getResult(getHref(), getStaticBaseURIString(), context, resolver, resolveAgainstStaticBase);
+                closeableResult = getResult(getHref(), getStaticBaseURIString(), context, resolver, resolveAgainstStaticBase);
+                result = closeableResult.result;
             } catch (XPathException e) {
                 e.maybeSetLocation(getLocation());
                 throw e;
@@ -484,20 +491,27 @@ public class ResultDocument extends Instruction
         }
         context.setReceiver(saved);
         context.setCurrentOutputUri(savedOutputUri);
-        if (resolver != null) {
+        if (closeableResult != null) {
             try {
-                resolver.close(result);
+                closeableResult.close();
             } catch (TransformerException e) {
                 throw XPathException.makeXPathException(e);
             }
         }
     }
 
-    public static Result getResult(Expression href, String baseURI,
+    public static class CloseableResult {
+        public Result result;
+        public void close() throws TransformerException {};
+    }
+
+    public static CloseableResult getResult(Expression href, String baseURI,
                                    XPathContext context, OutputURIResolver resolver,
                                    boolean resolveAgainstStaticBase) throws XPathException {
         String resultURI;
         Result result;
+        CloseableResult closeable;
+        boolean resultFromResolver = false;
         Controller controller = context.getController();
         if (href == null) {
             result = controller.getPrincipalResult();
@@ -505,6 +519,8 @@ public class ResultDocument extends Instruction
             if (resultURI == null) {
                 resultURI = Controller.ANONYMOUS_PRINCIPAL_OUTPUT_URI;
             }
+            closeable = new CloseableResult();
+            closeable.result = result;
         } else {
             try {
                 String base;
@@ -521,6 +537,8 @@ public class ResultDocument extends Instruction
                     if (resultURI == null) {
                         resultURI = Controller.ANONYMOUS_PRINCIPAL_OUTPUT_URI;
                     }
+                    closeable = new CloseableResult();
+                    closeable.result = result;
                 } else {
                     try {
                         result = resolver == null ? null : resolver.resolve(hrefValue, base);
@@ -535,6 +553,14 @@ public class ResultDocument extends Instruction
                         resolver = StandardOutputResolver.getInstance();
                         result = resolver.resolve(hrefValue, base);
                     }
+                    final OutputURIResolver finalResolver = resolver;
+                    closeable = new CloseableResult() {
+                        @Override
+                        public void close() throws TransformerException {
+                            finalResolver.close(result);
+                        }
+                    };
+                    closeable.result = result;
                     resultURI = result.getSystemId();
                     if (resultURI == null) {
                         try {
@@ -551,7 +577,7 @@ public class ResultDocument extends Instruction
         }
         checkAcceptableUri(context, resultURI);
         traceDestination(context, result);
-        return result;
+        return closeable;
     }
 
     public static void traceDestination(XPathContext context, Result result) {
@@ -737,7 +763,7 @@ public class ResultDocument extends Instruction
                                                 String value, /*@Nullable*/ NamespaceResolver nsResolver,
                                                 boolean prevalidated, Configuration config)
             throws XPathException {
-
+        SerializerFactory sf = config.getSerializerFactory();
         if (uri.isEmpty() || NamespaceConstant.SAXON.equals(uri)) {
             if (lname.equals("method")) {
                 value = Whitespace.trim(value);
@@ -796,10 +822,10 @@ public class ResultDocument extends Instruction
             } else if (lname.equals("attribute-order")) {
                 processListOfNodeNames(details, SaxonOutputKeys.ATTRIBUTE_ORDER, value, nsResolver, false, prevalidated);
             } else if (lname.equals("next-in-chain")) {
-                XPathException e = new XPathException("saxon:next-in-chain property is available only on xsl:output");
-                e.setErrorCodeQName(
-                        new StructuredQName("saxon", NamespaceConstant.SAXON, SaxonErrorCode.SXWN9004));
-                throw e;
+//                XPathException e = new XPathException("saxon:next-in-chain property is available only on xsl:output");
+//                e.setErrorCodeQName(
+//                        new StructuredQName("saxon", NamespaceConstant.SAXON, SaxonErrorCode.SXWN9004));
+//                throw e;
             } else {
                 // all other properties in the default or Saxon namespaces
                 if (lname.equals("output-version")) {
@@ -815,7 +841,7 @@ public class ResultDocument extends Instruction
                             // TODO: whitespace rules seem to vary for different interfaces
                             value = Whitespace.trim(value);
                         }
-                        value = SaxonOutputKeys.checkOutputProperty(clarkName, value, config, true); //TODO: test whether xslt3.0 is allowed
+                        value = sf.checkOutputProperty(clarkName, value);
                     } catch (XPathException err) {
                         err.maybeSetErrorCode("SEPM0016");
                         throw err;
@@ -910,34 +936,18 @@ public class ResultDocument extends Instruction
         }
     }
 
-//#ifdefined BYTECODE
 
     /**
-     * Return the bytecode compiler for the ResultDocument expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new ResultDocumentCompiler();
+    public String getStreamerName() {
+        return "ResultDocument";
     }
 
-//#endif
-
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public ResultDocumentAdjunct getStreamingAdjunct() {
-        return new ResultDocumentAdjunct();
-    }
-
-
-//#endif
 
     public Expression getHref() {
         return hrefOp == null ? null : hrefOp.getChildExpression();
