@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -11,13 +11,13 @@ package net.sf.saxon.expr.instruct;
 
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
-import net.sf.saxon.evpull.EventIterator;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.expr.sort.AtomicComparer;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.NodeTest;
-import net.sf.saxon.query.Annotation;
+import net.sf.saxon.query.AnnotationList;
+import net.sf.saxon.style.StylesheetPackage;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trace.LocationKind;
 import net.sf.saxon.trans.FunctionStreamability;
@@ -33,8 +33,6 @@ import net.sf.saxon.type.SpecificFunctionType;
 import net.sf.saxon.value.EmptySequence;
 import net.sf.saxon.value.SequenceType;
 
-import java.util.List;
-
 /**
  * This object represents the compiled form of a user-written function
  * (the source can be either an XSLT stylesheet function or an XQuery function).
@@ -46,11 +44,9 @@ import java.util.List;
  * convert the supplied arguments.
  */
 
-public class UserFunction extends ComponentCode<UserFunction> implements Function, ContextOriginator {
+public class UserFunction extends Actor implements Function, ContextOriginator {
 
-    public enum Determinism {DETERMINISTIC, PROACTIVE, ELIDABLE}
-
-    ;
+    public enum Determinism {DETERMINISTIC, PROACTIVE, ELIDABLE};
 
     private StructuredQName functionName;  // null for an anonymous function
     private boolean tailCalls = false;
@@ -58,15 +54,17 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     private boolean tailRecursive = false;
     // indicates that the function contains tail calls on itself
     private UserFunctionParameter[] parameterDefinitions;
-    private SequenceType declaredResultType;
     private SequenceType resultType;
+    private SequenceType declaredResultType;
     protected int evaluationMode = ExpressionTool.UNDECIDED;
     private boolean isUpdating = false;
     private int inlineable = -1; // 0:no 1:yes -1:don't know
-    private List<Annotation> annotations;
+    private boolean overrideExtensionFunction = true;
+    private AnnotationList annotations;
     private FunctionStreamability declaredStreamability = FunctionStreamability.UNCLASSIFIED;
     private Controller preallocatedController = null;
     private Determinism determinism = Determinism.PROACTIVE;
+    private int refCount = 0;
 
 
     /**
@@ -121,8 +119,8 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
         return functionName;
     }
 
-    public SymbolicName getSymbolicName() {
-        return new SymbolicName(StandardNames.XSL_FUNCTION, functionName, getArity());
+    public SymbolicName.F getSymbolicName() {
+        return new SymbolicName.F(functionName, getArity());
     }
 
 
@@ -141,7 +139,6 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
         return new SpecificFunctionType(argTypes, resultType);
     }
 
-
     /**
      * Get the roles of the arguments, for the purposes of streaming
      *
@@ -198,6 +195,15 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
         return false;
     }
 
+    public boolean isOverrideExtensionFunction() {
+        return overrideExtensionFunction;
+    }
+
+    public void setOverrideExtensionFunction(boolean overrideExtensionFunction) {
+        this.overrideExtensionFunction = overrideExtensionFunction;
+    }
+
+
     /**
      * Supply the controller to be used when evaluating this function. This is used when the function is dynamically loaded
      * when using the load-xquery-module function.
@@ -212,19 +218,19 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     /**
      * Supply a set of annotations
      *
-     * @param map the new set of annotations, which will replace any previous annotations on the function
+     * @param list the new set of annotations, which will replace any previous annotations on the function
      */
 
-    public void setAnnotations(List<Annotation> map) {
-        this.annotations = map;
+    public void setAnnotations(AnnotationList list) {
+        this.annotations = list;
     }
 
     /**
-     * Get the list of annotations defined on this function
-     * @return the list of annotations
+     * Get the annotations defined on this function
+     * @return the list of annotations defined on this function
      */
 
-    public List<Annotation> getAnnotations() {
+    public AnnotationList getAnnotations() {
         return annotations;
     }
 
@@ -258,6 +264,8 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
             // the caller needs to know whether a tail call was returned or not: if we
             // return a Closure, the tail call escapes into the wild and can reappear anywhere...
             evaluationMode = ExpressionTool.eagerEvaluationMode(getBody());
+//        } else if ((getBody().getSpecialProperties() & StaticProperty.ALL_NODES_NEWLY_CREATED) != 0) {
+//            evaluationMode = ExpressionTool.PROCESS;
         } else {
             evaluationMode = ExpressionTool.lazyEvaluationMode(getBody());
         }
@@ -333,7 +341,7 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     /**
      * Set the declared result type of the function
      *
-     * @param resultType the declared return type (or item()* if no explicit type was declared)
+     * @param resultType the declared return type
      */
 
     public void setResultType(SequenceType resultType) {
@@ -392,7 +400,7 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     public boolean isUpdating() {
         return isUpdating;
     }
-
+    
     /**
      * Set the declared streamability (XSLT 3.0 attribute)
      *
@@ -421,9 +429,7 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
      */
 
     public SequenceType getResultType() {
-        Visibility vis = getDeclaredVisibility();
-        if (resultType == SequenceType.ANY_SEQUENCE && getBody() != null &&
-                (vis == Visibility.FINAL || vis == Visibility.PRIVATE)) {
+        if (resultType == SequenceType.ANY_SEQUENCE && getBody() != null) {
             // see if we can infer a more precise result type. We don't do this if the function contains
             // calls on further functions, to prevent infinite regress.
             if (!containsUserFunctionCalls(getBody())) {
@@ -437,7 +443,7 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     /**
      * Get the declared result type
      *
-     * @return the declared result type, or item()* if the type was not explicitly declared
+     * @return the declared result type
      */
 
     public SequenceType getDeclaredResultType() {
@@ -531,7 +537,7 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
                         new RoleDiagnostic(RoleDiagnostic.FUNCTION_RESULT,
                                            functionName == null ? "" : functionName.getDisplayName(), 0);
                 role.setErrorCode(getPackageData().getHostLanguage() == Configuration.XSLT && getFunctionName() != null ? "XTTE0780" : "XPTY0004");
-                exp2 = TypeChecker.staticTypeCheck(exp2, resultType, false, role, visitor);
+                exp2 = visitor.getConfiguration().getTypeChecker(false).staticTypeCheck(exp2, resultType, role, visitor);
             }
         } catch (XPathException err) {
             err.maybeSetLocation(getLocation());
@@ -564,6 +570,8 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
         return c2;
     }
 
+    //private static int pullCalls = 0, pushCalls = 0;
+
 
     /**
      * Call this function to return a value.
@@ -581,6 +589,19 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
 
     public Sequence call(XPathContext context, Sequence[] actualArgs)
             throws XPathException {
+//        if (pullCalls++ % 1000 == 0) {
+//            System.err.println(pullCalls + " pull function calls");
+//        }
+//        if(onFirstCall != null) {
+//                if(callsUntilOptimize-- <=0) {
+//                    synchronized (this){
+//                        body = body.optimize(null, null); //TODO action to be performed as a Closure - holds the deferred ExpressionVisitor visitor, ContextItemStaticInfo cit
+//                        optimized = true;
+//                    }
+//                }
+//
+//
+//        }
         if (evaluationMode == ExpressionTool.UNDECIDED) {
             // should have been done at compile time
             computeEvaluationMode();
@@ -622,31 +643,13 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
 
     public void process(Sequence[] actualArgs, XPathContextMajor context)
             throws XPathException {
+//        if (pushCalls++ % 1000 == 0) {
+//            System.err.println(pushCalls + " push function calls");
+//        }
         context.setStackFrame(getStackFrameMap(), actualArgs);
         getBody().process(context);
     }
 
-    /**
-     * Call this function in "pull" mode, returning the results as a sequence of PullEvents.
-     *
-     * @param actualArgs the arguments supplied to the function. These must have the correct
-     *                   types required by the function signature (it is the caller's responsibility to check this).
-     *                   It is acceptable to supply a {@link net.sf.saxon.value.Closure} to represent a value whose
-     *                   evaluation will be delayed until it is needed. The array must be the correct size to match
-     *                   the number of arguments: again, it is the caller's responsibility to check this.
-     * @param context    This provides the run-time context for evaluating the function. It is the caller's
-     *                   responsibility to allocate a "clean" context for the function to use; the context that is provided
-     *                   will be overwritten by the function.
-     * @return an iterator over the results of the function call
-     * @throws net.sf.saxon.trans.XPathException if a dynamic error occurs while evaluating the function
-     */
-
-    public EventIterator iterateEvents(Sequence[] actualArgs, XPathContextMajor context)
-            throws XPathException {
-        context.setStackFrame(getStackFrameMap(), actualArgs);
-        return getBody().iterateEvents(context);
-    }
-
     /**
      * Call this function. This method allows an XQuery function to be called directly from a Java
      * application. It creates the environment needed to achieve this
@@ -741,12 +744,12 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
                 break;
         }
         presenter.emitAttribute("flags", flags);
-        presenter.emitAttribute("as", getDeclaredResultType().toString());
+        presenter.emitAttribute("as", getDeclaredResultType().toExportString());
         presenter.emitAttribute("slots", getStackFrameMap().getNumberOfVariables() + "");
         for (UserFunctionParameter p : parameterDefinitions) {
             presenter.startElement("arg");
             presenter.emitAttribute("name", p.getVariableQName());
-            presenter.emitAttribute("as", p.getRequiredType().toString());
+            presenter.emitAttribute("as", p.getRequiredType().toExportString());
             presenter.endElement();
         }
         presenter.setChildRole("body");
@@ -754,6 +757,13 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
         presenter.endElement();
     }
 
+    @Override
+    public boolean isExportable() {
+        return refCount > 0 ||
+                (getDeclaredVisibility() != null && getDeclaredVisibility() != Visibility.PRIVATE) ||
+                ((StylesheetPackage) getPackageData()).isRetainUnusedFunctions();
+    }
+
     public boolean isTrustedResultType() {
         return false;
     }
@@ -935,6 +945,19 @@ public class UserFunction extends ComponentCode<UserFunction> implements Functio
     public AtomicSequence atomize() throws XPathException {
         throw new XPathException("Functions cannot be atomized", "FOTY0013");
     }
+
+    public void incrementReferenceCount() {
+        refCount++;
+    }
+
+    public int getReferenceCount() {
+        return refCount;
+    }
+
+    public void prepareForStreaming() throws XPathException {
+        // method is defined in streamable subclass
+    }
+
 }
 
 
