@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -9,18 +9,17 @@ package net.sf.saxon.expr.parser;
 
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.*;
-import net.sf.saxon.expr.instruct.Choose;
-import net.sf.saxon.expr.instruct.NamedTemplate;
-import net.sf.saxon.expr.instruct.TemplateRule;
+import net.sf.saxon.expr.instruct.*;
 import net.sf.saxon.expr.sort.DocumentSorter;
 import net.sf.saxon.lib.FeatureKeys;
 import net.sf.saxon.lib.Logger;
 import net.sf.saxon.om.Sequence;
 import net.sf.saxon.om.SequenceIterator;
 import net.sf.saxon.pattern.Pattern;
+import net.sf.saxon.style.XSLFunction;
 import net.sf.saxon.style.XSLTemplate;
 import net.sf.saxon.trans.GlobalVariableManager;
-import net.sf.saxon.trans.RuleTarget;
+import net.sf.saxon.trans.rules.RuleTarget;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.TypeHierarchy;
 
@@ -36,8 +35,8 @@ public class Optimizer {
     public static final int FULL_OPTIMIZATION = 10;
 
     /*@NotNull*/ protected Configuration config;
-    private int optimizationLevel = FULL_OPTIMIZATION;
-    private boolean tracing;
+    private OptimizerOptions optimizerOptions = OptimizerOptions.FULL_EE_OPTIMIZATION;
+    protected boolean tracing;
 
     /**
      * Create an Optimizer.
@@ -62,35 +61,38 @@ public class Optimizer {
     }
 
     /**
-     * Set the optimization level
+     * Set the optimizer options
      *
-     * @param level the optimization level, between 0 (no optimization) and 10 (full optimization).
-     *              Currently all values greater than zero have the same effect as full optimization
+     * @param options the optimizer options
      */
 
-    public void setOptimizationLevel(int level) {
-        if (level < NO_OPTIMIZATION || level > FULL_OPTIMIZATION) {
-            throw new IllegalArgumentException("Optimization level");
-        }
-        optimizationLevel = level;
+    public void setOptimizerOptions(OptimizerOptions options) {
+        optimizerOptions = options;
     }
 
     /**
-     * Get the optimization level
+     * Get the optimizer options
      *
-     * @return the optimization level, between 0 (no optimization) and 10 (full optimization).
-     *         Currently all values greater than zero have the same effect as full optimization
+     * @return the optimizer options
      */
 
-    public int getOptimizationLevel() {
-        return optimizationLevel;
+    public OptimizerOptions getOptimizerOptions() {
+        return optimizerOptions;
+    }
+
+    /**
+     * Ask whether a particular optimizer option is set
+     * @param option the code identifying the option, e.g. {@link OptimizerOptions#BYTE_CODE}
+     *            @return true if the option is set
+     */
+
+    public boolean isOptionSet(int option) {
+        return optimizerOptions.isSet(option);
     }
 
     /**
      * Simplify a GeneralComparison expression
      *
-     *
-     *
      * @param visitor             the expression visitor
      * @param gc                  the GeneralComparison to be simplified
      * @param backwardsCompatible true if in 1.0 compatibility mode
@@ -98,20 +100,20 @@ public class Optimizer {
      */
 
     public Expression optimizeGeneralComparison(ExpressionVisitor visitor,
-                         GeneralComparison gc, boolean backwardsCompatible, ContextItemStaticInfo contextItemType) {
+                                                GeneralComparison gc, boolean backwardsCompatible, ContextItemStaticInfo contextItemType) {
         return gc;
     }
 
     /**
-     * Attempt to optimize a copy operation. Return null if no optimization is possible.
+     * Attempt to optimize a call on saxon:stream(). Return null if no optimization is possible.
      *
      * @param select the expression that selects the items to be copied
      * @return null if no optimization is possible, or an expression that does an optimized
-     *         copy of these items otherwise
+     * copy of these items otherwise
      */
 
     /*@Nullable*/
-    public Expression optimizeCopy(ExpressionVisitor visitor, ContextItemStaticInfo cisi, Expression select) throws XPathException {
+    public Expression optimizeSaxonStreamFunction(ExpressionVisitor visitor, ContextItemStaticInfo cisi, Expression select) throws XPathException {
         if (select.getItemType().isPlainType()) {
             return select;
         }
@@ -170,7 +172,7 @@ public class Optimizer {
      *
      * @param filter the predicate expression
      * @return 0 if not indexable; +1 if the predicate is in the form expression=value; -1 if it is in
-     *         the form value=expression
+     * the form value=expression
      */
 
     public int isIndexableFilter(Expression filter) {
@@ -189,6 +191,19 @@ public class Optimizer {
         throw new UnsupportedOperationException("Indexing requires Saxon-EE");
     }
 
+    /**
+     * Evaluate the streaming (first) argument of a streamable stylesheet function
+     *
+     * @param expr    the expression supplied for the value of the streaming argument
+     * @param context the XPath evaluation context of the caller
+     * @return the (nominal) result of the evaluation
+     */
+
+    public Sequence evaluateStreamingArgument(Expression expr, XPathContext context) throws XPathException {
+        // non-streaming fallback implementation
+        return ExpressionTool.eagerEvaluate(expr, context);
+    }
+
     /**
      * Determine whether it is possible to rearrange an expression so that all references to a given
      * variable are replaced by a reference to ".". This is true of there are no references to the variable
@@ -228,7 +243,7 @@ public class Optimizer {
      * @param sorter the document sorter
      * @param path   the path expression
      * @return the original sorter unchanged when no optimization is possible, which is always the
-     *         case in Saxon-B
+     * case in Saxon-B
      */
 
     public Expression makeConditionalDocumentSorter(DocumentSorter sorter, SlashExpression path) throws XPathException {
@@ -243,8 +258,8 @@ public class Optimizer {
      * @param visitor         the expression visitor
      * @param contextItemType the context item type
      * @return either the original expression unchanged, or an expression that consists of the inlined
-     *         function body, with all function parameters bound as required. In Saxon-B, function inlining is
-     *         not supported, so the original functionCall is always returned unchanged
+     * function body, with all function parameters bound as required. In Saxon-B, function inlining is
+     * not supported, so the original functionCall is always returned unchanged
      */
 
     public Expression tryInlineFunctionCall(
@@ -259,7 +274,7 @@ public class Optimizer {
      * @param body    the body of the template or function
      * @param visitor the expression visitor
      * @return the expression after subexpressions have been promoted to global variables; or null if
-     *         nothing has changed
+     * nothing has changed
      */
 
     public Expression promoteExpressionsToGlobal(Expression body, GlobalVariableManager gvManager, ExpressionVisitor visitor)
@@ -267,14 +282,27 @@ public class Optimizer {
         return null;
     }
 
+    /**
+     * Eliminate common subexpressions. Rewrites (contains(X, 'x') and contains(X, 'y')) as
+     * (let $vv:C := X return (contains($vv:C, 'x') and contains($vv:C, 'y'))).
+     * Dummy method; the optimization happens only in Saxon-EE.
+     * @param in the expression to be optimized
+     * @return out the optimized expression (possibly the same as the input).
+     */
+
+    public Expression eliminateCommonSubexpressions(Expression in) {
+        return in;
+    }
+
     /**
      * Try to convert a Choose expression into a switch
      *
      * @param choose the Choose expression
+     * @param visitor
      * @return the result of optimizing this (the original expression if no optimization was possible)
      */
 
-    public Expression trySwitch(Choose choose) {
+    public Expression trySwitch(Choose choose, ExpressionVisitor visitor) {
         return choose;
     }
 
@@ -282,15 +310,14 @@ public class Optimizer {
     /**
      * Try to convert an Or expression into a comparison with Literal sequence
      *
-     *
-     * @param visitor  an expression visitor
+     * @param visitor         an expression visitor
      * @param contextItemType the static type of "." at the point where this expression is invoked.
      *                        The parameter is set to null if it is known statically that the context item will be undefined.
      *                        If the type of the context item is not known statically, the argument is set to
      *                        {@link net.sf.saxon.type.Type#ITEM_TYPE}
-     * @param orExpr           the expression to be converted
+     * @param orExpr          the expression to be converted
      * @return the result of optimizing the Or expression (the original expression if no optimization was possible)
-     * */
+     */
 
     public Expression tryGeneralComparison(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType, OrExpression orExpr) throws XPathException {
         return orExpr;
@@ -305,10 +332,10 @@ public class Optimizer {
      * @return a reference to the new global variable if a variable has been created, or null if not
      */
 
-    public Expression extractGlobalVariables(Expression body, ExpressionVisitor visitor, PromotionOffer offer)
-            throws XPathException {
-        return null;
-    }
+//    public Expression extractGlobalVariables(Expression body, ExpressionVisitor visitor, PromotionOffer offer)
+//            throws XPathException {
+//        return null;
+//    }
 
     /**
      * Generate the inversion of the expression comprising the body of a template rules.
@@ -371,7 +398,6 @@ public class Optimizer {
     }
 
 
-
     /**
      * Generate a multi-threaded version of an instruction.
      * Supported in Saxon-EE only; ignored with no action in Saxon-HE and Saxon-PE
@@ -387,16 +413,33 @@ public class Optimizer {
     /**
      * Generate Java byte code for an expression
      *
+     * @param compilerService
      * @param expr              the expression to be compiled
      * @param objectName        the name of the object (e.g. function) being compiled
      * @param evaluationMethods The evaluation modes for which code is generated. Currently a subset of
-     *                          {@link net.sf.saxon.expr.Expression#PROCESS_METHOD}, {@link net.sf.saxon.expr.Expression#ITERATE_METHOD}. If no code is generated for
+     *                          {@link Expression#PROCESS_METHOD}, {@link Expression#ITERATE_METHOD}. If no code is generated for
      */
 
-    public Expression compileToByteCode(Expression expr, String objectName, int evaluationMethods) {
+    public Expression compileToByteCode(ICompilerService compilerService, Expression expr, String objectName, int evaluationMethods) {
         return null;
     }
 
+    public Expression makeByteCodeCandidate(ExpressionOwner owner, Expression expr, String objectName, int requiredEvaluationModes) {
+        return null;
+    }
+
+    public void injectByteCodeCandidates(Expression exp) throws XPathException {
+
+    };
+
+    public Expression optimizeNumberInstruction(NumberInstruction ni, ContextItemStaticInfo contextInfo) {
+        return null;
+    }
+
+    public void assessFunctionStreamability(XSLFunction sourceFunction, UserFunction compiledFunction) throws XPathException {
+        throw new XPathException("Streamable stylesheet functions are not supported in Saxon-HE", "XTSE3430");
+    }
+
     /**
      * Trace optimization actions
      *
@@ -407,28 +450,20 @@ public class Optimizer {
     public void trace(String message, Expression exp) {
         if (tracing) {
             Logger err = getConfiguration().getLogger();
-            err.info("OPT ======================================");
-            err.info("OPT : At line " + exp.getLocation().getLineNumber() + " of " + exp.getSystemId());
+            err.info("OPT : At line " + exp.getLocation().getLineNumber() + " of " + exp.getLocation().getSystemId());
             err.info("OPT : " + message);
-            err.info("OPT ====== Expression after rewrite ======");
-            exp.explain(err);
-            err.info("\nOPT ======================================");
+            err.info("OPT : Expression after rewrite: " + exp.toString());
         }
     }
 
-    /**
-     * Trace optimization actions
-     *
-     * @param message the message to be displayed
-     */
-
-    public void trace(String message) {
-        if (tracing) {
-            Logger err = getConfiguration().getLogger();
-            err.info("OPT ======================================");
+    public static void trace(Configuration config, String message, Expression exp) {
+        if (config.getBooleanProperty(FeatureKeys.TRACE_OPTIMIZER_DECISIONS)) {
+            Logger err = config.getLogger();
+            err.info("OPT : At line " + exp.getLocation().getLineNumber() + " of " + exp.getLocation().getSystemId());
             err.info("OPT : " + message);
-            err.info("OPT ======================================");
+            err.info("OPT : Expression after rewrite: " + exp.toString());
         }
     }
+
 }
 
