@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -49,10 +49,6 @@ public final class Tokenizer {
 
     public static final int OPERATOR_STATE = 3;
 
-    /**
-     * The starting line number (for XPath in XSLT, the line number in the stylesheet)
-     */
-    public int startLineNumber;
     /**
      * The number identifying the most recently read token
      */
@@ -134,6 +130,12 @@ public final class Tokenizer {
 
     public int languageLevel = 20;
 
+    /**
+     * Flag to allow Saxon extensions
+     */
+
+    public boolean allowSaxonExtensions = false;
+
     public Tokenizer() {
     }
 
@@ -604,7 +606,7 @@ public final class Tokenizer {
                     if (inputOffset < inputLength
                             && input.charAt(inputOffset) == ':'
                             && inputOffset + 1 < inputLength
-                            && (input.charAt(inputOffset+1) > 127 || NameChecker.isNCNameStartChar(input.charAt(inputOffset + 1)))) {
+                            && (input.charAt(inputOffset + 1) > 127 || NameChecker.isNCNameStartChar(input.charAt(inputOffset + 1)))) {
                         inputOffset++;
                         nextToken = Token.SUFFIX;
                         return;
@@ -787,10 +789,10 @@ public final class Tokenizer {
                     nextToken = Token.STRING_LITERAL;
                     return;
                 case '`':
-                    if (inputOffset < inputLength-1
+                    if (inputOffset < inputLength - 1
                             && input.charAt(inputOffset) == '`'
-                            && input.charAt(inputOffset+1) == '[') {
-                        inputOffset+=2;
+                            && input.charAt(inputOffset + 1) == '[') {
+                        inputOffset += 2;
                         int j = inputOffset;
                         int newlines = 0;
                         while (true) {
@@ -799,17 +801,17 @@ public final class Tokenizer {
                             }
                             if (input.charAt(j) == '\n') {
                                 newlines++;
-                            } else if (input.charAt(j) == '`' && j+1 < inputLength && input.charAt(j+1) == '{') {
+                            } else if (input.charAt(j) == '`' && j + 1 < inputLength && input.charAt(j + 1) == '{') {
                                 nextToken = Token.STRING_TEMPLATE_INITIAL;
                                 nextTokenValue = input.substring(inputOffset, j);
-                                inputOffset = j+2;
+                                inputOffset = j + 2;
                                 incrementLineNumber(newlines);
                                 return;
                             } else if (input.charAt(j) == ']' && j + 2 < inputLength
                                     && input.charAt(j + 1) == '`' && input.charAt(j + 2) == '`') {
                                 nextToken = Token.STRING_TEMPLATE_COMPLETE;
                                 nextTokenValue = input.substring(inputOffset, j);
-                                inputOffset = j+3;
+                                inputOffset = j + 3;
                                 incrementLineNumber(newlines);
                                 return;
                             }
@@ -828,10 +830,11 @@ public final class Tokenizer {
                     if (inputOffset < inputLength && input.charAt(inputOffset) == '{') {
                         // EQName, revised syntax as per bug 15399
                         int close = input.indexOf('}', inputOffset++);
-                        if(close<inputOffset) {
+                        if (close < inputOffset) {
                             throw new XPathException("Missing closing brace in EQName");
                         }
                         String uri = input.substring(inputOffset, close);
+                        uri = Whitespace.collapseWhitespace(uri).toString(); // Bug 29708
                         inputOffset = close + 1;
                         int start = inputOffset;
                         boolean isStar = false;
@@ -930,6 +933,7 @@ public final class Tokenizer {
         }
     }
 
+
     /**
      * Identify a binary operator
      *
@@ -938,7 +942,7 @@ public final class Tokenizer {
      *         known operator
      */
 
-    private int getBinaryOp(String s) {
+    public int getBinaryOp(String s) {
         switch (s.length()) {
             case 2:
                 if (s.equals("or")) {
@@ -1030,11 +1034,17 @@ public final class Tokenizer {
                 if (s.equals("modify")) {
                     return Token.MODIFY;
                 }
+                if (s.equals("orElse") && allowSaxonExtensions) {
+                    return Token.MODIFY;
+                }
                 break;
             case 7:
                 if (s.equals("default")) {
                     return Token.DEFAULT;
                 }
+                if (s.equals("andAlso") && allowSaxonExtensions) {
+                    return Token.AND_ALSO;
+                }
                 break;
             case 9:
                 if (s.equals("intersect")) {
@@ -1083,6 +1093,12 @@ public final class Tokenizer {
                 if (s.equals("array")) {
                     return languageLevel == 20 ? Token.FUNCTION : Token.NODEKIND;  // array() in XPath 3.1
                 }
+                if (s.equals("tuple")) {
+                    return allowSaxonExtensions ? Token.NODEKIND : Token.FUNCTION; // Saxon tuple(...) types
+                }
+                if (s.equals("union")) {
+                    return allowSaxonExtensions ? Token.NODEKIND : Token.FUNCTION; // Saxon union(...) types
+                }
                 break;
             case 6:
                 if (s.equals("switch")) {
