@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -9,24 +9,24 @@ package net.sf.saxon.expr.parser;
 
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.*;
-import net.sf.saxon.functions.SystemFunction;
+import net.sf.saxon.lib.ConversionRules;
 import net.sf.saxon.om.*;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
+import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.SequenceExtent;
 import net.sf.saxon.value.SequenceType;
 
 /**
- * This class provides Saxon's type checking capability. It contains a static method,
+ * This class provides Saxon's type checking capability. It contains a method,
  * staticTypeCheck, which is called at compile time to perform type checking of
- * an expression. This class is never instantiated.
+ * an expression.
  */
 
-public final class TypeChecker {
+public class TypeChecker {
 
-    // Class is not instantiated
-    private TypeChecker() {
+    public TypeChecker() {
     }
 
     /**
@@ -55,33 +55,32 @@ public final class TypeChecker {
      *
      * @param supplied            The expression to be type-checked
      * @param req                 The required type for the context in which the expression is used
-     * @param backwardsCompatible True if XPath 1.0 backwards compatibility mode is applicable
      * @param role                Information about the role of the subexpression within the
      *                            containing expression, used to provide useful error messages
      * @param visitor             An expression visitor
      * @return The original expression if it is type-safe, or the expression
-     *         wrapped in a run-time type checking expression if not.
+     * wrapped in a run-time type checking expression if not.
      * @throws XPathException if the supplied type is statically inconsistent with the
      *                        required type (that is, if they have no common subtype)
      */
 
-    public static Expression staticTypeCheck(Expression supplied,
-                                             SequenceType req,
-                                             boolean backwardsCompatible,
-                                             RoleDiagnostic role,
-                                             final ExpressionVisitor visitor)
+    public Expression staticTypeCheck(Expression supplied,
+                                      SequenceType req,
+                                      RoleDiagnostic role,
+                                      final ExpressionVisitor visitor)
             throws XPathException {
 
         // System.err.println("Static Type Check on expression (requiredType = " + req + "):"); supplied.display(10);
 
         if (supplied.implementsStaticTypeCheck()) {
-            return supplied.staticTypeCheck(req, backwardsCompatible, role, visitor);
+            return supplied.staticTypeCheck(req, false, role, visitor);
         }
 
         Expression exp = supplied;
         final StaticContext env = visitor.getStaticContext();
         final Configuration config = env.getConfiguration();
         final TypeHierarchy th = config.getTypeHierarchy();
+        final ContextItemStaticInfo defaultContextInfo = config.getDefaultContextItemStaticInfo();
 
         final ItemType reqItemType = req.getPrimaryType();
         int reqCard = req.getCardinality();
@@ -115,62 +114,15 @@ public final class TypeChecker {
 //                // supplied type is empty-sequence(): this can violate a cardinality constraint but not an item type constraint
 //                itemTypeOK = true;
 //            } else {
-                if (reqItemType == null || suppliedItemType == null) {
-                    throw new NullPointerException();
-                }
-                int relation = th.relationship(reqItemType, suppliedItemType);
-                itemTypeOK = relation == TypeHierarchy.SAME_TYPE || relation == TypeHierarchy.SUBSUMES;
+            if (reqItemType == null || suppliedItemType == null) {
+                throw new NullPointerException();
+            }
+            int relation = th.relationship(reqItemType, suppliedItemType);
+            itemTypeOK = relation == TypeHierarchy.SAME_TYPE || relation == TypeHierarchy.SUBSUMES;
 //            }
         }
 
 
-        // Handle the special rules for 1.0 compatibility mode
-
-        if (backwardsCompatible && !allowsMany) {
-            // rule 1
-            if (Cardinality.allowsMany(suppliedCard)) {
-                Expression cexp = FirstItemExpression.makeFirstItemExpression(exp);
-                cexp.adoptChildExpression(exp);
-                exp = cexp;
-                suppliedCard = StaticProperty.ALLOWS_ZERO_OR_ONE;
-                cardOK = Cardinality.subsumes(reqCard, suppliedCard);
-            }
-            if (!itemTypeOK) {
-                // rule 2
-                if (reqItemType.equals(BuiltInAtomicType.STRING)) {
-                    final RetainedStaticContext rsc = new RetainedStaticContext(config);
-                    Expression fn = SystemFunction.makeCall("string", rsc, exp);
-                    try {
-                        exp = fn.simplify().typeCheck(visitor, ContextItemStaticInfo.DEFAULT);
-                    } catch (XPathException err) {
-                        err.maybeSetLocation(exp.getLocation());
-                        err.setIsStaticError(true);
-                        throw err;
-                    }
-                    suppliedItemType = BuiltInAtomicType.STRING;
-                    suppliedCard = StaticProperty.EXACTLY_ONE;
-                    cardOK = Cardinality.subsumes(reqCard, suppliedCard);
-                    itemTypeOK = true;
-                }
-                // rule 3
-                if (reqItemType.equals(NumericType.getInstance()) || reqItemType.equals(BuiltInAtomicType.DOUBLE)) {
-                    final RetainedStaticContext rsc = new RetainedStaticContext(config);
-                    Expression fn = SystemFunction.makeCall("number", rsc, exp);
-                    try {
-                        exp = fn.simplify().typeCheck(visitor, ContextItemStaticInfo.DEFAULT);
-                    } catch (XPathException err) {
-                        err.maybeSetLocation(exp.getLocation());
-                        err.setIsStaticError(true);
-                        throw err;
-                    }
-                    suppliedItemType = BuiltInAtomicType.DOUBLE;
-                    suppliedCard = StaticProperty.EXACTLY_ONE;
-                    cardOK = Cardinality.subsumes(reqCard, suppliedCard);
-                    itemTypeOK = true;
-                }
-            }
-        }
-
         if (!itemTypeOK) {
             // Now apply the conversions needed in 2.0 mode
 
@@ -180,11 +132,16 @@ public final class TypeChecker {
                 if (!suppliedItemType.isPlainType() &&
                         !(suppliedCard == StaticProperty.EMPTY)) {
                     if (!suppliedItemType.isAtomizable()) {
-                        XPathException err = new XPathException("An atomic value is required for the " + role.getMessage() +
-                            ", but the supplied value cannot be atomized", "FOTY0013", supplied.getLocation());
+                        XPathException err = new XPathException(
+                                "An atomic value is required for the " + role.getMessage() +
+                                        ", but the supplied value cannot be atomized", "FOTY0013", supplied.getLocation());
                         err.setIsTypeError(true);
                         throw err;
                     }
+
+                    if (exp instanceof ContextItemExpression && exp.getRetainedStaticContext() == null) {
+                        exp.setRetainedStaticContextLocally(env.makeRetainedStaticContext());
+                    }
                     Expression cexp = Atomizer.makeAtomizer(exp);
                     ExpressionTool.copyLocationInfo(exp, cexp);
                     exp = cexp;
@@ -215,7 +172,7 @@ public final class TypeChecker {
                     try {
                         if (exp instanceof Literal) {
                             exp = Literal.makeLiteral(
-                                    new SequenceExtent(cexp.iterate(visitor.makeDynamicContext())));
+                                    SequenceExtent.makeSequenceExtent(cexp.iterate(visitor.makeDynamicContext())));
                             ExpressionTool.copyLocationInfo(cexp, exp);
                         } else {
                             exp = cexp;
@@ -247,7 +204,7 @@ public final class TypeChecker {
                     try {
                         if (exp instanceof Literal) {
                             exp = Literal.makeLiteral(
-                                    new SequenceExtent(conversion.iterate(visitor.makeDynamicContext())));
+                                    SequenceExtent.makeSequenceExtent(conversion.iterate(visitor.makeDynamicContext())));
                             ExpressionTool.copyLocationInfo(supplied, exp);
                         } else {
                             exp = conversion;
@@ -270,7 +227,7 @@ public final class TypeChecker {
                         ExpressionTool.copyLocationInfo(exp, cexp);
                         exp = cexp;
                         try {
-                            exp = exp.simplify().typeCheck(visitor, ContextItemStaticInfo.DEFAULT);
+                            exp = exp.simplify().typeCheck(visitor, defaultContextInfo);
                         } catch (XPathException err) {
                             err.maybeSetLocation(exp.getLocation());
                             err.setIsStaticError(true);
@@ -286,7 +243,7 @@ public final class TypeChecker {
                         ExpressionTool.copyLocationInfo(exp, cexp);
                         exp = cexp;
                         try {
-                            exp = exp.simplify().typeCheck(visitor, ContextItemStaticInfo.DEFAULT);
+                            exp = exp.simplify().typeCheck(visitor, defaultContextInfo);
                         } catch (XPathException err) {
                             err.maybeSetLocation(exp.getLocation());
                             err.setIsStaticError(true);
@@ -319,16 +276,16 @@ public final class TypeChecker {
 
             } else if (reqItemType instanceof JavaExternalObjectType &&
                     /*Sequence.class.isAssignableFrom(((JavaExternalObjectType) reqItemType).getJavaClass()) &&  */
-                reqCard == StaticProperty.EXACTLY_ONE) {
+                    reqCard == StaticProperty.EXACTLY_ONE) {
 
                 if (Sequence.class.isAssignableFrom(((JavaExternalObjectType) reqItemType).getJavaClass())) {
                     // special case: allow an extension function to call an instance method on the implementation type of an XDM value
                     // we leave the conversion to be sorted out at run-time
                     itemTypeOK = true;
-                } else if (supplied instanceof FunctionCall){
+                } else if (supplied instanceof FunctionCall) {
                     // adjust the required type of the Java extension function call
                     // this does nothing unless supplied is an instanceof JavaExtensionFunctionCall
-                    if(((FunctionCall)supplied).adjustRequiredType((JavaExternalObjectType) reqItemType)) {
+                    if (((FunctionCall) supplied).adjustRequiredType((JavaExternalObjectType) reqItemType)) {
                         itemTypeOK = true;
                         cardOK = true;
                     }
@@ -380,12 +337,12 @@ public final class TypeChecker {
             if (Cardinality.allowsZero(suppliedCard) &&
                     Cardinality.allowsZero(reqCard)) {
                 if (suppliedCard != StaticProperty.EMPTY) {
-                    String msg = role.composeErrorMessage(reqItemType, suppliedItemType);
+                    String msg = role.composeErrorMessage(reqItemType, supplied, suppliedItemType);
                     msg += ". The expression can succeed only if the supplied value is an empty sequence.";
                     visitor.issueWarning(msg, supplied.getLocation());
                 }
             } else {
-                String msg = role.composeErrorMessage(reqItemType, suppliedItemType);
+                String msg = role.composeErrorMessage(reqItemType, supplied, suppliedItemType);
                 XPathException err = new XPathException(msg, role.getErrorCode(), supplied.getLocation());
                 err.setIsTypeError(role.isTypeError());
                 throw err;
@@ -398,22 +355,27 @@ public final class TypeChecker {
 
         if (!(relation == TypeHierarchy.SAME_TYPE || relation == TypeHierarchy.SUBSUMED_BY)) {
             if (exp instanceof Literal) {
-                String msg = role.composeErrorMessage(reqItemType, suppliedItemType);
+                // Try a more detailed check, since for maps, functions etc getItemType() can be imprecise
+                if (req.matches(((Literal) exp).getValue(), th)) {
+                    return exp;
+                }
+                String msg = role.composeErrorMessage(reqItemType, supplied, suppliedItemType);
                 XPathException err = new XPathException(msg, role.getErrorCode(), supplied.getLocation());
                 err.setIsTypeError(role.isTypeError());
                 throw err;
+            } else {
+                Expression cexp = new ItemChecker(exp, reqItemType, role);
+                ExpressionTool.copyLocationInfo(exp, cexp);
+                exp = cexp;
             }
-            Expression cexp = new ItemChecker(exp, reqItemType, role);
-            ExpressionTool.copyLocationInfo(exp, cexp);
-            exp = cexp;
         }
 
         if (!cardOK) {
             if (exp instanceof Literal) {
                 XPathException err = new XPathException("Required cardinality of " + role.getMessage() +
-                    " is " + Cardinality.toString(reqCard) +
-                    "; supplied value has cardinality " +
-                    Cardinality.toString(suppliedCard), role.getErrorCode(), supplied.getLocation());
+                                                                " is " + Cardinality.toString(reqCard) +
+                                                                "; supplied value has cardinality " +
+                                                                Cardinality.toString(suppliedCard), role.getErrorCode(), supplied.getLocation());
                 err.setIsTypeError(role.isTypeError());
                 throw err;
             } else {
@@ -426,6 +388,18 @@ public final class TypeChecker {
         return exp;
     }
 
+    public Expression makeArithmeticExpression(Expression lhs, int operator, Expression rhs) {
+        return new ArithmeticExpression(lhs, operator, rhs);
+    }
+
+    public Expression makeGeneralComparison(Expression lhs, int operator, Expression rhs) {
+        return new GeneralComparison20(lhs, operator, rhs);
+    }
+
+    public Expression processValueOf(Expression select, Configuration config) {
+        return select;
+    }
+
     private static Expression makeFunctionSequenceCoercer(
             Expression exp, FunctionItemType reqItemType, ExpressionVisitor visitor, RoleDiagnostic role) throws XPathException {
         // Apply function coercion as defined in XPath 3.0.
@@ -446,7 +420,7 @@ public final class TypeChecker {
      * @param env      The static context containing the types being checked. At present
      *                 this is used only to locate a NamePool
      * @return The original expression if it is type-safe, or the expression
-     *         wrapped in a run-time type checking expression if not.
+     * wrapped in a run-time type checking expression if not.
      * @throws XPathException if the supplied type is statically inconsistent with the
      *                        required type (that is, if they have no common subtype)
      */
@@ -533,13 +507,13 @@ public final class TypeChecker {
                 if (suppliedCard != StaticProperty.EMPTY) {
                     String msg = "Required item type of " + role.getMessage() +
                             " is " + reqItemType.toString() +
-                            "; supplied value has item type " +
+                            "; supplied value (" + supplied.toShortString() + ") has item type " +
                             suppliedItemType.toString() +
                             ". The expression can succeed only if the supplied value is an empty sequence.";
                     env.issueWarning(msg, supplied.getLocation());
                 }
             } else {
-                String msg = role.composeErrorMessage(reqItemType, suppliedItemType);
+                String msg = role.composeErrorMessage(reqItemType, supplied, suppliedItemType);
                 XPathException err = new XPathException(msg, role.getErrorCode(), supplied.getLocation());
                 err.setIsTypeError(role.isTypeError());
                 throw err;
@@ -559,9 +533,9 @@ public final class TypeChecker {
         if (!cardOK) {
             if (exp instanceof Literal) {
                 XPathException err = new XPathException("Required cardinality of " + role.getMessage() +
-                    " is " + Cardinality.toString(reqCard) +
-                    "; supplied value has cardinality " +
-                    Cardinality.toString(suppliedCard), role.getErrorCode(), supplied.getLocation());
+                                                                " is " + Cardinality.toString(reqCard) +
+                                                                "; supplied value has cardinality " +
+                                                                Cardinality.toString(suppliedCard), role.getErrorCode(), supplied.getLocation());
                 err.setIsTypeError(role.isTypeError());
                 throw err;
             } else {
@@ -581,7 +555,7 @@ public final class TypeChecker {
      * @param requiredType the required type
      * @param context      XPath dynamic context
      * @return an XPathException describing the error condition if the value doesn't conform;
-     *         or null if it does.
+     * or null if it does.
      * @throws XPathException if a failure occurs reading the value
      */
 
@@ -597,7 +571,7 @@ public final class TypeChecker {
             count++;
             if (!reqItemType.matches(item, context.getConfiguration().getTypeHierarchy())) {
                 XPathException err = new XPathException("Required type is " + reqItemType +
-                        "; supplied value has type " + UType.getUType(SequenceTool.toGroundedValue(val)));
+                                                                "; supplied value has type " + UType.getUType(SequenceTool.toGroundedValue(val)));
                 err.setIsTypeError(true);
                 err.setErrorCode("XPTY0004");
                 return err;
@@ -661,18 +635,25 @@ public final class TypeChecker {
     }
 
     private static Expression makePromoterToDouble(Expression exp) {
-        AtomicSequenceConverter asc = new AtomicSequenceConverter(exp, BuiltInAtomicType.DOUBLE);
-        Converter converter = new Converter.PromoterToDouble();
-        converter.setConversionRules(exp.getConfiguration().getConversionRules());
-        asc.setConverter(converter);
-        ExpressionTool.copyLocationInfo(exp, asc);
-        return asc;
+        return makePromoter(exp, new Converter.PromoterToDouble(), BuiltInAtomicType.DOUBLE);
     }
 
     private static Expression makePromoterToFloat(Expression exp) {
-        AtomicSequenceConverter asc = new AtomicSequenceConverter(exp, BuiltInAtomicType.FLOAT);
-        Converter converter = new Converter.PromoterToFloat();
-        converter.setConversionRules(exp.getConfiguration().getConversionRules());
+        return makePromoter(exp, new Converter.PromoterToFloat(), BuiltInAtomicType.FLOAT);
+    }
+
+    private static Expression makePromoter(Expression exp, Converter converter, BuiltInAtomicType type) {
+        ConversionRules rules = exp.getConfiguration().getConversionRules();
+        converter.setConversionRules(rules);
+        if (exp instanceof Literal && ((Literal) exp).getValue() instanceof AtomicValue) {
+            ConversionResult result = converter.convert((AtomicValue) ((Literal) exp).getValue());
+            if (result instanceof AtomicValue) {
+                Literal converted = Literal.makeLiteral((AtomicValue) result);
+                ExpressionTool.copyLocationInfo(exp, converted);
+                return converted;
+            }
+        }
+        AtomicSequenceConverter asc = new AtomicSequenceConverter(exp, type);
         asc.setConverter(converter);
         ExpressionTool.copyLocationInfo(exp, asc);
         return asc;
