@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -9,7 +9,7 @@ package net.sf.saxon.expr.sort;
 
 import net.sf.saxon.om.AtomicSequence;
 import net.sf.saxon.om.Item;
-import net.sf.saxon.om.SequenceIterator;
+import net.sf.saxon.om.MemoSequence;
 
 /**
  * This class is a specialization of ObjectToBeSorted for use when the sequence
@@ -20,7 +20,7 @@ import net.sf.saxon.om.SequenceIterator;
 public class GroupToBeSorted extends ObjectToBeSorted<Item> {
 
     /*@Nullable*/ public AtomicSequence currentGroupingKey;
-    public SequenceIterator currentGroupIterator;
+    public MemoSequence currentGroup;
 
     public GroupToBeSorted(int numberOfSortKeys) {
         super(numberOfSortKeys);
