@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,10 +7,6 @@
 
 package net.sf.saxon.expr.sort;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.SortExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.SortExpressionAdjunct;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.parser.*;
 import net.sf.saxon.om.SequenceIterator;
@@ -22,8 +18,6 @@ import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.SequenceType;
 
-import java.util.Map;
-
 /**
  * Expression equivalent to the imaginary syntax
  * expr sortby (sort-key)+
@@ -196,11 +190,12 @@ public class SortExpression extends Expression
             comparators = new AtomicComparer[getSortKeyDefinitionList().size()];
         }
 
+        TypeChecker tc = visitor.getConfiguration().getTypeChecker(false);
         for (int i = 0; i < getSortKeyDefinitionList().size(); i++) {
             SortKeyDefinition sortKeyDef = getSortKeyDefinition(i);
             Expression sortKey = sortKeyDef.getSortKey();
             if (sortKeyDef.isSetContextForSortKey()) {
-                ContextItemStaticInfo cit = new ContextItemStaticInfo(sortedItemType, false);
+                ContextItemStaticInfo cit = visitor.getConfiguration().makeContextItemStaticInfo(sortedItemType, false);
                 sortKey = sortKey.typeCheck(visitor, cit);
             } else {
                 sortKey = sortKey.typeCheck(visitor, contextInfo);
@@ -211,7 +206,7 @@ public class SortExpression extends Expression
                 RoleDiagnostic role =
                         new RoleDiagnostic(RoleDiagnostic.INSTRUCTION, "xsl:sort/select", 0);
                 role.setErrorCode("XTTE1020");
-                sortKey = TypeChecker.staticTypeCheck(sortKey, SequenceType.OPTIONAL_ATOMIC, false, role, visitor);
+                sortKey = tc.staticTypeCheck(sortKey, SequenceType.OPTIONAL_ATOMIC, role, visitor);
                 //sortKey = CardinalityChecker.makeCardinalityChecker(sortKey, StaticProperty.ALLOWS_ZERO_OR_ONE, role);
             }
             sortKeyDef.setSortKey(sortKey, sortKeyDef.isSetContextForSortKey());
@@ -258,7 +253,7 @@ public class SortExpression extends Expression
         ContextItemStaticInfo cit;
         if (getSortKeyDefinition(0).isSetContextForSortKey()) {
             ItemType sortedItemType = getSelect().getItemType();
-            cit = new ContextItemStaticInfo(sortedItemType, false);
+            cit = visitor.getConfiguration().makeContextItemStaticInfo(sortedItemType, false);
         } else {
             cit = contextItemType;
         }
@@ -294,41 +289,6 @@ public class SortExpression extends Expression
         return se2;
     }
 
-    /**
-     * Offer promotion for this subexpression. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is not accepted - this is appropriate in the case of simple expressions
-     * such as constant values and variable references where promotion would give no performance
-     * advantage. This method is always called at compile time.
-     *
-     *
-     * @param offer  details of the offer, for example the offer to move
-     *               expressions that don't depend on the context to an outer level in
-     *               the containing expression
-     * @return if the offer is not accepted, return this expression unchanged.
-     *         Otherwise return the result of rewriting the expression to promote
-     *         this subexpression
-     * @throws net.sf.saxon.trans.XPathException
-     *          if any error is detected
-     */
-
-    public Expression promote(PromotionOffer offer) throws XPathException {
-        Expression exp = offer.accept(this);
-        if (exp != null) {
-            return exp;
-        } else {
-            setSelect(doPromotion(getSelect(), offer));
-            for (SortKeyDefinition sortKeyDefinition : getSortKeyDefinitionList()) {
-                final Expression sk2 = sortKeyDefinition.getSortKey().promote(offer);
-                sortKeyDefinition.setSortKey(sk2, true);
-                for (Operand o : sortKeyDefinition.operands()) {
-                    o.setChildExpression(o.getChildExpression().promote(offer));
-                }
-            }
-            return this;
-        }
-    }
-
     /**
      * Test whether a given expression is one of the sort keys
      *
@@ -424,7 +384,7 @@ public class SortExpression extends Expression
             }
         }
         iter = new SortedIterator(context, iter, this, comps, getSortKeyDefinition(0).isSetContextForSortKey());
-        ((SortedIterator) iter).setHostLanguage(getHostLanguage());
+        ((SortedIterator) iter).setHostLanguage(getPackageData().getHostLanguage());
         return iter;
     }
 
@@ -441,32 +401,17 @@ public class SortExpression extends Expression
         return "sort(" + getBaseExpression().toShortString() + ")";
     }
 
-//#ifdefined BYTECODE
 
     /**
-     * Return the bytecode compiler for the Sort expression
+     * Get the (partial) name of a class that supports streaming of this kind of expression
      *
-     * @return the relevant ExpressionCompiler
+     * @return the partial name of a class that can be instantiated to provide streaming support in Saxon-EE,
+     * or null if there is no such class
      */
     @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new SortExpressionCompiler();
+    public String getStreamerName() {
+        return "SortExpression";
     }
-//#endif
-
-//#ifdefined STREAM
-
-    /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
-     */
-    @Override
-    public StreamingAdjunct getStreamingAdjunct() {
-        return new SortExpressionAdjunct();
-    }
-
-//#endif
 
     /**
      * Diagnostic print of expression structure. The abstract expression tree
