@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,8 +7,6 @@
 
 package net.sf.saxon.functions;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.bytecode.RoundingCompiler;
 import net.sf.saxon.expr.XPathContext;
 import net.sf.saxon.om.Item;
 import net.sf.saxon.value.AtomicValue;
@@ -25,18 +23,9 @@ public final class Abs extends ScalarSystemFunction {
         return ((NumericValue)arg).abs();
     }
 
-    //#ifdefined BYTECODE
-
-    /**
-     * Return the compiler of the Round expression
-     *
-     * @return the relevant ExpressionCompiler
-     */
-    @Override
-    public ExpressionCompiler getExpressionCompiler() {
-        return new RoundingCompiler();
+    public String getCompilerName() {
+        return "RoundingCompiler";
     }
-//#endif
 
 
 }
