@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,8 +7,6 @@
 
 package net.sf.saxon.functions;
 
-import com.saxonica.ee.stream.adjunct.DeepEqualAdjunct;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.XPathContext;
 import net.sf.saxon.expr.sort.AtomicComparer;
@@ -274,11 +272,12 @@ public class DeepEqual extends CollatingFunctionFixed {
                 }
                 AxisIterator a1 = n1.iterateAxis(AxisInfo.ATTRIBUTE);
                 AxisIterator a2 = n2.iterateAxis(AxisInfo.ATTRIBUTE);
-                if (Count.count(a1.getAnother()) != Count.count(a2)) {
+                if (!SequenceTool.sameLength(a1, a2)) {
                     explain(config, "elements have different number of attributes", flags, n1, n2);
                     return false;
                 }
                 NodeInfo att1;
+                a1 = n1.iterateAxis(AxisInfo.ATTRIBUTE);
                 while ((att1 = a1.next()) != null) {
                     AxisIterator a2iter = n2.iterateAxis(AxisInfo.ATTRIBUTE,
                             new SameNameTest(att1));
@@ -587,14 +586,9 @@ public class DeepEqual extends CollatingFunctionFixed {
         return BooleanValue.get(b);
     }
 
-//#ifdefined STREAM
-
-
-    //@Override
-    public StreamingAdjunct getStreamingAdjunct() {
-        return new DeepEqualAdjunct();
+    public String getStreamerName() {
+        return "DeepEqual";
     }
 
-//#endif
 }
 
