@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -30,7 +30,6 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.tiny.Statistics;
 import net.sf.saxon.tree.tiny.TinyBuilder;
 import net.sf.saxon.value.Cardinality;
-import net.sf.saxon.value.Whitespace;
 
 import javax.xml.transform.Source;
 import javax.xml.transform.SourceLocator;
@@ -48,6 +47,8 @@ import java.net.URISyntaxException;
 
 public class DocumentFn extends SystemFunction implements Callable {
 
+    private Location location = null;
+
     /**
      * Determine the static cardinality
      */
@@ -79,11 +80,11 @@ public class DocumentFn extends SystemFunction implements Callable {
         // with the same arguments will produce identical results.
     }
 
-
     @Override
     public Expression makeFunctionCall(Expression... arguments) {
+        location = arguments[0].getLocation();
         Expression expr = Doc.maybePreEvaluate(this, arguments);
-               return expr == null ? super.makeFunctionCall(arguments) : expr;
+        return expr == null ? super.makeFunctionCall(arguments) : expr;
     }
 
     /**
@@ -113,7 +114,7 @@ public class DocumentFn extends SystemFunction implements Callable {
         map.baseURI = baseURI;
         map.stylesheetURI = getStaticBaseUriString();
         map.packageData = getRetainedStaticContext().getPackageData();
-        //map.locator = this;
+        map.locator = location;
 
         ItemMappingIterator iter = new ItemMappingIterator(hrefSequence, map);
 
@@ -151,9 +152,7 @@ public class DocumentFn extends SystemFunction implements Callable {
             } catch (XPathException xerr) {
                 if (xerr.getErrorCodeLocalPart().equals("XTRE1160")) {
                     // Invalid fragment identifier: error code changes in XSLT 3.0
-                    if (context.getController().getExecutable().isAllowXPath30()) {
-                        xerr.setErrorCode("XTDE1160");
-                    }
+                    xerr.setErrorCode("XTDE1160");
                     // Report a "recoverable error"
                     try {
                         context.getController().recoverableError(xerr);
@@ -314,8 +313,8 @@ public class DocumentFn extends SystemFunction implements Callable {
                     eh.setSilent(true);
                     options.setErrorHandler(eh);
                 }
-                if (controller.getExecutable().stripsInputTypeAnnotations()) {
-                    s = controller.getConfiguration().getAnnotationStripper(s);
+                if (packageData instanceof StylesheetPackage && ((StylesheetPackage)packageData).isStripsTypeAnnotations()) {
+                    s = config.getAnnotationStripper(s);
                 }
 
                 PathMap map = controller.getPathMapForDocumentProjection();
@@ -453,7 +452,8 @@ public class DocumentFn extends SystemFunction implements Callable {
      * @param baseURI  the base URI
      * @param packageData
      * @param resolver the URIResolver
-     * @param strip true if the document is subject to whitespace-stripping (true for source documents, false for stylesheets)
+     * @param strip true if the document is subject to whitespace stripping (typically a source document), false
+     *              otherwise (typically a stylesheet module)
      */
 
     public static DocumentURI computeDocumentKey(String href, String baseURI, PackageData packageData, URIResolver resolver, boolean strip) throws XPathException {
@@ -603,42 +603,6 @@ public class DocumentFn extends SystemFunction implements Callable {
     }
 
 
-    /**
-     * Copy the documents identified by this expression to a given Receiver. This method is used only when it is
-     * known that the documents are being copied, because there is then no problem about node identity.
-     *
-     * @param context the XPath dynamic context
-     * @param args the arguments to the document() function
-     * @param out     the destination to which the documents will be sent
-     */
-
-    public void sendDocuments(XPathContext context, Expression[] args, Receiver out) throws XPathException {
-        SequenceIterator hrefSequence = args[0].iterate(context);
-        String explicitBaseURI = null;
-        if (getArity() == 2) {
-            // we can trust the type checking: it must be a node
-            NodeInfo base = (NodeInfo) args[1].evaluateItem(context);
-            explicitBaseURI = base.getBaseURI();
-        }
-        ParseOptions options = new ParseOptions();
-        while (true) {
-            Item href = hrefSequence.next();
-            if (href == null) {
-                break;
-            }
-            String base;
-            if (explicitBaseURI == null) {
-                if (href instanceof NodeInfo) {
-                    base = ((NodeInfo) href).getBaseURI();
-                } else {
-                    base = getStaticBaseUriString();
-                }
-            } else {
-                base = explicitBaseURI;
-            }
-            sendDoc(href.getStringValue(), base, context, null, out, options);
-        }
-    }
 
     /**
      * Supporting routine to push one external document given a URI (href) and a baseURI to a given Receiver.
@@ -722,18 +686,13 @@ public class DocumentFn extends SystemFunction implements Callable {
                 throw xerr;
             }
         }
-        ParseOptions options = new ParseOptions(parseOptions);
-        options.setStripSpace(Whitespace.XSLT);
 
-        if (controller.getExecutable().stripsInputTypeAnnotations()) {
-            out = controller.getConfiguration().getAnnotationStripper(out);
-        }
         if (controller.getConfiguration().isTiming()) {
             controller.getConfiguration().getLogger().info("Streaming input document " + source.getSystemId());
         }
         out.setPipelineConfiguration(pipe);
         try {
-            Sender.send(source, out, options);
+            Sender.send(source, out, parseOptions);
         } catch (XPathException e) {
             e.maybeSetLocation(locator);
             e.maybeSetErrorCode("FODC0002");
