@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,12 +7,12 @@
 
 package net.sf.saxon.functions;
 
-import com.saxonica.ee.bytecode.ExpressionCompiler;
-import com.saxonica.ee.stream.adjunct.StreamingAdjunct;
 import net.sf.saxon.expr.*;
+import net.sf.saxon.expr.instruct.UserFunction;
 import net.sf.saxon.expr.parser.ContextItemStaticInfo;
 import net.sf.saxon.expr.parser.ExpressionVisitor;
 import net.sf.saxon.expr.parser.RetainedStaticContext;
+import net.sf.saxon.functions.registry.BuiltInFunctionSet;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.om.*;
 import net.sf.saxon.trace.ExpressionPresenter;
@@ -20,6 +20,7 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.FunctionItemType;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.SpecificFunctionType;
+import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.IntegerValue;
 import net.sf.saxon.value.SequenceType;
 
@@ -33,7 +34,7 @@ import java.util.Properties;
 public abstract class SystemFunction extends AbstractFunction {
 
     private int arity;
-    private StandardFunction.Entry details;
+    private BuiltInFunctionSet.Entry details;
     private RetainedStaticContext retainedStaticContext;
 
     /**
@@ -52,7 +53,9 @@ public abstract class SystemFunction extends AbstractFunction {
         if (f == null) {
             return null;
         }
-        return f.makeFunctionCall(arguments);
+        Expression expr = f.makeFunctionCall(arguments);
+        expr.setRetainedStaticContext(rsc);
+        return expr;
     }
 
     /**
@@ -65,23 +68,16 @@ public abstract class SystemFunction extends AbstractFunction {
      */
 
     public static SystemFunction makeFunction(String name, RetainedStaticContext rsc, int arity) {
-        StandardFunction.Entry entry = StandardFunction.getFunction(name, arity);
-        if (entry == null) {
-            return null;
+        if (rsc == null) {
+            throw new NullPointerException();
         }
-
-        Class functionClass = entry.implementationClass;
-        try {
-            SystemFunction f = (SystemFunction) functionClass.newInstance();
-            f.arity = arity;
-            f.setDetails(entry);
-            f.setRetainedStaticContext(rsc);
-            return f;
-        } catch (IllegalAccessException err) {
-            return null;
-        } catch (InstantiationException err) {
-            return null;
+        SystemFunction fn = rsc.getConfiguration().makeSystemFunction(name, arity);
+        if (fn == null) {
+            rsc.getConfiguration().makeSystemFunction(name, arity);
+            throw new IllegalStateException(name);
         }
+        fn.setRetainedStaticContext(rsc);
+        return fn;
     }
 
     /**
@@ -107,6 +103,17 @@ public abstract class SystemFunction extends AbstractFunction {
         this.arity = arity;
     }
 
+    /**
+     * Get an estimate of the net cost of evaluating the function, excluding the cost of evaluating
+     * its arguments. The result is 0 for very simple functions like position() and exists(), 1 by
+     * default, and higher values for particularly expensive functions.
+     * @return the estimated cost
+     */
+
+    public int getNetCost() {
+        return 1;
+    }
+
     /**
      * Allow the function to create an optimized call based on the values of the actual arguments
      * @param visitor the expression visitor
@@ -169,13 +176,22 @@ public abstract class SystemFunction extends AbstractFunction {
         this.retainedStaticContext = retainedStaticContext;
     }
 
+    /**
+     * Ask whether the result of the function depends on the context item
+     * @return true if the function depends on the context item
+     */
+
+    public boolean dependsOnContextItem() {
+        return (details.properties & (BuiltInFunctionSet.CITEM | BuiltInFunctionSet.CDOC)) != 0;
+    }
+
     /**
      * Set the details of this type of function
      *
      * @param entry information giving details of the function signature and other function properties
      */
 
-    public void setDetails(StandardFunction.Entry entry) {
+    public void setDetails(BuiltInFunctionSet.Entry entry) {
         details = entry;
     }
 
@@ -185,7 +201,7 @@ public abstract class SystemFunction extends AbstractFunction {
      * @return information about the function signature and other function properties
      */
 
-    public StandardFunction.Entry getDetails() {
+    public BuiltInFunctionSet.Entry getDetails() {
         return details;
     }
 
@@ -196,7 +212,7 @@ public abstract class SystemFunction extends AbstractFunction {
      */
 
     public StructuredQName getFunctionName() {
-        return new StructuredQName("", NamespaceConstant.FN, details.name);
+        return details.name;
     }
 
     /**
@@ -207,7 +223,7 @@ public abstract class SystemFunction extends AbstractFunction {
      * @return a description of the function for use in error messages
      */
     public String getDescription() {
-        return "fn:" + details.name;
+        return details.name.getDisplayName();
     }
 
     /**
@@ -356,9 +372,9 @@ public abstract class SystemFunction extends AbstractFunction {
      */
 
     public ItemType getResultItemType(Expression[] args) {
-        if ((details.properties & StandardFunction.AS_ARG0) != 0) {
+        if ((details.properties & BuiltInFunctionSet.AS_ARG0) != 0) {
             return args[0].getItemType();
-        } else if ((details.properties & StandardFunction.AS_PRIM_ARG0) != 0) {
+        } else if ((details.properties & BuiltInFunctionSet.AS_PRIM_ARG0) != 0) {
             return args[0].getItemType().getPrimitiveItemType();
         } else {
             return details.itemType;
@@ -372,21 +388,31 @@ public abstract class SystemFunction extends AbstractFunction {
      */
 
     public int getCardinality(Expression[] args) {
-        return details.cardinality;
+        int c = details.cardinality;
+        if (c == BuiltInFunctionSet.OPT && (details.properties & BuiltInFunctionSet.CARD0) != 0 &&
+                !Cardinality.allowsZero(args[0].getCardinality())) {
+            return StaticProperty.EXACTLY_ONE;
+        } else {
+            return c;
+        }
     }
 
     /**
      * Determine the special properties of this function. The general rule
-     * is that a system function call is non-creative if its return type is
-     * atomic, or if all its arguments are non-creative. This is overridden
-     * for the generate-id() function, which is considered creative if
-     * its operand is creative (because the result depends on the
-     * identity of the operand)
+     * is that a system function call is non-creative unless more details
+     * are defined in a subclass.
      * @param arguments the actual arguments supplied in a call to the function
      */
 
     public int getSpecialProperties(Expression[] arguments) {
-        return details==null ? 0 : StaticProperty.NON_CREATIVE;
+        if ((details.properties & BuiltInFunctionSet.NEW) != 0) {
+            return StaticProperty.ALL_NODES_NEWLY_CREATED;
+        }
+        int p = StaticProperty.NON_CREATIVE;
+        if ((details.properties & BuiltInFunctionSet.SIDE) != 0) {
+             p |= StaticProperty.HAS_SIDE_EFFECTS;
+        }
+        return p;
     }
 
     /**
@@ -413,6 +439,23 @@ public abstract class SystemFunction extends AbstractFunction {
         }
     }
 
+    /**
+     * Make a dynamic call to a supplied argument function (convenience method for use by implementations)
+     *
+     * @param context the XPath dynamic evaluation context
+     * @param args    the actual arguments to be supplied
+     * @return the result of invoking the function
+     * @throws XPathException if a dynamic error occurs within the function
+     */
+
+    public static Sequence dynamicCall(Function f, XPathContext context, Sequence[] args) throws XPathException {
+        if (f instanceof UserFunction) {
+            context = ((UserFunction)f).makeNewContext(context);
+        }
+        return f.call(context, args);
+    }
+
+
     /**
      * Diagnostic print of expression structure. The abstract expression tree
      * is written to the supplied output destination.
@@ -420,9 +463,11 @@ public abstract class SystemFunction extends AbstractFunction {
 
     public void export(ExpressionPresenter out) throws XPathException {
         out.startElement("fnRef");
-        out.emitAttribute("name", getFunctionName().getLocalPart());
+        StructuredQName qName = getFunctionName();
+        String name = qName.hasURI(NamespaceConstant.FN) ? qName.getLocalPart() : qName.getEQName();
+        out.emitAttribute("name", name);
         out.emitAttribute("arity", getArity() + "");
-        if ((getDetails().properties & StandardFunction.DEPENDS_ON_STATIC_CONTEXT) != 0) {
+        if ((getDetails().properties & BuiltInFunctionSet.DEPENDS_ON_STATIC_CONTEXT) != 0) {
             out.emitRetainedStaticContext(getRetainedStaticContext(), null);
         }
         out.endElement();
@@ -437,6 +482,10 @@ public abstract class SystemFunction extends AbstractFunction {
         return caller;
     }
 
+    public boolean isTrustedResultType() {
+        return true;
+    }
+
     public String getStaticBaseUriString() {
         return getRetainedStaticContext().getStaticBaseUriString();
     }
@@ -449,8 +498,12 @@ public abstract class SystemFunction extends AbstractFunction {
 
     public void exportAttributes(ExpressionPresenter out) {};
 
-    public boolean isTrustedResultType() {
-        return true;
+    /**
+     * Export any implicit arguments held in optimized form within the SystemFunction call
+     * @param out the export destination
+     */
+
+    public void exportAdditionalArguments(SystemFunctionCall call, ExpressionPresenter out) throws XPathException {
     }
 
     /**
@@ -462,27 +515,30 @@ public abstract class SystemFunction extends AbstractFunction {
 
     public void importAttributes(Properties attributes) throws XPathException {}
 
-//#ifdefined BYTECODE
     /**
-     * Return the bytecode compiler for a static call to the function in question.
-     *
-     * @return the relevant ExpressionCompiler for the corresponding function call
+     * Get the (local) name of a class that can be used to generate bytecode for this
+     * system function
+     * @return the name of a bytecode generation class, or null if there is no bytecode
+     * support for this function
      */
-    public ExpressionCompiler getExpressionCompiler() {
+    public String getCompilerName() {
         return null;
     }
-//#endif
-//#ifdefined STREAM
+
     /**
-     * Get a class that supports streamed evaluation of this expression
-     *
-     * @return the relevant StreamingAdjunct, or null if none is available
+     * Get a name that identifies the class that can be used to evaluate this function
+     * in streaming mode, that is, supplying one item or event at a time as it is
+     * delivered by the parser. Special streaming support is generally available in
+     * Saxon-EE for all system functions that take a sequence as their argument.
+     * @return a name that identifies a class that supports streamed evaluation of this system
+     * function, or null if no such class is available. The mapping of names to classes is
+     * defined in the Saxon-EE class {@code StreamerMap}. In non-streaming Saxon editions
+     * (HE and PE) the method always returns null.
      */
 
-    public StreamingAdjunct getStreamingAdjunct() {
+    public String getStreamerName() {
         return null;
     }
-//#endif
 
 }
 
