@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -43,7 +43,7 @@ import java.util.Comparator;
  * javax.xml.transform.TransformerFactory and javax.xml.transform.sax.SAXTransformerFactory
  * classes.</p>
  * <p/>
- * <p>In Saxon 9.6, the JAXP transformation interface is re-implemented as a layer
+ * <p>Since Saxon 9.6, the JAXP transformation interface is re-implemented as a layer
  * on top of the s9api interface. This will affect applications that attempt to
  * down-cast from JAXP interfaces to the underlying implementation classes.</p>
  */
@@ -82,7 +82,8 @@ public class SaxonTransformerFactory extends SAXTransformerFactory {
     /**
      * Get the configuration. This can also be done using the JAXP method
      * getAttribute, with the attribute name {@link net.sf.saxon.lib.FeatureKeys#CONFIGURATION}
-     *
+     * This is a trapdoor method that provides access to underlying implementation details that
+     * may change in subsequent Saxon releases.
      * @return the Saxon configuration
      */
 
@@ -464,7 +465,7 @@ public class SaxonTransformerFactory extends SAXTransformerFactory {
         if (!(templates instanceof TemplatesImpl)) {
             throw new TransformerConfigurationException("Supplied Templates object was not created using Saxon");
         }
-        TransformerImpl transformer = (TransformerImpl) templates.newTransformer();
+        AbstractTransformerImpl transformer = (AbstractTransformerImpl) templates.newTransformer();
         return transformer.newXMLFilter();
     }
 
@@ -519,5 +520,15 @@ public class SaxonTransformerFactory extends SAXTransformerFactory {
         }
     }
 
+    /**
+     * Get the underlying s9api Processor. (Trapdoor method providing access to underlying
+     * implementation details which may change in subsequent releases)
+     * @since 9.8.0.5
+     */
+
+    public Processor getProcessor() {
+        return processor;
+    }
+
 }
 
