@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -11,10 +11,9 @@ import net.sf.saxon.Configuration;
 import net.sf.saxon.om.NameChecker;
 import net.sf.saxon.om.NamespaceResolver;
 import net.sf.saxon.om.QNameException;
-import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.ValidationException;
-import net.sf.saxon.value.DecimalValue;
+import net.sf.saxon.value.BigDecimalValue;
 
 import javax.xml.transform.OutputKeys;
 import java.math.BigDecimal;
@@ -98,6 +97,15 @@ public class SaxonOutputKeys {
 
     /*@NotNull*/ public static final String LINE_LENGTH = "{http://saxon.sf.net/}line-length";
 
+    /**
+     * saxon:single-quotes = boolean.
+     * <p/>
+     * <p>Indicates that attributes should be delimited by apostrophes rather than quotation marks</p>
+     */
+
+    /*@NotNull*/ public static final String SINGLE_QUOTES = "{http://saxon.sf.net/}single-quotes";
+
+
     /**
      * suppress-indentation = list of element names
      * <p/>
@@ -220,6 +228,17 @@ public class SaxonOutputKeys {
     /*@NotNull*/ public static final String NEXT_IN_CHAIN = "{http://saxon.sf.net/}next-in-chain";
     /*@NotNull*/ public static final String NEXT_IN_CHAIN_BASE_URI = "{http://saxon.sf.net/}next-in-chain-base-uri";
 
+    /**
+     * parameter-document = URI.
+     * <p/>
+     * <p>Indicates that the output is to be piped into another XSLT stylesheet
+     * to perform another transformation. The auxiliary property NEXT_IN_CHAIN_BASE_URI
+     * records the base URI of the stylesheet element where this attribute was found.</p>
+     */
+    /*@NotNull*/ public static final String PARAMETER_DOCUMENT = "parameter-document";
+    /*@NotNull*/ public static final String PARAMETER_DOCUMENT_BASE_URI = "{http://saxon.sf.net/}parameter-document-base-uri";
+
+
     /**
      * byte-order-mark = yes|no.
      * <p/>
@@ -296,212 +315,12 @@ public class SaxonOutputKeys {
     public static final String UNFAILING = "{http://saxon.sf.net/}unfailing";
 
 
-    /**
-     * Check that a supplied output property is valid, and normalize the value (specifically in the case of boolean
-     * values where yes|true|1 are normalized to "yes", and no|false|0 are normalized to "no").
-     *
-     *
-     * @param key    the name of the property, in Clark format
-     * @param value  the value of the property. This may be set to null, in which case no validation takes place.
-     *               The value must be in JAXP format, that is, with lexical QNames expanded to Clark names
-     * @param config the Saxon Configuration. May be null, in which case validation may be incomplete
-     * @param allow30 true if serializer xslt 3.0 rules are followed
-     * @return     normalized value of the property, or null if the supplied value is null
-     *
-     * @throws XPathException if the property name or value is invalid
-     */
-
-    public static String checkOutputProperty(/*@NotNull*/ String key, /*@Nullable*/ String value, /*@Nullable*/ Configuration config, boolean allow30) throws XPathException {
-        if (!key.startsWith("{") || key.startsWith("{http://saxon.sf.net/}")) {
-            if (key.equals(ALLOW_DUPLICATE_NAMES)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(BUILD_TREE)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(BYTE_ORDER_MARK)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(OutputKeys.CDATA_SECTION_ELEMENTS)) {
-                if (value != null) {
-                    checkListOfClarkNames(key, value);
-                }
-            } else if (key.equals(OutputKeys.DOCTYPE_PUBLIC)) {
-                if (value != null) {
-                    checkPublicIdentifier(value);
-                }
-            } else if (key.equals(OutputKeys.DOCTYPE_SYSTEM)) {
-                if (value != null) {
-                    checkSystemIdentifier(value);
-                }
-            } else if (key.equals(OutputKeys.ENCODING)) {
-                // no constraints
-            } else if (key.equals(ESCAPE_URI_ATTRIBUTES)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(INCLUDE_CONTENT_TYPE)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(OutputKeys.INDENT)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(OutputKeys.MEDIA_TYPE)) {
-                // no constraints
-            } else if (key.equals(OutputKeys.METHOD) || key.equals(JSON_NODE_OUTPUT_METHOD)) {
-                if (value != null) {
-                    checkMethod(key, value, allow30, config);
-                }
-            } else if (key.equals(NEWLINE)) {
-                // no constraints
-            } else if (key.equals(NORMALIZATION_FORM)) {
-                if (value != null) {
-                    checkNormalizationForm(value);
-                }
-            } else if (key.equals(OutputKeys.OMIT_XML_DECLARATION)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(OutputKeys.STANDALONE)) {
-                if (value != null && !value.equals("omit")) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(UNDECLARE_PREFIXES)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(USE_CHARACTER_MAPS)) {
-                if (value != null) {
-                    checkListOfClarkNames(key, value);
-                }
-            } else if (key.equals(OutputKeys.VERSION)) {
-                // no constraints
-            } else if (key.equals(STYLESHEET_VERSION)) {
-                // no constraints
-            } else if (key.equals(INDENT_SPACES)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    checkNonNegativeInteger(key, value);
-                }
-            } else if (key.equals(LINE_LENGTH)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    checkNonNegativeInteger(key, value);
-                }
-            } else if (key.equals(CHARACTER_REPRESENTATION)) {
-                checkExtensions(key, config);
-            } else if (key.equals(NEXT_IN_CHAIN)) {
-                checkExtensions(key, config);
-            } else if (key.equals(NEXT_IN_CHAIN_BASE_URI)) {
-                // no validation performed
-            } else if (key.equals(REQUIRE_WELL_FORMED)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(RECOGNIZE_BINARY)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(SUPPRESS_INDENTATION)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    checkListOfClarkNames(key, value);
-                }
-
-            } else if (key.equals(DOUBLE_SPACE)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    checkListOfClarkNames(key, value);
-                }
-
-            } else if (key.equals(WRAP)) {
-                if (value != null) {
-                    checkExtensions(key, config);
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(SUPPLY_SOURCE_LOCATOR)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(HTML_VERSION)) {
-                if (value != null) {
-                    checkDecimal(key, value);
-                }
-            } else if (key.equals(UNFAILING)) {
-                if (value != null) {
-                    value = checkYesOrNo(key, value, allow30);
-                }
-            } else if (key.equals(ITEM_SEPARATOR)) {
-                // no checking needed
-            } else {
-                throw new XPathException("Unknown serialization parameter " + Err.wrap(key), "XQST0109");
-            }
-        } else {
-            //return;
-        }
-        return value;
-    }
-
     private static void checkExtensions(String key, /*@Nullable*/ Configuration config) throws XPathException {
         if (config != null) {
             config.checkLicensedFeature(Configuration.LicenseFeature.PROFESSIONAL_EDITION, "custom serialization " + key, -1);
         }
     }
 
-    private static String checkYesOrNo(String key, String value, boolean allow30) throws XPathException {
-        if (allow30){
-            if ("yes".equals(value)||"true".equals(value)||"1".equals(value)){
-                return "yes";
-            } else if ("no".equals(value)||"false".equals(value)||"0".equals(value)){
-                return "no";
-            }
-            else {
-                throw new XPathException("Serialization parameter " + Err.wrap(key) + " must have the value yes|no, true|false, or 1|0", "SEPM0016");
-            }
-        } else if ("yes".equals(value) || "no".equals(value)) {
-            return value;
-        } else {
-            throw new XPathException("Serialization parameter " + Err.wrap(key) + " must have the value yes or no", "SEPM0016");
-        }
-    }
-
-    private static void checkMethod(String key, String value, boolean allow30, Configuration config) throws XPathException {
-        if (!"xml".equals(value) && !"html".equals(value) && !"xhtml".equals(value) && !"text".equals(value)) {
-            if (allow30 && !JSON_NODE_OUTPUT_METHOD.equals(key) && ("json".equals(value) || "adaptive".equals(value))) {
-                return;
-            }
-            if (isValidClarkName(value)) {
-                checkExtensions(value, config);
-            } else {
-                throw new XPathException("Invalid value for serialization method: " +
-                        "must be xml, html, xhtml, text, or a QName in '{uri}local' form", "SEPM0016");
-            }
-        }
-
-    }
-
-    private static void checkNormalizationForm(String value) throws XPathException {
-        if (!NameChecker.isValidNmtoken(value)) {
-            throw new XPathException("Invalid value for normalization-form: " +
-                    "must be NFC, NFD, NFKC, NFKD, fully-normalized, or none", "SEPM0016");
-        }
-//        if ("NFC".equals(value)) return;
-//        if ("NFD".equals(value)) return;
-//        if ("NFKC".equals(value)) return;
-//        if ("NFKD".equals(value)) return;
-//        if ("fully-normalized".equals(value)) return;
-//        if ("none".equals(value)) return;
-//        throw new XPathException("Invalid value for normalization-form: " +
-//                "must be NFC, NFD, NFKC, NFKD, fully-normalized, or none");
-
-    }
-
     private static boolean isValidClarkName(/*@NotNull*/ String value) {
         if (value.charAt(0) != '{') {
             return false;
@@ -512,51 +331,8 @@ public class SaxonOutputKeys {
                 NameChecker.isValidNCName(value.substring(closer + 1));
     }
 
-    private static void checkNonNegativeInteger(String key, String value) throws XPathException {
-        try {
-            int n = Integer.parseInt(value);
-            if (n < 0) {
-                throw new XPathException("Value of " + Err.wrap(key) + " must be a non-negative integer", "SEPM0016");
-            }
-        } catch (NumberFormatException err) {
-            throw new XPathException("Value of " + Err.wrap(key) + " must be a non-negative integer", "SEPM0016");
-        }
-    }
-
-    private static void checkDecimal(String key, String value) throws XPathException {
-        if (!DecimalValue.castableAsDecimal(value)) {
-            throw new XPathException("Value of " + Err.wrap(key) +
-                    " must be a decimal number", "SEPM0016");
-        }
-    }
-
-    private static void checkListOfClarkNames(String key, String value) throws XPathException {
-        StringTokenizer tok = new StringTokenizer(value, " \t\n\r", false);
-        while (tok.hasMoreTokens()) {
-            String s = tok.nextToken();
-            if (isValidClarkName(s) || NameChecker.isValidNCName(s)) {
-                // ok
-            } else {
-                throw new XPathException("Value of " + Err.wrap(key) +
-                        " must be a list of QNames in '{uri}local' notation", "SEPM0016");
-            }
-        }
-    }
-
     private static Pattern publicIdPattern = Pattern.compile("^[\\s\\r\\na-zA-Z0-9\\-'()+,./:=?;!*#@$_%]*$");
 
-    private static void checkPublicIdentifier(String value) throws XPathException {
-        if (!publicIdPattern.matcher(value).matches()) {
-            throw new XPathException("Invalid character in doctype-public parameter", "SEPM0016");
-        }
-    }
-
-    private static void checkSystemIdentifier(/*@NotNull*/ String value) throws XPathException {
-        if (value.contains("'") && value.contains("\"")) {
-            throw new XPathException("The doctype-system parameter must not contain both an apostrophe and a quotation mark", "SEPM0016");
-        }
-    }
-
     /**
      * Process a serialization property whose value is a list of element names, for example cdata-section-elements
      *
@@ -612,7 +388,7 @@ public class SaxonOutputKeys {
         String htmlVersion = properties.getProperty(SaxonOutputKeys.HTML_VERSION);
         try {
             return htmlVersion != null &&
-                    ((DecimalValue) DecimalValue.makeDecimalValue(htmlVersion, false).asAtomic())
+                    ((BigDecimalValue) BigDecimalValue.makeDecimalValue(htmlVersion, false).asAtomic())
                             .getDecimalValue().equals(BigDecimal.valueOf(5));
         } catch (ValidationException e) {
             return false;
@@ -635,7 +411,7 @@ public class SaxonOutputKeys {
         }
         if (htmlVersion != null) {
             try {
-                return ((DecimalValue) DecimalValue.makeDecimalValue(htmlVersion, false).asAtomic())
+                return ((BigDecimalValue) BigDecimalValue.makeDecimalValue(htmlVersion, false).asAtomic())
                                 .getDecimalValue().equals(BigDecimal.valueOf(5));
             } catch (ValidationException e) {
                 return false;
