@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,23 +7,35 @@
 
 package net.sf.saxon.lib;
 
+import net.sf.saxon.serialize.XHTMLPrefixRemover;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.event.*;
+import net.sf.saxon.om.NameChecker;
+import net.sf.saxon.om.NamespaceResolver;
+import net.sf.saxon.om.QNameException;
+import net.sf.saxon.om.TreeInfo;
 import net.sf.saxon.query.SequenceWrapper;
 import net.sf.saxon.serialize.*;
+import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.SaxonErrorCode;
 import net.sf.saxon.trans.XPathException;
+import net.sf.saxon.value.BigDecimalValue;
 
 import javax.xml.transform.OutputKeys;
 import javax.xml.transform.Result;
+import javax.xml.transform.Source;
+import javax.xml.transform.TransformerException;
 import javax.xml.transform.sax.SAXResult;
 import javax.xml.transform.stream.StreamResult;
 import java.io.OutputStream;
 import java.io.OutputStreamWriter;
 import java.io.UnsupportedEncodingException;
 import java.io.Writer;
+import java.util.Enumeration;
 import java.util.List;
 import java.util.Properties;
+import java.util.StringTokenizer;
+import java.util.regex.Pattern;
 
 /**
  * Helper class to construct a serialization pipeline for a given result destination
@@ -62,6 +74,9 @@ public class SerializerFactory {
         this.config = config;
     }
 
+    public Configuration getConfiguration() {
+        return config;
+    }
     /**
      * Create a serializer with given output properties, and return
      * an XMLStreamWriter that can be used to feed events to the serializer.
@@ -160,6 +175,39 @@ public class SerializerFactory {
             sansNext.setProperty(SaxonOutputKeys.NEXT_IN_CHAIN, "");
             return prepareNextStylesheet(pipe, href, base, result);
         }
+        String paramDoc = props.getProperty(SaxonOutputKeys.PARAMETER_DOCUMENT);
+        if (paramDoc != null) {
+            String base = props.getProperty(SaxonOutputKeys.PARAMETER_DOCUMENT_BASE_URI);
+            if (base == null) {
+                base = result.getSystemId();
+            }
+            Properties props2 = new Properties(props);
+            Source source;
+            try {
+                source = pipe.getConfiguration().getURIResolver().resolve(paramDoc, base);
+            } catch (TransformerException e) {
+                throw XPathException.makeXPathException(e);
+            }
+            ParseOptions options = new ParseOptions();
+            options.setSchemaValidationMode(Validation.LAX);
+            TreeInfo doc = pipe.getConfiguration().buildDocumentTree(source);
+            SerializationParamsHandler ph = new SerializationParamsHandler();
+            ph.setSerializationParams(doc.getRootNode());
+            Properties paramDocProps = ph.getSerializationProperties();
+            Enumeration<?> names = paramDocProps.propertyNames();
+            while (names.hasMoreElements()) {
+                String name = (String)names.nextElement();
+                String value = paramDocProps.getProperty(name);
+                props2.setProperty(name, value);
+            }
+            props = props2;
+            CharacterMap characterMap = ph.getCharacterMap();
+            if (characterMap != null) {
+                CharacterMapIndex index = new CharacterMapIndex();
+                charMapIndex.putCharacterMap(characterMap.getName(), characterMap);
+                props.setProperty(SaxonOutputKeys.USE_CHARACTER_MAPS, characterMap.getName().getClarkName());
+            }
+        }
         if (result instanceof Emitter) {
             if (((Emitter) result).getOutputProperties() == null) {
                 ((Emitter) result).setOutputProperties(props);
@@ -211,7 +259,7 @@ public class SerializerFactory {
             SequenceReceiver target;
             String method = props.getProperty(OutputKeys.METHOD);
             if (method == null) {
-                return newUncommittedSerializer(result, new Sink(pipe), props);
+                return newUncommittedSerializer(result, new Sink(pipe), props, charMapIndex);
             }
 
             Emitter emitter = null;
@@ -310,6 +358,7 @@ public class SerializerFactory {
             return target;
 
         } else if (staxResultClass != null && staxResultClass.isAssignableFrom(result.getClass())) {
+            // TODO: dynamic loading is no longer needed here
             StAXResultHandler handler = (StAXResultHandler) config.getDynamicLoader().getInstance("net.sf.saxon.stax.StAXResultHandlerImpl", getClass().getClassLoader());
             Receiver r = handler.getReceiver(result, props);
             r.setPipelineConfiguration(pipe);
@@ -384,7 +433,7 @@ public class SerializerFactory {
             target = newCDATAFilter(target, props);
         }
 
-        if (SaxonOutputKeys.isXhtmlHtmlVersion5(props)) {
+        if (SaxonOutputKeys.isHtmlVersion5(props)) {
             target = addHtml5Component(target, props);
         }
 
@@ -524,10 +573,20 @@ public class SerializerFactory {
         return makeSequenceNormalizer(target, props);
     }
 
-    public Receiver addHtml5Component(Receiver next, Properties outputProperties) {
-        return next;
-    }
+    /**
+     * This method constructs a step in the output pipeline to perform namespace-related
+     * tasks for HTML5 serialization. The default implementation adds a NamespaceReducer
+     * and an XHTMLPrefixRemover
+     * @param target the Receiver that receives the output of this step
+     * @param outputProperties the serialization properties
+     * @return a new Receiver to perform HTML5-related namespace manipulation
+     */
 
+    public Receiver addHtml5Component(Receiver target, Properties outputProperties) {
+        target = new NamespaceReducer(target);
+        target = new XHTMLPrefixRemover(target);
+        return target;
+    }
 
     /**
      * Create a serialization pipeline to implement the XML output method. This method is protected
@@ -633,8 +692,8 @@ public class SerializerFactory {
      * @return the newly created UncommittedSerializer.
      */
 
-    protected UncommittedSerializer newUncommittedSerializer(Result result, Receiver next, Properties properties) {
-        return new UncommittedSerializer(result, next, properties);
+    protected UncommittedSerializer newUncommittedSerializer(Result result, Receiver next, Properties properties, CharacterMapIndex charMap) {
+        return new UncommittedSerializer(result, next, properties, charMap);
     }
 
     /**
@@ -830,7 +889,9 @@ public class SerializerFactory {
 
     /**
      * Create a new AttributeSorter, responsible for sorting of attributes into a specified order.
-     * This method exists so that it can be overridden in a subclass.
+     * This method exists so that it can be overridden in a subclass. The Saxon-HE version of
+     * this method returns the supplied receiver unchanged (attribute sorting is not supported
+     * in Saxon-HE)
      *
      * @param next             the next receiver in the pipeline
      * @param outputProperties the serialization parameters
@@ -839,10 +900,8 @@ public class SerializerFactory {
      *          if an error occurs
      */
 
-    protected ProxyReceiver newAttributeSorter(Receiver next, Properties outputProperties) throws XPathException {
-        AttributeSorter r = new AttributeSorter(next);
-        r.setOutputProperties(outputProperties);
-        return r;
+    protected Receiver newAttributeSorter(Receiver next, Properties outputProperties) throws XPathException {
+        return next;
     }
 
 
@@ -918,5 +977,257 @@ public class SerializerFactory {
         return new SequenceWrapper(destination);
     }
 
+    /**
+     * Check that a supplied output property is valid, and normalize the value (specifically in the case of boolean
+     * values where yes|true|1 are normalized to "yes", and no|false|0 are normalized to "no").
+     *
+     * @param key     the name of the property, in Clark format
+     * @param value   the value of the property. This may be set to null, in which case no validation takes place.
+     *                The value must be in JAXP format, that is, with lexical QNames expanded to Clark names
+     * @return normalized value of the property, or null if the supplied value is null
+     * @throws XPathException if the property name or value is invalid
+     */
+
+    public String checkOutputProperty(String key, String value) throws XPathException {
+        if (!key.startsWith("{")) {
+            if (key.equals(SaxonOutputKeys.ALLOW_DUPLICATE_NAMES)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.BUILD_TREE)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.BYTE_ORDER_MARK)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(OutputKeys.CDATA_SECTION_ELEMENTS)) {
+                if (value != null) {
+                    checkListOfClarkNames(key, value);
+                }
+            } else if (key.equals(OutputKeys.DOCTYPE_PUBLIC)) {
+                if (value != null) {
+                    checkPublicIdentifier(value);
+                }
+            } else if (key.equals(OutputKeys.DOCTYPE_SYSTEM)) {
+                if (value != null) {
+                    checkSystemIdentifier(value);
+                }
+            } else if (key.equals(OutputKeys.ENCODING)) {
+                // no constraints
+            } else if (key.equals(SaxonOutputKeys.ESCAPE_URI_ATTRIBUTES)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.HTML_VERSION)) {
+                if (value != null) {
+                    checkDecimal(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.INCLUDE_CONTENT_TYPE)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(OutputKeys.INDENT)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.ITEM_SEPARATOR)) {
+                // no checking needed
+
+            } else if (key.equals(OutputKeys.METHOD) || key.equals(SaxonOutputKeys.JSON_NODE_OUTPUT_METHOD)) {
+                if (value != null) {
+                    checkMethod(key, value);
+                }
+            } else if (key.equals(OutputKeys.MEDIA_TYPE)) {
+                // no constraints
+            } else if (key.equals(SaxonOutputKeys.NORMALIZATION_FORM)) {
+                if (value != null) {
+                    checkNormalizationForm(value);
+                }
+            } else if (key.equals(OutputKeys.OMIT_XML_DECLARATION)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(OutputKeys.STANDALONE)) {
+                if (value != null && !value.equals("omit")) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.SUPPRESS_INDENTATION)) {
+                if (value != null) {
+                    checkListOfClarkNames(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.UNDECLARE_PREFIXES)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.USE_CHARACTER_MAPS)) {
+                if (value != null) {
+                    checkListOfClarkNames(key, value);
+                }
+            } else if (key.equals(OutputKeys.VERSION)) {
+                // no constraints
+
+            } else if (key.equals(SaxonOutputKeys.PARAMETER_DOCUMENT)) {
+                // no checking
+            } else {
+                throw new XPathException("Unknown serialization parameter " + Err.wrap(key), "XQST0109");
+            }
+        } else if (key.startsWith("{http://saxon.sf.net/}")) {
+            // Some Saxon serialization parameters are recognized in HE if they are used for internal purposes
+            if (key.equals(SaxonOutputKeys.STYLESHEET_VERSION)) {
+                // return
+            } else if (key.equals(SaxonOutputKeys.PARAMETER_DOCUMENT_BASE_URI)) {
+                // return
+            } else if (key.equals(SaxonOutputKeys.SUPPLY_SOURCE_LOCATOR)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else if (key.equals(SaxonOutputKeys.UNFAILING)) {
+                if (value != null) {
+                    value = checkYesOrNo(key, value);
+                }
+            } else {
+                throw new XPathException("Serialization parameter " + Err.wrap(key) + " not available in Saxon-HE", "XQST0109");
+            }
+        } else {
+            //return;
+        }
+        return value;
+    }
+
+    protected static String checkYesOrNo(String key, String value) throws XPathException {
+        if ("yes".equals(value) || "true".equals(value) || "1".equals(value)) {
+            return "yes";
+        } else if ("no".equals(value) || "false".equals(value) || "0".equals(value)) {
+            return "no";
+        } else {
+            throw new XPathException("Serialization parameter " + Err.wrap(key) + " must have the value yes|no, true|false, or 1|0", "SEPM0016");
+        }
+    }
+
+    private void checkMethod(String key, String value) throws XPathException {
+        if (!"xml".equals(value) && !"html".equals(value) && !"xhtml".equals(value) && !"text".equals(value)) {
+            if (!SaxonOutputKeys.JSON_NODE_OUTPUT_METHOD.equals(key) && ("json".equals(value) || "adaptive".equals(value))) {
+                return;
+            }
+            if (isValidClarkName(value)) {
+                checkExtensions(value);
+            } else {
+                throw new XPathException("Invalid value for serialization method: " +
+                                                 "must be xml|html|xhtml|text|json|adaptive, or a QName in '{uri}local' form", "SEPM0016");
+            }
+        }
+
+    }
+
+    private static void checkNormalizationForm(String value) throws XPathException {
+        if (!NameChecker.isValidNmtoken(value)) {
+            throw new XPathException("Invalid value for normalization-form: " +
+                                             "must be NFC, NFD, NFKC, NFKD, fully-normalized, or none", "SEPM0016");
+        }
+    }
+
+    private static boolean isValidClarkName(/*@NotNull*/ String value) {
+        if (value.isEmpty() || value.charAt(0) != '{') {
+            return false;
+        }
+        int closer = value.indexOf('}');
+        return closer >= 1 &&
+                closer != value.length() - 1 &&
+                NameChecker.isValidNCName(value.substring(closer + 1));
+    }
+
+    protected static void checkNonNegativeInteger(String key, String value) throws XPathException {
+        try {
+            int n = Integer.parseInt(value);
+            if (n < 0) {
+                throw new XPathException("Value of " + Err.wrap(key) + " must be a non-negative integer", "SEPM0016");
+            }
+        } catch (NumberFormatException err) {
+            throw new XPathException("Value of " + Err.wrap(key) + " must be a non-negative integer", "SEPM0016");
+        }
+    }
+
+    private static void checkDecimal(String key, String value) throws XPathException {
+        if (!BigDecimalValue.castableAsDecimal(value)) {
+            throw new XPathException("Value of " + Err.wrap(key) +
+                                             " must be a decimal number", "SEPM0016");
+        }
+    }
+
+    protected static void checkListOfClarkNames(String key, String value) throws XPathException {
+        StringTokenizer tok = new StringTokenizer(value, " \t\n\r", false);
+        while (tok.hasMoreTokens()) {
+            String s = tok.nextToken();
+            if (isValidClarkName(s) || NameChecker.isValidNCName(s)) {
+                // ok
+            } else {
+                throw new XPathException("Value of " + Err.wrap(key) +
+                                                 " must be a list of QNames in '{uri}local' notation", "SEPM0016");
+            }
+        }
+    }
+
+    private static Pattern publicIdPattern = Pattern.compile("^[\\s\\r\\na-zA-Z0-9\\-'()+,./:=?;!*#@$_%]*$");
+
+    private static void checkPublicIdentifier(String value) throws XPathException {
+        if (!publicIdPattern.matcher(value).matches()) {
+            throw new XPathException("Invalid character in doctype-public parameter", "SEPM0016");
+        }
+    }
+
+    private static void checkSystemIdentifier(/*@NotNull*/ String value) throws XPathException {
+        if (value.contains("'") && value.contains("\"")) {
+            throw new XPathException("The doctype-system parameter must not contain both an apostrophe and a quotation mark", "SEPM0016");
+        }
+    }
+
+    /**
+     * Process a serialization property whose value is a list of element names, for example cdata-section-elements
+     *
+     * @param value        The value of the property as written
+     * @param nsResolver   The namespace resolver to use; may be null if prevalidated is set or if names are supplied
+     *                     in Clark format
+     * @param useDefaultNS
+     * @param prevalidated true if the property has already been validated
+     * @param errorCode    The error code to return in the event of problems
+     * @return The list of element names with lexical QNames replaced by Clark names, starting with a single space
+     * @throws XPathException if any error is found in the list of element names, for example, an undeclared namespace prefix
+     */
+
+    /*@NotNull*/
+    public static String parseListOfNodeNames(
+            String value, /*@Nullable*/ NamespaceResolver nsResolver, boolean useDefaultNS, boolean prevalidated, /*@NotNull*/  String errorCode)
+            throws XPathException {
+        String s = "";
+        StringTokenizer st = new StringTokenizer(value, " \t\n\r", false);
+        while (st.hasMoreTokens()) {
+            String displayname = st.nextToken();
+            if (prevalidated || (nsResolver == null)) {
+                s += ' ' + displayname;
+            } else if (displayname.startsWith("Q{")) {
+                s += ' ' + displayname.substring(1);
+            } else {
+                try {
+                    String[] parts = NameChecker.getQNameParts(displayname);
+                    String muri = nsResolver.getURIForPrefix(parts[0], useDefaultNS);
+                    if (muri == null) {
+                        throw new XPathException("Namespace prefix '" + parts[0] + "' has not been declared", errorCode);
+                    }
+                    s += " {" + muri + '}' + parts[1];
+                } catch (QNameException err) {
+                    throw new XPathException("Invalid element name. " + err.getMessage(), errorCode);
+                }
+            }
+        }
+        return s;
+    }
+
+    protected void checkExtensions(String key /*@Nullable*/) throws XPathException {
+        throw new XPathException("Serialization property " + Err.wrap(key) + " is not available in Saxon-HE");
+    }
+
+
 }
 
