@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -28,6 +28,7 @@ import net.sf.saxon.tree.AttributeLocation;
 import net.sf.saxon.tree.util.Navigator;
 import net.sf.saxon.type.Type;
 import net.sf.saxon.type.ValidationException;
+import net.sf.saxon.type.ValidationFailure;
 import org.xml.sax.SAXException;
 
 import javax.xml.transform.SourceLocator;
@@ -222,9 +223,10 @@ public class StandardErrorListener implements UnfailingErrorListener {
         String message;
         if (exception instanceof ValidationException) {
             String explanation = getExpandedMessage(exception);
-            String constraintReference = ((ValidationException) exception).getConstraintReferenceMessage();
-            String validationLocation = ((ValidationException) exception).getValidationLocationText();
-            String contextLocation = ((ValidationException) exception).getContextLocationText();
+            ValidationFailure failure = ((ValidationException) exception).getValidationFailure();
+            String constraintReference = failure.getConstraintReferenceMessage();
+            String validationLocation = failure.getValidationLocationText();
+            String contextLocation = failure.getContextLocationText();
             message = "Validation error " +
                     getLocationMessage(exception) +
                     "\n  " +
@@ -232,7 +234,7 @@ public class StandardErrorListener implements UnfailingErrorListener {
                     wordWrap(validationLocation.isEmpty() ? "" : "\n  " + validationLocation) +
                     wordWrap(contextLocation.isEmpty() ? "" : "\n  " + contextLocation) +
                     wordWrap(constraintReference == null ? "" : "\n  " + constraintReference) +
-                    getOffenderListText(((ValidationException) exception).offendingNodes);
+                    getOffenderListText(failure);
         } else {
             String prefix = recoveryPolicy == Configuration.RECOVER_WITH_WARNINGS ?
                     "Recoverable error " : "Error ";
@@ -350,7 +352,6 @@ public class StandardErrorListener implements UnfailingErrorListener {
 
             } else {
 
-
                 int line = nestedLoc.getLocalLineNumber();
                 int column = nestedLoc.getColumnNumber();
                 String innerLoc = "";
@@ -361,7 +362,7 @@ public class StandardErrorListener implements UnfailingErrorListener {
                     innerLoc += "at " + (line < 0 ? "char " : "column ") + (column+1) + ' ';
                 }
                 if (outerLoc.getLineNumber() > 1) {
-                    innerLoc += "(" + langText + " on line " + outerLoc.getLineNumber() + ") ";
+                    innerLoc += "(" + langText + "on line " + outerLoc.getLineNumber() + ") ";
                 }
                 if (outerLoc.getSystemId() != null) {
                     innerLoc += "of " + outerLoc.getSystemId() + " ";
@@ -379,7 +380,8 @@ public class StandardErrorListener implements UnfailingErrorListener {
 
         } else if (xe instanceof ValidationException) {
             String explanation = getExpandedMessage(exception);
-            String constraintReference = ((ValidationException) exception).getConstraintReferenceMessage();
+            ValidationFailure failure = ((ValidationException)xe).getValidationFailure();
+            String constraintReference = failure.getConstraintReferenceMessage();
             if (constraintReference != null) {
                 explanation += " (" + constraintReference + ')';
             }
@@ -387,6 +389,7 @@ public class StandardErrorListener implements UnfailingErrorListener {
                     getLocationMessage(exception) +
                     "\n  " +
                     wordWrap(explanation);
+            message += getOffenderListText(failure);
         } else {
             message = kind +
                     getLocationMessage(exception) +
@@ -397,6 +400,33 @@ public class StandardErrorListener implements UnfailingErrorListener {
         return message;
     }
 
+    public static String getOffenderListText(ValidationFailure failure) {
+        String message = "";
+        List<NodeInfo> offendingNodes = failure.getOffendingNodes();
+        if (!offendingNodes.isEmpty()) {
+            message += "\n  Nodes for which the assertion fails:";
+            for (NodeInfo offender : offendingNodes) {
+                String nodeDesc = Type.displayTypeName(offender);
+                if (offender.getNodeKind() == Type.TEXT) {
+                    nodeDesc += " " + Err.wrap(offender.getStringValueCS(), Err.VALUE);
+                }
+                if (offender.getLineNumber() != -1) {
+                    nodeDesc += " on line " + offender.getLineNumber();
+                    if (offender.getColumnNumber() != -1) {
+                        nodeDesc += " column " + offender.getColumnNumber();
+                    }
+                    if (offender.getSystemId() != null) {
+                        nodeDesc += " of " + offender.getSystemId();
+                    }
+                } else {
+                    nodeDesc += " at " + Navigator.getPath(offender);
+                }
+                message += "\n  * " + nodeDesc;
+            }
+        }
+        return message;
+    }
+
     /**
      * Generate a stack trace. This method is protected so it can be overridden in a subclass.
      *
@@ -467,7 +497,7 @@ public class StandardErrorListener implements UnfailingErrorListener {
             }
             systemId = loc.getSystemId();
             lineNumber = loc.getLineNumber();
-        } else if (loc instanceof ComponentCode) {
+        } else if (loc instanceof Actor) {
             String kind = "procedure";
             if (loc instanceof UserFunction) {
                 kind = "function";
@@ -553,10 +583,8 @@ public class StandardErrorListener implements UnfailingErrorListener {
     public String getExpandedMessage(TransformerException err) {
 
         StructuredQName qCode = null;
-        String additionalLocationText = null;
         if (err instanceof XPathException) {
             qCode = ((XPathException) err).getErrorCodeQName();
-            additionalLocationText = ((XPathException) err).getAdditionalLocationText();
         }
         if (qCode == null && err.getException() instanceof XPathException) {
             qCode = ((XPathException) err.getException()).getErrorCodeQName();
@@ -570,10 +598,6 @@ public class StandardErrorListener implements UnfailingErrorListener {
             }
         }
 
-        if (additionalLocationText != null) {
-            message += " " + additionalLocationText;
-        }
-
         if (err instanceof XPathException) {
             Sequence errorObject = ((XPathException) err).getErrorObject();
             if (errorObject != null) {
@@ -748,31 +772,5 @@ public class StandardErrorListener implements UnfailingErrorListener {
         }
     }
 
-    public static String getOffenderListText(List<NodeInfo> offendingNodes) {
-        String message = "";
-        if (offendingNodes != null && !offendingNodes.isEmpty()) {
-            message += "\n  Nodes for which the assertion fails:";
-            for (NodeInfo offender : offendingNodes) {
-                String nodeDesc = Type.displayTypeName(offender);
-                if (offender.getNodeKind() == Type.TEXT) {
-                    nodeDesc += " " + Err.wrap(offender.getStringValueCS(), Err.VALUE);
-                }
-                if (offender.getLineNumber() != -1) {
-                    nodeDesc += " on line " + offender.getLineNumber();
-                    if (offender.getColumnNumber() != -1) {
-                        nodeDesc += " column " + offender.getColumnNumber();
-                    }
-                    if (offender.getSystemId() != null) {
-                        nodeDesc += " of " + offender.getSystemId();
-                    }
-                } else {
-                    nodeDesc += " at " + Navigator.getPath(offender);
-                }
-                message += "\n  * " + nodeDesc;
-            }
-        }
-        return message;
-    }
-
 }
 
