@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,25 +7,17 @@
 
 package net.sf.saxon.ma.json;
 
-import net.sf.saxon.expr.StaticProperty;
 import net.sf.saxon.expr.XPathContext;
-import net.sf.saxon.expr.parser.ExplicitLocation;
-import net.sf.saxon.expr.parser.RoleDiagnostic;
-import net.sf.saxon.ma.map.HashTrieMap;
-import net.sf.saxon.ma.map.MapItem;
+import net.sf.saxon.om.NameChecker;
 import net.sf.saxon.om.Sequence;
-import net.sf.saxon.om.SequenceTool;
 import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.XPathException;
-import net.sf.saxon.tree.iter.AtomicIterator;
 import net.sf.saxon.tree.util.FastStringBuffer;
-import net.sf.saxon.type.SpecificFunctionType;
 import net.sf.saxon.type.StringToDouble;
-import net.sf.saxon.type.TypeHierarchy;
-import net.sf.saxon.value.*;
-import net.sf.saxon.z.IntPredicate;
+import net.sf.saxon.value.BooleanValue;
+import net.sf.saxon.value.StringValue;
+import net.sf.saxon.value.Whitespace;
 
-import java.util.HashMap;
 import java.util.Map;
 
 /**
@@ -33,8 +25,6 @@ import java.util.Map;
  */
 public class JsonParser {
 
-    private IntPredicate charChecker;
-
     public static final int ESCAPE = 1;
     public static final int ALLOW_ANY_TOP_LEVEL = 2;
     public static final int LIBERAL = 4;
@@ -54,19 +44,17 @@ public class JsonParser {
 
     /**
      * Create a JSON parser
-     *
-     * @param charChecker a predicate that determines which characters are accepted
      */
 
-    public JsonParser(IntPredicate charChecker) {
-        this.charChecker = charChecker;
+    public JsonParser() {
     }
 
     /**
      * Parse the JSON string according to supplied options
      *
+     *
      * @param input   JSON input string
-     * @param flags   options for the conversion as a map of xs:string : value pairs
+     * @param flags options for the conversion as a map of xs:string : value pairs
      * @param handler event handler to which parsing events are notified
      * @param context XPath evaluation context
      * @throws XPathException if the syntax of the input is incorrect
@@ -79,7 +67,7 @@ public class JsonParser {
         JsonTokenizer t = new JsonTokenizer(input);
         t.next();
 
-        parseConstruct(handler, t, flags, true, context);
+        parseConstruct(handler, t, flags, context);
 
         if (t.next() != JsonTokenizer.EOF) {
             invalidJSON("Unexpected token beyond end of JSON input", ERR_GRAMMAR);
@@ -88,29 +76,29 @@ public class JsonParser {
     }
 
 
-    public static int getFlags(MapItem options, XPathContext context, Boolean allowValidate) throws XPathException {
+    public static int getFlags(Map<String, Sequence> options, XPathContext context, Boolean allowValidate) throws XPathException {
         int flags = 0;
-        if (getBooleanOption(options, "debug", false)) {
+        BooleanValue debug = (BooleanValue) options.get("debug");
+        if (debug != null && debug.getBooleanValue()) {
             flags |= DEBUG;
         }
 
-        boolean escape = getBooleanOption(options, "escape", false);
+        boolean escape = ((BooleanValue) options.get("escape")).getBooleanValue();
         if (escape) {
             flags |= ESCAPE;
-            if (options.get(new StringValue("fallback")) != null) {
+            if (options.get("fallback") != null) {
                 throw new XPathException("Cannot specify a fallback function when escape=true", "FOJS0005");
             }
         }
 
-
-        if (getBooleanOption(options, "liberal", false)) {
+        if (((BooleanValue) options.get("liberal")).getBooleanValue()) {
             flags |= LIBERAL;
             flags |= ALLOW_ANY_TOP_LEVEL;
         }
 
         boolean validate = false;
         if (allowValidate) {
-            validate = getBooleanOption(options, "validate", false);
+            validate = ((BooleanValue) options.get("validate")).getBooleanValue();
             if (validate) {
                 if (!context.getController().getExecutable().isSchemaAware()) {
                     error("Requiring validation on non-schema-aware processor", ERR_SCHEMA);
@@ -119,8 +107,8 @@ public class JsonParser {
             }
         }
 
-        String duplicates = getStringOption(options, "duplicates", null, context.getConfiguration().getTypeHierarchy());
-        if (duplicates != null) {
+        if (options.containsKey("duplicates")) {
+            String duplicates = ((StringValue) options.get("duplicates")).getStringValue();
             if ("reject".equals(duplicates)) {
                 flags |= DUPLICATES_REJECTED;
             } else if ("use-last".equals(duplicates)) {
@@ -132,77 +120,25 @@ public class JsonParser {
             } else {
                 error("Invalid value for 'duplicates' option", ERR_OPTIONS);
             }
-        }
-
-        if (validate && "retain".equals(duplicates)) {
-            error("The options validate:true and duplicates:retain cannot be used together", ERR_OPTIONS);
+            if (validate && "retain".equals(duplicates)) {
+                error("The options validate:true and duplicates:retain cannot be used together", ERR_OPTIONS);
+            }
         }
         return flags;
     }
 
-    /**
-     * Get the value of an option setting (as a string)
-     *
-     * @param options      the set of options provided
-     * @param option       the name of the option required
-     * @param defaultValue the default to be used if the option has not been specified; may be null
-     * @return the setting of the option
-     * @throws net.sf.saxon.trans.XPathException if the options cannot be read
-     */
-
-    private static String getStringOption(MapItem options, String option, String defaultValue, TypeHierarchy th)
-            throws XPathException {
-        StringValue ov = new StringValue(option);
-        Sequence val = options.get(ov);
-        if (val != null) {
-            RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.OPTION, option, 0);
-            role.setErrorCode(ERR_OPTIONS);
-            Sequence converted = th.applyFunctionConversionRules(
-                    val, SequenceType.SINGLE_STRING, role, ExplicitLocation.UNKNOWN_LOCATION);
-            converted = SequenceTool.toGroundedValue(converted);
-            return converted.head().getStringValue();
-        } else {
-            return defaultValue;
-        }
-    }
-
-    /**
-     * Get the value of an option setting (as a boolean)
-     *
-     * @param options      the set of options provided
-     * @param option       the name of the option required
-     * @param defaultValue the default to be used if the option has not been specified
-     * @return the setting of the option
-     * @throws net.sf.saxon.trans.XPathException if the options cannot be read
-     */
-    private static boolean getBooleanOption(MapItem options, String option, boolean defaultValue)
-            throws XPathException {
-        StringValue ov = new StringValue(option);
-        Sequence val = options.get(ov);
-        if (val == null) {
-            return defaultValue;
-        } else if (val instanceof BooleanValue) {
-            return ((BooleanValue) val).getBooleanValue();
-        } else {
-            error("Value of option '" + option + "' is not xs:boolean", ERR_OPTIONS);
-            return defaultValue;
-        }
-    }
-
     /**
      * Parse a JSON construct (top-level or nested)
      *
      * @param handler   the handler to generate the result
      * @param tokenizer the tokenizer, positioned at the first token of the construct to be read
      * @param flags     parsing options
-     * @param toplevel  true if this is the outermost construct of the JSON text to be parsed
      * @param context   XPath evaluation context
      * @throws net.sf.saxon.trans.XPathException if a dynamic error occurs (for example, invalid JSON input)
      */
 
-    private void parseConstruct(JsonHandler handler, JsonTokenizer tokenizer, int flags, boolean toplevel, XPathContext context) throws XPathException {
+    private void parseConstruct(JsonHandler handler, JsonTokenizer tokenizer, int flags, XPathContext context) throws XPathException {
         boolean debug = (flags & DEBUG) != 0;
-        boolean checkAllowedAtTopLevel = false; // toplevel && ((flags & ALLOW_ANY_TOP_LEVEL) == 0);
         if (debug) {
             System.err.println("token:" + tokenizer.currentToken + " :" + tokenizer.currentTokenValue);
         }
@@ -216,38 +152,23 @@ public class JsonParser {
                 break;
 
             case JsonTokenizer.NUMERIC_LITERAL:
-                if (checkAllowedAtTopLevel) {
-                    disallowedAtTopLevel();
-                }
                 double d = parseNumericLiteral(tokenizer.currentTokenValue.toString(), flags);
                 handler.writeNumeric(tokenizer.currentTokenValue.toString(), d);
                 break;
 
             case JsonTokenizer.TRUE:
-                if (checkAllowedAtTopLevel) {
-                    disallowedAtTopLevel();
-                }
                 handler.writeBoolean(true);
                 break;
 
             case JsonTokenizer.FALSE:
-                if (checkAllowedAtTopLevel) {
-                    disallowedAtTopLevel();
-                }
                 handler.writeBoolean(false);
                 break;
 
             case JsonTokenizer.NULL:
-                if (checkAllowedAtTopLevel) {
-                    disallowedAtTopLevel();
-                }
                 handler.writeNull();
                 break;
 
             case JsonTokenizer.STRING_LITERAL:
-                if (checkAllowedAtTopLevel) {
-                    disallowedAtTopLevel();
-                }
                 String literal = tokenizer.currentTokenValue.toString();
                 handler.writeString(unescape(literal, flags, ERR_GRAMMAR));
                 break;
@@ -257,10 +178,6 @@ public class JsonParser {
         }
     }
 
-    private void disallowedAtTopLevel() throws XPathException {
-        invalidJSON("Top-level JSON value must be an object or array", ERR_GRAMMAR);
-    }
-
     /**
      * Parse a JSON object (or map), i.e. construct delimited by curly braces
      *
@@ -276,12 +193,12 @@ public class JsonParser {
         handler.startMap();
         int tok = tokenizer.next();
         while (tok != JsonTokenizer.RCURLY) {
-            if (tok != JsonTokenizer.STRING_LITERAL) {
+            if (tok != JsonTokenizer.STRING_LITERAL && !(tok == JsonTokenizer.UNQUOTED_STRING && liberal)) {
                 invalidJSON("Property name must be a string literal", ERR_GRAMMAR);
             }
             String key = tokenizer.currentTokenValue.toString();
             key = unescape(key, flags, ERR_GRAMMAR);
-            String reEscaped = handler.reEscape(key, true);
+            String reEscaped = handler.reEscape(key);
             tok = tokenizer.next();
             if (tok != JsonTokenizer.COLON) {
                 invalidJSON("Missing colon after \"" + Err.wrap(key) + "\"", ERR_GRAMMAR);
@@ -291,13 +208,13 @@ public class JsonParser {
             if (duplicate && ((flags & DUPLICATES_REJECTED) != 0)) {
                 invalidJSON("Duplicate key value \"" + Err.wrap(key) + "\"", ERR_DUPLICATE);
             }
-            if (!duplicate || ((flags & (DUPLICATES_LAST | DUPLICATES_RETAINED)) != 0)) {
-                parseConstruct(handler, tokenizer, flags, false, context);
+            if (!duplicate || ((flags & (DUPLICATES_LAST|DUPLICATES_RETAINED)) != 0)) {
+                parseConstruct(handler, tokenizer, flags, context);
             } else {
                 // retain first: parse the duplicate value but discard it
                 JsonHandler h2 = new JsonHandler();
                 h2.setContext(context);
-                parseConstruct(h2, tokenizer, flags, false, context);
+                parseConstruct(h2, tokenizer, flags, context);
             }
             tok = tokenizer.next();
             if (tok == JsonTokenizer.COMMA) {
@@ -337,7 +254,7 @@ public class JsonParser {
             return;
         }
         while (true) {
-            parseConstruct(handler, tokenizer, flags, false, context);
+            parseConstruct(handler, tokenizer, flags, context);
             tok = tokenizer.next();
             if (tok == JsonTokenizer.COMMA) {
                 tok = tokenizer.next();
@@ -378,7 +295,7 @@ public class JsonParser {
                         t = t.substring(1);
                     }
                     if (t.startsWith("0") &&
-                            !(t.equals("0") || t.startsWith("0.") || t.startsWith("0e") || t.startsWith("0E"))) {
+                        !(t.equals("0") || t.startsWith("0.") || t.startsWith("0e") || t.startsWith("0E"))) {
                         invalidJSON("Redundant leading zeroes not allowed: " + token, ERR_GRAMMAR);
                     }
                     if (t.endsWith(".") || t.contains(".e") || t.contains(".E")) {
@@ -396,26 +313,11 @@ public class JsonParser {
         }
     }
 
-//    private CharSequence reEscape(String literal, int flags, XPathContext context) throws XPathException {
-//        try {
-//            literal = unescape(literal, flags, ERR_GRAMMAR);
-//            return JsonReceiver.escape(literal, true, new IntPredicate() {
-//                public boolean matches(int value) {
-//                    return value < 32 ||
-//                        (value >= 127 && value < 160);
-//                }
-//            });
-//        } catch (XPathException e) {
-//            e.setErrorCode(ERR_GRAMMAR);
-//            throw e;
-//        }
-//    }
-
     /**
      * Unescape a JSON string literal,
      *
-     * @param literal   the string literal to be processed
-     * @param flags     parsing options
+     * @param literal the string literal to be processed
+     * @param flags   parsing options
      * @param errorCode Error code
      * @return the result of parsing and conversion to XDM
      * @throws net.sf.saxon.trans.XPathException if a dynamic error occurs (such as invalid JSON input)
@@ -474,7 +376,7 @@ public class JsonParser {
                         break;
                     default:
                         if (liberal) {
-                            buffer.append(c);
+                            buffer.append(literal.charAt(i));
                         } else {
                             throw new XPathException("Unknown escape sequence \\" + literal.charAt(i), errorCode);
                         }
@@ -486,7 +388,6 @@ public class JsonParser {
         return buffer.toString();
     }
 
-
     /**
      * Throw an error
      *
@@ -536,6 +437,7 @@ public class JsonParser {
         public static final int NULL = 9;
         public static final int COLON = 10;
         public static final int COMMA = 11;
+        public static final int UNQUOTED_STRING = 12; // for "liberal" parsing only
         public static final int EOF = 999;
 
         public JsonTokenizer(String input) {
@@ -582,8 +484,8 @@ public class JsonParser {
                         if (afterBackslash && c == 'u') {
                             try {
                                 String hex = input.substring(position, position + 4);
-                                int code = Integer.parseInt(hex, 16);
-                                // TODO: we aren't using the value
+                                //noinspection ResultOfMethodCallIgnored
+                                Integer.parseInt(hex, 16);
                             } catch (Exception e) {
                                 invalidJSON("\\u must be followed by four hex characters", ERR_GRAMMAR);
                             }
@@ -604,6 +506,8 @@ public class JsonParser {
                 case ',':
                     return COMMA;
                 case '-':
+                case '+': // for liberal parsing
+                case '.': // for liberal parsing
                 case '0':
                 case '1':
                 case '2':
@@ -630,87 +534,40 @@ public class JsonParser {
                         }
                     }
                     return NUMERIC_LITERAL;
-                case 't':
-                case 'f':
-                case 'n':
-                    currentTokenValue.setLength(0);
-                    currentTokenValue.append(ch);
-                    while (true) {
-                        char c = input.charAt(position);
-                        if (c >= 'a' && c <= 'z') {
-                            currentTokenValue.append(c);
-                            if (++position >= input.length()) {
+                default:
+                    // Allow unquoted strings in liberal mode
+                    if (NameChecker.isNCNameChar(ch)) {
+                        currentTokenValue.setLength(0);
+                        currentTokenValue.append(ch);
+                        while (position < input.length()) {
+                            char c = input.charAt(position);
+                            if (NameChecker.isNCNameChar(c)) {
+                                currentTokenValue.append(c);
+                                position++;
+                            } else {
                                 break;
                             }
-                        } else {
-                            break;
                         }
-                    }
-                    String val = currentTokenValue.toString();
-                    if (val.equals("true")) {
-                        return TRUE;
-                    } else if (val.equals("false")) {
-                        return FALSE;
-                    } else if (val.equals("null")) {
-                        return NULL;
+                        String val = currentTokenValue.toString();
+                        if (val.equals("true")) {
+                            return TRUE;
+                        } else if (val.equals("false")) {
+                            return FALSE;
+                        } else if (val.equals("null")) {
+                            return NULL;
+                        } else {
+                            return UNQUOTED_STRING;
+                        }
                     } else {
-                        error("Unknown constant " + currentTokenValue, ERR_GRAMMAR);
+                        char c = input.charAt(--position);
+                        invalidJSON("Unexpected character '" + c + "' (\\u" +
+                                            Integer.toHexString(c) + ") at position " + position, ERR_GRAMMAR);
+                        return -1;
                     }
-                default:
-                    char c = input.charAt(--position);
-                    invalidJSON("Unexpected character '" + c + "' (\\u" +
-                                        Integer.toHexString(c) + ") at position " + position, ERR_GRAMMAR);
-                    return -1;
             }
         }
     }
 
-
-    private final static Map<String, SequenceType> requiredTypes = new HashMap<String, SequenceType>(20);
-
-    static {
-        requiredTypes.put("liberal", SequenceType.SINGLE_BOOLEAN);
-        requiredTypes.put("duplicates", SequenceType.SINGLE_STRING);
-        requiredTypes.put("validate", SequenceType.SINGLE_BOOLEAN);
-        requiredTypes.put("escape", SequenceType.SINGLE_BOOLEAN);
-        requiredTypes.put("fallback", SequenceType.makeSequenceType(new SpecificFunctionType(new SequenceType[]{SequenceType.SINGLE_STRING}, SequenceType.SINGLE_STRING), StaticProperty.EXACTLY_ONE));
-    }
-
-    private static String[] optionNames = new String[]{
-            "liberal", "duplicates", "validate", "escape", "fallback"
-    };
-
-    private static boolean isOptionName(String string) {
-        for (String s : optionNames) {
-            if (s.equals(string)) {
-                return true;
-            }
-        }
-        return false;
-    }
-
-    public static MapItem checkOptions(MapItem map, XPathContext context) throws XPathException {
-        HashTrieMap result = new HashTrieMap(context);
-        TypeHierarchy th = context.getConfiguration().getTypeHierarchy();
-
-        AtomicIterator keysIterator = map.keys();
-        AtomicValue key;
-        while ((key = keysIterator.next()) != null) {
-            if (key instanceof StringValue) {
-                String keyName = key.getStringValue();
-                if (isOptionName(keyName)) {
-                    RoleDiagnostic role = new RoleDiagnostic(RoleDiagnostic.OPTION, keyName, 0);
-                    role.setErrorCode("XPTY0004");
-                    Sequence converted = th.applyFunctionConversionRules(
-                            map.get(key), requiredTypes.get(keyName), role, ExplicitLocation.UNKNOWN_LOCATION);
-                    converted = SequenceTool.toGroundedValue(converted);
-                    result = result.addEntry(key, converted);
-                }
-            }
-        }
-        return result;
-    }
-
 }
 
-// Copyright (c) 2015 Saxonica Limited. All rights reserved.
+// Copyright (c) 2017 Saxonica Limited.
