@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -12,7 +12,7 @@ import net.sf.saxon.event.Receiver;
 import net.sf.saxon.event.StartTagBuffer;
 import net.sf.saxon.expr.parser.Location;
 import net.sf.saxon.lib.NamespaceConstant;
-import net.sf.saxon.om.NamespaceBinding;
+import net.sf.saxon.om.NamespaceBindingSet;
 import net.sf.saxon.om.NodeName;
 import net.sf.saxon.trans.Err;
 import net.sf.saxon.trans.XPathException;
@@ -97,7 +97,7 @@ public class JsonReceiver implements Receiver {
         boolean inMap = "map".equals(parent);
         stack.push(elemName);
         //started.push(false);
-        if (!elemName.hasURI(NamespaceConstant.JSON)) {
+        if (!elemName.hasURI(NamespaceConstant.FN)) {
             throw new XPathException("xml-to-json: element found in wrong namespace: " +
                                              elemName.getStructuredQName().getEQName(), ERR_INPUT);
         }
@@ -111,6 +111,9 @@ public class JsonReceiver implements Receiver {
             if (startTagBuffer == null) {
                 startTagBuffer = (StartTagBuffer) pipe.getComponent(StartTagBuffer.class.getName());
             }
+            if (startTagBuffer == null) {
+                throw new IllegalStateException();
+            }
             String key = startTagBuffer.getAttribute("", "key");
             if (key == null) {
                 throw new XPathException("xml-to-json: Child elements of <map> must have a key attribute", ERR_INPUT);
@@ -192,7 +195,7 @@ public class JsonReceiver implements Receiver {
         }
     }
 
-    public void namespace(NamespaceBinding namespaceBinding, int properties) throws XPathException {
+    public void namespace(NamespaceBindingSet namespaceBindings, int properties) throws XPathException {
         // no action
     }
 
@@ -200,16 +203,21 @@ public class JsonReceiver implements Receiver {
         if (attName.hasURI("") && (attName.getLocalPart().equals("key") || attName.getLocalPart().equals("escaped-key"))) {
             boolean inMap = stack.size() == 1 || stack.get(stack.size() - 2).getLocalPart().equals("map");
             if (!inMap) {
-                throw new XPathException("xml-to-json: The " + attName.getLocalPart() +
-                                                 " attribute is allowed only on elements within a map", ERR_INPUT);
+                throw new XPathException(
+                        "xml-to-json: The " + attName.getLocalPart() +
+                                " attribute is allowed only on elements within a map", ERR_INPUT);
             }
         } else if (attName.hasURI("") && attName.getLocalPart().equals("escaped")) {
-            boolean inString = !stack.empty() && stack.peek().getLocalPart().equals("string");
-            if (!inString) {
-                throw new XPathException("xml-to-json: The escaped" +
-                                                 " attribute is allowed only on the <string> element", ERR_INPUT);
+            boolean allowed = stack.size() == 1 ||
+                    (stack.size() > 1 && stack.peek().getLocalPart().equals("string"));
+            // See bugs 29917 and 30077: at the top level, the escaped attribute is ignored
+            // whatever element it appears on
+            if (!allowed) {
+                throw new XPathException(
+                        "xml-to-json: The escaped attribute is allowed only on the <string> element",
+                        ERR_INPUT);
             }
-        } else if (attName.hasURI("") || attName.hasURI(NamespaceConstant.JSON)) {
+        } else if (attName.hasURI("") || attName.hasURI(NamespaceConstant.FN)) {
             throw new XPathException("xml-to-json: Disallowed attribute in input: " + attName.getDisplayName(), ERR_INPUT);
         }
         // Attributes in other namespaces are ignored
@@ -237,7 +245,7 @@ public class JsonReceiver implements Receiver {
                 }
                 output.append(new DoubleValue(d).getStringValueCS());
             } catch (NumberFormatException e) {
-                throw new XPathException("xml-to-json: Invalid number: " + textBuffer);
+                throw new XPathException("xml-to-json: Invalid number: " + textBuffer, ERR_INPUT);
             }
         } else if (local.equals("string")) {
             output.append('"');
@@ -262,13 +270,22 @@ public class JsonReceiver implements Receiver {
         atStart = false;
     }
 
+    /**
+     * Handle a string that is already escaped, and that should remain escaped, while normalizing
+     * escape sequences to standard format
+     * @param str the input string
+     * @return the result string
+     * @throws XPathException if the input contains invalid escape sequences
+     */
+
     private static CharSequence handleEscapedString(String str) throws XPathException {
         // check that escape sequences are valid
         unescape(str);
         FastStringBuffer out = new FastStringBuffer(str.length() * 2);
+        boolean afterEscapeChar = false;
         for (int i = 0; i < str.length(); i++) {
             char c = str.charAt(i);
-            if (c == '"' && (i == 0 || str.charAt(i - 1) != '\\')) {
+            if (c == '"' && !afterEscapeChar) {
                 out.append("\\\"");
             } else if (c < 32 || (c >= 127 && c < 160)) {
                 if (c == '\b') {
@@ -289,15 +306,25 @@ public class JsonReceiver implements Receiver {
                     }
                     out.append(hex);
                 }
-            } else if (c == '/') {
+            } else if (c == '/' && !afterEscapeChar) {
                 out.append("\\/");
             } else {
                 out.append(c);
             }
+            afterEscapeChar = c == '\\' && !afterEscapeChar;
         }
         return out;
     }
 
+    /**
+     * Escape a string using backslash escape sequences as defined in JSON
+     *
+     * @param in         the input string
+     * @param forXml     true if the output is for the json-to-xml functino
+     * @param hexEscapes a predicate identifying characters that should be output as hex escapes using \ u XXXX notation.
+     * @return the escaped string
+     */
+
     public static CharSequence escape(CharSequence in, boolean forXml, IntPredicate hexEscapes) throws XPathException {
         FastStringBuffer out = new FastStringBuffer(in.length());
         for (int i = 0; i < in.length(); i++) {
@@ -461,4 +488,4 @@ public class JsonReceiver implements Receiver {
     }
 }
 
-// Copyright (c) 2015 Saxonica Limited. All rights reserved
\ No newline at end of file
+// Copyright (c) 2017 Saxonica Limited. All rights reserved
\ No newline at end of file
