@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,21 +7,23 @@
 
 package net.sf.saxon.ma.map;
 
-import net.sf.saxon.ma.trie.ImmutableHashTrieMap;
-import net.sf.saxon.ma.trie.ImmutableMap;
-import net.sf.saxon.ma.trie.Option;
-import net.sf.saxon.ma.trie.Tuple2;
 import net.sf.saxon.expr.OperandRole;
 import net.sf.saxon.expr.StaticProperty;
 import net.sf.saxon.expr.XPathContext;
 import net.sf.saxon.expr.sort.AtomicComparer;
 import net.sf.saxon.expr.sort.AtomicMatchKey;
 import net.sf.saxon.functions.DeepEqual;
+import net.sf.saxon.ma.trie.ImmutableHashTrieMap;
+import net.sf.saxon.ma.trie.ImmutableMap;
+import net.sf.saxon.ma.trie.Tuple2;
 import net.sf.saxon.om.*;
+import net.sf.saxon.pattern.NodeKindTest;
+import net.sf.saxon.pattern.NodeTest;
+import net.sf.saxon.query.AnnotationList;
 import net.sf.saxon.trace.ExpressionPresenter;
-import net.sf.saxon.trans.NoDynamicContextException;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.iter.AtomicIterator;
+import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.Cardinality;
@@ -46,28 +48,43 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
 
     private ImmutableMap<AtomicMatchKey, KeyValuePair> imap;
 
-    // The following values are null if unknown; they are then maintained incrementally when
-    // entries are added to the map, but are reset to null when values are removed from the map.
-    private AtomicType keyType;
-    private ItemType valueType;
-    private int valueCardinality;
-    private int entries = -1;
-    private XPathContext context; // used only to get the TypeHierarchy for the Configuration
+    // The following values are maintained incrementally when
+    // entries are added to the map. They are not changed when entries are removed,
+    // so the actual type may be narrower than these values suggest. The purpose of
+    // keeping this information is to try and avoid dynamic checking of the map
+    // contents against a required type wherever possible. We therefore have
+    // a guarantee that all entries in the map conform to the types maintained here;
+    // but there is no guarantee that they do not also conform to some more specific
+    // type.
 
-    // Count of CalendarValue keys: positive if they all have a timezone, negative if they all have no timezone
-    //private int timezoneCount = 0;
+    // The set of UTypes of keys in the map
+    private UType keyUType = UType.VOID;
+
+    // The set of UTypes of items in the values in the map
+    private UType valueUType = UType.VOID;
+
+    // The atomic type of all keys, if they are homogeneous; otherwise null
+    private AtomicType keyAtomicType = ErrorType.getInstance();
+
+    // The item type of all items in values, if they are homogeneous; otherwise null
+    private ItemType valueItemType = ErrorType.getInstance();
+
+    // The "envelope" cardinality of values in the map (0:1, 0:0, 1:1, 0:many, 1:many)
+    private int valueCardinality = 0;
+
+    // The number of entries in the map; -1 if unknown
+    private int entries = -1;
 
     /**
      * Create an empty map
-     * @param context The XPath dynamic context. Gives access to type information.
      */
 
-    public HashTrieMap(XPathContext context) {
-        setContext(context);
+    public HashTrieMap() {
         this.imap = ImmutableHashTrieMap.empty();
         this.entries = 0;
     }
 
+
     /**
      * Create a singleton map with a single key and value
      * @param key the key value
@@ -78,19 +95,15 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      */
 
     public static HashTrieMap singleton(AtomicValue key, Sequence value, XPathContext context) throws XPathException {
-        return new HashTrieMap(context).addEntry(key, value);
+        return new HashTrieMap().addEntry(key, value);
     }
 
     /**
      * Create a map whose contents are a copy of an existing immutable map
      * @param imap the map to be copied
-     * @param context The XPath dynamic context. Used to define the implicit timezone for comparing
-     * keys in this map, and also gives access to type information.
-     * @throws NoDynamicContextException if the supplied context has no implicit timezone available
      */
 
-    public HashTrieMap(ImmutableMap<AtomicMatchKey, KeyValuePair> imap, XPathContext context)  throws NoDynamicContextException {
-        setContext(context);
+    public HashTrieMap(ImmutableMap<AtomicMatchKey, KeyValuePair> imap) {
         this.imap = imap;
         entries = -1;
     }
@@ -99,26 +112,17 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      * Create a map whose entries are copies of the entries in an existing MapItem
      *
      * @param map       the existing map to be copied
-     * @param context   the XPath dynamic context
      * @return the new map
-     * @throws net.sf.saxon.trans.XPathException
-     *          if a dynamic error occurs
      */
 
-    public static HashTrieMap copy(MapItem map, XPathContext context) throws XPathException {
+    public static HashTrieMap copy(MapItem map) {
         if (map instanceof HashTrieMap) {
             return (HashTrieMap)map;
         }
-        HashTrieMap m2 = new HashTrieMap(context);
-        ImmutableHashTrieMap<AtomicMatchKey, KeyValuePair> imap  = ImmutableHashTrieMap.empty();
-        int entries = 0;
+        HashTrieMap m2 = new HashTrieMap();
         for (KeyValuePair pair : map) {
-            imap = imap.put(m2.makeKey(pair.key), pair);
-            entries++;
+            m2 = m2.addEntry(pair.key, pair.value);
         }
-        m2.imap = imap;
-        m2.entries = entries;
-        m2.computeTypeInformation();
         return m2;
     }
 
@@ -140,6 +144,18 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
         return true;
     }
 
+    /**
+     * Get the function annotations (as defined in XQuery). Returns an empty
+     * list if there are no function annotations.
+     *
+     * @return the function annotations
+     */
+
+    @Override
+    public AnnotationList getAnnotations() {
+        return AnnotationList.EMPTY;
+    }
+
     /**
      * Atomize the item.
      *
@@ -150,20 +166,106 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
         throw new XPathException("Maps cannot be atomized", "FOTY0013");
     }
 
-    private void setContext(XPathContext context) {
-        this.context = context;
-    }
+    /**
+     * After adding an entry to the map, update the cached type information
+     * @param key the new key
+     * @param val the new associated value
+     * @param wasEmpty true if the map was empty before adding these values
+     */
 
-    private void updateTypeInformation(AtomicValue key, Sequence val) {
-        TypeHierarchy th = context.getConfiguration().getTypeHierarchy();
-        if (keyType == null) {
-            keyType = key.getItemType();
-            valueType = SequenceTool.getItemType(val, th);
+    private void updateTypeInformation(AtomicValue key, Sequence val, boolean wasEmpty) {
+        if (wasEmpty) {
+            keyUType = key.getUType();
+            valueUType = SequenceTool.getUType(val);
+            keyAtomicType = key.getItemType();
+            valueItemType = getItemTypeOfSequence(val);
             valueCardinality = SequenceTool.getCardinality(val);
         } else {
-            keyType = (AtomicType) Type.getCommonSuperType(keyType, key.getItemType(), th);
-            valueType = Type.getCommonSuperType(valueType, SequenceTool.getItemType(val, th), th);
+            keyUType = keyUType.union(key.getUType());
+            valueUType = valueUType.union(SequenceTool.getUType(val));
             valueCardinality = Cardinality.union(valueCardinality, SequenceTool.getCardinality(val));
+            if (key.getItemType() != keyAtomicType) {
+                keyAtomicType = null;
+            }
+            if (!isKnownToConform(val, valueItemType)) {
+                valueItemType = null;
+            }
+        }
+    }
+
+    /**
+     * Ask whether all the items in a sequence are known to conform to a given item type
+     * @param value the sequence
+     * @param itemType the given item type
+     * @return true if all the items conform; false if not, or if the information cannot
+     * be efficiently determined
+     */
+
+    private static boolean isKnownToConform(Sequence value, ItemType itemType) {
+        // Problem is we don't have access to a TypeHierarchy object...
+        if (itemType == AnyItemType.getInstance()) {
+            return true;
+        }
+        try {
+            SequenceIterator iter = value.iterate();
+            Item item;
+            while ((item = iter.next()) != null) {
+                if (item instanceof AtomicValue) {
+                    if (itemType instanceof AtomicType) {
+                        if (!Type.isSubType(((AtomicValue)item).getItemType(), (AtomicType)itemType)) {
+                            return false;
+                        }
+                    } else {
+                        return false;
+                    }
+                } else if (item instanceof NodeInfo) {
+                    if (itemType instanceof NodeTest) {
+                        if (!((NodeTest)itemType).matchesNode((NodeInfo)item)) {
+                            return false;
+                        }
+                    } else {
+                        return false;
+                    }
+                } else {
+                    // functions, maps, arrays: give up (this is only an optimization)
+                    return false;
+                }
+            }
+            return true;
+        } catch (XPathException e) {
+            return false;
+        }
+    }
+
+    /**
+     * Get an item type to which all the values in a sequence are known to conform
+     * @param val the sequence
+     * @return the type of the first item in the sequence, provided that all subsequent
+     * values in the sequence are known to conform to this type; otherwise item().
+     */
+
+    private ItemType getItemTypeOfSequence(Sequence val) {
+        try {
+            Item first = val.head();
+            if (first == null) {
+                return AnyItemType.getInstance();
+            } else {
+                ItemType type;
+                if (first instanceof AtomicValue) {
+                    type = ((AtomicValue)first).getItemType();
+                } else if (first instanceof NodeInfo) {
+                    type = NodeKindTest.makeNodeKindTest(((NodeInfo)first).getNodeKind());
+                } else {
+                    type = AnyFunctionType.getInstance();
+                }
+                if (isKnownToConform(val, type)) {
+                    return type;
+                } else {
+                    return AnyItemType.getInstance();
+                }
+            }
+        } catch (XPathException e) {
+            return AnyItemType.getInstance();
         }
     }
 
@@ -193,76 +295,124 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
     }
 
     /**
-     * Get the lowest common item type of the keys in the map
+     * Ask whether the map conforms to a given map type
      *
-     * @return the most specific type to which all the keys belong. If the map is
-     *         empty, return AnyAtomicType by convention
+     * @param requiredKeyType   the required keyType
+     * @param requiredValueType the required valueType
+     * @param th        the type hierarchy cache for the configuration
+     * @return true if the map conforms to the required type
      */
-    public AtomicType getKeyType() {
+    @Override
+    public boolean conforms(AtomicType requiredKeyType, SequenceType requiredValueType, TypeHierarchy th) {
         if (isEmpty()) {
-            return ErrorType.getInstance();
+            return true;
+        }
+        if (keyAtomicType == requiredKeyType && valueItemType == requiredValueType.getPrimaryType() &&
+                Cardinality.subsumes(requiredValueType.getCardinality(), valueCardinality)) {
+            return true;
+        }
+        boolean needFullCheck = false;
+        if (requiredKeyType != BuiltInAtomicType.ANY_ATOMIC) {
+            ItemType upperBoundKeyType = keyUType.toItemType();
+            int rel = th.relationship(requiredKeyType, upperBoundKeyType);
+            if (rel == TypeHierarchy.SAME_TYPE || rel == TypeHierarchy.SUBSUMES) {
+                // The key type is matched
+            } else if (rel == TypeHierarchy.DISJOINT) {
+                return false;
+            } else {
+                needFullCheck = true;
+            }
+        }
+        ItemType requiredValueItemType = requiredValueType.getPrimaryType();
+        if (requiredValueItemType != BuiltInAtomicType.ANY_ATOMIC) {
+            ItemType upperBoundValueType = valueUType.toItemType();
+            int rel = th.relationship(requiredValueItemType, upperBoundValueType);
+            if (rel == TypeHierarchy.SAME_TYPE || rel == TypeHierarchy.SUBSUMES) {
+                // The value type is matched
+            } else if (rel == TypeHierarchy.DISJOINT) {
+                return false;
+            } else {
+                needFullCheck = true;
+            }
+        }
+        int requiredValueCard = requiredValueType.getCardinality();
+        if (!Cardinality.subsumes(requiredValueCard, valueCardinality)) {
+            needFullCheck = true;
         }
-        computeTypeInformation();
-        return keyType;
-    }
 
-    /**
-     * Get the lowest common item type of the keys in the map
-     *
-     * @return the most specific type to which all the keys belong. If the map is
-     *         empty, return ErrorType.getInstance() (the type with no instances)
-     */
-    public UType getKeyUType() {
-        // temporary
-        return UType.ANY_ATOMIC.except(UType.QNAME).except(UType.NOTATION);
-    }
-
-//    /**
-//     * Ask whether the map includes calendar values with a timezone, calendar values without a timezone, or neither.
-//     * Maps are not allowed to mix keys that have a timezone and keys that do not. Here "calendar values" means any
-//     * of the types dateTime, date, time, gYear, gYearMonth, gMonth, gMonthDay, or gDay
-//     *
-//     * @return the time zone status of the map contents.
-//     */
-//    public TimeZoneStatus getTimeZoneStatus() {
-//        if (timezoneCount == 0) {
-//            return TimeZoneStatus.CONTAINS_NO_CALENDAR_VALUES;
-//        } else if (timezoneCount > 0) {
-//            return TimeZoneStatus.CONTAINS_CALENDAR_VALUES_WITH_TIMEZONE;
-//        } else {
-//            return TimeZoneStatus.CONTAINS_CALENDAR_VALUES_WITHOUT_TIMEZONE;
-//        }
-//    }
-
-    private void computeTypeInformation() {
-        if (keyType == null) {
-            TypeHierarchy th = context.getConfiguration().getTypeHierarchy();
-            for (KeyValuePair kvp : this) {
-                AtomicValue key = kvp.key;
-                Sequence value = kvp.value;
-                if (keyType == null) {
-                    keyType = key.getItemType();
-                    valueType = SequenceTool.getItemType(value, th);
-                    valueCardinality = SequenceTool.getCardinality(value);
-                } else {
-                    updateTypeInformation(key, value);
+        if (needFullCheck) {
+            // we need to test the entries individually
+            AtomicIterator keyIter = keys();
+            AtomicValue key;
+            while ((key = keyIter.next()) != null) {
+                if (!requiredKeyType.matches(key, th)) {
+                    return false;
+                }
+                Sequence val = get(key);
+                try {
+                    if (!requiredValueType.matches(val, th)) {
+                        return false;
+                    }
+                } catch (XPathException e) {
+                    throw new AssertionError(e); // cannot happen with a grounded sequence
                 }
             }
         }
+
+        return true;
+
     }
 
     /**
-     * Get the lowest common sequence type of all the values in the map
+     * Get the type of the map. This method is used largely for diagnostics, to report
+     * the type of a map when it differs from the required type. This method has the side-effect
+     * of updating the internal type information held within the map.
      *
-     * @return the most specific sequence type to which all the values belong. If the map
-     *         is empty, return SequenceType.VOID
+     * @param th The type hierarchy cache for the configuration
+     * @return the type of this map
      */
-    public SequenceType getValueType() {
-        if (isEmpty()) {
-            return SequenceType.VOID;
+    @Override
+    public MapType getItemType(TypeHierarchy th) {
+        AtomicType keyType = null;
+        ItemType valueType = null;
+        int valueCard = 0;
+        // we need to test the entries individually
+        AtomicIterator keyIter = keys();
+        AtomicValue key;
+        while ((key = keyIter.next()) != null) {
+            Sequence val = get(key);
+            if (keyType == null) {
+                keyType = key.getItemType();
+                valueType = SequenceTool.getItemType(val, th);
+                valueCard = SequenceTool.getCardinality(val);
+            } else {
+                keyType = (AtomicType)Type.getCommonSuperType(keyType, key.getItemType(), th);
+                valueType = Type.getCommonSuperType(valueType, SequenceTool.getItemType(val, th), th);
+                valueCard = Cardinality.union(valueCard, SequenceTool.getCardinality(val));
+            }
         }
-        computeTypeInformation();
-        return SequenceType.makeSequenceType(valueType, valueCardinality);
+        if (keyType == null) {
+            // implies the map is empty
+            this.keyUType = UType.VOID;
+            this.valueUType = UType.VOID;
+            this.valueCardinality = 0;
+            return MapType.ANY_MAP_TYPE;
+        } else {
+            this.keyUType = keyType.getUType();
+            this.valueUType = valueType.getUType();
+            this.valueCardinality = valueCard;
+            return new MapType(keyType, SequenceType.makeSequenceType(valueType, valueCard));
+        }
+    }
+
+    /**
+     * Get the lowest common item type of the keys in the map
+     *
+     * @return the most specific type to which all the keys belong. If the map is
+     *         empty, return UType.VOID (the type with no instances)
+     */
+    public UType getKeyUType() {
+        return keyUType;
     }
 
     /**
@@ -282,39 +432,18 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      * @param key     the key of the new entry
      * @param value   the value associated with the new entry
      * @return the new map containing the additional entry
-     * @throws XPathException if adding the entry violates the rule against mixing keys with timezones
-     * and keys with no timezone
      */
 
-    public HashTrieMap addEntry(AtomicValue key, Sequence value) throws XPathException {
+    public HashTrieMap addEntry(AtomicValue key, Sequence value) {
+        boolean empty = isEmpty();
         ImmutableMap<AtomicMatchKey, KeyValuePair> imap2 = imap.put(makeKey(key), new KeyValuePair(key, value));
-        HashTrieMap t2;
-        try {
-            t2 = new HashTrieMap(imap2, context);
-        } catch (NoDynamicContextException e) {
-            // Can't happen, we've already checked the context for suitability
-            throw new IllegalStateException(e);
-        }
+        HashTrieMap t2 = new HashTrieMap(imap2);
         t2.valueCardinality = this.valueCardinality;
-        t2.keyType = keyType;
-        t2.valueType = valueType;
-        t2.updateTypeInformation(key, value);
-//        t2.timezoneCount = timezoneCount;
-//        if (key instanceof CalendarValue) {
-//            if (((CalendarValue)key).hasTimezone()) {
-//                if (timezoneCount >= 0) {
-//                    t2.timezoneCount = timezoneCount + 1;
-//                } else {
-//                    throw new XPathException("Cannot mix values with timezone and values without timezone in the same map", "XTDE3368");
-//                }
-//            } else {
-//                if (timezoneCount <= 0) {
-//                    t2.timezoneCount = timezoneCount - 1;
-//                } else {
-//                    throw new XPathException("Cannot mix values with timezone and values without timezone in the same map", "XTDE3368");
-//                }
-//            }
-//        }
+        t2.keyUType = keyUType;
+        t2.valueUType = valueUType;
+        t2.keyAtomicType = keyAtomicType;
+        t2.valueItemType = valueItemType;
+        t2.updateTypeInformation(key, value, empty);
         return t2;
     }
 
@@ -329,9 +458,10 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      */
 
     public boolean initialPut(AtomicValue key, Sequence value) {
+        boolean empty = isEmpty();
         boolean exists = get(key) != null;
         imap = imap.put(makeKey(key), new KeyValuePair(key, value));
-        updateTypeInformation(key, value);
+        updateTypeInformation(key, value, empty);
         entries = -1;
         return exists;
     }
@@ -349,23 +479,18 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      *         unchanged if the specified key was not present
      */
 
-    public HashTrieMap remove(AtomicValue key) throws XPathException {
+    public HashTrieMap remove(AtomicValue key) {
         ImmutableMap<AtomicMatchKey, KeyValuePair> m2 = imap.remove(makeKey(key));
         if (m2 == imap) {
             // The key is not present; the map is unchanged
             return this;
         }
-        HashTrieMap map = new HashTrieMap(m2, context);
-//        if (key instanceof CalendarValue) {
-//            if (((CalendarValue)key).hasTimezone()) {
-//                map.timezoneCount = this.timezoneCount - 1;
-//            } else {
-//                map.timezoneCount = this.timezoneCount + 1;
-//            }
-//        } else {
-//            map.timezoneCount = this.timezoneCount;
-//        }
-        return map;
+        HashTrieMap result = new HashTrieMap(m2);
+        result.keyUType = keyUType;
+        result.valueUType = valueUType;
+        result.valueCardinality = valueCardinality;
+        result.entries = entries-1;
+        return result;
     }
 
     /**
@@ -376,12 +501,8 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      */
 
     public Sequence get(AtomicValue key)  {
-        Option<KeyValuePair> o = imap.get(makeKey(key));
-        if (o.isDefined()) {
-            return o.get().value;
-        } else {
-            return null;
-        }
+        KeyValuePair o = imap.get(makeKey(key));
+        return o==null ? null : o.value;
     }
 
     /**
@@ -392,12 +513,7 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
      */
 
     public KeyValuePair getKeyValuePair(AtomicValue key) {
-        Option<KeyValuePair> kvp = imap.get(makeKey(key));
-        if (kvp.isDefined()) {
-            return kvp.get();
-        } else {
-            return null;
-        }
+        return imap.get(makeKey(key));
     }
 
     /**
@@ -408,24 +524,16 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
     public AtomicIterator keys() {
         return new AtomicIterator() {
 
-            int pos = 0;
             Iterator<Tuple2<AtomicMatchKey, KeyValuePair>> base = imap.iterator();
 
             public AtomicValue next() {
                 if (base.hasNext()) {
-                    AtomicValue curr = base.next()._2.key;
-                    pos++;
-                    return curr;
+                    return base.next()._2.key;
                 } else {
-                    pos = -1;
                     return null;
                 }
             }
 
-            public AtomicIterator getAnother() {
-                return keys();
-            }
-
             public void close() {
             }
 
@@ -625,6 +733,26 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
         throw new XPathException("A map item has no effective boolean value");
     }
 
+    /**
+     * Returns a string representation of the object.
+     * @return a string representation of the object.
+     */
+    @Override
+    public String toString() {
+        FastStringBuffer buffer = new FastStringBuffer(256);
+        buffer.append("map{");
+        for (KeyValuePair pair : this) {
+            if (buffer.length() > 4) {
+                buffer.append(",");
+            }
+            buffer.append(pair.key.toString());
+            buffer.append(":");
+            buffer.append(pair.value.toString());
+        }
+        buffer.append("}");
+        return buffer.toString();
+    }
+
     /**
      * Output information about this function item to the diagnostic explain() output
      */
@@ -635,11 +763,25 @@ public class HashTrieMap extends AbstractItem implements MapItem, GroundedValue
     }
 
     public boolean isTrustedResultType() {
-        return false;
+        return true;
+    }
+
+    public void diagnosticDump() {
+        System.err.println("Map details:");
+        Iterator<Tuple2<AtomicMatchKey, KeyValuePair>> iter = imap.iterator();
+        while (iter.hasNext()) {
+            Tuple2<AtomicMatchKey, KeyValuePair> entry = iter.next();
+            AtomicMatchKey k1 = entry._1;
+            AtomicValue k2 = entry._2.key;
+            Sequence v = entry._2.value;
+            System.err.println(k1.getClass() + " " + k1 +
+                                       " #:" + k1.hashCode() +
+                                       " = (" + k2.getClass() + " " + k2 + " : " + v + ")");
+        }
     }
 }
 
-// Copyright (c) 2015 Saxonica Limited. All rights reserved.
+// Copyright (c) 2017 Saxonica Limited.
 
 
 
