@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -14,16 +14,14 @@ import net.sf.saxon.expr.sort.AtomicMatchKey;
 import net.sf.saxon.expr.sort.CodepointMatchKey;
 import net.sf.saxon.functions.Count;
 import net.sf.saxon.om.*;
+import net.sf.saxon.query.AnnotationList;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.iter.AtomicIterator;
 import net.sf.saxon.tree.iter.SingletonIterator;
 import net.sf.saxon.tree.iter.UnfailingIterator;
-import net.sf.saxon.type.AtomicType;
-import net.sf.saxon.type.BuiltInAtomicType;
-import net.sf.saxon.type.UType;
+import net.sf.saxon.type.*;
 import net.sf.saxon.value.*;
-import net.sf.saxon.value.StringValue;
 
 import java.util.ArrayList;
 import java.util.Iterator;
@@ -41,14 +39,12 @@ public class RangeKey implements MapItem {
 
     private CodepointMatchKey min;
     private CodepointMatchKey max;
-    private XPathContext context;
     private TreeMap<AtomicMatchKey, List<NodeInfo>> index;
 
-    public RangeKey(String min, String max, XPathContext context, TreeMap<AtomicMatchKey, List<NodeInfo>> index) throws XPathException {
+    public RangeKey(String min, String max, TreeMap<AtomicMatchKey, List<NodeInfo>> index) throws XPathException {
         this.min = min == null ? null : new CodepointMatchKey(min);
         this.max = max == null ? null : new CodepointMatchKey(max);
         this.index = index;
-        this.context = context;
     }
 
     /**
@@ -69,6 +65,19 @@ public class RangeKey implements MapItem {
         return true;
     }
 
+    /**
+     * Get the function annotations (as defined in XQuery). Returns an empty
+     * list if there are no function annotations.
+     *
+     * @return the function annotations
+     */
+
+    @Override
+    public AnnotationList getAnnotations() {
+        return AnnotationList.EMPTY;
+    }
+
+
     /**
      * Get the roles of the arguments, for the purposes of streaming
      *
@@ -92,8 +101,6 @@ public class RangeKey implements MapItem {
 
     /**
      * Get an entry from the Map
-     *
-     *
      * @param key     the value of the key
      * @return the value associated with the given key, or null if the key is not present in the map
      */
@@ -142,17 +149,6 @@ public class RangeKey implements MapItem {
         return new RangeKeyIterator();
     }
 
-//    /**
-//     * Ask whether the map includes calendar values with a timezone, calendar values without a timezone, or neither.
-//     * Maps are not allowed to mix keys that have a timezone and keys that do not. Here "calendar values" means any
-//     * of the types dateTime, date, time, gYear, gYearMonth, gMonth, gMonthDay, or gDay
-//     * @return the time zone status of the map contents: always {@link TimeZoneStatus#CONTAINS_NO_CALENDAR_VALUES}
-//     */
-//
-//    public TimeZoneStatus getTimeZoneStatus() {
-//        return TimeZoneStatus.CONTAINS_NO_CALENDAR_VALUES;
-//    }
-
     /**
      * Get the set of all key-value pairs in the map
      *
@@ -174,22 +170,9 @@ public class RangeKey implements MapItem {
      * @param key     the key of the entry to be removed
      * @return a new map in which the requested entry has been removed; or this map
      *         unchanged if the specified key was not present
-     * @throws net.sf.saxon.trans.XPathException
-     *          if a dynamic error occurs
      */
-    public MapItem remove(AtomicValue key) throws XPathException {
-        HashTrieMap copy = HashTrieMap.copy(this, context);
-        return copy.remove(key);
-    }
-
-    /**
-     * Get the lowest common item type of the keys in the map
-     *
-     * @return the most specific type to which all the keys belong. If the map is
-     *         empty, return AnyAtomicType by convention
-     */
-    public AtomicType getKeyType() {
-        return BuiltInAtomicType.STRING;
+    public MapItem remove(AtomicValue key) {
+        return HashTrieMap.copy(this).remove(key);
     }
 
     /**
@@ -203,13 +186,54 @@ public class RangeKey implements MapItem {
     }
 
     /**
-     * Get the lowest common sequence type of all the values in the map
+     * Create a new map containing the existing entries in the map plus an additional entry,
+     * without modifying the original. If there is already an entry with the specified key,
+     * this entry is replaced by the new entry.
      *
-     * @return the most specific sequence type to which all the values belong. If the map
-     *         is empty, return ANY_SEQUENCE by convention
+     * @param key   the key of the new entry
+     * @param value the value associated with the new entry
+     * @return the new map containing the additional entry
      */
-    public SequenceType getValueType() {
-        return SequenceType.NODE_SEQUENCE;
+    @Override
+    public MapItem addEntry(AtomicValue key, Sequence value) {
+        return HashTrieMap.copy(this).addEntry(key, value);
+    }
+
+    /**
+     * Ask whether the map conforms to a given map type
+     *
+     * @param keyType   the required keyType
+     * @param valueType the required valueType
+     * @param th        the type hierarchy cache for the configuration
+     * @return true if the map conforms to the required type
+     */
+    @Override
+    public boolean conforms(AtomicType keyType, SequenceType valueType, TypeHierarchy th) {
+        AtomicIterator keyIter = keys();
+        AtomicValue key;
+        while ((key = keyIter.next()) != null) {
+            Sequence value = get(key);
+            try {
+                if (!valueType.matches(value, th)) {
+                    return false;
+                }
+            } catch (XPathException e) {
+                throw new AssertionError(e);
+            }
+        }
+        return true;
+    }
+
+    /**
+     * Get the type of the map. This method is used largely for diagnostics, to report
+     * the type of a map when it differs from the required type.
+     *
+     * @param th the type hierarchy (not used)
+     * @return the type of this map
+     */
+    @Override
+    public MapType getItemType(TypeHierarchy th) {
+        return new MapType(BuiltInAtomicType.STRING, SequenceType.NODE_SEQUENCE);
     }
 
     /**
@@ -459,22 +483,10 @@ public class RangeKey implements MapItem {
                 return null;
             } else {
                 pos++;
-                return new StringValue(curr.toString());
+                return new StringValue(curr.getValue());
             }
         }
 
-        /**
-         * Get another iterator over the same sequence of items, positioned at the
-         * start of the sequence. It must be possible to call this method at any time, whether
-         * none, some, or all of the items in the original iterator have been read. The method
-         * is non-destructive: it does not change the state of the original iterator.
-         *
-         * @return a new iterator over the same sequence
-         */
-        public AtomicIterator getAnother() {
-            return new RangeKeyIterator();
-        }
-
         /**
          * Close the iterator. This indicates to the supplier of the data that the client
          * does not require any more items to be delivered by the iterator. This may enable the
@@ -508,4 +520,4 @@ public class RangeKey implements MapItem {
     }
 }
 
-// Copyright (c) 2012 Saxonica Limited. All rights reserved.
\ No newline at end of file
+// Copyright (c) 2012 Saxonica Limited.
\ No newline at end of file
