@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -454,7 +454,7 @@ public final class AxisInfo {
 
     public static UType getTargetUType(UType origin, byte axis) {
         UType resultType = UType.VOID;
-        Set<PrimitiveUType> origins = origin.decompose();
+        Set<PrimitiveUType> origins = origin.intersection(UType.ANY_NODE).decompose();
         for (PrimitiveUType u : origins) {
             UType r = axisTransitions.get(makeKey(u, axis));
             resultType = resultType.union(r);
@@ -502,11 +502,11 @@ public final class AxisInfo {
         e(PrimitiveUType.ATTRIBUTE, CHILD, UType.VOID);
         e(PrimitiveUType.ATTRIBUTE, DESCENDANT, UType.VOID);
         e(PrimitiveUType.ATTRIBUTE, DESCENDANT_OR_SELF, UType.ATTRIBUTE);
-        e(PrimitiveUType.ATTRIBUTE, FOLLOWING, UType.CHILD_NODE_KINDS);  // TODO
+        e(PrimitiveUType.ATTRIBUTE, FOLLOWING, UType.CHILD_NODE_KINDS);
         e(PrimitiveUType.ATTRIBUTE, FOLLOWING_SIBLING, UType.VOID);
         e(PrimitiveUType.ATTRIBUTE, NAMESPACE, UType.VOID);
         e(PrimitiveUType.ATTRIBUTE, PARENT, UType.ELEMENT);
-        e(PrimitiveUType.ATTRIBUTE, PRECEDING, UType.CHILD_NODE_KINDS); // TODO
+        e(PrimitiveUType.ATTRIBUTE, PRECEDING, UType.CHILD_NODE_KINDS);
         e(PrimitiveUType.ATTRIBUTE, PRECEDING_SIBLING, UType.VOID);
         e(PrimitiveUType.ATTRIBUTE, SELF, UType.ATTRIBUTE);
 
@@ -558,11 +558,11 @@ public final class AxisInfo {
         e(PrimitiveUType.NAMESPACE, CHILD, UType.VOID);
         e(PrimitiveUType.NAMESPACE, DESCENDANT, UType.VOID);
         e(PrimitiveUType.NAMESPACE, DESCENDANT_OR_SELF, UType.NAMESPACE);
-        e(PrimitiveUType.NAMESPACE, FOLLOWING, UType.CHILD_NODE_KINDS);  // TODO
+        e(PrimitiveUType.NAMESPACE, FOLLOWING, UType.CHILD_NODE_KINDS);
         e(PrimitiveUType.NAMESPACE, FOLLOWING_SIBLING, UType.VOID);
         e(PrimitiveUType.NAMESPACE, NAMESPACE, UType.VOID);
         e(PrimitiveUType.NAMESPACE, PARENT, UType.ELEMENT);
-        e(PrimitiveUType.NAMESPACE, PRECEDING, UType.CHILD_NODE_KINDS); // TODO
+        e(PrimitiveUType.NAMESPACE, PRECEDING, UType.CHILD_NODE_KINDS);
         e(PrimitiveUType.NAMESPACE, PRECEDING_SIBLING, UType.VOID);
         e(PrimitiveUType.NAMESPACE, SELF, UType.NAMESPACE);
     }
