@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -24,10 +24,6 @@ public abstract class CopyOptions {
 
     public static final int FOR_UPDATE = 8;
 
-    public static final int ACCUMULATORS = 16;
-
-    public static final int IS_XQUERY = 32; // either XSLT or XQUERY: only affects error codes
-
     public static boolean includes(int options, int option) {
         return (options & option) == option;
     }
