@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -12,6 +12,7 @@ import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.value.Whitespace;
 
 import javax.xml.namespace.QName;
+import java.util.Arrays;
 
 /**
  * This class provides an economical representation of a QName triple (prefix, URI, and localname).
@@ -28,6 +29,12 @@ public class StructuredQName implements IdentityComparable {
     private int prefixStart;
     private int cachedHashCode = -1;
 
+    private StructuredQName(char[] content, int localNameStart, int prefixStart) {
+        this.content = content;
+        this.localNameStart = localNameStart;
+        this.prefixStart = prefixStart;
+    }
+
     /**
      * Construct a StructuredQName from a prefix, URI, and local name. This method performs no validation.
      *
@@ -52,14 +59,13 @@ public class StructuredQName implements IdentityComparable {
     }
 
     /**
-     * Constructor to copy an existing StructuredQName, for use only from subclasses
-     * @param in the QName to be copied
+     * Make a StructuredQName from a StructuredQName holding the local name and URI, plus a new prefix
      */
 
-    protected StructuredQName(StructuredQName in) {
-        content = in.content;
-        localNameStart = in.localNameStart;
-        prefixStart = in.prefixStart;
+    public StructuredQName addPrefix(String prefix) {
+        char[] content2 = Arrays.copyOf(content, content.length + prefix.length());
+        prefix.getChars(0, prefix.length(), content2, content.length);
+        return new StructuredQName(content2, localNameStart, content.length);
     }
 
     /**
