@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,31 +7,31 @@
 
 package net.sf.saxon.pattern;
 
+import com.saxonica.ee.stream.Streamability;
 import com.saxonica.ee.stream.Sweep;
+import com.saxonica.ee.trans.ContextItemStaticInfoEE;
+import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.instruct.SlotManager;
-import net.sf.saxon.expr.parser.*;
-import net.sf.saxon.om.AxisInfo;
-import net.sf.saxon.om.Item;
-import net.sf.saxon.om.NodeInfo;
+import net.sf.saxon.expr.parser.ContextItemStaticInfo;
+import net.sf.saxon.expr.parser.ExpressionTool;
+import net.sf.saxon.expr.parser.ExpressionVisitor;
+import net.sf.saxon.expr.parser.RebindingMap;
+import net.sf.saxon.om.*;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.iter.ManualIterator;
 import net.sf.saxon.tree.util.Navigator;
+import net.sf.saxon.type.AnyItemType;
 import net.sf.saxon.type.ErrorType;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.Type;
 import net.sf.saxon.type.UType;
 import net.sf.saxon.value.NumericValue;
 
-import java.util.Iterator;
-import java.util.Map;
-
 /**
  * A GeneralPositionalPattern is a pattern of the form A[P] where A is an axis expression using the child axis
- * and P is an expression that depends on the position. When this kind of pattern is used for matching streamed nodes,
- * it relies on the histogram data of preceding siblings maintained as part of a
- * {@link com.saxonica.ee.stream.om.FleetingParentNode}
+ * and P is an expression that depends on the position.
  * <p/>
  * This class handles cases where the predicate P is arbitrarily complex. Simple comparisons of position() against
  * an integer value are handled by the class SimplePositionalPattern.
@@ -111,7 +111,7 @@ public class GeneralPositionalPattern extends Pattern {
     public Pattern typeCheck(ExpressionVisitor visitor, ContextItemStaticInfo contextItemType) throws XPathException {
 
         // analyze each component of the pattern
-        ContextItemStaticInfo cit = new ContextItemStaticInfo(getItemType(), false);
+        ContextItemStaticInfo cit = visitor.getConfiguration().makeContextItemStaticInfo(getItemType(), false);
 
         positionExpr = positionExpr.typeCheck(visitor, cit);
         positionExpr = ExpressionTool.unsortedIfHomogeneous(positionExpr, false);
@@ -139,7 +139,8 @@ public class GeneralPositionalPattern extends Pattern {
      */
     @Override
     public Pattern optimize(ExpressionVisitor visitor, ContextItemStaticInfo contextInfo) throws XPathException {
-        ContextItemStaticInfo cit = new ContextItemStaticInfo(getItemType(), false);
+        Configuration config = visitor.getConfiguration();
+        ContextItemStaticInfo cit = config.makeContextItemStaticInfo(getItemType(), false);
         positionExpr = positionExpr.optimize(visitor, cit);
 
         if (Literal.isConstantBoolean(positionExpr, true)) {
@@ -154,7 +155,7 @@ public class GeneralPositionalPattern extends Pattern {
         }
 
         // See if the expression is now known to be non-positional (see bugs 1908, 1992, test mode-0011)
-        if (!FilterExpression.isPositionalFilter(positionExpr, getConfiguration().getTypeHierarchy())) {
+        if (!FilterExpression.isPositionalFilter(positionExpr, config.getTypeHierarchy())) {
             byte axis = AxisInfo.CHILD;
             if (nodeTest.getPrimitiveType() == Type.ATTRIBUTE) {
                 axis = AxisInfo.ATTRIBUTE;
@@ -163,7 +164,7 @@ public class GeneralPositionalPattern extends Pattern {
             }
             AxisExpression ae = new AxisExpression(axis, nodeTest);
             FilterExpression fe = new FilterExpression(ae, positionExpr);
-            return PatternMaker.fromExpression(fe, getConfiguration(), true)
+            return PatternMaker.fromExpression(fe, config, true)
                 .typeCheck(visitor, contextInfo);
         }
         return this;
@@ -179,15 +180,6 @@ public class GeneralPositionalPattern extends Pattern {
         return positionExpr.getDependencies() & StaticProperty.DEPENDS_ON_LOCAL_VARIABLES;
     }
 
-    /**
-     * Iterate over the subexpressions within this pattern
-     */
-
-    /*@NotNull*/
-    public Iterator<Expression> iterateSubExpressions() {
-        return new MonoIterator<Expression>(positionExpr);
-    }
-
     /**
      * Allocate slots to any variables used within the pattern
      *
@@ -199,32 +191,6 @@ public class GeneralPositionalPattern extends Pattern {
         return ExpressionTool.allocateSlots(positionExpr, nextFree, slotManager);
     }
 
-    /**
-     * Offer promotion for subexpressions within this pattern. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is not accepted - this is appropriate in the case of simple expressions
-     * such as constant values and variable references where promotion would give no performance
-     * advantage. This method is always called at compile time.
-     * <p/>
-     * <p>Unlike the corresponding method on {@link net.sf.saxon.expr.Expression}, this method does not return anything:
-     * it can make internal changes to the pattern, but cannot return a different pattern. Only certain
-     * kinds of promotion are applicable within a pattern: specifically, promotions affecting local
-     * variable references within the pattern.
-     *
-     * @param offer  details of the offer, for example the offer to move
-     *               expressions that don't depend on the context to an outer level in
-     *               the containing expression
-     * @param parent the parent expression
-     * @throws net.sf.saxon.trans.XPathException if any error is detected
-     */
-
-    public void promote(PromotionOffer offer, Expression parent) throws XPathException {
-
-        Binding[] savedBindingList = offer.bindingList;
-        positionExpr = positionExpr.promote(offer);
-        offer.bindingList = savedBindingList;
-    }
-
     /**
      * Determine whether the pattern matches a given item.
      *
@@ -256,7 +222,6 @@ public class GeneralPositionalPattern extends Pattern {
      */
 
     private boolean internalMatches(NodeInfo node, NodeInfo anchor, XPathContext context) throws XPathException {
-        // System.err.println("Matching node type and fingerprint");
         if (!nodeTest.matchesNode(node)) {
             return false;
         }
@@ -267,8 +232,10 @@ public class GeneralPositionalPattern extends Pattern {
 
         try {
             XPathContext c = c2;
+            int actualPosition = -1;
             if (usesPosition) {
-                ManualIterator man = new ManualIterator(node, getActualPosition(node, Integer.MAX_VALUE));
+                actualPosition = getActualPosition(node, Integer.MAX_VALUE, context.getCurrentIterator());
+                ManualIterator man = new ManualIterator(node, actualPosition);
                 XPathContext c3 = c2.newMinorContext();
                 c3.setCurrentIterator(man);
                 c = c3;
@@ -276,12 +243,11 @@ public class GeneralPositionalPattern extends Pattern {
             Item predicate = positionExpr.evaluateItem(c);
             if (predicate instanceof NumericValue) {
                 NumericValue position = (NumericValue) positionExpr.evaluateItem(context);
-                if (position.isWholeNumber() && position.compareTo(0) > 0) {
-                    int requiredPos = (int) position.longValue();
-                    return getActualPosition(node, requiredPos) == requiredPos;
-                } else {
-                    return false;
+                int requiredPos = position.asSubscript();
+                if (actualPosition < 0 && requiredPos != -1) {
+                    actualPosition = getActualPosition(node, requiredPos, context.getCurrentIterator());
                 }
+                return requiredPos != -1 && actualPosition == requiredPos;
             } else {
                 return ExpressionTool.effectiveBooleanValue(predicate);
             }
@@ -302,7 +268,11 @@ public class GeneralPositionalPattern extends Pattern {
         }
     }
 
-    private int getActualPosition(NodeInfo node, int max) throws XPathException {
+    private int getActualPosition(NodeInfo node, int max, FocusIterator iterator) {
+        if (iterator instanceof FocusTrackingIterator) {
+            // This path makes use of cached information
+            return ((FocusTrackingIterator)iterator).getSiblingPosition(node, nodeTest, max);
+        }
         return Navigator.getSiblingPosition(node, nodeTest, max);
     }
 
@@ -366,14 +336,15 @@ public class GeneralPositionalPattern extends Pattern {
      * without repositioning the input stream. This is a necessary condition for patterns used
      * as the match pattern of a streamed template rule.
      *
-     * @param allowExtensions true if Saxon streamability extensions are allowed in the analysis
      * @return true if the pattern is motionless, that is, if it can be evaluated against a streamed
      * node without changing the position in the streamed input file
      */
 
-    public boolean isMotionless(boolean allowExtensions) {
-        positionExpr.getStreamability(allowExtensions, new ContextItemStaticInfo(getItemType(), false, true), null);
-        return positionExpr.getSweep() == Sweep.MOTIONLESS;
+    public boolean isMotionless() {
+        ContextItemStaticInfoEE csi = (ContextItemStaticInfoEE)getConfiguration().makeContextItemStaticInfo(getItemType(), false);
+        csi.setContextPostureStriding();
+        Streamability.getStreamability(positionExpr, csi, null);
+        return Streamability.getSweep(positionExpr) == Sweep.MOTIONLESS;
     }
 //#endif
 
@@ -404,6 +375,15 @@ public class GeneralPositionalPattern extends Pattern {
     public void export(ExpressionPresenter presenter) throws XPathException {
         presenter.startElement("p.genPos");
         presenter.emitAttribute("type", nodeTest.toString());
+        if ("JS".equals(presenter.getOption("target"))) {
+            try {
+                int targetVersion = presenter.getIntOption("targetVersion", 1);
+                presenter.emitAttribute("jsTest", nodeTest.generateJavaScriptItemTypeTest(AnyItemType.getInstance(), targetVersion));
+            } catch (XPathException e) {
+                e.maybeSetLocation(getLocation());
+                throw e;
+            }
+        }
         positionExpr.export(presenter);
         presenter.endElement();
     }
