@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -7,28 +7,22 @@
 
 package net.sf.saxon.pattern;
 
-import com.saxonica.ee.stream.Posture;
-import com.saxonica.ee.stream.PostureAndSweep;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.expr.*;
 import net.sf.saxon.expr.instruct.SlotManager;
-import net.sf.saxon.expr.parser.*;
+import net.sf.saxon.expr.parser.ContextItemStaticInfo;
+import net.sf.saxon.expr.parser.ExpressionVisitor;
+import net.sf.saxon.expr.parser.RebindingMap;
+import net.sf.saxon.expr.parser.XPathParser;
 import net.sf.saxon.functions.Current;
 import net.sf.saxon.om.*;
 import net.sf.saxon.style.ExpressionContext;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trans.XPathException;
-import net.sf.saxon.tree.iter.AxisIterator;
-import net.sf.saxon.tree.iter.EmptyIterator;
-import net.sf.saxon.tree.iter.PrependIterator;
-import net.sf.saxon.tree.iter.SingletonIterator;
+import net.sf.saxon.tree.iter.*;
 import net.sf.saxon.type.ItemType;
 import net.sf.saxon.type.UType;
 
-import java.util.Collections;
-import java.util.Iterator;
-import java.util.List;
-
 /**
  * A Pattern represents the result of parsing an XSLT pattern string. <br>
  * Patterns are created by calling the static method Pattern.make(string). <br>
@@ -51,13 +45,10 @@ public abstract class Pattern extends PseudoExpression {
 
     public static Pattern make(String pattern, StaticContext env, PackageData packageData) throws XPathException {
 
-        int languageLevel = 20;
-        if (packageData.getXPathVersion() >= 30 && env.getXPathVersion() >= 30) {
-            languageLevel = 30;
-        }
+        int languageLevel = 30;
         int lineNumber = env instanceof ExpressionContext ? ((ExpressionContext) env).getStyleElement().getLineNumber() : -1;
-        PatternParser parser = (PatternParser) env.getConfiguration().newExpressionParser("PATTERN", false, languageLevel);
-        ((XPathParser) parser).setLanguage(XPathParser.XSLT_PATTERN, env.getXPathVersion());
+        PatternParser parser = (PatternParser) env.getConfiguration().newExpressionParser("PATTERN", false, 30);
+        ((XPathParser) parser).setLanguage(XPathParser.XSLT_PATTERN, 30);
         Pattern pat = parser.parsePattern(pattern, env, lineNumber);
         pat.setRetainedStaticContext(env.makeRetainedStaticContext());
         // System.err.println("Simplified [" + pattern + "] to " + pat.getClass() + " default prio = " + pat.getDefaultPriority());
@@ -86,6 +77,27 @@ public abstract class Pattern extends PseudoExpression {
         }
     }
 
+    /**
+     * Ask whether a pattern has dependencies on local variables
+     * @param pattern the pattern (typically a pattern in an xsl:number or xsl:for-each-group); or null
+     * @return true if the pattern is non-null and has dependencies on local variables
+     */
+
+    public static boolean patternContainsVariable (Pattern pattern) {
+        return pattern != null &&
+                (pattern.getDependencies() & StaticProperty.DEPENDS_ON_LOCAL_VARIABLES) != 0;
+    }
+
+    /**
+     * Ask whether the expression can be lifted out of a loop, assuming it has no dependencies
+     * on the controlling variable/focus of the loop
+     * @param forStreaming
+     */
+
+    public boolean isLiftable(boolean forStreaming) {
+        return false;
+    }
+
     /**
      * Replace any calls on current() by a variable reference bound to the supplied binding
      */
@@ -94,6 +106,17 @@ public abstract class Pattern extends PseudoExpression {
         // default: no action
     }
 
+    /**
+     * Ask whether the pattern is anchored on a call on current-group()
+     * @return true if calls on matchesBeneathAnchor should test with all nodes in the
+     * current group as anchor nodes. If false, only the first node in a group is
+     * treated as the anchor node
+     */
+
+    public boolean matchesCurrentGroup() {
+        return false;
+    }
+
     /**
      * Set the original text of the pattern for use in diagnostics
      *
@@ -139,30 +162,6 @@ public abstract class Pattern extends PseudoExpression {
         return 0;
     }
 
-    /**
-     * Iterate over the subexpressions within this pattern
-     *
-     * @return an iterator over the subexpressions. Default implementation returns an empty sequence
-     */
-
-    /*@NotNull*/
-    public Iterator<Expression> iterateSubExpressions() {
-        List<Expression> list = Collections.emptyList();
-        return list.iterator();
-    }
-
-    /**
-     * Set the static context for use in evaluating subexpressions
-     * @param rsc the static context
-     */
-//
-//    public void setRetainedStaticContext(RetainedStaticContext rsc) {
-//        for (Iterator<Expression> iter = iterateSubExpressions(); iter.hasNext();) {
-//            Expression exp = iter.next();
-//            exp.setRetainedStaticContext(rsc);
-//        }
-//    }
-
     /**
      * Allocate slots to any variables used within the pattern
      *
@@ -175,43 +174,16 @@ public abstract class Pattern extends PseudoExpression {
         return nextFree;
     }
 
-    /**
-     * Offer promotion for subexpressions within this pattern. The offer will be accepted if the subexpression
-     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
-     * By default the offer is not accepted - this is appropriate in the case of simple expressions
-     * such as constant values and variable references where promotion would give no performance
-     * advantage. This method is always called at compile time.
-     * <p/>
-     * <p>Unlike the corresponding method on {@link Expression}, this method does not return anything:
-     * it can make internal changes to the pattern, but cannot return a different pattern. Only certain
-     * kinds of promotion are applicable within a pattern: specifically, promotions affecting local
-     * variable references within the pattern.
-     *
-     * @param offer  details of the offer, for example the offer to move
-     *               expressions that don't depend on the context to an outer level in
-     *               the containing expression
-     * @param parent
-     * @throws net.sf.saxon.trans.XPathException if any error is detected
-     */
-
-    public void promote(PromotionOffer offer, Expression parent) throws XPathException {
-        // default implementation does nothing
-    }
-
     /**
      * Test whether a pattern is motionless, that is, whether it can be evaluated against a node
      * without repositioning the input stream. This is a necessary condition for patterns used
      * as the match pattern of a streamed template rule.
      *
-     * @param allowExtensions if false, the result is determined strictly according to the W3C
-     *                        "guaranteed streamability rules. If true, Saxon extensions are permitted: that is, constructs
-     *                        may be recognized as motionless by Saxon even if they are not recognized as motionless by
-     *                        the W3C rules.
      * @return true if the pattern is motionless, that is, if it can be evaluated against a streamed
      * node without changing the position in the streamed input file
      */
 
-    public boolean isMotionless(boolean allowExtensions) {
+    public boolean isMotionless() {
         // default implementation for subclasses
         return true;
     }
@@ -259,7 +231,7 @@ public abstract class Pattern extends PseudoExpression {
     }
 
     /**
-     * Select nodes in a document using this PatternFinder.
+     * Select all nodes in a document that match this pattern.
      *
      * @param document     the document
      * @param context the dynamic evaluation context
@@ -277,15 +249,15 @@ public abstract class Pattern extends PseudoExpression {
             }
         } else if (UType.ATTRIBUTE.subsumes(uType)) {
             AxisIterator allElements = doc.iterateAxis(AxisInfo.DESCENDANT, NodeKindTest.ELEMENT);
-            MappingFunction<NodeInfo, NodeInfo> atts = new MappingFunction<NodeInfo, NodeInfo>() {
-                public SequenceIterator map(NodeInfo item) {
-                    return item.iterateAxis(AxisInfo.ATTRIBUTE);
+            MappingFunction atts = new MappingFunction() {
+                public SequenceIterator map(Item item) {
+                    return ((NodeInfo)item).iterateAxis(AxisInfo.ATTRIBUTE);
                 }
             };
             SequenceIterator allAttributes =
-                    new MappingIterator<NodeInfo, NodeInfo>(allElements, atts);
-            ItemMappingFunction<NodeInfo, NodeInfo> test = new ItemMappingFunction<NodeInfo, NodeInfo>() {
-                public NodeInfo mapItem(NodeInfo item) throws XPathException {
+                    new MappingIterator(allElements, atts);
+            ItemMappingFunction test = new ItemMappingFunction() {
+                public Item mapItem(Item item) throws XPathException {
                     if (matches(item, context)) {
                         return item;
                     } else {
@@ -296,15 +268,15 @@ public abstract class Pattern extends PseudoExpression {
             return new ItemMappingIterator(allAttributes, test);
         } else if (UType.NAMESPACE.subsumes(uType)) {
             AxisIterator allElements = doc.iterateAxis(AxisInfo.DESCENDANT, NodeKindTest.ELEMENT);
-            MappingFunction<NodeInfo, NodeInfo> atts = new MappingFunction<NodeInfo, NodeInfo>() {
-                public SequenceIterator map(NodeInfo item) {
-                    return item.iterateAxis(AxisInfo.NAMESPACE);
+            MappingFunction atts = new MappingFunction() {
+                public SequenceIterator map(Item item) {
+                    return ((NodeInfo)item).iterateAxis(AxisInfo.NAMESPACE);
                 }
             };
             SequenceIterator allNamespaces =
-                    new MappingIterator<NodeInfo, NodeInfo>(allElements, atts);
-            ItemMappingFunction<NodeInfo, NodeInfo> test = new ItemMappingFunction<NodeInfo, NodeInfo>() {
-                public NodeInfo mapItem(NodeInfo item) throws XPathException {
+                    new MappingIterator(allElements, atts);
+            ItemMappingFunction test = new ItemMappingFunction() {
+                public Item mapItem(Item item) throws XPathException {
                     if (matches(item, context)) {
                         return item;
                     } else {
@@ -327,15 +299,21 @@ public abstract class Pattern extends PseudoExpression {
             };
             return new ItemMappingIterator(allChildren, test);
         } else {
-            // TODO: handle patterns that match namespaces as well as other nodes
-            AxisIterator allChildren = doc.iterateAxis(AxisInfo.DESCENDANT);
-            MappingFunction<NodeInfo, NodeInfo> attsOrSelf = new MappingFunction<NodeInfo, NodeInfo>() {
-                public SequenceIterator map(NodeInfo item) {
-                    return new PrependIterator(item, item.iterateAxis(AxisInfo.ATTRIBUTE));
+            byte axis = uType.subsumes(UType.DOCUMENT) ? AxisInfo.DESCENDANT_OR_SELF : AxisInfo.DESCENDANT;
+            AxisIterator allChildren = doc.iterateAxis(axis);
+            MappingFunction processElement = new MappingFunction() {
+                public SequenceIterator map(Item item) {
+                    AxisIterator mapper = SingleNodeIterator.makeIterator((NodeInfo)item);
+                    if (uType.subsumes(UType.NAMESPACE)) {
+                        mapper = new ConcatenatingAxisIterator(mapper, ((NodeInfo) item).iterateAxis(AxisInfo.NAMESPACE));
+                    }
+                    if (uType.subsumes(UType.ATTRIBUTE)) {
+                        mapper = new ConcatenatingAxisIterator(mapper, ((NodeInfo) item).iterateAxis(AxisInfo.ATTRIBUTE));
+                    }
+                    return mapper;
                 }
             };
-            SequenceIterator attributesOrSelf =
-                    new MappingIterator<NodeInfo, NodeInfo>(allChildren, attsOrSelf);
+            SequenceIterator attributesOrSelf = new MappingIterator(allChildren, processElement);
             ItemMappingFunction test = new ItemMappingFunction() {
                 public Item mapItem(Item item) throws XPathException {
                     if (matches(item, context)) {
@@ -478,26 +456,27 @@ public abstract class Pattern extends PseudoExpression {
         return this;
     }
 
+
     @Override
     public String toShortString() {
         return toString();
     }
 
-    //#ifdefined STREAM
-
-    @Override
-    public PostureAndSweep getStreamability(boolean allowExtensions, ContextItemStaticInfo contextInfo, List<String> reasons) {
-        if (contextInfo.getContextItemPosture() == Posture.GROUNDED || isMotionless(allowExtensions)) {
-            return PostureAndSweep.GROUNDED_AND_MOTIONLESS;
-        } else {
-            if (reasons != null) {
-                reasons.add("The pattern " + toShortString() + " is not motionless");
-            }
-            return PostureAndSweep.ROAMING_AND_FREE_RANGING;
-        }
-    }
-
-    //#endif
+//    //#ifdefined STREAM
+//
+//    @Override
+//    public PostureAndSweep getStreamability(ContextItemStaticInfo contextInfo, List<String> reasons) {
+//        if (contextInfo.getContextItemPosture() == Posture.GROUNDED || isMotionless()) {
+//            return PostureAndSweep.GROUNDED_AND_MOTIONLESS;
+//        } else {
+//            if (reasons != null) {
+//                reasons.add("The pattern " + toShortString() + " is not motionless");
+//            }
+//            return PostureAndSweep.ROAMING_AND_FREE_RANGING;
+//        }
+//    }
+//
+//    //#endif
 
 
 }
