@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -102,7 +102,7 @@ public class StaticQueryContext {
      */
 
     public StaticQueryContext(/*@NotNull*/ Configuration config) {
-        this(config.getDefaultStaticQueryContext());
+        //copyFrom(config.getDefaultStaticQueryContext());
         this.config = config;
         namePool = config.getNamePool();
         errorListener = config.getErrorListener();
@@ -125,30 +125,6 @@ public class StaticQueryContext {
      * @param initial if set, this is the StaticQueryContext owned by the Configuration
      */
 
-    public StaticQueryContext(/*@NotNull*/ Configuration config, boolean initial) {
-        if (initial) {
-            this.config = config;
-            namePool = config.getNamePool();
-            errorListener = config.getErrorListener();
-            //moduleURIResolver = config.getModuleURIResolver();
-            if (errorListener instanceof StandardErrorListener) {
-                errorListener = ((StandardErrorListener) errorListener).makeAnother(Configuration.XQUERY);
-                ((StandardErrorListener) errorListener).setRecoveryPolicy(Configuration.DO_NOT_RECOVER);
-            }
-            //collationMap = new CollationMap(config.getCollationMap());
-            //locationMap = new LocationMap();
-            //schemaAware = (Boolean)config.getConfigurationProperty(FeatureKeys.XQUERY_SCHEMA_AWARE);
-//             executable = new Executable(config);
-//             executable.setCollationTable(new CollationMap(config.getCollationMap()));
-//             executable.setHostLanguage(Configuration.XQUERY);
-//             executable.setLocationMap(new LocationMap());
-            reset();
-        } else {
-            copyFrom(config.getDefaultStaticQueryContext());
-        }
-    }
-
-
     /**
      * Create a copy of a supplied StaticQueryContext
      *
@@ -159,7 +135,7 @@ public class StaticQueryContext {
         copyFrom(c);
     }
 
-    protected void copyFrom(/*@NotNull*/ StaticQueryContext c) {
+    public void copyFrom(/*@NotNull*/ StaticQueryContext c) {
         config = c.config;
         namePool = c.namePool;
         baseURI = c.baseURI;
@@ -183,16 +159,6 @@ public class StaticQueryContext {
         errorListener = c.errorListener;
         codeInjector = c.codeInjector;
         isUpdating = c.isUpdating;
-        languageVersion = c.languageVersion;
-        //collationMap = new CollationMap(c.collationMap);
-        //locationMap = new LocationMap();
-        //schemaAware = c.schemaAware;
-        //streaming = c.streaming;
-//        executable = new Executable(config);
-//        executable.setCollationTable(new CollationMap(c.executable.getCollationTable()));
-//        executable.setHostLanguage(Configuration.XQUERY);
-//        executable.setLocationMap(new LocationMap());
-//        executable.setSchemaAware(c.executable.isSchemaAware());
     }
 
 
@@ -220,7 +186,6 @@ public class StaticQueryContext {
         moduleURIResolver = null;
         clearNamespaces();
         isUpdating = false;
-        languageVersion = 30;
 
     }
 
@@ -262,7 +227,7 @@ public class StaticQueryContext {
     public Executable makeExecutable() {
         Executable executable = new Executable(config);
         executable.setSchemaAware(isSchemaAware());
-        executable.setHostLanguage(Configuration.XQUERY, getLanguageVersion() >= 30);
+        executable.setHostLanguage(Configuration.XQUERY);
         return executable;
     }
 
@@ -365,12 +330,13 @@ public class StaticQueryContext {
      *
      * @param version The XQuery language version. Must be 10 (="1.0") or 30 (="3.0") or 31 (="3.1").
      * @since 9.2; changed in 9.3 to expect a DecimalValue rather than a string. Changed in 9.7 to
-     * accept an int (30 = "3.0") and to allow "3.1".
+     * accept an int (30 = "3.0") and to allow "3.1".  From 9.8.0.3 the supplied value is ignored
+     * and the language version is always set to "3.1".
      */
 
     public void setLanguageVersion(int version) {
         if (version==10 || version==30 || version==31) {
-            languageVersion = version;
+            // value is ignored
         } else {
             throw new IllegalArgumentException("languageVersion = " + version);
         }
@@ -381,11 +347,12 @@ public class StaticQueryContext {
      *
      * @return the language version. Either "1.0" or "1.1". Default is "1.0".
      * @since 9.2; changed in 9.3 to return a DecimalValue rather than a string;
-     * changed in 9.7 to return an int (30 = "3.0" and so on)
+     * changed in 9.7 to return an int (30 = "3.0" and so on). Changed in 9.8.0.3 to
+     * always return 31.
      */
 
     public int getLanguageVersion() {
-        return languageVersion;
+        return 31;
     }
 
     /**
@@ -575,7 +542,7 @@ public class StaticQueryContext {
 
     /*@NotNull*/
     public XQueryExpression compileQuery(/*@NotNull*/ String query) throws XPathException {
-        XQueryParser qp = (XQueryParser) config.newExpressionParser("XQ", isUpdating, languageVersion);
+        XQueryParser qp = (XQueryParser) config.newExpressionParser("XQ", isUpdating, 31);
         if (codeInjector != null) {
             qp.setCodeInjector(codeInjector);
         } else if (config.isCompileWithTracing()) {
@@ -583,9 +550,7 @@ public class StaticQueryContext {
         }
         qp.setStreaming(isStreaming());
         QueryModule mainModule = new QueryModule(this);
-        if (languageVersion >= 30) {
-            qp.setDisableCycleChecks(true);
-        }
+        qp.setDisableCycleChecks(true);
         return qp.makeXQueryExpression(query, mainModule, config);
     }
 
@@ -718,6 +683,10 @@ public class StaticQueryContext {
         return null;
     }
 
+    public Collection<QueryLibrary> getCompiledLibraries() {
+        return Collections.emptySet();
+    }
+
 
     /**
      * Declare a namespace whose prefix can be used in expressions. This is
@@ -1025,15 +994,27 @@ public class StaticQueryContext {
     /**
      * Set the default collation.
      *
-     * @param name The collation name, as specified in the query prolog. The name
-     *             is not validated until it is used.
-     * @throws NullPointerException if the supplied value is null
-     * @since 8.4. Changed in 8.6 so it no longer validates the collation name: this is
-     *        because the base URI is not necessarily known at the point where the default
-     *        collation is declared.
+     * @param name The collation name, as specified in the query prolog. Must be the name
+     *             of a known registered collation.
+     * @throws NullPointerException  if the supplied value is null
+     * @throws IllegalStateException if the supplied value is not a known collation URI registered with the
+     *                               configuration.
+     * @since 8.4. Changed in 9.7.0.15 so that it validates the collation name: see bug 3121.
      */
 
     public void declareDefaultCollation(String name) {
+        if (name == null) {
+            throw new NullPointerException();
+        }
+        StringCollator c;
+        try {
+            c = getConfiguration().getCollation(name);
+        } catch (XPathException e) {
+            c = null;
+        }
+        if (c == null) {
+            throw new IllegalStateException("Unknown collation " + name);
+        }
         this.defaultCollationName = name;
     }
 
