@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -17,6 +17,7 @@ import net.sf.saxon.om.StructuredQName;
 import net.sf.saxon.trace.ExpressionPresenter;
 import net.sf.saxon.trace.InstructionInfo;
 import net.sf.saxon.trace.LocationKind;
+import net.sf.saxon.trans.FunctionStreamability;
 import net.sf.saxon.trans.GlobalVariableManager;
 import net.sf.saxon.trans.SymbolicName;
 import net.sf.saxon.trans.XPathException;
@@ -40,7 +41,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
     private NamespaceResolver namespaceResolver;
     private QueryModule staticContext;
     private boolean isUpdating = false;
-    private List<Annotation> annotations = new ArrayList<Annotation>();
+    private AnnotationList annotations = AnnotationList.EMPTY;
 
     /**
      * Create an XQuery function
@@ -151,7 +152,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
 
     /*@NotNull*/
     public SymbolicName getIdentificationKey() {
-        return new SymbolicName(StandardNames.XSL_FUNCTION, functionName, arguments.size());
+        return new SymbolicName.F(functionName, arguments.size());
     }
 
     /**
@@ -163,7 +164,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
      */
 
     public static SymbolicName getIdentificationKey(StructuredQName qName, int arity) {
-        return new SymbolicName(StandardNames.XSL_FUNCTION, qName, arity);
+        return new SymbolicName.F(qName, arity);
     }
 
     /**
@@ -289,26 +290,36 @@ public class XQueryFunction implements InstructionInfo, Declaration {
      * @param annotations the annotations, indexed by annotation name
      */
 
-    public void setAnnotations(List<Annotation> annotations) {
+    public void setAnnotations(AnnotationList annotations) {
         this.annotations = annotations;
         if (compiledFunction != null) {
             compiledFunction.setAnnotations(annotations);
         }
-        if (Annotation.existsAnnotation(annotations, Annotation.UPDATING)) {
+        if (annotations.includes(Annotation.UPDATING)) {
             setUpdating(true);
         }
     }
 
     /**
-     * Ask whether an annotation with a particular name exists
-     * @param name the name of the required annotation
-     * @return true if there is an annotation with this name            
+     * Get the annotations defined on this function
+     *
+     * @return the list of annotations defined on this function
      */
-    
-    public boolean hasAnnotation(StructuredQName name) {
-        return Annotation.existsAnnotation(annotations, name);
+
+    public AnnotationList getAnnotations() {
+        return annotations;
     }
 
+    /**
+     * Ask whether the function has an annotation with a particular name
+     * @param name the name of the required annotation
+     */
+
+    public boolean hasAnnotation(StructuredQName name) {
+        return annotations.includes(name);
+    }
+
+
     /**
      * Ask whether this is a private function (as defined in XQuery 3.0)
      *
@@ -358,8 +369,8 @@ public class XQueryFunction implements InstructionInfo, Declaration {
                 RoleDiagnostic role =
                         new RoleDiagnostic(RoleDiagnostic.FUNCTION_RESULT, functionName.getDisplayName(), 0);
                 //role.setSourceLocator(this);
-                body = TypeChecker.staticTypeCheck(body, resultType, false, role, visitor);
-                if (config.isCompileWithTracing()) {
+                body = config.getTypeChecker(false).staticTypeCheck(body, resultType, role, visitor);
+                if (staticContext.getUserQueryContext().isCompileWithTracing()) {
                     namespaceResolver = staticContext.getNamespaceResolver();
                     TraceExpression trace = new TraceExpression(body);
                     trace.setConstructType(StandardNames.XSL_FUNCTION);
@@ -369,7 +380,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
                     body = trace;
                 }
 
-                compiledFunction = config.newUserFunction(memoFunction);
+                compiledFunction = config.newUserFunction(memoFunction, FunctionStreamability.UNCLASSIFIED);
                 compiledFunction.setRetainedStaticContext(rsc);
                 compiledFunction.setPackageData(staticContext.getPackageData());
                 compiledFunction.setBody(body);
@@ -382,27 +393,11 @@ public class XQueryFunction implements InstructionInfo, Declaration {
                 compiledFunction.setUpdating(isUpdating);
                 compiledFunction.setAnnotations(annotations);
 
-//                for (UserFunctionParameter param : params) {
-//                    int refs = ExpressionTool.getReferenceCount(body, param, false);
-//                    param.setReferenceCount(refs);
-//                }
             }
             // bind all references to this function to the UserFunction object
 
             fixupReferences();
 
-            // register this function with the function library available at run-time (e.g. for saxon:evaluate())
-
-            // TODO: FIXME
-//            FunctionLibraryList libList = executable.getFunctionLibrary();
-//            if (libList.getLibraryList().size() == 1 && libList.getLibraryList().get(0) instanceof ExecutableFunctionLibrary) {
-//                ExecutableFunctionLibrary lib  = (ExecutableFunctionLibrary)libList.getLibraryList().get(0);
-//                lib.addFunction(compiledFunction);
-//            } else {
-////                throw new AssertionError("executable.getFunctionLibrary() is an instance of " +
-////                        executable.getFunctionLibrary().getClass().getName());
-//            }
-
         } catch (XPathException e) {
             e.maybeSetLocation(this);
             throw e;
@@ -439,25 +434,13 @@ public class XQueryFunction implements InstructionInfo, Declaration {
         Configuration config = staticContext.getConfiguration();
         Optimizer opt = config.obtainOptimizer();
         int arity = arguments.size();
-        if (opt.getOptimizationLevel() != Optimizer.NO_OPTIMIZATION) {
-            body = body.optimize(visitor, ContextItemStaticInfo.DEFAULT);
+        if (opt.getOptimizerOptions().isSet(OptimizerOptions.MISCELLANEOUS)) {
+            //body.verifyParentPointers();
+            body = body.optimize(visitor, ContextItemStaticInfo.ABSENT);
+            //body.verifyParentPointers();
+            body = LoopLifter.process(body, visitor, ContextItemStaticInfo.ABSENT);
             final Executable exec = ((QueryModule)getStaticContext()).getExecutable();
             GlobalVariableManager manager = new GlobalVariableManager() {
-
-                public Collection<StructuredQName> getGlobalVariableNames() {
-                    return getPackageData().getGlobalSlotManager().getVariableMap();
-                }
-
-                public GlobalVariable getGlobalVariable(StructuredQName name) {
-                    // TODO: PERF: avoid sequential search
-                    for (GlobalVariable var : getPackageData().getGlobalVariableList()) {
-                        if (var.getVariableQName().equals(name)) {
-                            return var;
-                        }
-                    }
-                    return null;
-                }
-
                 public void addGlobalVariable(GlobalVariable variable) throws XPathException {
                     PackageData pd = staticContext.getPackageData();
                     variable.setPackageData(pd);
@@ -467,6 +450,11 @@ public class XQueryFunction implements InstructionInfo, Declaration {
                     variable.compile(exec, slot);
                     pd.addGlobalVariable(variable);
                 }
+
+                @Override
+                public GlobalVariable getEquivalentVariable(Expression select) {
+                    return null;
+                }
             };
 
             // Try to extract new global variables from the body of the function
@@ -478,8 +466,6 @@ public class XQueryFunction implements InstructionInfo, Declaration {
             // mark tail calls within the function body
 
             if (!isUpdating) {
-                // TODO: implement tail call optimization for updating functions. Requires TailCallLoop to
-                // be an updating expression (like Block)
                 int tailCalls = ExpressionTool.markTailFunctionCalls(body, functionName, arity);
                 if (tailCalls != 0) {
                     compiledFunction.setBody(body);
@@ -487,14 +473,14 @@ public class XQueryFunction implements InstructionInfo, Declaration {
                     body = new TailCallLoop(compiledFunction, body);
                 }
             }
-            //body.setContainer(this);
             compiledFunction.setBody(body);
         }
         compiledFunction.computeEvaluationMode();
         ExpressionTool.allocateSlots(body, arity, compiledFunction.getStackFrameMap());
         if (config.isGenerateByteCode(Configuration.XQUERY)) {
-            Expression cbody = opt.compileToByteCode(body, getFunctionName().getDisplayName(),
-                    Expression.PROCESS_METHOD | Expression.ITERATE_METHOD);
+            ICompilerService compilerService = config.makeCompilerService(Configuration.XQUERY);
+            Expression cbody = opt.compileToByteCode(compilerService, body, getFunctionName().getDisplayName(),
+                                                     Expression.PROCESS_METHOD | Expression.ITERATE_METHOD);
             if (cbody != null) {
                 body = cbody;
             }
@@ -510,9 +496,6 @@ public class XQueryFunction implements InstructionInfo, Declaration {
     public void fixupReferences() throws XPathException {
         for (UserFunctionResolvable ufc : references) {
             ufc.setFunction(compiledFunction);
-            if (ufc instanceof UserFunctionCall) {
-                ((UserFunctionCall) ufc).computeArgumentEvaluationModes();
-            }
         }
     }
 
@@ -527,7 +510,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
             if (ufr instanceof UserFunctionCall) {
                 UserFunctionCall ufc = (UserFunctionCall) ufr;
                 ufc.checkFunctionCall(compiledFunction, visitor);
-                ufc.computeArgumentEvaluationModes();
+                //ufc.computeArgumentEvaluationModes();
             }
         }
 
@@ -551,7 +534,7 @@ public class XQueryFunction implements InstructionInfo, Declaration {
             if (compiledFunction.isMemoFunction()) {
                 out.emitAttribute("memo", "true");
             }
-            out.emitAttribute("tailRecursive", (compiledFunction.isTailRecursive() ? "true" : "false"));
+            out.emitAttribute("tailRecursive", compiledFunction.isTailRecursive() ? "true" : "false");
             body.export(out);
         }
         out.endElement();
@@ -676,8 +659,8 @@ public class XQueryFunction implements InstructionInfo, Declaration {
      */
 
     /*@NotNull*/
-    public Iterator getProperties() {
-        return new PairIterator("name", "as");
+    public Iterator<String> getProperties() {
+        return new PairIterator<String>("name", "as");
     }
 
     /**
