@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -9,10 +9,17 @@ package net.sf.saxon.regex;
 
 
 import net.sf.saxon.expr.sort.EmptyIntIterator;
+import net.sf.saxon.regex.charclass.CharacterClass;
+import net.sf.saxon.regex.charclass.EmptyCharacterClass;
+import net.sf.saxon.regex.charclass.IntSetCharacterClass;
+import net.sf.saxon.regex.charclass.SingletonCharacterClass;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.z.*;
 
-import java.util.*;
+import java.util.Iterator;
+import java.util.LinkedList;
+import java.util.List;
+import java.util.Stack;
 
 /**
  * Represents an operation or instruction in the regular expression program. The class Operation
@@ -58,6 +65,30 @@ public abstract class Operation {
 
     public abstract boolean matchesEmptyString();
 
+    /**
+     * Ask whether the expression contains any capturing sub-expressions
+     * @return true if the expression contains any capturing sub-expressions (but not
+     * if it is a capturing expression itself, unless it contains nested capturing
+     * expressions)
+     */
+
+    public boolean containsCapturingExpressions() {
+        return false;
+    }
+
+    /**
+     * Get a CharacterClass identifying the set of characters that can appear as the first
+     * character of a non-empty string that matches this term. This is allowed to be an
+     * over-estimate (that is, the returned Character class must match every character
+     * that can legitimately appear at the start of the matched string, but it can
+     * also match other characters).
+     * @param caseBlind true if case-blind matching is in force ("i" flag)
+     */
+
+    public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+        return EmptyCharacterClass.getComplement();
+    }
+
     /**
      * Optimize the operation
      * @return the optimized operation
@@ -123,6 +154,26 @@ public abstract class Operation {
             return false;
         }
 
+        @Override
+        public boolean containsCapturingExpressions() {
+            for (Operation o : branches) {
+                if (o instanceof OpCapture || o.containsCapturingExpressions()) {
+                    return true;
+                }
+            }
+            return false;
+        }
+
+        @Override
+        public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+            CharacterClass result = EmptyCharacterClass.getInstance();
+            for (Operation o : branches) {
+                result = RECompiler.makeUnion(result, o.getInitialCharacterClass(caseBlind));
+            }
+            return result;
+        }
+
+
         @Override
         public Operation optimize(REProgram program, REFlags flags) {
             for (int i=0; i<branches.size(); i++) {
@@ -240,6 +291,28 @@ public abstract class Operation {
             return true;
         }
 
+        @Override
+        public boolean containsCapturingExpressions() {
+            for (Operation o : operations) {
+                if (o instanceof OpCapture || o.containsCapturingExpressions()) {
+                    return true;
+                }
+            }
+            return false;
+        }
+
+        @Override
+        public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+            CharacterClass result = EmptyCharacterClass.getInstance();
+            for (Operation o : operations) {
+                result = RECompiler.makeUnion(result, o.getInitialCharacterClass(caseBlind));
+                if (!o.matchesEmptyString()) {
+                    return result;
+                }
+            }
+            return result;
+        }
+
         /**
          * Display the operation as a regular expression, possibly in abbreviated form
          *
@@ -262,9 +335,6 @@ public abstract class Operation {
             } else if (operations.size() == 1) {
                 return operations.get(0);
             } else {
-//                if (operations.get(0) instanceof OpBOL) {
-//                    program.optimizationFlags |= REProgram.OPT_HASBOL;
-//                }
                 for (int i=0; i<operations.size()-1; i++) {
                     Operation o1 = operations.get(i);
                     Operation o2 = o1.optimize(program, flags);
@@ -293,6 +363,8 @@ public abstract class Operation {
 
             // A stack of iterators, one for each piece in the sequence
             final Stack<IntIterator> iterators = new Stack<IntIterator>();
+            final REMatcher.State savedState =
+                    containsCapturingExpressions() ? matcher.captureState() : null;
 
             return new IntIterator() {
 
@@ -324,7 +396,9 @@ public abstract class Operation {
                         }
                         iterators.pop();
                     }
-                    //System.err.println("Clear to " + position);
+                    if (savedState != null) {
+                        matcher.resetState(savedState);
+                    }
                     //matcher.clearCapturedGroupsBeyond(position);
                     return -1;
                 }
@@ -372,6 +446,15 @@ public abstract class Operation {
             return false;
         }
 
+        @Override
+        public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+            if (predicate instanceof CharacterClass) {
+                return (CharacterClass)predicate;
+            } else {
+                return super.getInitialCharacterClass(caseBlind);
+            }
+        }
+
         @Override
         public IntIterator iterateMatches(REMatcher matcher, int position) {
             //System.err.println("CharClass " + predicate + " matching at position " + position);
@@ -442,6 +525,26 @@ public abstract class Operation {
             return len == 0;
         }
 
+        @Override
+        public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+            if (len == 0) {
+                return EmptyCharacterClass.getInstance();
+            } else if (caseBlind) {
+                IntSet set;
+                int ch = atom.uCharAt(0);
+                int[] variants = CaseVariants.getCaseVariants(ch);
+                if (variants.length > 0) {
+                    set = new IntHashSet(variants.length);
+                    set.add(ch);
+                    for (int v : variants) {
+                        set.add(v);
+                    }
+                    return new IntSetCharacterClass(set);
+                }
+            }
+            return new SingletonCharacterClass(atom.uCharAt(0));
+        }
+
         @Override
         public IntIterator iterateMatches(REMatcher matcher, int position) {
             UnicodeString in = matcher.search;
@@ -636,6 +739,16 @@ public abstract class Operation {
             return min == 0 || op.matchesEmptyString();
         }
 
+        @Override
+        public boolean containsCapturingExpressions() {
+            return op instanceof OpCapture || op.containsCapturingExpressions();
+        }
+
+        @Override
+        public CharacterClass getInitialCharacterClass(boolean caseBlind) {
+            return op.getInitialCharacterClass(caseBlind);
+        }
+
         @Override
         public int getMatchLength() {
             return min == max && op.getMatchLength() >= 0 ? min * op.getMatchLength() : -1;
@@ -1030,13 +1143,12 @@ public abstract class Operation {
                 public int next() {
                     int next = base.next();
                     // Increase valid paren count
-                    if (groupNr >= matcher.parenCount) {
-                        matcher.parenCount = groupNr + 1;
+                    if (groupNr >= matcher.captureState.parenCount) {
+                        matcher.captureState.parenCount = groupNr + 1;
                     }
 
                     // Don't set paren if already set later on
                     //if (matcher.getParenStart(groupNr) == -1) {
-                        //System.err.println("Set group " + groupNr + " to (" + position + ", " + next + ")");
                         matcher.setParenStart(groupNr, position);
                         matcher.setParenEnd(groupNr, next);
                     //}
