@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,14 +8,12 @@
 package net.sf.saxon.s9api;
 
 import net.sf.saxon.Configuration;
-import net.sf.saxon.Controller;
-import net.sf.saxon.expr.JPConverter;
-import net.sf.saxon.expr.XPathContext;
 import net.sf.saxon.expr.parser.Token;
 import net.sf.saxon.lib.NamespaceConstant;
+import net.sf.saxon.ma.arrays.ArrayItemType;
+import net.sf.saxon.ma.map.MapType;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.*;
-import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.ObjectValue;
@@ -346,7 +344,7 @@ public class ItemTypeFactory {
         NameTest nameTest = null;
         ContentTypeTest contentTest = null;
         if (name != null) {
-            int elementFP = config.getNamePool().allocate("", name.getNamespaceURI(), name.getLocalName());
+            int elementFP = config.getNamePool().allocateFingerprint(name.getNamespaceURI(), name.getLocalName());
             nameTest = new NameTest(Type.ELEMENT, elementFP, config.getNamePool());
         }
         if (schemaType != null) {
@@ -426,7 +424,7 @@ public class ItemTypeFactory {
         ContentTypeTest contentTest = null;
         Configuration config = processor.getUnderlyingConfiguration();
         if (name != null) {
-            int attributeFP = config.getNamePool().allocate("", name.getNamespaceURI(), name.getLocalName());
+            int attributeFP = config.getNamePool().allocateFingerprint(name.getNamespaceURI(), name.getLocalName());
             nameTest = new NameTest(Type.ATTRIBUTE, attributeFP, config.getNamePool());
         }
         if (schemaType != null) {
@@ -507,6 +505,32 @@ public class ItemTypeFactory {
         return (XdmItem) XdmItem.wrap(new ObjectValue<Object>(object));
     }
 
+    /**
+     * Obtain a map type, that is a type for XDM maps with a given key type and value type
+     * @param keyType the type of the keys in the map
+     * @param valueType the type of the values in the map
+     * @return the required map type
+     */
+
+    public ItemType getMapType(ItemType keyType, SequenceType valueType) {
+        if (!(keyType.getUnderlyingItemType() instanceof AtomicType)) {
+            throw new IllegalArgumentException("Map key must be atomic");
+        }
+        return new ConstructedItemType(
+                new MapType((AtomicType)keyType.getUnderlyingItemType(), valueType.getUnderlyingSequenceType()), processor);
+    }
+
+    /**
+     * Obtain an array type, that is a type for XDM arrays with a given member type
+     *
+     * @param memberType the type of the members in the array
+     * @return the required array type
+     */
+
+    public ItemType getArrayType(SequenceType memberType) {
+        return new ConstructedItemType(new ArrayItemType(memberType.getUnderlyingSequenceType()), processor);
+    }
+
     /**
      * Factory method to construct a map item from a Java map
      * @param map the input map. The contents of this map are shallow-copied, so subsequent changes to the
@@ -518,19 +542,8 @@ public class ItemTypeFactory {
      * @since 9.6
      */
 
-    public XdmFunctionItem newMap(Map<?, ?> map) throws SaxonApiException {
-        XdmFunctionItem fi = null;
-        //#if EE==true || PE==true
-        try {
-            Controller controller = new Controller(processor.getUnderlyingConfiguration());
-            XPathContext context = controller.newXPathContext();
-            Sequence result = new JPConverter.FromMap().convert(map, context);
-            fi = new XdmFunctionItem((Function)result);
-        } catch (XPathException e) {
-            throw new SaxonApiException(e);
-        }
-        //#endif
-        return fi;
+    public XdmMap newMap(Map<?, ?> map) throws SaxonApiException {
+        return XdmMap.makeMap(map);
     }
 
     /**
