@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -17,7 +17,10 @@ import net.sf.saxon.expr.parser.ExplicitLocation;
 import net.sf.saxon.expr.sort.RuleBasedSubstringMatcher;
 import net.sf.saxon.expr.sort.SimpleCollation;
 import net.sf.saxon.lib.*;
-import net.sf.saxon.om.*;
+import net.sf.saxon.om.Item;
+import net.sf.saxon.om.NodeInfo;
+import net.sf.saxon.om.Sequence;
+import net.sf.saxon.om.StructuredQName;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.value.SequenceExtent;
 import net.sf.saxon.value.SequenceType;
@@ -27,6 +30,7 @@ import java.io.File;
 import java.io.OutputStream;
 import java.io.Writer;
 import java.text.RuleBasedCollator;
+import java.util.Comparator;
 
 /**
  * The <code>Processor</code> class serves three purposes: it allows global Saxon configuration options to be set;
@@ -369,14 +373,14 @@ public class Processor {
      * Bind a collation URI to a collation
      *
      * @param uri       the absolute collation URI
-     * @param collation a {@link java.text.Collator} object that implements the required collation
+     * @param collation a {@link Comparator} object that implements the required collation
      * @throws IllegalArgumentException if an attempt is made to rebind the standard URI
      *                                  for the Unicode codepoint collation or the HTML5 case-blind
      *                                  collation
-     * @since 9.6
+     * @since 9.6. Changed in 9.8 to allow any Comparator to be supplied as a collation
      */
 
-    public void declareCollation(String uri, final java.text.Collator collation) {
+    public void declareCollation(String uri, final Comparator collation) {
         if (uri.equals(NamespaceConstant.CODEPOINT_COLLATION_URI)) {
             throw new IllegalArgumentException("Cannot redeclare the Unicode codepoint collation URI");
         }
