@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -10,21 +10,16 @@ package net.sf.saxon.s9api;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.PreparedStylesheet;
 import net.sf.saxon.lib.ErrorGatherer;
-import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.lib.StringCollator;
-import net.sf.saxon.om.AxisInfo;
-import net.sf.saxon.om.NodeInfo;
-import net.sf.saxon.om.StandardNames;
-import net.sf.saxon.pattern.NameTest;
 import net.sf.saxon.style.*;
 import net.sf.saxon.trace.XSLTTraceCodeInjector;
-import net.sf.saxon.trans.*;
-import net.sf.saxon.tree.iter.AxisIterator;
+import net.sf.saxon.trans.CompilerInfo;
+import net.sf.saxon.trans.XPathException;
+import net.sf.saxon.trans.packages.IPackageLoader;
+import net.sf.saxon.trans.packages.PackageDetails;
+import net.sf.saxon.trans.packages.PackageLibrary;
+import net.sf.saxon.trans.packages.VersionedPackageName;
 import net.sf.saxon.tree.linked.DocumentImpl;
-import net.sf.saxon.tree.linked.ElementImpl;
-import net.sf.saxon.type.Type;
-import net.sf.saxon.value.NestedIntegerValue;
-import net.sf.saxon.value.Whitespace;
 
 import javax.xml.transform.ErrorListener;
 import javax.xml.transform.Source;
@@ -43,8 +38,9 @@ import java.util.List;
  * <p/>
  * <p>An XsltCompiler may be used repeatedly to compile multiple queries. Any changes made to the
  * XsltCompiler (that is, to the static context) do not affect queries that have already been compiled.
- * An XsltCompiler may be used concurrently in multiple threads, but it should not then be modified once
- * initialized.</p>
+ * An XsltCompiler may in principle be used concurrently in multiple threads, but in practice this
+ * is best avoided because all instances will share the same ErrorListener and it may therefore be
+ * difficult to establish which error messages are associated with each compilation.</p>
  *
  * @since 9.0
  */
@@ -68,6 +64,7 @@ public class XsltCompiler {
         compilerInfo = new CompilerInfo(config.getDefaultXsltCompilerInfo());
         compilerInfo.setGenerateByteCode(config.isGenerateByteCode(Configuration.XSLT));
         compilerInfo.setTargetEdition(config.getEditionCode());
+        compilerInfo.setJustInTimeCompilation(config.isJITEnabled(Configuration.XSLT));
     }
 
     /**
@@ -185,15 +182,39 @@ public class XsltCompiler {
         return compilerInfo.isSchemaAware();
     }
 
+    /**
+     * Ask whether any package produced by this compiler can be deployed to a different location, with a different base URI
+     *
+     * @return if true then static-base-uri() represents the deployed location of the package,
+     *                    rather than its compile time location
+     * @since 9.8
+     */
+
+    public boolean isRelocatable() {
+        return compilerInfo.isRelocatable();
+    }
+
+    /**
+     * Say whether any package produced by this compiler can be deployed to a different location, with a different base URI
+     *
+     * @param relocatable if true then static-base-uri() represents the deployed location of the package,
+     *                    rather than its compile time location
+     * @since 9.8
+     * */
+
+    public void setRelocatable(boolean relocatable) {
+        compilerInfo.setRelocatable(relocatable);
+    }
+
     /**
      * Set the target edition under which the stylesheet will be executed.
      *
-     * @param edition the Saxon edition for the run-time environment. One of "EE", "PE", "HE", or "JS".
+     * @param edition the Saxon edition for the run-time environment. One of "EE", "PE", "HE", "JS", or "JS2".
      * @since 9.7.0.5. Experimental and subject to change.
      */
 
     public void setTargetEdition(String edition) {
-        if (!("EE".equals(edition) || "PE".equals(edition) || "HE".equals(edition) || "JS".equals(edition))) {
+        if (!("EE".equals(edition) || "PE".equals(edition) || "HE".equals(edition) || "JS".equals(edition) || "JS2".equals(edition))) {
             throw new IllegalArgumentException("Unknown Saxon edition " + edition);
         }
         compilerInfo.setTargetEdition(edition);
@@ -202,7 +223,7 @@ public class XsltCompiler {
     /**
      * Get the target edition under which the stylesheet will be executed.
      *
-     * @return the Saxon edition for the run-time environment. One of "EE", "PE", "HE", or "JS".
+     * @return the Saxon edition for the run-time environment. One of "EE", "PE", "HE", "JS", or "JS2".
      * @since 9.7.0.5. Experimental and subject to change.
      */
 
@@ -229,8 +250,9 @@ public class XsltCompiler {
     /**
      * Declare the default collation
      *
-     * @param uri the absolute URI of the default collation. This URI must have been bound to a collation
-     *            using the method {@link #declareCollation(String, java.text.Collator)}, or it must be a
+     * @param uri the absolute URI of the default collation. Either this URI must have been bound to a collation
+     *            using the method {@link Configuration#registerCollation(String, StringCollator)}
+     *            Collation(String, java.text.Collator)}, or it must be a
      *            collation that is recognized implicitly, such as a UCA collation
      * @throws IllegalStateException if the collation URI has not been registered, unless it is the standard
      *                               Unicode codepoint collation which is registered implicitly
@@ -262,54 +284,28 @@ public class XsltCompiler {
     }
 
     /**
-     * Set the XSLT (and XPath) language level to be supported by the processor.
+     * Set the XSLT (and XPath) language level to be supported by the processor. This has no effect
+     * from Saxon 9.8: the processor is an XSLT 3.0 processor regardless of the language level
+     * requested.
      *
-     * @param version the language level to be supported. The value 2.0 requests a processor conforming to the
-     *                XSLT 2.0 specification; 3.0 requests an XSLT 3.0 processor, while the value "0.0" (which is the
-     *                default setting in the absence of a call on this method) gets a processor according to the
-     *                value of the version attribute on the xsl:stylesheet element.
-     *                <p>Although this interface can be used to run a 1.0 stylesheet, it is not possible to request
-     *                a 1.0 processor; Saxon will handle it as a 2.0 processor in backwards-compatibility mode, which
-     *                is not quite the same thing.</p>
-     *                <p>Note that Saxon-HE does not support XSLT 3.0. However, requesting 3.0
-     *                via this interface does not cause an error; an error only occurs if the stylesheet
-     *                actually uses XSLT 3.0 features.</p>
+     * @param version the language level to be supported. The value is ignored.
      * @throws IllegalArgumentException if the value is not equal to 0.0, 2.0, or 3.0
-     * @since 9.3
+     * @deprecated Has no effect from Saxon 9.8.
+     * @since 9.3. Has no effect from Saxon 9.8.
      */
 
     public void setXsltLanguageVersion(String version) {
-        int v = 0;
-        if ("2.0".equals(version)) {
-            v = 20;
-        } else if ("3.0".equals(version)) {
-            v = 30;
-        } else if ("0.0".equals(version)) {
-            v = 0;
-        } else {
-            throw new IllegalArgumentException("XSLT version must be 0.0, 2.0, or 3.0");
-        }
-        compilerInfo.setXsltVersion(v);
     }
 
     /**
      * Get the XSLT (and XPath) language level supported by the processor.
      *
-     * @return the language level supported. The value 2.0 represents a processor conforming to the
-     * XSLT 2.0 specification; 3.0 represents an XSLT 3.0 processor, while the value zero (which is the
-     * default setting in the absence of a call on {@link #setXsltLanguageVersion} represents a processor
-     * selected according to the value of the version attribute on the xsl:stylesheet element.
+     * @return the language level supported. From Saxon 9.8 this always returns "3.0".
      * @since 9.3
      */
 
     public String getXsltLanguageVersion() {
-        int v = compilerInfo.getXsltVersion();
-        switch (v) {
-            case 0: return "0.0";
-            case 20: return "2.0";
-            case 30: return "3.0";
-            default: return "" + v;
-        }
+        return "3.0";
     }
 
     /**
@@ -394,20 +390,16 @@ public class XsltCompiler {
     }
 
 
-//    /**
-//     * Import a compiled XQuery function library.
-//     * @param queryCompiler An XQueryCompiler that has been used to compile a library of XQuery functions
-//     * (by using one of the overloaded methods named <code>compileLibrary</code>).
-//     * @param namespace The namespace of the functions that are to be made available to the stylesheet.
-//     * All the global functions with this namespace that are
-//     * defined in libraries that have been compiled using this XQueryCompiler are added to the static context
-//     * of the XSLT stylesheet. The stylesheet does not need to (and should not) contain a call on
-//     * <code>saxon:import-query</code> to import these functions.
-//     */
-//
-//    public void importXQueryLibrary(XQueryCompiler queryCompiler, String namespace) {
-//        compilerInfo.importXQueryLibrary(queryCompiler.)
-//    }
+    /**
+     * Import a compiled XQuery library. This makes pre-compiled XQuery library modules available
+     * to the <code>saxon:import-query</code> declaration.
+     * @param queryCompiler An XQueryCompiler that has been used to compile a library of XQuery functions
+     * (by using one of the overloaded methods named <code>compileLibrary</code>).
+     */
+
+    public void importXQueryEnvironment(XQueryCompiler queryCompiler) {
+        compilerInfo.setXQueryLibraries(queryCompiler.getUnderlyingStaticContext().getCompiledLibraries());
+    }
 
 
     /**
@@ -461,8 +453,7 @@ public class XsltCompiler {
      * <p/>
      * <p>The source argument identifies an XML file containing an &lt;xsl:package&gt; element. Any packages
      * on which this package depends must have been made available to the <code>XsltCompiler</code>
-     * by importing them either using {@link #importPackage}
-     * or added previously through {@link #addCompilePackages(Iterable, boolean)}}.</p>
+     * by importing them using {@link #importPackage}.</p>
      *
      * @param source identifies an XML document holding the the XSLT package to be compiled
      * @return the XsltPackage that results from the compilation. Note that this package
@@ -482,65 +473,36 @@ public class XsltCompiler {
             } else {
                 compilation = new Compilation(config, compilerInfo);
             }
+            compilation.setLibraryPackage(true);
             XsltPackage pack = new XsltPackage(processor, compilation.compilePackage(source).getStylesheetPackage());
             if (compilation.getErrorCount() > 0) {
                 throw new SaxonApiException("Package compilation failed: " + compilation.getErrorCount() + " errors reported");
             }
             return pack;
         } catch (XPathException e) {
-            throw new SaxonApiException(e.getMessage());
+            throw new SaxonApiException(e);
         }
     }
 
-    /**
-     * Initialise a fresh package library for this compiler
-     * This permits incremental additions to the library
-     */
-    public void newPackageLibrary() {
-        compilerInfo.setPackageLibrary(new PackageLibrary());
-    }
-
-    public PackageLibrary getPackageLibrary() {
+    private PackageLibrary getPackageLibrary() {
         return compilerInfo.getPackageLibrary();
     }
-    public void setPackageLibrary(PackageLibrary packageLibrary) {
-        compilerInfo.setPackageLibrary(packageLibrary);
-    }
+
 
     /**
      * Compile a list of packages.
      *
      * @param sources the collection of packages to be compiled, in the form of an Iterable
      * @return the collection of compiled packages, in the form of an Iterable.
-     * @throws SaxonApiException if the source cannot be read or if static errors are found during the
-     *                           compilation. Any such errors will have been notified to the registered <code>ErrorListener</code>
-     *                           if there is one, or reported on the <code>System.err</code> output stream otherwise.
+     * @throws UnsupportedOperationException - always. This method is no longer available
+     * from Saxon 9.8
      * @since 9.6
+     * @deprecated since 9.8. Multiple packages may be supplied in the form of a {@link PackageLibrary} registered
+     * with the underlying {@link CompilerInfo}, or may be listed in the configuration file. Alternatively they
+     * can be imported explicitly (taking care over the order of importing) using {@link #importPackage(XsltPackage)}.
      */
-    public Iterable<XsltPackage> compilePackages(Iterable<Source> sources) throws SaxonApiException, XPathException {
-
-        Compilation compilation = new Compilation(config, compilerInfo);
-        compilerInfo.setPackageLibrary(new PackageLibrary());
-        List<XsltPackage> result = new ArrayList<XsltPackage>();
-        ArrayList<PackDepends> depends = new ArrayList<PackDepends>();
-
-        for (Source s : sources) {
-            DocumentImpl document = StylesheetModule.loadStylesheetModule(s, true, compilation, NestedIntegerValue.TWO);
-            ElementImpl packageElement = document.getDocumentElement();
-            String packageName = packageElement.getAttributeValue("", "name");
-            if (packageName == null) {
-                throw new SaxonApiException("Outermost element must be an xsl:package element with a name attribute");
-            }
-            depends.add(new PackDepends(document));
-        }
-
-        for (PackDepends p : resolveDependencies(depends)) {
-            XsltPackage packagei = compilePackage(p.doc);
-            compilerInfo.getPackageLibrary().addPackage(packagei.getName(), packagei.getUnderlyingPreparedPackage());
-            result.add(packagei);
-        }
-
-        return result;
+    public Iterable<XsltPackage> compilePackages(Iterable<Source> sources) throws SaxonApiException {
+        throw new UnsupportedOperationException("XsltCompiler#compilePackages() is dropped in Saxon 9.8");
     }
 
     /**
@@ -557,192 +519,13 @@ public class XsltCompiler {
      *                           compilation. Any such errors will have been notified to the registered <code>ErrorListener</code>
      *                           if there is one, or reported on the <code>System.err</code> output stream otherwise.
      * @throws XPathException    if there are XPath errors in the stylesheet
+     * @since 9.6
+     * @deprecated since 9.8. Multiple packages may be supplied in the form of a {@link PackageLibrary} registered
+     * with the underlying {@link CompilerInfo}, or may be listed in the configuration file. Alternatively they
+     * can be imported explicitly (taking care over the order of importing) using {@link #importPackage(XsltPackage)}.
      */
     public Iterable<XsltPackage> addCompilePackages(Iterable<Source> sources, boolean link) throws SaxonApiException, XPathException {
-
-        Compilation compilation = new Compilation(config, compilerInfo);
-        List<XsltPackage> result = new ArrayList<XsltPackage>();
-        ArrayList<PackDepends> depends = new ArrayList<PackDepends>();
-
-        for (Source s : sources) {
-            DocumentImpl document = StylesheetModule.loadStylesheetModule(s, true, compilation, NestedIntegerValue.TWO);
-//            ElementImpl packageElement = document.getDocumentElement();
-//            String packageName = packageElement.getAttributeValue("", "name");
-//            if (packageName == null) {
-//                throw new SaxonApiException("Outermost element must be an xsl:package element with a name attribute");
-//            }
-            depends.add(new PackDepends(document));
-        }
-
-//        // Find the packages that have already been processed and thus have had dependencies resolved.
-//        ArrayList<PackDepends> alreadyResolved = new ArrayList<PackDepends>();
-//        for (StylesheetPackage p : compilerInfo.getPackageLibrary().getPackages()) {
-//            alreadyResolved.add(new PackDepends(p));
-//        }
-
-        PackageLibrary pl = compilerInfo.getPackageLibrary();
-        // Compile the newly provided packages in dependency order and add them to the package library
-
-        ArrayList<PackDepends> resolved = resolveDependencies(depends);
-        for (PackDepends p : resolved) {
-            if (p.doc != null) {
-                XsltPackage packagei;
-                ElementImpl top = p.doc.getDocumentElement();
-                if (top.getURI().equals(NamespaceConstant.SAXON_XSLT_EXPORT)) {
-                    IPackageLoader loader = config.makePackageLoader();
-                    StylesheetPackage pack = loader.loadPackageDoc(p.doc);
-                    packagei = new XsltPackage(processor, pack);
-                } else {
-                    packagei = compilePackage(p.doc);
-                }
-                pl.addPackage(packagei.getName(), packagei.getUnderlyingPreparedPackage());
-                if (link) {
-                    packagei.link();
-                }
-                result.add(packagei);
-            }
-        }
-        return result;
-    }
-
-    /* At present we don't need to track the package versions, but we could to pre-evaluate and avoid
-     * compiling packages that aren't used. */
-
-    /**
-     * A class to accommodate a package and its dependencies
-     * - it can either hold a pre-compiled or a post-compiled version
-     */
-    private class PackDepends {
-        DocumentImpl doc;
-        StylesheetPackage pack;
-        String packageName;
-        PackageVersion packageVersion;
-        ArrayList<UsePack> uses;
-
-        PackDepends(StylesheetPackage p) {
-            pack = p;
-            packageName = p.getPackageName();
-            packageVersion = p.getPackageVersion();
-            /* We shouldn't need to track the package uses of a compiled package */
-            //p.getUsedPackages();
-        }
-
-        PackDepends(DocumentImpl d) throws XPathException {
-            doc = d;
-            ElementImpl packageElement = doc.getDocumentElement();
-
-            uses = new ArrayList<UsePack>();
-            if (packageElement.getURI().equals(NamespaceConstant.XSLT)) {
-                packageName = packageElement.getAttributeValue("", "name");
-                if (packageName == null) {
-                    packageName = "#unnamedPackage";
-                }
-                String pv = packageElement.getAttributeValue("", "package-version");
-                packageVersion = pv == null ? PackageVersion.ONE : new PackageVersion(pv);
-                // TODO: this is flawed - see bug 2779
-                AxisIterator iter = doc.iterateAxis(AxisInfo.DESCENDANT,
-                                                    new NameTest(Type.ELEMENT, StandardNames.XSL_USE_PACKAGE, config.getNamePool()));
-                NodeInfo current;
-                while ((current = iter.next()) != null) {
-                    uses.add(new UsePack(current));
-                }
-            } else {
-                packageName = packageElement.getAttributeValue("", "name");
-                String pv = packageElement.getAttributeValue("", "packageVersion");
-                packageVersion = pv == null ? PackageVersion.ONE : new PackageVersion(pv);
-            }
-        }
-
-        // For checking that this package is suitable for a given use
-        // typically via a contains() invocation
-        boolean equals(UsePack u) {
-            return u.packageName.equals(packageName) && u.ranges.contains(packageVersion);
-        }
-    }
-
-
-    /**
-     * Describes an xsl:use-package requirement from within a package,
-     * in terms of a name and a set of version ranges
-     */
-    private class UsePack {
-        String packageName;
-        PackageVersionRanges ranges;
-
-        UsePack(NodeInfo current) throws XPathException {
-            packageName = Whitespace.trim(current.getAttributeValue("", "name"));
-            String pvr = Whitespace.trim(current.getAttributeValue("", "package-version"));
-            ranges = new PackageVersionRanges(pvr == null ? "*" : pvr);
-        }
-    }
-
-    /**
-     * Reorder the packages to hopefully place them in increasing dependency order.
-     *
-     * @param in packages to be analysed
-     * @return packages in increasing dependency order
-     * @throws SaxonApiException if there appears to be circular dependency
-     */
-
-    ArrayList<PackDepends> resolveDependencies(ArrayList<PackDepends> in) throws SaxonApiException {
-        ArrayList<PackDepends> resolved = new ArrayList<PackDepends>();
-        resolveDependencies(in, resolved);
-        return resolved;
-    }
-
-    /**
-     * Reorder the packages to hopefully place them in increasing dependency order.
-     *
-     * @param in       packages remaining to be analysed
-     * @param resolved packages whose dependencies have already been resolved
-     * @return packages in increasing dependency order, i.e.
-     * it should be possible to compile and add to the package library each in sequence
-     * @throws SaxonApiException if there appears to be circular dependency,
-     *                           which is effectively that at some pass every stylesheet has a dependency on one
-     *                           whose dependency order is still to be determined.
-     */
-    private void resolveDependencies(ArrayList<PackDepends> in,
-                                               ArrayList<PackDepends> resolved) throws SaxonApiException {
-        if (in.isEmpty()) {
-            return;
-        }
-        ArrayList<PackDepends> unresolved = new ArrayList<PackDepends>();
-        boolean changed = false;
-        for (PackDepends p : in) {
-            if (allDependenciesSatisfied(p, resolved)) {
-                resolved.add(p);
-                changed = true;
-            } else {
-                unresolved.add(p);
-            }
-        }
-        if (changed) {
-            resolveDependencies(unresolved, resolved);
-        } else {
-            String message = "Unable to resolve package dependencies for " + unresolved.get(0).packageName;
-            throw new SaxonApiException(message);
-            // TODO: be more specific if there's a circularity
-        }
-    }
-
-    private boolean allDependenciesSatisfied(PackDepends packDepends, List<PackDepends> resolved) {
-        PackageLibrary lib = getPackageLibrary();
-        for (UsePack u : packDepends.uses) {
-            boolean found = false;
-            if (lib.getPackage(u.packageName, u.ranges) != null) {
-                continue;
-            }
-            for (PackDepends pd : resolved) {
-                if (pd.packageName.equals(u.packageName) && u.ranges.contains(pd.packageVersion)) {
-                    found = true;
-                    break;
-                }
-            }
-            if (!found) {
-                return false;
-            }
-        }
-        return true;
+        throw new UnsupportedOperationException("XsltCompiler.addCompilePackages is dropped from Saxon 9.8");
     }
 
     /**
@@ -814,9 +597,107 @@ public class XsltCompiler {
 
     public void importPackage(XsltPackage thePackage) throws SaxonApiException {
         if (thePackage.getProcessor() != processor) {
-            throw new SaxonApiException("The imported package and the XsltCompiler must belong to the same Processor");
+            throw new SaxonApiException(
+                    "The imported package and the XsltCompiler must belong to the same Processor");
+        }
+        compilerInfo.getPackageLibrary().addPackage(thePackage.getUnderlyingPreparedPackage());
+    }
+
+    /**
+     * Import a library package, changing the package name and/or version. Calling this method
+     * makes the supplied package available for reference in the <code>xsl:use-package</code>
+     * declaration of subsequent compilations performed using the <code>XsltCompiler</code>. The
+     * supplied package name and version are used in place of the name and version used in the XSLT
+     * source code. This provides a level of indirection: for example the same source package can
+     * be compiled twice, with different settings for the static parameter values, and the two
+     * different compiled versions can then be selected from xsl:use-package declarations. distinguishing
+     * them by the new package name and/or version.
+     *
+     * @param thePackage the package to be imported
+     * @param packageName the new package name to be used. If null, the original package name is used
+     *  unchanged
+     * @param version the new package version number to be used. If null, the original package version
+     *  number is used unchanged
+     * @throws SaxonApiException if the imported package was created under a different {@link Processor},
+     * or if the supplied version number is invalid
+     * @since 9.8
+     */
+
+    public void importPackage(XsltPackage thePackage, String packageName, String version) throws SaxonApiException {
+        try {
+            if (thePackage.getProcessor() != processor) {
+                throw new SaxonApiException("The imported package and the XsltCompiler must belong to the same Processor");
+            }
+            PackageDetails details = new PackageDetails();
+            if (packageName == null) {
+                packageName = thePackage.getName();
+            }
+            if (version == null) {
+                version = thePackage.getVersion();
+            }
+            details.nameAndVersion = new VersionedPackageName(packageName, version);
+            details.loadedPackage = thePackage.getUnderlyingPreparedPackage();
+            compilerInfo.getPackageLibrary().addPackage(details);
+        } catch (XPathException e) {
+            throw new SaxonApiException(e);
+        }
+    }
+
+    /**
+     * Import a named package, together with all the packages on which it depends, recursively.
+     * The package must be identified in the package library for this XsltCompiler, which defaults
+     * to the package library defined in the Configuration, typically set up by loading a configuration
+     * file.
+     * @param packageName the name of the required package. This is the name under which it is registered
+     *                    in the package library, which is not necessarily the same as the name appearing
+     *                    in the XSLT source code.
+     * @param versionRange the required version of the package, or range of versions, in the format
+     *                     of the package-version attribute of xsl:use-package.
+     * @return the best matching package if there is one, or null otherwise. The name of the package
+     * must match; if there are multiple versions, then the version chosen is based first on the
+     * priority attached to this package/version in the library, and if the priorities are equal (or
+     * there are no explicit priorities) then the one with highest version number is taken.
+     * @since 9.8
+     */
+
+    public XsltPackage obtainPackage(String packageName, String versionRange) throws SaxonApiException {
+        try {
+            PackageVersionRanges pvr = new PackageVersionRanges(versionRange);
+            PackageDetails details = getPackageLibrary().findPackage(packageName, pvr);
+            if (details != null) {
+                if (details.loadedPackage != null) {
+                    return new XsltPackage(processor, details.loadedPackage);
+                } else if (details.sourceLocation != null) {
+                    XsltPackage pack = compilePackage(details.sourceLocation);
+                    details.loadedPackage = pack.getUnderlyingPreparedPackage();
+                    return pack;
+                }
+            }
+            return null;
+        } catch (XPathException e) {
+            throw new SaxonApiException(e);
+        }
+    }
+
+    /**
+     * Import a package from the configuration file (or more generally, from the packageLibrary
+     * of this XsltCompiler) given an alias used to identify it
+     * @param alias the alias of the package/version in the configuration file
+     * @return the referenced package
+     * @throws SaxonApiException the package does not exist, or if compiling the package fails
+     */
+
+    public XsltPackage obtainPackageWithAlias(String alias) throws SaxonApiException {
+        PackageDetails details = getPackageLibrary().findDetailsForAlias(alias);
+        if (details == null) {
+            throw new SaxonApiException("No package with alias " + alias + " found in package library");
+        }
+        try {
+            StylesheetPackage pack = getPackageLibrary().obtainLoadedPackage(details, new ArrayList<VersionedPackageName>());
+            return new XsltPackage(processor, pack);
+        } catch (XPathException e) {
+            throw new SaxonApiException(e);
         }
-        compilerInfo.getPackageLibrary().addPackage(thePackage.getName(), thePackage.getUnderlyingPreparedPackage());
     }
 
 
@@ -838,9 +719,8 @@ public class XsltCompiler {
      * <li>{@link javax.xml.transform.sax.SAXSource}, allowing the stylesheet to be supplied as a stream
      * of SAX events from a SAX2-compliant XML parser (or any other source of SAX events)</li>
      * <li>{@link javax.xml.transform.dom.DOMSource}, allowing the stylesheet to be supplied as a
-     * DOM tree. This option is available only if saxon9-dom.jar is on the classpath.</li>
-     * <li>Document wrappers for XOM, JDOM, or DOM4J trees, provided the appropriate support libraries
-     * are on the classpath</li>
+     * DOM tree</li>
+     * <li>Document wrappers for XOM, JDOM, or DOM4J trees</li>
      * <li>A Saxon NodeInfo, representing the root of a tree in any of the native tree formats supported
      * by Saxon</li>
      * <li>An {@link XdmNode} representing the document node of the stylesheet module</li>
@@ -888,5 +768,35 @@ public class XsltCompiler {
         compilerInfo.setErrorListener(new ErrorGatherer(errorList));
     }
 
+    /**
+     * Say whether just-in-time compilation of template rules should be used.
+     * @param jit true if just-in-time compilation is to be enabled. With this option enabled,
+     *            static analysis of a template rule is deferred until the first time that the
+     *            template is matched. This can improve performance when many template
+     *            rules are rarely used during the course of a particular transformation; however,
+     *            it means that static errors in the stylesheet may go undetected.
+     */
+
+    public void setJustInTimeCompilation(boolean jit){
+        if (jit && !config.isLicensedFeature(Configuration.LicenseFeature.ENTERPRISE_XSLT)){
+            throw new UnsupportedOperationException("XSLT just-in-time compilation requires a Saxon-EE license");
+        }
+        compilerInfo.setJustInTimeCompilation(jit);
+    }
+
+    /**
+     * Ask whether just-in-time compilation of template rules should be used.
+     *
+     * @return true if just-in-time compilation is enabled. With this option enabled,
+     *            static analysis of a template rule is deferred until the first time that the
+     *            template is matched. This can improve performance when many template
+     *            rules are rarely used during the course of a particular transformation; however,
+     *            it means that static errors in the stylesheet may go undetected.
+     */
+
+    public boolean isJustInTimeCompilation(){
+        return compilerInfo.isJustInTimeCompilation();
+    }
+
 }
 
