@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -25,6 +25,7 @@ import net.sf.saxon.type.SchemaType;
 import net.sf.saxon.type.Untyped;
 
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.Iterator;
 import java.util.List;
 
@@ -89,11 +90,10 @@ public class LiteralResultElement extends StyleElement {
 
             for (int i = 0; i < num; i++) {
 
-                int anameCode = atts.getNameCode(i);
-                String attURI = namePool.getURI(anameCode);
+                int fp = atts.getFingerprint(i);
+                String attURI = namePool.getURI(fp);
 
                 if (attURI.equals(NamespaceConstant.XSLT)) {
-                    int fp = anameCode & NamePool.FP_MASK;
 
                     if (fp == StandardNames.XSL_USE_ATTRIBUTE_SETS) {
                         // deal with this later
@@ -101,6 +101,8 @@ public class LiteralResultElement extends StyleElement {
                         // already dealt with
                     } else if (fp == StandardNames.XSL_DEFAULT_MODE) {
                         // already dealt with
+                    } else if (fp == StandardNames.XSL_DEFAULT_VALIDATION) {
+                        // already dealt with
                     } else if (fp == StandardNames.XSL_EXTENSION_ELEMENT_PREFIXES) {
                         // already dealt with
                     } else if (fp == StandardNames.XSL_EXCLUDE_RESULT_PREFIXES) {
@@ -119,11 +121,13 @@ public class LiteralResultElement extends StyleElement {
                         // deal with this later
                     } else if (fp == StandardNames.XSL_INHERIT_NAMESPACES) {
                         inheritNamespaces = processBooleanAttribute("xsl:inherit-namespaces", atts.getValue(i));
+                    } else if (forwardsCompatibleModeIsEnabled()) {
+                        // the attribute is ignored
                     } else {
-                        compileError("Unknown XSL attribute " + namePool.getDisplayName(anameCode), "XTSE0805");
+                        compileError("Unknown XSLT attribute " + atts.getNodeName(i).getDisplayName(), "XTSE0805");
                     }
                 } else {
-                    attributeNames[numberOfAttributes] = new FingerprintedQName(atts.getPrefix(i), atts.getURI(i), atts.getLocalName(i), anameCode);
+                    attributeNames[numberOfAttributes] = new FingerprintedQName(atts.getPrefix(i), atts.getURI(i), atts.getLocalName(i), fp);
                     Expression exp = makeAttributeValueTemplate(atts.getValue(i), i);
                     attributeValues[numberOfAttributes] = exp;
                     numberOfAttributes++;
@@ -134,14 +138,8 @@ public class LiteralResultElement extends StyleElement {
             // attributes that weren't copied into the arrays.
 
             if (numberOfAttributes < attributeNames.length) {
-
-                NodeName[] attributeNames2 = new NodeName[numberOfAttributes];
-                System.arraycopy(attributeNames, 0, attributeNames2, 0, numberOfAttributes);
-                attributeNames = attributeNames2;
-
-                Expression[] attributeValues2 = new Expression[numberOfAttributes];
-                System.arraycopy(attributeValues, 0, attributeValues2, 0, numberOfAttributes);
-                attributeValues = attributeValues2;
+                attributeNames = Arrays.copyOf(attributeNames, numberOfAttributes);
+                attributeValues = Arrays.copyOf(attributeValues, numberOfAttributes);
             }
         }
         resultNodeName = getNodeName();
@@ -315,6 +313,7 @@ public class LiteralResultElement extends StyleElement {
                         attributeNames[i],
                         Validation.STRIP,
                         null);
+                att.setRetainedStaticContext(makeRetainedStaticContext());
                 att.setSelect(attributeValues[i]);
                 att.setLocation(allocateLocation());
                 Expression exp = att;
@@ -376,7 +375,7 @@ public class LiteralResultElement extends StyleElement {
                 message = "The supplied file does not appear to be a stylesheet";
             }
             XPathException err = new XPathException(message);
-            err.setLocation(getLocation());
+            err.setLocation(allocateLocation());
             err.setErrorCode("XTSE0150");
             err.setIsStaticError(true);
             //noinspection EmptyCatchBlock
@@ -393,7 +392,7 @@ public class LiteralResultElement extends StyleElement {
             XPathException err = new XPathException("Simplified stylesheet: xsl:version attribute is missing");
             err.setErrorCode("XTSE0150");
             err.setIsStaticError(true);
-            err.setLocation(getLocation());
+            err.setLocation(allocateLocation());
             //noinspection EmptyCatchBlock
             compileError(err);
             throw err;
@@ -410,13 +409,13 @@ public class LiteralResultElement extends StyleElement {
 
             int st = StandardNames.XSL_STYLESHEET;
             final Location loc = ExplicitLocation.UNKNOWN_LOCATION;
-            builder.startElement(new CodedName(st, getNamePool()), Untyped.getInstance(), loc, 0);
+            builder.startElement(new CodedName(st, "xsl", getNamePool()), Untyped.getInstance(), loc, 0);
             builder.namespace(new NamespaceBinding("xsl", NamespaceConstant.XSLT), 0);
             builder.attribute(new NoNamespaceName("version"), BuiltInAtomicType.UNTYPED_ATOMIC, version, loc, 0);
             builder.startContent();
 
             int te = StandardNames.XSL_TEMPLATE;
-            builder.startElement(new CodedName(te, getNamePool()), Untyped.getInstance(), loc, 0);
+            builder.startElement(new CodedName(te, "xsl", getNamePool()), Untyped.getInstance(), loc, 0);
             builder.attribute(new NoNamespaceName("match"), BuiltInAtomicType.UNTYPED_ATOMIC, "/", loc, 0);
             builder.startContent();
 
@@ -432,7 +431,7 @@ public class LiteralResultElement extends StyleElement {
             return newRoot;
         } catch (XPathException err) {
             //TransformerConfigurationException e = new TransformerConfigurationException(err);
-            err.setLocation(getLocation());
+            err.setLocation(allocateLocation());
             throw err;
         }
 
