@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is “Incompatible With Secondary Licenses”, as defined by the Mozilla Public License, v. 2.0.
@@ -8,10 +8,12 @@
 package net.sf.saxon.testdriver;
 
 import net.sf.saxon.Configuration;
+import net.sf.saxon.dom.DocumentWrapper;
 import net.sf.saxon.expr.XPathContext;
 import net.sf.saxon.expr.sort.SimpleCollation;
 import net.sf.saxon.lib.*;
 import net.sf.saxon.om.*;
+import net.sf.saxon.query.QueryResult;
 import net.sf.saxon.resource.AbstractResourceCollection;
 import net.sf.saxon.resource.XmlResource;
 import net.sf.saxon.s9api.*;
@@ -19,7 +21,12 @@ import net.sf.saxon.trans.DecimalFormatManager;
 import net.sf.saxon.trans.DecimalSymbols;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.util.FastStringBuffer;
+import org.w3c.dom.Document;
+import org.xml.sax.InputSource;
+import org.xml.sax.SAXException;
 
+import javax.xml.parsers.DocumentBuilderFactory;
+import javax.xml.parsers.ParserConfigurationException;
 import javax.xml.transform.Source;
 import javax.xml.transform.TransformerException;
 import javax.xml.transform.URIResolver;
@@ -68,6 +75,7 @@ public class Environment implements URIResolver {
     public boolean outputTree = true;
     public boolean outputSerialize = false;
     public Map<String, CatalogResource> catalogResources = new HashMap<String, CatalogResource>();
+    public Map<String, List<String>> jsCollections;
 
 
     /**
@@ -90,11 +98,15 @@ public class Environment implements URIResolver {
         if (testDriver.runWithJS) {
             environment.xsltCompiler.setTargetEdition("JS");
         }
+        if (environment.processor.getSaxonEdition().equals("EE")) {
+            environment.xsltCompiler.setJustInTimeCompilation(testDriver.jitFlag);
+        }
+        environment.xsltCompiler.setRelocatable(testDriver.relocatable);
         if (spec == Spec.XT30) {
-            environment.xsltCompiler.setXsltLanguageVersion("3.0");
+            //environment.xsltCompiler.setXsltLanguageVersion("3.0");
             environment.xpathCompiler.setLanguageVersion("3.1");
         } else if (spec == Spec.XT20) {
-            environment.xsltCompiler.setXsltLanguageVersion("2.0");
+            throw new AssertionError("Unsupported spec: XSLT 2.0");
         }
         if (unfolded) {
             testDriver.addInjection(environment.xqueryCompiler);
@@ -103,6 +115,7 @@ public class Environment implements URIResolver {
             environment.xsltCompiler.setCompileWithTracing(true);
         }
         environment.processor.getUnderlyingConfiguration().setDefaultCollection(null);
+        environment.processor.setConfigurationProperty(FeatureKeys.STABLE_UNPARSED_TEXT, true);
         return environment;
     }
 
@@ -122,7 +135,7 @@ public class Environment implements URIResolver {
             throws SaxonApiException {
         Environment environment = new Environment();
         String name = ((XdmNode) env).getAttributeValue(new QName("name"));
-        if (name != null) {
+        if (name != null && !driver.quiet) {
             System.err.println("Loading environment " + name);
         }
         environment.processor = new Processor(!driver.driverProc.getSaxonEdition().equals("HE"));
@@ -144,13 +157,16 @@ public class Environment implements URIResolver {
         environment.xpathCompiler.setBaseURI(((XdmNode) env).getBaseURI());
         environment.xqueryCompiler = environment.processor.newXQueryCompiler();
         environment.xqueryCompiler.setBaseURI(((XdmNode) env).getBaseURI());
-        environment.xqueryCompiler.setLanguageVersion(driver.spec.version.equals("2.0") ? "1.0" : driver.spec.version);
         if (driver.spec.shortSpecName.equals("XT")) {
             environment.xsltCompiler = environment.processor.newXsltCompiler();
-            environment.xsltCompiler.setXsltLanguageVersion(driver.spec.version);
+            //environment.xsltCompiler.setXsltLanguageVersion(driver.spec.version);
             if (driver.runWithJS) {
                 environment.xsltCompiler.setTargetEdition("JS");
             }
+            if (environment.processor.getSaxonEdition().equals("EE")) {
+                environment.xsltCompiler.setJustInTimeCompilation(driver.jitFlag);
+            }
+            environment.xsltCompiler.setRelocatable(driver.relocatable);
         }
 
         if (driver.unfolded) {
@@ -199,7 +215,7 @@ public class Environment implements URIResolver {
         // load the requested schema documents
 
         SchemaManager manager = environment.processor.getSchemaManager();
-        boolean validateSources = loadSchemaDocuments(xpc, env, environment, manager);
+        boolean validateSources = loadSchemaDocuments(driver, xpc, env, environment, manager);
 
         // load the requested source documents
 
@@ -234,18 +250,18 @@ public class Environment implements URIResolver {
         // compile any stylesheet packages defined as part of the environment
         // Support this only in EE - an unusable environment in PE/HE
         for (XdmItem stylesheet : xpc.evaluate("package[@role='secondary']", env)) {
-            if (!"EE".equals(environment.processor.getSaxonEdition())) {
-                environment.usable = false;
-                break;
-            }
+//            if (!"EE".equals(environment.processor.getSaxonEdition())) {
+//                environment.usable = false;
+//                break;
+//            }
             String fileName = ((XdmNode) stylesheet).getAttributeValue(new QName("file"));
             Source styleSource = new StreamSource(((XdmNode) env).getBaseURI().resolve(fileName).toString());
             try {
                 XsltPackage pkg = environment.xsltCompiler.compilePackage(styleSource);
                 environment.xsltCompiler.importPackage(pkg);
             } catch (SaxonApiException e) {
-                //e.printStackTrace();
-                driver.println("**** failure while compiling environment-defined stylesheet package " + fileName + ": " + e.getMessage());
+                e.printStackTrace();
+                driver.println("**** failure while compiling environment-defined stylesheet package " + fileName);
                 environment.failedToBuild = true;
                 environment.usable = false;
             } catch (Exception e) {
@@ -262,16 +278,25 @@ public class Environment implements URIResolver {
             String fileName = ((XdmNode) stylesheet).getAttributeValue(new QName("file"));
             Source styleSource = new StreamSource(((XdmNode) env).getBaseURI().resolve(fileName).toString());
             try {
-                environment.xsltExecutable = environment.xsltCompiler.compile(styleSource);
+                if (driver.export) {
+                    if (driver.runWithJS) {
+                        String sourceFile = ((XdmNode) env).getBaseURI().resolve(fileName).toString();
+                        environment.exportedStylesheet = driver.exportStylesheet(environment.xsltCompiler, sourceFile);
+                    } else {
+                        File exportFile = new File(driver.resultsDir + "/export/" + name + ".sef");
+                        XsltPackage compiledPack = environment.xsltCompiler.compilePackage(styleSource);
+                        compiledPack.save(exportFile);
+                        environment.xsltExecutable = environment.xsltCompiler.loadExecutablePackage(exportFile.toURI());
+                    }
+                } else {
+                    environment.xsltExecutable = environment.xsltCompiler.compile(styleSource);
+                }
             } catch (SaxonApiException e) {
+                e.printStackTrace();
                 driver.println("**** failure while compiling environment-defined stylesheet " + fileName);
                 environment.failedToBuild = true;
                 environment.usable = false;
             }
-            if (driver.runWithJS) {
-                String sourceFile = ((XdmNode) env).getBaseURI().resolve(fileName).toString();
-                environment.exportedStylesheet = driver.exportStylesheet(environment.xsltCompiler, sourceFile);
-            }
         }
 
 
@@ -280,18 +305,23 @@ public class Environment implements URIResolver {
 
     private static void configureByteCode(TestDriver driver, Environment environment, int generate) {
         if (environment.processor.getSaxonEdition().equals("EE")) {
-            if (generate == 1) {
+            if (generate == 0) {
                 environment.processor.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, "true");
                 environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "false");
-            } else if (generate == 2) {
-                environment.processor.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, "true");
-                environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "true");
-                environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE_DIR, "debugByteCode");
+                environment.processor.setConfigurationProperty(FeatureKeys.MAX_COMPILED_CLASSES, 1000000);
+            } else  if (generate > 0) {
+                environment.processor.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, ""+generate);
+                environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "false");
+                environment.processor.setConfigurationProperty(FeatureKeys.MAX_COMPILED_CLASSES, 1000000);
             } else {
                 environment.processor.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, "false");
                 environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "false");
                 //environment.processor.setConfigurationProperty(FeatureKeys.ALLOW_MULTITHREADING, "false"); // TODO should have a separate option for this
             }
+            if (generate >= 0 && driver.isDebugByteCode()) {
+                environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "true");
+                environment.processor.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE_DIR, "debugByteCode");
+            }
         }
     }
 
@@ -308,7 +338,7 @@ public class Environment implements URIResolver {
             if (uri == null || "#UNDEFINED".equals(uri)) {
                 /*environment.xpathCompiler.setBaseURI(null);
                 environment.xqueryCompiler.setBaseURI(null); */
-                driver.println("**** Error: The BaseURI null and #UNDEFINED is not supported");
+                driver.println("**** Error: The BaseURI values null and #UNDEFINED are not supported in this test driver");
             } else {
                 try {
                     environment.xpathCompiler.setBaseURI(new URI(uri));
@@ -339,9 +369,7 @@ public class Environment implements URIResolver {
             if (defaultAtt != null && (defaultAtt.trim().equals("true") || defaultAtt.trim().equals("1"))) {
                 environment.xpathCompiler.declareDefaultCollation(uri);
                 environment.xqueryCompiler.declareDefaultCollation(uri);
-                if (environment.xsltCompiler != null) {
-                    environment.xsltCompiler.declareDefaultCollation(uri);
-                }
+                environment.xsltCompiler.declareDefaultCollation(uri);
             }
         }
     }
@@ -367,18 +395,25 @@ public class Environment implements URIResolver {
         }
     }
 
-    private static boolean loadSchemaDocuments(XPathCompiler xpc, XdmItem env, Environment environment, SchemaManager manager) throws SaxonApiException {
+    private static boolean loadSchemaDocuments(TestDriver driver, XPathCompiler xpc, XdmItem env, Environment environment, SchemaManager manager) throws SaxonApiException {
         boolean validateSources = false;
         for (XdmItem schema : xpc.evaluate("schema", env)) {
             String role = ((XdmNode) schema).getAttributeValue(new QName("role"));
             String xsdVersion = ((XdmNode) schema).getAttributeValue(new QName("xsd-version"));
             if (manager == null) {
-                System.err.println("*** Processor is not schema aware");
+                driver.println("*** Processor is not schema aware");
+                environment.usable = false;
+                return false;
+            }
+            if (!driver.treeModel.isSchemaAware()) {
+                driver.println("*** Tree model is not schema aware");
                 environment.usable = false;
                 return false;
             }
             if (xsdVersion != null) {
                 manager.setXsdVersion(xsdVersion);
+            } else {
+                manager.setXsdVersion("1.0");
             }
             if (!"secondary".equals(role)) {
                 String href = ((XdmNode) schema).getAttributeValue(new QName("file"));
@@ -388,7 +423,7 @@ public class Environment implements URIResolver {
                         Source[] sources = manager.getSchemaURIResolver().resolve(ns, null, new String[0]);
                         manager.load(sources[0]);
                     } catch (Exception e) {
-                        System.err.println("*** Failed to load schema by URI: " + ns + " - " + e.getMessage());
+                        driver.println("*** Failed to load schema by URI: " + ns + " - " + e.getMessage());
                         e.printStackTrace();
                     }
                 } else {
@@ -396,10 +431,10 @@ public class Environment implements URIResolver {
                     try {
                         manager.load(new StreamSource(file));
                     } catch (SaxonApiException err) {
-                        System.err.println("*** Failed to load schema: " + href + " - " + err.getMessage());
+                        driver.println("*** Failed to load schema: " + href + " - " + err.getMessage());
                     } catch (NullPointerException err) {
                         err.printStackTrace();
-                        System.err.println("*** NPE: Failed to load schema: " + href + " - " + err.getMessage());
+                        driver.println("*** NPE: Failed to load schema: " + href + " - " + err.getMessage());
                     }
                 }
                 xpc.importSchemaNamespace(ns);
@@ -428,12 +463,9 @@ public class Environment implements URIResolver {
                 value = sourceDoc;
             } else {
                 String select = ((XdmNode) param).getAttributeValue(new QName("select"));
-                XPathSelector xpe = environment.xpathCompiler.compile(select).load();
-                for (Map.Entry<QName, XdmValue> e : environment.params.entrySet()) {
-                    xpe.setVariable(e.getKey(), e.getValue());
-                }
-                value = xpe.evaluate();
+                value = xpc.evaluate(select, null);
             }
+            boolean isStatic = "true".equals(((XdmNode) param).getAttributeValue(new QName("static")));
             QName varQName;
             int colon = varName.indexOf(':');
             if (colon >= 0) {
@@ -451,6 +483,10 @@ public class Environment implements URIResolver {
             } else {
                 environment.paramDeclarations.append("declare variable $" + varName + " external; ");
             }
+            if (isStatic) {
+                environment.xsltCompiler.setParameter(varQName, value);
+                System.err.println("set " + varQName + " = " + value);
+            }
         }
     }
 
@@ -790,6 +826,30 @@ public class Environment implements URIResolver {
             });
 
         }
+
+        if (driver.runWithJS && collections != null) {
+            environment.jsCollections = new HashMap<String, List<String>>();
+            for (Map.Entry<String, ResourceCollection> entry : collections.entrySet()) {
+                String uri = entry.getKey();
+                List<String> lexicalContents = new ArrayList<String>();
+                ResourceCollection contents = entry.getValue();
+                XPathContext context = environment.processor.getUnderlyingConfiguration().getConversionContext();
+                try {
+                    Iterator<? extends Resource> iter = contents.getResources (context);
+                    while (iter.hasNext()) {
+                        Resource resource = iter.next();
+                        Item item = resource.getItem(context);
+                        if (item instanceof NodeInfo) {
+                            String lexicalXml = QueryResult.serialize((NodeInfo)item);
+                            lexicalContents.add(lexicalXml);
+                        }
+                    }
+                } catch (XPathException e) {
+                    e.printStackTrace();
+                }
+                environment.jsCollections.put(uri, lexicalContents);
+            }
+        }
     }
 
     private static void loadSourceDocuments(TestDriver driver, XPathCompiler xpc, XdmItem env, Environment environment, DocumentBuilder builder, SchemaManager manager, boolean validateSources) throws SaxonApiException {
@@ -797,9 +857,12 @@ public class Environment implements URIResolver {
 
             CatalogResource res = new CatalogResource();
             String rawUri = ((XdmNode) source).getAttributeValue(new QName("uri"));
+            String role = ((XdmNode) source).getAttributeValue(new QName("role"));
             if (rawUri != null) {
                 res.uri = ((XdmNode) env).getBaseURI().resolve(rawUri).toString();
                 environment.catalogResources.put(res.uri, res);
+            } else if (".".equals(role)) {
+                environment.catalogResources.put("", res);
             }
             res.mediaType = ((XdmNode) source).getAttributeValue(new QName("media-type"));
             if (res.mediaType != null && "application/xml".equals(res.mediaType)) {
@@ -830,9 +893,26 @@ public class Environment implements URIResolver {
                 }
             }
 
-            String role = ((XdmNode) source).getAttributeValue(new QName("role"));
             String href = ((XdmNode) source).getAttributeValue(new QName("file"));
-            if ("true".equals(streaming) || driver.runWithJS) {
+            String select = ((XdmNode) source).getAttributeValue(new QName("select"));
+            if ("true".equals(streaming)) {
+                if (".".equals(role)) {
+                    if (href == null) {
+                        environment.streamedContent = xpc.evaluate("string(content)", source).toString();
+                    } else {
+                        environment.streamedFile = new File(((XdmNode) env).getBaseURI().resolve(href));
+                        res.file = environment.streamedFile;
+                    }
+                } else {
+                    File file = new File(((XdmNode) env).getBaseURI().resolve(href));
+                    environment.streamedSecondaryDocs.put(file.toURI().toString(), validation);
+                    res.file = file;
+                }
+            } else if (driver.runWithJS || driver.isAltova) {
+                if (select != null) {
+                    driver.println("source select=path not supported by this test driver");
+                    environment.usable = false;
+                }
                 if (".".equals(role)) {
                     if (href == null) {
                         environment.streamedContent = xpc.evaluate("string(content)", source).toString();
@@ -847,9 +927,13 @@ public class Environment implements URIResolver {
                 }
             } else {
                 Source ss;
+                File file = null;
+                FileInputStream fileInputStream = null;
+                StringReader stringReader = null;
+                String baseUri = null;
                 if (href != null) {
                     URI fileLoc = ((XdmNode) env).getBaseURI().resolve(href);
-                    File file = null;
+                    baseUri = fileLoc.toString();
                     if (fileLoc.getScheme().equals("file")) {
                         file = new File(((XdmNode) env).getBaseURI().resolve(href));
                         if (res.uri == null) {
@@ -858,38 +942,77 @@ public class Environment implements URIResolver {
                         res.file = file;
                     }
                     try {
-                        ss = (file == null ? new StreamSource(fileLoc.toString())
-                                : new StreamSource(new FileInputStream(file), res.uri));
+                        if (file == null) {
+                            ss = new StreamSource(fileLoc.toString());
+                        } else {
+                            fileInputStream = new FileInputStream(file);
+                            ss = new StreamSource(fileInputStream, res.uri);
+                        }
                     } catch (FileNotFoundException e) {
                         driver.println("*** failed to find source document " + href);
                         continue;
                     }
                 } else {
                     // content is inline in the catalog
-                    String baseUri = res.uri;
+                    baseUri = res.uri;
                     if (res.uri == null) {
                         baseUri = ((XdmNode) env).getBaseURI().toString();
                     }
                     String content = xpc.evaluate("string(content)", source).toString();
-                    ss = new StreamSource(new StringReader(content), baseUri);
+                    stringReader = new StringReader(content);
+                    ss = new StreamSource(stringReader, baseUri);
                 }
 
+                if (selectedTreeModel.getName().equals("DOM") || selectedTreeModel.getName().equals("DOMINO")) {
+                    // create the DOM directly. Using a SAX parser plus document builder loses
+                    // entities and doctype nodes, which will be encountered in real life
+                    Document doc = null;
+                    try {
+                        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
+                        factory.setNamespaceAware(true);
+                        javax.xml.parsers.DocumentBuilder db = factory.newDocumentBuilder();
+                        if (file != null) {
+                            doc = db.parse(file);
+                        } else if (fileInputStream != null) {
+                            InputSource is = new InputSource(fileInputStream);
+                            is.setSystemId(baseUri);
+                            doc = db.parse(is);
+                        } else if (stringReader != null) {
+                            InputSource is = new InputSource(stringReader);
+                            is.setSystemId(baseUri);
+                            doc = db.parse(is);
+                        }
+                    } catch (ParserConfigurationException e) {
+                        e.printStackTrace();
+                    } catch (IOException e) {
+                        e.printStackTrace();
+                    } catch (SAXException e) {
+                        e.printStackTrace();
+                    }
 
-                try {
-                    builder.setLineNumbering(true);
-                    res.doc = builder.build(ss);
+                    if (selectedTreeModel.getName().equals("DOM")) {
+                        NodeInfo root = new DocumentWrapper(doc, res.uri, environment.processor.getUnderlyingConfiguration()).getRootNode();
+                        res.doc = (XdmNode)XdmValue.wrap(root);
+                    } else {
+                        res.doc = driver.makeDominoTree(doc, environment.processor.getUnderlyingConfiguration(), res.uri);
+                    }
                     environment.sourceDocs.put(res.uri, res.doc);
-                    environment.sourceDocs.put(rawUri, res.doc);
+                } else {
+                    try {
+                        builder.setLineNumbering(true);
+                        res.doc = builder.build(ss);
+                        environment.sourceDocs.put(res.uri, res.doc);
+                        environment.sourceDocs.put(rawUri, res.doc);
 
-                } catch (SaxonApiException e) {
-                    //e.printStackTrace();
-                    driver.println("*** failed to build source document " + href + " - " + e.getMessage());
+                    } catch (SaxonApiException e) {
+                        //e.printStackTrace();
+                        driver.println("*** failed to build source document " + href + " - " + e.getMessage());
+                    }
                 }
 
                 builder.setTreeModel(selectedTreeModel);
 
                 XdmItem selectedItem = res.doc;
-                String select = ((XdmNode) source).getAttributeValue(new QName("select"));
                 if (select != null) {
                     XPathSelector selector = environment.xpathCompiler.compile(select).load();
                     selector.setContextItem(selectedItem);
@@ -960,6 +1083,10 @@ public class Environment implements URIResolver {
         }
     }
 
+    public CatalogResource getContextDocument() {
+        return catalogResources.get("");
+    }
+
 
     public abstract static class ResetAction {
         public abstract void reset(Environment env);
