@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -148,7 +148,7 @@ public class QT3TestDriverHE extends TestDriver {
                 return false;
             }
             if (value.contains("1.1") && !inverse) {
-                if (treeModel.getName().equals("JDOM") || treeModel.getName().equals("JDOM2") || treeModel.getName().equals("XOM")) {
+                if (treeModel.getName().equals("JDOM") || treeModel.getName().equals("JDOM2") || treeModel.getName().equals("XOM") || treeModel.getName().equals("Axiom")) {
                     return false;
                 }
                 if (env != null) {
@@ -215,15 +215,15 @@ public class QT3TestDriverHE extends TestDriver {
         } else if ("unicode-normalization-form".equals(type)) {
             return value.equalsIgnoreCase("FULLY-NORMALIZED") ? inverse : !inverse;
         } else if ("feature".equals(type)) {
+            String edition = env.processor.getSaxonEdition();
             if ("namespace-axis".equals(value)) {
                 return !inverse;
             } else if ("higherOrderFunctions".equals(value)) {
-                String edition = env.processor.getSaxonEdition();
                 return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else if ("schemaImport".equals(value) || "schemaValidation".equals(value) || "schemaAware".equals(value)) {
                 return makeSchemaAware(env, inverse);
             } else if ("xpath-1.0-compatibility".equals(value)) {
-                if (env != null) {
+                if (env != null && !env.processor.getSaxonEdition().equals("HE")) {
                     env.xpathCompiler.setBackwardsCompatible(!inverse);
                     return true;
                 } else {
@@ -231,10 +231,11 @@ public class QT3TestDriverHE extends TestDriver {
                 }
             } else if ("staticTyping".equals(value)) {
                 return inverse;
+            } else if ("remote_http".equals(value)) {
+                return !inverse;
             } else if ("moduleImport".equals(value)) {
                 return !inverse;
             } else if ("schema-location-hint".equals(value)) {
-                String edition = env.processor.getUnderlyingConfiguration().getEditionCode();
                 return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else if ("infoset-dtd".equals(value)) {
                 if (treeModel == TreeModel.TINY_TREE || treeModel == TreeModel.LINKED_TREE || treeModel == TreeModel.TINY_TREE_CONDENSED) {
@@ -251,16 +252,15 @@ public class QT3TestDriverHE extends TestDriver {
             } else if ("fn-transform-XSLT".equals(value)) {
                 return !inverse;
             } else if ("fn-transform-XSLT30".equals(value)) {
-                String edition = env.processor.getSaxonEdition();
                 return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else if ("fn-load-xquery-module".equals(value)) {
-                String edition = env.processor.getSaxonEdition();
                 return edition.equals("EE") ^ inverse;
             } else if ("fn-format-integer-CLDR".equals(value)) {
-                String edition = env.processor.getSaxonEdition();
                 return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else if ("simple-uca-fallback".equals(value)) {
                 return !inverse;
+            } else if ("advanced-uca-fallback".equals(value)) {
+                return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else {
                 println("**** feature = " + value + "  ????");
                 return false;
@@ -315,12 +315,13 @@ public class QT3TestDriverHE extends TestDriver {
 
         Environment env = getEnvironment(testCase, xpc);
         if (env == null) {
+            writeTestcaseElement(testCaseName, "notRun", "test catalog error");
             notrun++;
             return;
         }
         if (env.failedToBuild) {
             writeTestcaseElement(testCaseName, "fail", "unable to build environment");
-            failures++;
+            noteFailure(testSetName, testCaseName);
             return;
         }
         if (!env.usable) {
@@ -386,8 +387,11 @@ public class QT3TestDriverHE extends TestDriver {
                 specOpt = Spec.XP30;
             } */
             if (!ensureDependencySatisfied((XdmNode) dependency, env)) {
-                println("*** Dependency not satisfied: " + ((XdmNode) dependency).getAttributeValue(new QName("type")));
-                writeTestcaseElement(testCaseName, "n/a", "Dependency not satisfied");
+                if ("false".equals(((XdmNode) dependency).getAttributeValue(new QName("", "satisfied")))) {
+                    type = "!" + type;
+                }
+                println("*** Dependency not satisfied: " + type + ":" + value);
+                writeTestcaseElement(testCaseName, "n/a", "dependency not satisfied: " + type + ":" + value);
                 //run = false;
                 notrun++;
                 return;
@@ -402,7 +406,7 @@ public class QT3TestDriverHE extends TestDriver {
             XdmItem config = xpath.evaluateSingle("configuration", exceptionElement);
 
             String runAtt = exceptionElement.getAttributeValue(new QName("run"));
-            String reasonMsg = exceptionElement.getAttributeValue(new QName("reason"));
+            String reasonMsg = exceptionElement.axisIterator(Axis.CHILD, new QName("reason")).next().getStringValue();
             if (reasonMsg == null) {
                 reasonMsg = "no reason given";
             }
@@ -479,7 +483,7 @@ public class QT3TestDriverHE extends TestDriver {
                     } catch (Exception err) {
                         println(err.getMessage());
                         err.printStackTrace();
-                        writeTestcaseElement(testCaseName, "fail", "*** crashed: + err.getMessage()");
+                        writeTestcaseElement(testCaseName, "fail", "*** crashed: " + err.getMessage());
                         //notrun++;
                         //return;
                     }
@@ -519,18 +523,24 @@ public class QT3TestDriverHE extends TestDriver {
                     try {
                         testXqc.getUnderlyingStaticContext().setModuleLocation(
                             new ExplicitLocation(testXqc.getBaseURI().toString(), testCase.getLineNumber(), -1));
-                        XQueryExecutable q = testXqc.compile(exp);
+                        XQueryExecutable q = null;
+                        try {
+                            q = testXqc.compile(exp);
+                        } catch (SaxonApiException e) {
+                            println("Static error in query");
+                            throw e;
+                        }
                         if (optimization != null) {
                             // Test whether required optimizations have been performed
                             XdmDestination expDest = new XdmDestination();
                             Configuration config = driverProc.getUnderlyingConfiguration();
-                            ExpressionPresenter presenter = new ExpressionPresenter(config, expDest.getReceiver(config));
                             try {
+                                ExpressionPresenter presenter = new ExpressionPresenter(config, expDest.getReceiver(config));
                                 q.getUnderlyingCompiledQuery().explain(presenter);
+                                presenter.close();
                             } catch (XPathException e) {
                                 e.printStackTrace();
                             }
-                            presenter.close();
                             XdmNode explanation = expDest.getXdmNode();
                             XdmItem optResult = xpc.evaluateSingle(optimization.getAttributeValue(new QName("assert")), explanation);
                             if (((XdmAtomicValue) optResult).getBooleanValue()) {
@@ -539,7 +549,7 @@ public class QT3TestDriverHE extends TestDriver {
                                 println("Failed optimization test");
                                 driverProc.writeXdmValue(explanation, driverProc.newSerializer(System.err));
                                 writeTestcaseElement(testCaseName, "fail", "Failed optimization assertions");
-                                failures++;
+                                noteFailure(testSetName, testCaseName);
                                 return;
                             }
 
@@ -604,7 +614,7 @@ public class QT3TestDriverHE extends TestDriver {
             if (assertion == null) {
                 println("*** No assertions found for test case " + testCaseName);
                 writeTestcaseElement(testCaseName, "disputed", "No assertions in test case");
-                failures++;
+                noteFailure(testSetName, testCaseName);
                 return;
             }
             XPathCompiler assertXpc = env.processor.newXPathCompiler();
@@ -614,7 +624,7 @@ public class QT3TestDriverHE extends TestDriver {
             assertXpc.declareNamespace("math", NamespaceConstant.MATH);
             assertXpc.declareNamespace("map", NamespaceConstant.MAP_FUNCTIONS);
             assertXpc.declareNamespace("array", NamespaceConstant.ARRAY_FUNCTIONS);
-            assertXpc.declareNamespace("j", NamespaceConstant.JSON);
+            assertXpc.declareNamespace("j", NamespaceConstant.FN);
             assertXpc.declareVariable(new QName("result"));
             assertXpc.setBaseURI(assertion.getBaseURI());
 
@@ -640,13 +650,13 @@ public class QT3TestDriverHE extends TestDriver {
                         println("*** TEST WRONG ERRORCODE: result " + outcome.getException().getErrorCode() +
                                 " Expected error:" + expectedError.getStringValue());
                     } else {
-                        failures++;
+                        noteFailure(testSetName, testCaseName);
                         writeTestcaseElement(testCaseName, "fail", "Expected success, got " + outcome.getException().getErrorCode());
                         println("*** TEST-FAILURE: result " + outcome.getException().getErrorCode() +
                                 " Expected success.");
                     }
                 } else {
-                    failures++;
+                    noteFailure(testSetName, testCaseName);
                     if (expectedError == null){
 //                      if (debug) {
 //                            outcome.getException().printStackTrace(System.out);
@@ -695,15 +705,6 @@ public class QT3TestDriverHE extends TestDriver {
         return applicable;
     }
 
-    private void copySchemaNamespaces(Environment env, XPathCompiler testXpc) {
-        Configuration config = env.xpathCompiler.getProcessor().getUnderlyingConfiguration();
-        //if (config instanceof EnterpriseConfiguration) {
-        for (String s : config.getImportedNamespaces()) {
-            testXpc.importSchemaNamespace(s);
-        }
-        //}
-    }
-
     private String truncate(String in) {
         if (in.length() > 80) {
             return in.substring(0, 80) + "...";
@@ -719,6 +720,17 @@ public class QT3TestDriverHE extends TestDriver {
 
     protected void writeResultFilePostamble() throws XMLStreamException {
         resultsDoc.writeResultFilePostamble();
+        if (requestedTestSet != null) {
+            System.err.println("Failing tests: ");
+            for (Map.Entry<String, Integer> entry : failSummary.entrySet()) {
+                System.err.println("  " + entry.getKey());
+            }
+        } else {
+            System.err.println("Failures by Test Set: ");
+            for (Map.Entry<String, Integer> entry : failSummary.entrySet()) {
+                System.err.println("  " + entry.getKey() + " : " + entry.getValue());
+            }
+        }
     }
 
 
@@ -815,7 +827,6 @@ public class QT3TestDriverHE extends TestDriver {
 
     private void buildDependencyMap(Processor processor, XdmNode catalog, Environment env) throws SaxonApiException {
         XQueryCompiler xqCompiler = processor.newXQueryCompiler();
-        xqCompiler.setLanguageVersion("3.0");
         xqCompiler.setBaseURI(new File(System.getProperty("user.dir")).toURI());
         XQueryEvaluator eval = xqCompiler.compile(
                 "        declare namespace fots = \"http://www.w3.org/2010/09/qt-fots-catalog\";\n" +
