@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is “Incompatible With Secondary Licenses”, as defined by the Mozilla Public License, v. 2.0.
@@ -8,24 +8,24 @@
 package net.sf.saxon.testdriver;
 
 
+import net.sf.saxon.Configuration;
 import net.sf.saxon.Version;
 import net.sf.saxon.dom.DOMObjectModel;
 import net.sf.saxon.lib.FeatureKeys;
 import net.sf.saxon.om.TreeModel;
 import net.sf.saxon.s9api.*;
+import org.w3c.dom.Document;
 
 import javax.xml.stream.XMLStreamException;
 import java.io.File;
 import java.io.IOException;
-import java.util.HashMap;
-import java.util.HashSet;
-import java.util.Map;
-import java.util.Set;
+import java.util.*;
 import java.util.regex.Pattern;
 
 public abstract class TestDriver {
 
     protected String catalogFileName;
+    protected String exceptionsFileName = "exceptions.xml";
     protected String resultsDir = null;
     protected TestReport resultsDoc;
     protected int successes = 0;
@@ -37,18 +37,20 @@ public abstract class TestDriver {
     protected boolean saveResults = false;
     protected boolean runPostureAndSweepTests = true;
     protected int generateByteCode = 1;
+    protected boolean debugByteCode = false;
+    protected boolean jitFlag = false;
     protected int streaming = 1;
     protected TreeModel treeModel = TreeModel.TINY_TREE;
     protected boolean debug = false;
     protected boolean export = false;
     protected boolean runWithJS = false;
+    protected boolean relocatable = false;
     protected String jsBase = null;
     protected Pattern testPattern = null;
     protected String requestedTestSet = null;
     protected String testSuiteDir;
     protected Processor driverProc = null;
     protected Serializer driverSerializer = null;
-    protected String exceptionsFileName = null;
     protected HashMap<String, XdmNode> exceptionsMap = new HashMap<String, XdmNode>();
     protected HashMap<String, String> optimizationAssertions = new HashMap<String, String>();
     protected Map<String, Environment> globalEnvironments = new HashMap<String, Environment>();
@@ -58,6 +60,9 @@ public abstract class TestDriver {
     protected String lang;
     protected boolean useXslt30Transformer = true;  // Temporary for controlling test processor
     protected boolean tracing = false;
+    protected boolean isAltova = false;
+    protected Map<String, Integer> failSummary = new TreeMap<String, Integer>();
+    protected boolean quiet = false;
 
     static Set<String> unsharedEnvironments = new HashSet<String>();
     static {
@@ -81,10 +86,8 @@ public abstract class TestDriver {
         System.err.println("Testing " + getProductEdition() + " " + Version.getProductVersion());
         System.err.println("Java version " + System.getProperty("java.version"));
 
-
         testSuiteDir = args[0];
         String catalog = args[1];
-
         char separatorChar = '/';
         if (File.separatorChar != '/') {
             separatorChar = '\\';
@@ -93,7 +96,6 @@ public abstract class TestDriver {
             testSuiteDir = testSuiteDir + separatorChar;
         }
         catalogFileName = catalog;
-
         catalog = testSuiteDir+catalog;
 
         for (int i = 2; i < args.length; i++) {
@@ -103,17 +105,20 @@ public abstract class TestDriver {
             if (args[i].startsWith("-s:")) {
                 requestedTestSet = args[i].substring(3);
             }
+            if (args[i].equals("-quiet")) {
+                quiet = true;
+            }
             if (args[i].startsWith("-o:")) {
                 resultsDir = args[i].substring(3);
             }
             if (args[i].startsWith("-debug")) {
                 debug = true;
             }
-            if (args[i].startsWith("-export")) {
+            if (args[i].startsWith("-export") && !args[i].equals("-export:off")) {
                 export = true;
             }
-            if (args[i].startsWith("-exceptions:")) {
-                exceptionsFileName = args[i].substring(12);
+            if (args[i].startsWith("-jit") && !args[i].equals("-jit:off")) {
+                jitFlag = true;
             }
             if (args[i].startsWith("-js")) {
                 export = true;
@@ -129,25 +134,35 @@ public abstract class TestDriver {
             if (args[i].equals("-save")) {
                 saveResults = true;
             }
+            if (args[i].startsWith("-exceptions:")) {
+                exceptionsFileName = args[i].substring(12);
+            }
             if (args[i].equals("-T")) {
                 tracing = true;
             }
             if (args[i].startsWith("-bytecode:")) {
+                String value = args[i].substring(10);
                 if (args[i].substring(10).equals("on")) {
-                    generateByteCode = 1;
-                } else if (args[i].substring(10).equals("debug")) {
-                    generateByteCode = 2;
-                } else {
                     generateByteCode = 0;
+                } else if (args[i].substring(10).equals("debug")) {
+                    debugByteCode = true;
+                    generateByteCode = 0;
+                } else if (args[i].substring(10).equals("off")) {
+                    generateByteCode = -1;
+
+                } else {
+                    generateByteCode = Integer.parseInt(value);
                 }
             }
             if (args[i].startsWith("-streaming:")) {
                 if (args[i].substring(11).equals("off")) {
                     streaming = 0;
-                } else if (args[i].substring(11).equals("std")) {
+                } else if (args[i].substring(11).equals("on")) {
                     streaming = 1;
-                } else if (args[i].substring(11).equals("ext")) {
-                    streaming = 2;
+                } else if (args[i].substring(11).equals("strict")) {
+                    streaming = 3;
+                } else {
+                    throw new Exception("-streaming must be off|on|strict");
                 }
             }
             if (args[i].startsWith("-tree")) {
@@ -180,6 +195,14 @@ public abstract class TestDriver {
                     runPostureAndSweepTests = false;
                 }
             }
+            // Determine whether static base uri should be exported
+            if (args[i].startsWith("-relocatable:")) {
+                if (args[i].substring(13).equals("on")) {
+                    relocatable = true;
+                } else if (args[i].substring(13).equals("off")) {
+                    relocatable = false;
+                }
+            }
         }
         if (resultsDoc == null) {
             printError("No result document: missing -lang option", "");
@@ -195,16 +218,21 @@ public abstract class TestDriver {
         }
 
         System.err.println("UsingXslt30Transformer: " + useXslt30Transformer);
+        System.err.println("Relocatable: " + relocatable);
 
         driverSerializer.setOutputStream(System.err);
         driverSerializer.setOutputProperty(Serializer.Property.METHOD, "adaptive");
         driverSerializer.setOutputProperty(Serializer.Property.OMIT_XML_DECLARATION, "yes");
         driverSerializer.setOutputProperty(Serializer.Property.INDENT, "yes");
+        driverProc.setConfigurationProperty(FeatureKeys.STABLE_UNPARSED_TEXT, true);
         processCatalog(new File(catalog));
-        System.err.println("Writing results to " + resultsDir + "/results" + Version.getProductVersion() + ".xml");
         printResults(resultsDir + "/results" + Version.getProductVersion() + ".xml");
     }
 
+    public boolean isDebugByteCode(){
+        return debugByteCode;
+    }
+
     /**
      * Return the appropriate tree model to use
      *
@@ -225,6 +253,10 @@ public abstract class TestDriver {
         return tree;
     }
 
+    public XdmNode makeDominoTree(Document doc, Configuration config, String baseUri) throws SaxonApiException {
+        throw new UnsupportedOperationException("Domino needs PE+");
+    }
+
     public String getResultsDir() {
         return resultsDir;
     }
@@ -232,22 +264,29 @@ public abstract class TestDriver {
     public abstract void processSpec(String specStr);
 
     public boolean isByteCode() {
-        return generateByteCode != 0;
+        return generateByteCode == 0;
     }
 
     public String getProductEdition() {
         return "Saxon-" + driverProc.getSaxonEdition();
     }
 
+    public String getProductVersion() {
+        return Version.getProductVersion();
+    }
+
     public void prepareForSQL(Processor processor) {}
 
 
     protected void processCatalog(File catalogFile) throws SaxonApiException {
         if (driverProc.getSaxonEdition().equals("EE")) {
-            if (generateByteCode == 1) {
+            if (generateByteCode == 0 && !debugByteCode) {
                 driverProc.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, "true");
                 driverProc.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "false");
-            } else if (generateByteCode == 2) {
+            } else if (generateByteCode >0) {
+                driverProc.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, ""+generateByteCode);
+                driverProc.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "false");
+            } else if (generateByteCode == 0 && debugByteCode) {
                 driverProc.setConfigurationProperty(FeatureKeys.GENERATE_BYTE_CODE, "true");
                 driverProc.setConfigurationProperty(FeatureKeys.DEBUG_BYTE_CODE, "true");
             } else {
@@ -300,11 +339,7 @@ public abstract class TestDriver {
     }
 
     protected String getNameOfExceptionsFile() {
-        if (exceptionsFileName != null) {
-            return exceptionsFileName;
-        } else {
-            return "exceptions.xml";
-        }
+        return exceptionsFileName;
     }
 
     /**
@@ -445,7 +480,7 @@ public abstract class TestDriver {
                 if (env == null) {
                     println("*** Unknown environment " + envName);
                     //writeTestcaseElement(testCaseName, "fail", "Environment " + envName + " not found", null);
-                    failures++;
+                    noteFailure(testCase.getParent().getAttributeValue(new QName("name")), testCaseName);
                     return null;
                 }
 
@@ -480,6 +515,20 @@ public abstract class TestDriver {
         // added in subclasses
     }
 
+    public void noteFailure(String testSet, String testCase) {
+        failures++;
+        if (requestedTestSet != null) {
+            failSummary.put(testCase, 1);
+        } else {
+            Integer fails = failSummary.get(testSet);
+            if (fails != null) {
+                failSummary.put(testSet, fails + 1);
+            } else {
+                failSummary.put(testSet, 1);
+            }
+        }
+    }
+
     protected void writeResultFilePreamble(Processor processor, XdmNode catalog)
             throws IOException, SaxonApiException, XMLStreamException, Exception {
         resultsDoc.writeResultFilePreamble(processor, catalog);
@@ -488,6 +537,17 @@ public abstract class TestDriver {
     protected void writeResultFilePostamble()
             throws XMLStreamException {
         resultsDoc.writeResultFilePostamble();
+        if (requestedTestSet != null) {
+            System.err.println("Failing tests: ");
+            for (Map.Entry<String, Integer> entry : failSummary.entrySet()) {
+                System.err.println("  " + entry.getKey());
+            }
+        } else {
+            System.err.println("Failures by Test Set: ");
+            for (Map.Entry<String, Integer> entry : failSummary.entrySet()) {
+                System.err.println("  " + entry.getKey() + " : " + entry.getValue());
+            }
+        }
     }
 
     protected void startTestSetElement(XdmNode testSetNode) {
@@ -507,7 +567,7 @@ public abstract class TestDriver {
         XdmNode testSetDocNode = catbuilder.build(testSetFile);
         createLocalEnvironments(testSetDocNode);
         boolean run = true;
-        // TODO: this won't pick up any test-set level dependencies in the XSLT 3.0 catalog
+        // pick up any test-set level dependencies in the QT3 catalog
         if (((XdmAtomicValue) xpc.evaluate("exists(/test-set/dependency)", testSetDocNode).itemAt(0)).getBooleanValue()) {
             for (XdmItem dependency : xpc.evaluate("/test-set/dependency", testSetDocNode)) {
                 if (!ensureDependencySatisfied((XdmNode) dependency, localEnvironments.get("default"))) {
@@ -521,6 +581,30 @@ public abstract class TestDriver {
                 }
             }
         }
+        // pick up any test-set level dependencies in the XSLT 3.0 catalog
+        if (((XdmAtomicValue) xpc.evaluate("exists(/test-set/dependencies)", testSetDocNode).itemAt(0)).getBooleanValue()) {
+            for (XdmItem dependency : xpc.evaluate("/test-set/dependencies/*", testSetDocNode)) {
+                if (!ensureDependencySatisfied((XdmNode) dependency, localEnvironments.get("default"))) {
+                    for (XdmItem testCase : xpc.evaluate("//test-case", testSetDocNode)) {
+                        String type = ((XdmNode) dependency).getNodeName().getLocalName();
+                        String value = ((XdmNode) dependency).getAttributeValue(new QName("", "value"));
+                        if (value == null) {
+                            value = type;
+                        } else {
+                            value = type + ":" + value;
+                        }
+                        if ("false".equals(((XdmNode) dependency).getAttributeValue(new QName("", "satisfied")))) {
+                            value = "!" + value;
+                        }
+                        String testCaseName = ((XdmNode) testCase).getAttributeValue(new QName("name"));
+                        resultsDoc.writeTestcaseElement(testCaseName, "n/a", "test-set dependencies not satisfied: " + value);
+                        notrun++;
+                    }
+                    run = false;
+                    break;
+                }
+            }
+        }
         if (run) {
             if (testPattern == null) {
                 for (XdmItem env : xpc.evaluate("//environment[@name]", testSetDocNode)) {
@@ -563,7 +647,9 @@ public abstract class TestDriver {
     }
 
     public void println(String data) {
-        shell.println(data);
+        if (!quiet) {
+            shell.println(data);
+        }
     }
 
     public void printResults(String resultsFileStr) {
@@ -595,5 +681,14 @@ public abstract class TestDriver {
 
     public abstract boolean ensureDependencySatisfied(XdmNode dependency, Environment env);
 
+    protected void copySchemaNamespaces(Environment env, XPathCompiler testXpc) {
+        Configuration config = env.xpathCompiler.getProcessor().getUnderlyingConfiguration();
+        //if (config instanceof EnterpriseConfiguration) {
+        for (String s : config.getImportedNamespaces()) {
+            testXpc.importSchemaNamespace(s);
+        }
+        //}
+    }
+
 }
 
