@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is “Incompatible With Secondary Licenses”, as defined by the Mozilla Public License, v. 2.0.
@@ -27,6 +27,7 @@ import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.Type;
 import net.sf.saxon.value.AtomicValue;
+import net.sf.saxon.value.SequenceExtent;
 import net.sf.saxon.value.StringValue;
 import net.sf.saxon.value.Whitespace;
 
@@ -46,6 +47,7 @@ public class TestOutcome {
     public static class SingleResultDoc {
         public XdmValue value;
         public String serialization;
+        public boolean wellFormed = true;
 
         public SingleResultDoc() {
         }
@@ -115,7 +117,7 @@ public class TestOutcome {
         return principalResult;
     }
 
-    public synchronized void setSecondaryResult(URI uri, XdmNode value, String serialization) {
+    public synchronized void setSecondaryResult(URI uri, XdmValue value, String serialization) {
         SingleResultDoc result = xslResultDocuments.get(uri);
         if (result == null) {
             result = new SingleResultDoc(value, serialization);
@@ -257,7 +259,7 @@ public class TestOutcome {
         } else {
             StringWriter sw = new StringWriter();
             Serializer s = p.newSerializer(sw);
-            s.setOutputProperty(Serializer.Property.METHOD, "xml");
+            s.setOutputProperty(Serializer.Property.METHOD, "adaptive");
             s.setOutputProperty(Serializer.Property.INDENT, "no");
             s.setOutputProperty(Serializer.Property.OMIT_XML_DECLARATION, "yes");
             try {
@@ -595,8 +597,8 @@ public class TestOutcome {
                             break;
                         }
                     }
-//                    driver.println("Actual results:");
-//                    driver.println(resultString);
+                    driver.println("Actual results:");
+                    driver.println(resultString);
                 } else {
                     driver.println("Serialized results differ");
                 }
@@ -605,6 +607,12 @@ public class TestOutcome {
         }
     }
 
+    private String resultVar;
+    public void setResultVar(String variableName) {
+        resultVar = variableName;
+    }
+
+
     private boolean assertXPath(XdmNode assertion, SingleResultDoc result, XPathCompiler assertXpc, boolean debug) throws SaxonApiException {
         if (isException()) {
             return false;
@@ -616,7 +624,65 @@ public class TestOutcome {
                     assertXpc.declareNamespace(namespace.getNodeName().getLocalName(), namespace.getStringValue());
                 }
             }
+            if (resultVar != null) {
+                assertXpc.declareVariable(new QName("result"));
+            }
+            driver.println("Testing " + assertion.getStringValue());
+            XPathExecutable exp = assertXpc.compile(assertion.getStringValue());
+            XPathSelector s = exp.load();
+            QName resultVarName = new QName("result");
+            if (resultVar != null) {
+                s.setVariable(new QName(resultVar), result.value);
+            } else if (exp.getRequiredCardinalityForVariable(resultVarName) == null) {
+                if (result.value instanceof XdmItem) { // this path used in XSLT tests
+                    s.setContextItem((XdmItem) result.value);
+                }
+            } else {
+                s.setVariable(resultVarName, result.value);
+            }
+            boolean b = s.effectiveBooleanValue();
+            if (!b && debug) {
+                driver.println("XPath assertion " + assertion.getStringValue() + " failed");
+                try {
+                    String ass = assertion.getStringValue();
+                    // Try to evaluate the expression on the lhs of an "=" operator in the assertion
+                    int eq = ass.indexOf("=");
+                    if (eq > 0) {
+                        ass = ass.substring(0, eq);
+                        exp = assertXpc.compile(ass);
+                        s = exp.load();
+                        if (exp.getRequiredCardinalityForVariable(resultVarName) == null) {
+                            if (result.value instanceof XdmItem) { // this path used in XSLT tests
+                                s.setContextItem((XdmItem) result.value);
+                            }
+                        } else {
+                            s.setVariable(resultVarName, result.value);
+                        }
+                        XdmValue val = s.evaluate();
+                        driver.println("Actual result of " + ass + ": " + val.toString());
+                    }
+                } catch (Exception err) {
+                    // Occurs for example with an assertion like /x[a = 2] where what precedes the '=' is not an expression
+                }
+                driver.println("Actual results: " + result.value);
+            }
+            return b;
+        }
+    }
 
+
+    private boolean assertXPathOLD(XdmNode assertion, SingleResultDoc result, XPathCompiler assertXpc, boolean debug) throws SaxonApiException {
+        if (isException()) {
+            return false;
+        } else {
+            XdmSequenceIterator iter = assertion.axisIterator(Axis.NAMESPACE);
+            while (iter.hasNext()) {
+                XdmNode namespace = (XdmNode) iter.next();
+                if (namespace.getNodeName() != null) {
+                    assertXpc.declareNamespace(namespace.getNodeName().getLocalName(), namespace.getStringValue());
+                }
+            }
+            driver.println("Testing " + assertion.getStringValue());
             XPathExecutable exp = assertXpc.compile(assertion.getStringValue());
             XPathSelector s = exp.load();
             QName resultVar = new QName("result");
@@ -723,7 +789,8 @@ public class TestOutcome {
                     driver.println("No serialized result available!");
                     return false;
                 }
-                RegularExpression re = Version.platform.compileRegularExpression(regex, flagsAtt, "XP30", warnings);
+                RegularExpression re = Version.platform.compileRegularExpression(
+                        xpath.getProcessor().getUnderlyingConfiguration(), regex, flagsAtt, "XP30", warnings);
                 if (re.containsMatch(principalSerializedResult)) {
                     return true;
                 } else {
@@ -807,20 +874,22 @@ public class TestOutcome {
                 } else {
                     iter0 = value.getUnderlyingValue().iterate();
                 }
+                GroundedValue val0 = SequenceExtent.makeSequenceExtent(iter0);
                 SequenceIterator iter1 = ((NodeInfo) expected.axisIterator(Axis.CHILD).next()
                         .getUnderlyingValue()).iterateAxis(AxisInfo.CHILD);
+                GroundedValue val1 = SequenceExtent.makeSequenceExtent(iter1);
                 boolean success = DeepEqual.deepEqual(
-                    iter0, iter1,
+                    val0.iterate(), val1.iterate(),
                     new GenericAtomicComparer(CodepointCollator.getInstance(), null),
                     assertXpc.getProcessor().getUnderlyingConfiguration().getConversionContext(), flag);
                 // if necessary try again ignoring whitespace nodes
                 if (!success) {
-                    iter0 = iter0.getAnother();
-                    iter1 = iter1.getAnother();
+                    iter0 = val0.iterate();
+                    iter1 = val1.iterate();
                     // deep-equals with the EXCLUDE_WHITESPACE flag doesn't ignore top-level whitespace, so we
                     // need to filter that out ourselves
-                    iter0 = new ItemMappingIterator<NodeInfo, NodeInfo>(iter0, new RemoveWhitespace());
-                    iter1 = new ItemMappingIterator<NodeInfo, NodeInfo>(iter1, new RemoveWhitespace());
+                    iter0 = new ItemMappingIterator(iter0, new RemoveWhitespace());
+                    iter1 = new ItemMappingIterator(iter1, new RemoveWhitespace());
                     success = DeepEqual.deepEqual(
                         iter0, iter1,
                         new GenericAtomicComparer(CodepointCollator.getInstance(), null),
@@ -831,17 +900,17 @@ public class TestOutcome {
                     }
                 }
                 if (!success) {
-                    System.err.println("assert-xml comparison failed");
+                    driver.println("assert-xml comparison failed");
                     if (debug) {
-                        System.err.println("assert-xml comparison failed");
-                        System.err.println("Reference results:");
+                        driver.println("assert-xml comparison failed");
+                        driver.println("Reference results:");
                         /*if(expectedDoctype != null) {
                             System.err.println(expectedDoctype);
                         }*/
-                        System.err.println(expected.toString());
-                        System.err.println("Actual results:");
+                        driver.println(expected.toString());
+                        driver.println("Actual results:");
                         //System.err.println(result.serialization);
-                        System.err.println(value.toString());
+                        driver.println(value.toString());
                     }
                 }
                 return success;
@@ -852,10 +921,10 @@ public class TestOutcome {
         }
     }
 
-    private class RemoveWhitespace implements ItemMappingFunction<NodeInfo, NodeInfo> {
-        public NodeInfo mapItem(NodeInfo item) throws XPathException {
-            boolean isWhite = item.getNodeKind() == Type.TEXT && Whitespace.isWhite(item.getStringValueCS());
-            return (isWhite ? null : item);
+    private class RemoveWhitespace implements ItemMappingFunction {
+        public NodeInfo mapItem(Item item) throws XPathException {
+            boolean isWhite = ((NodeInfo)item).getNodeKind() == Type.TEXT && Whitespace.isWhite(item.getStringValueCS());
+            return isWhite ? null : (NodeInfo) item;
         }
     }
 
@@ -919,9 +988,6 @@ public class TestOutcome {
             XdmAtomicValue item;
             try {
                 item = (XdmAtomicValue) s.evaluateSingle();
-//                if (item != null && !item.getBooleanValue()) {
-//                    System.err.println("Actual: " + result.value.toString());
-//                }
             } catch (SaxonApiException e) {
                 System.err.println("assert-eq failed - " + e.getMessage());
                 return false;
