@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is “Incompatible With Secondary Licenses”, as defined by the Mozilla Public License, v. 2.0.
@@ -7,7 +7,6 @@
 
 package net.sf.saxon.testdriver;
 
-
 import net.sf.saxon.Configuration;
 import net.sf.saxon.Controller;
 import net.sf.saxon.lib.*;
@@ -19,6 +18,7 @@ import net.sf.saxon.value.AtomicValue;
 import javax.xml.transform.Result;
 import javax.xml.transform.Source;
 import javax.xml.transform.SourceLocator;
+import javax.xml.transform.TransformerException;
 import javax.xml.transform.stream.StreamResult;
 import javax.xml.transform.stream.StreamSource;
 import java.io.*;
@@ -30,7 +30,6 @@ import java.util.*;
  * This class runs the W3C XSLT Test Suite, driven from the test catalog.
  */
 public class Xslt30TestSuiteDriverHE extends TestDriver {
-    protected boolean showSerialization = false;
 
 
     public Xslt30TestSuiteDriverHE() {
@@ -46,6 +45,13 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         new Xslt30TestSuiteDriverHE().go(args);
     }
 
+    public void go(String[] args) throws Exception {
+        if (driverProc == null) {
+            driverProc = new Processor(false);
+        }
+        super.go(args);
+    }
+
     protected static void usage() {
         System.err.println("java com.saxonica.testdriver.Xslt30TestSuiteDriver[HE] testsuiteDir catalog [-o:resultsdir] [-s:testSetName]" +
                                    " [-t:testNamePattern] [-bytecode:on|off|debug] [-export] [-tree] [-lang] [-save] [-streaming:off|std|ext]" +
@@ -88,9 +94,9 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
 
     protected boolean isSlow(String testName) {
         return testName.startsWith("regex-classes")
-                || testName.startsWith("unicode90")
                 || testName.equals("normalize-unicode-008")
-                || testName.equals("catalog-005");
+        || (testName.equals("function-1031") && driverProc.getSaxonEdition().equals("HE"));
+              //  || testName.equals("catalog-005");
     }
 
 
@@ -103,19 +109,23 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
 
         if (exceptionsMap.containsKey("$" + testSetName)) {
             notrun++;
-            resultsDoc.writeTestcaseElement(testName, "notRun", exceptionsMap.get("$" + testSetName).getAttributeValue(new QName("report")));
+            XdmNode exceptionElement = exceptionsMap.get("$" + testSetName);
+            resultsDoc.writeTestcaseElement(testName, exceptionElement.getAttributeValue(new QName("report")),
+                    exceptionElement.axisIterator(Axis.CHILD, new QName("reason")).next().getStringValue());
             return;
         }
 
         if (exceptionsMap.containsKey(testName)) {
             notrun++;
-            resultsDoc.writeTestcaseElement(testName, "notRun", exceptionsMap.get(testName).getAttributeValue(new QName("report")));
+            XdmNode exceptionElement = exceptionsMap.get(testName);
+            resultsDoc.writeTestcaseElement(testName, exceptionElement.getAttributeValue(new QName("report")),
+                    exceptionElement.axisIterator(Axis.CHILD, new QName("reason")).next().getStringValue());
             return;
         }
 
         if (isSlow(testName)) {
             notrun++;
-            resultsDoc.writeTestcaseElement(testName, "notRun", "requires excessive resources");
+            resultsDoc.writeTestcaseElement(testName, "tooBig", "requires excessive resources");
             return;
         }
 
@@ -128,19 +138,41 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             notrun++;
             return;
         }
-
-        if (!env.usable) {
-            resultsDoc.writeTestcaseElement(testName, "notRun", "environment dependencies not satisfied");
-            notrun++;
-            return;
+        for (XdmItem dep : xpath.evaluate("(/test-set/dependencies/*, ./dependencies/*)", testCase)) {
+            if (!ensureDependencySatisfied((XdmNode) dep, env)) {
+                notrun++;
+                String type = ((XdmNode) dep).getNodeName().getLocalName();
+                String value = ((XdmNode) dep).getAttributeValue(new QName("", "value"));
+                if (value == null) {
+                    value = type;
+                } else {
+                    value = type + ":" + value;
+                }
+                if ("false".equals(((XdmNode) dep).getAttributeValue(new QName("", "satisfied")))) {
+                    value = "!" + value;
+                }
+                String message = "dependency not satisfied: " + value;
+                if (value.startsWith("feature:")) {
+                    message = "requires optional " + value;
+                }
+                resultsDoc.writeTestcaseElement(testName, "n/a", message);
+                return;
+            }
         }
 
         if (env.failedToBuild) {
             resultsDoc.writeTestcaseElement(testName, "fail", "unable to build environment");
-            failures++;
+            noteFailure(testSetName, testName);
             return;
         }
 
+        if (!env.usable) {
+            resultsDoc.writeTestcaseElement(testName, "n/a", "environment dependencies not satisfied");
+            notrun++;
+            return;
+        }
+
+
         if (testName.contains("environment-variable")) {
             EnvironmentVariableResolver resolver = new EnvironmentVariableResolver() {
                 public Set<String> getAvailableEnvironmentVariables() {
@@ -188,27 +220,21 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         XdmNode principalPackage = (XdmNode) xpath.evaluateSingle("package[@role='principal']", testInput);
         XdmValue usedPackages = xpath.evaluate("package[@role='secondary']", testInput);
 
-
-        for (XdmItem dep : xpath.evaluate("(/test-set/dependencies/*, ./dependencies/*)", testCase)) {
-            if (!ensureDependencySatisfied((XdmNode) dep, env)) {
-                notrun++;
-                String type = ((XdmNode) dep).getNodeName().getLocalName();
-                String value = ((XdmNode) dep).getAttributeValue(new QName("", "value"));
-                if (value == null) {
-                    value = type;
-                } else {
-                    value = type + ":" + value;
-                }
-                if ("false".equals(((XdmNode) dep).getAttributeValue(new QName("", "satisfied")))) {
-                    value = "!" + value;
-                }
-                resultsDoc.writeTestcaseElement(testName, "notRun", "dependency not satisfied: " + value);
-                return;
-            }
-        }
-
         XsltExecutable sheet = env.xsltExecutable;
         ErrorCollector collector = new ErrorCollector();
+        if (quiet) {
+            collector.setLogger(new Logger() {
+                @Override
+                public void println(String message, int severity) {
+                    // no action
+                }
+
+                @Override
+                public StreamResult asStreamResult() {
+                    return null;
+                }
+            });
+        }
         String xsltLanguageVersion = spec.contains("XSLT30") || spec.contains("XSLT20+") ? "3.0" : "2.0";
 
         if (postureAndSweep != null && runPostureAndSweepTests) {
@@ -219,13 +245,34 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             return;
         }
 
+        boolean strictStreamability = xpath.evaluate("result//error[@code='XTSE3430']", testCase).size() > 0;
+        if (strictStreamability) {
+            env.processor.setConfigurationProperty(FeatureKeys.STRICT_STREAMABILITY, true);
+            env.resetActions.add(new Environment.ResetAction() {
+                @Override
+                public void reset(Environment env) {
+                    env.processor.setConfigurationProperty(FeatureKeys.STRICT_STREAMABILITY, false);
+                }
+            });
+        }
+
+        boolean staticError = xpath.evaluate("result//error[starts-with(@code, 'XTSE')]", testCase).size() > 0;
+        if (staticError && jitFlag) {
+            env.xsltCompiler.setJustInTimeCompilation(false);
+            env.resetActions.add(new Environment.ResetAction() {
+                @Override
+                public void reset(Environment env) {
+                    env.xsltCompiler.setJustInTimeCompilation(true);
+                }
+            });
+        }
+
         if (stylesheet != null) {
             String fileName = stylesheet.getAttributeValue(new QName("file"));
 
             Source styleSource = new StreamSource(testCase.getBaseURI().resolve(fileName).toString());
 
             XsltCompiler compiler = env.xsltCompiler;
-            setLanguageVersion(compiler, spec);
             compiler.setErrorListener(collector);
             initPatternOptimization(compiler);
             compiler.clearParameters();
@@ -266,7 +313,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             }
         } else if (principalPackage != null) {
             XsltCompiler compiler = env.xsltCompiler;
-            compiler.setXsltLanguageVersion(xsltLanguageVersion);
+            //compiler.setXsltLanguageVersion(xsltLanguageVersion);
             compiler.setErrorListener(collector);
             compiler.clearParameters();
             for (XdmItem param : xpath.evaluate("param[@static=('yes','true','1')]", testInput)) {
@@ -310,6 +357,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         }
 
         // TODO - trap multiples?
+        XdmValue initialMatchSelection = null;
         XdmNode initialMode = (XdmNode) xpath.evaluateSingle("initial-mode", testInput);
         XdmNode initialFunction = (XdmNode) xpath.evaluateSingle("initial-function", testInput);
         XdmNode initialTemplate = (XdmNode) xpath.evaluateSingle("initial-template", testInput);
@@ -324,21 +372,22 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             } else {
                 initialModeName = getQNameAttribute(xpath, testInput, "initial-mode/@name");
             }
+            String select = initialMode.getAttributeValue(new QName("select"));
+            if (select != null) {
+                initialMatchSelection = env.xpathCompiler.evaluate(select, null);
+                if (runWithJS) {
+                    resultsDoc.writeTestcaseElement(testName, "n/a", "Test driver limitation: cannot set initial match selection");
+                    notrun++;
+                    return;
+                }
+            }
         }
 
         QName initialTemplateName = getQNameAttribute(xpath, testInput, "initial-template/@name");
+        QName initialFunctionName = getQNameAttribute(xpath, testInput, "initial-function/@name");
 
         if (runWithJS && !outcome.isException()) {
-            clearGlobalParameters();
-            Map<QName, XdmValue> caseGlobalParams = getNamedParameters(xpath, testInput, false, false);
-            for (Map.Entry<QName, XdmValue> entry : caseGlobalParams.entrySet()) {
-                setGlobalParameter(entry.getKey(), entry.getValue());
-            }
-            URI baseOut = new File(resultsDir + "/results/output.xml").toURI();
-            runJSTransform(env, outcome, initialTemplateName, initialModeName, baseOut);
-            if ("*notJS*".equals(outcome.getComment())) {
-                return;
-            }
+            initializeForSaxonJS(xpath, outcome, testName, env, testInput, initialMode, initialTemplate, initialModeName, initialTemplateName, initialFunctionName);
         }
 
         if (sheet != null && !runWithJS) {
@@ -361,7 +410,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                                                    testSetName, env, testInput, sheet, collector,
                                                    xsltLanguageVersion, contextItem, initialMode,
                                                    initialFunction, initialTemplate, initialModeName,
-                                                   initialTemplateName, baseOutputURI);
+                                                   initialMatchSelection, initialTemplateName, baseOutputURI);
             } else {
                 failure = runWithXsltTransformer(xpath, outcome, testName, testSetName, env, testInput,
                                                  sheet, collector, contextItem, initialMode, initialModeName,
@@ -377,17 +426,21 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         }
         env.resetActions.clear();
 
+        boolean expectEarlyExit = ((XdmAtomicValue)xpath.evaluateSingle("result/@early-exit-possible = 'true'", testCase)).getBooleanValue();
+        if (expectEarlyExit != collector.isMadeEarlyExit()) {
+            outcome.setComment(expectEarlyExit ? "Failed to make early exit" : "Unexpected early exit");
+        }
+
         XdmNode assertion = (XdmNode) xpath.evaluateSingle("result/*", testCase);
         if (assertion == null) {
-            failures++;
+            noteFailure(testSetName, testName);
             resultsDoc.writeTestcaseElement(testName, "fail", "No test assertions found");
             return;
         }
         XPathCompiler assertionXPath = env.processor.newXPathCompiler();
         assertionXPath.setSchemaAware(true);
         assertionXPath.setBaseURI(assertion.getBaseURI());
-        //assertionXPath.declareVariable(new QName("result"));
-        //assertionXPath.setLanguageVersion("3.0");
+        copySchemaNamespaces(env, assertionXPath);  // ensure environment has schema namespaces
         boolean success = outcome.testAssertion(assertion, outcome.getPrincipalResultDoc(), assertionXPath, xpath, debug);
         if (success) {
             successes++;
@@ -399,7 +452,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 successes++;
                 resultsDoc.writeTestcaseElement(testName, "wrongError", outcome.getComment());
             } else {
-                failures++;
+                noteFailure(testSetName, testName);
                 // MHK: Nov 2015 - temporary diagnostics
                 if (outcome.getException() != null && outcome.getComment() == null) {
                     outcome.setComment(outcome.getException().getMessage());
@@ -409,7 +462,15 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         }
     }
 
-    public void runJSTransform(Environment env, TestOutcome outcome, QName initialTemplateName, QName initialModeName,
+    protected void initializeForSaxonJS(XPathCompiler xpath, TestOutcome outcome, String testName,
+                                        Environment env, XdmNode testInput, XdmNode initialMode,
+                                        XdmNode initialTemplate, QName initialModeName, QName initialTemplateName,
+                                        QName initialFunctionName) throws SaxonApiException {
+        // exists to be overridden
+    }
+
+    public void runJSTransform(Environment env, TestOutcome outcome, QName initialTemplateName,
+                               QName initialModeName, QName initialFunctionName,
                                URI baseOutputURI) {
         // exists to be overridden
     }
@@ -527,20 +588,33 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             }
         } catch (Exception err) {
             err.printStackTrace();
-            failures++;
+            noteFailure(testSetName, testName);
             resultsDoc.writeTestcaseElement(testName, "fail", "*** crashed " + err.getClass() + ": " + err.getMessage());
             return true;
         }
         return false;
     }
 
-    protected void clearGlobalParameters(){}
+    //protected void clearGlobalParameters(){}
 
     protected void setGlobalParameter(QName qName, XdmValue value) {
         // For overriding in Javascript driver
     }
 
-    protected boolean runWithXslt30Transformer(XdmNode testCase, XPathCompiler xpath, final TestOutcome outcome, String testName, String testSetName, Environment env, XdmNode testInput, XsltExecutable sheet, ErrorCollector collector, String xsltLanguageVersion, XdmItem contextItem, XdmNode initialMode, XdmNode initialFunction, XdmNode initialTemplate, QName initialModeName, QName initialTemplateName, String baseOutputURI) {
+//    protected void clearInitialFunctionArguments() {
+//    }
+//
+//    protected void addInitialFunctionArgument(XdmValue value) {
+//        // For overriding in Javascript driver
+//    }
+
+    protected boolean runWithXslt30Transformer(XdmNode testCase, XPathCompiler xpath, final TestOutcome outcome,
+                                               String testName, String testSetName, Environment env,
+                                               XdmNode testInput, XsltExecutable sheet, ErrorCollector collector,
+                                               String xsltLanguageVersion, XdmItem contextItem, XdmNode initialMode,
+                                               XdmNode initialFunction,
+                                               XdmNode initialTemplate, QName initialModeName, XdmValue initialMatchSelection,
+                                               QName initialTemplateName, String baseOutputURI) {
         try {
 
             boolean assertsSerial = xpath.evaluate("result//(assert-serialization|assert-serialization-error|serialization-matches)", testCase).size() > 0;
@@ -550,6 +624,10 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             if (needsTree != null) {
                 resultAsTree = needsTree.getStringValue().equals("yes");
             }
+            XdmNode resultVarAtt = (XdmNode) xpath.evaluateSingle("output/@result-var", testInput);
+            if (resultVarAtt != null) {
+                outcome.setResultVar(resultVarAtt.getStringValue());
+            }
             XdmNode needsSerialization = (XdmNode) xpath.evaluateSingle("output/@serialize", testInput);
             if (needsSerialization != null) {
                 serializationDeclared = needsSerialization.getStringValue().equals("yes");
@@ -587,8 +665,8 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
 
             transformer.setBaseOutputURI(baseOutputURI);
 
-            transformer.setMessageListener(new MessageListener() {
-                public void message(XdmNode content, boolean terminate, SourceLocator locator) {
+            transformer.setMessageListener(new MessageListener2() {
+                public void message(XdmNode content, QName errorCode, boolean terminate, SourceLocator locator) {
                     outcome.addXslMessage(content);
                     System.err.println(content.getStringValue());
                 }
@@ -611,6 +689,8 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 controller.setOutputURIResolver(
                         new OutputResolver(env.processor, outcome, false));
                 dest = new XdmDestination();
+            } else {
+                controller.setBuildTree(false);
             }
             if (resultSerialized) {
                 dest = serializer;
@@ -658,7 +738,6 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 }
             }
 
-
             if (initialTemplateName != null) {
                 transformer.setGlobalContextItem(contextItem);
                 if (dest == null) {
@@ -674,6 +753,12 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 } else {
                     transformer.callFunction(name, params, dest);
                 }
+            } else if (initialMatchSelection != null) {
+                if (dest == null) {
+                    result = transformer.applyTemplates(initialMatchSelection);
+                } else {
+                    transformer.applyTemplates(initialMatchSelection, dest);
+                }
             } else {
                 if (dest == null) {
                     result = transformer.applyTemplates(src);
@@ -731,17 +816,13 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             }
         } catch (Exception err) {
             err.printStackTrace();
-            failures++;
+            noteFailure(testSetName, testName);
             resultsDoc.writeTestcaseElement(testName, "fail", "*** crashed " + err.getClass() + ": " + err.getMessage());
             return true;
         }
         return false;
     }
 
-    protected void setLanguageVersion(XsltCompiler compiler, String spec) {
-        compiler.setXsltLanguageVersion("2.0"); // Always 2.0 for Saxon-HE
-    }
-
     protected void initPatternOptimization(XsltCompiler compiler) {
     }
 
@@ -749,7 +830,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         try {
             if (export) {
                 sheet = exportStylesheet(testName, testSetName, compiler, sheet, styleSource);
-            } else {
+            } else if (sheet == null) {
                 sheet = compiler.compile(styleSource);
             }
         } catch (SaxonApiException err) {
@@ -767,6 +848,29 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         return sheet;
     }
 
+    protected XsltExecutable exportStylesheet(String testName, String testSetName, XsltCompiler compiler, XsltExecutable sheet, Source styleSource) throws SaxonApiException {
+        try {
+            File exportFile = new File(resultsDir + "/export/" + testSetName + "/" + testName + ".sef");
+            XsltPackage compiledPack = compiler.compilePackage(styleSource);
+            compiledPack.save(exportFile);
+            sheet = reloadExportedStylesheet(compiler, exportFile);
+        } catch (SaxonApiException e) {
+            try {
+                compiler.getErrorListener().error(XPathException.makeXPathException(e));
+            } catch (TransformerException te) {
+                assert false;
+            }
+            //System.err.println(e.getMessage());
+            //e.printStackTrace();  //temporary, for debugging
+            throw e;
+        }
+        return sheet;
+    }
+
+    protected XsltExecutable reloadExportedStylesheet(XsltCompiler compiler, File exportFile) throws SaxonApiException {
+        return compiler.loadExecutablePackage(exportFile.toURI());
+    }
+
     private XsltPackage exportImportPackage(String testName, String testSetName, TestOutcome outcome, XsltCompiler compiler, XsltPackage pack, ErrorCollector collector) {
         try {
             if (export) {
@@ -775,7 +879,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                     pack.save(exportFile);
                     return compiler.loadLibraryPackage(exportFile.toURI());
                 } catch (SaxonApiException e) {
-                    System.err.println(e.getMessage());
+                    compiler.getErrorListener().error(XPathException.makeXPathException(e));
                     //e.printStackTrace();  //temporary, for debugging
                     throw e;
                 }
@@ -797,24 +901,6 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         return pack;
     }
 
-    protected XsltExecutable exportStylesheet(String testName, String testSetName, XsltCompiler compiler, XsltExecutable sheet, Source styleSource) throws SaxonApiException {
-        try {
-            File exportFile = new File(resultsDir + "/export/" + testSetName + "/" + testName + ".sef");
-            XsltPackage compiledPack = compiler.compilePackage(styleSource);
-            compiledPack.save(exportFile);
-            sheet = reloadExportedStylesheet(compiler, exportFile);
-        } catch (SaxonApiException e) {
-            System.err.println(e.getMessage());
-            //e.printStackTrace();  //temporary, for debugging
-            throw e;
-        }
-        return sheet;
-    }
-
-    protected XsltExecutable reloadExportedStylesheet(XsltCompiler compiler, File exportFile) throws SaxonApiException {
-        return compiler.loadExecutablePackage(exportFile.toURI());
-    }
-
     /**
      * Run streamability tests
      */
@@ -890,7 +976,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             }
             params.add(value);
         }
-        return params.toArray(new XdmValue[0]);
+        return params.toArray(new XdmValue[params.size()]);
     }
 
     @SuppressWarnings({"ResultOfMethodCallIgnored"})
@@ -922,7 +1008,7 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         XPathSelector exp = xpe.compile(assertion).load();
         exp.setContextItem(expressionTree);
         XdmAtomicValue bv = (XdmAtomicValue) exp.evaluateSingle();
-        if (!bv.getBooleanValue()) {
+        if (bv==null || !bv.getBooleanValue()) {
             println("** Optimization assertion failed");
             println(expressionTree.toString());
             throw new SaxonApiException("Expected optimization not performed");
@@ -940,10 +1026,14 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
 
     protected Set<String> alwaysOn = new HashSet<String>();
 
-    // Dependencies which are always satisfied in Saxon-EE but not in Saxon-HE
+    // Dependencies which are always satisfied in Saxon-EE but not in Saxon-HE or -PE
 
     protected Set<String> needsEE = new HashSet<String>();
 
+    // Dependencies which are always satisfied in Saxon-PE and -EE but not in Saxon-HE
+
+    protected Set<String> needsPE = new HashSet<String>();
+
     // Dependencies which are never satisfied in Saxon
 
     protected Set<String> alwaysOff = new HashSet<String>();
@@ -953,9 +1043,10 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         alwaysOn.add("feature/serialization");
         alwaysOn.add("feature/namespace_axis");
         alwaysOn.add("feature/dtd");
-        alwaysOn.add("feature/backwards_compatibility");
         alwaysOn.add("feature/built_in_derived_types");
+        alwaysOn.add("feature/remote_http");
         alwaysOn.add("feature/xsl-stylesheet-processing-instruction");
+        alwaysOn.add("feature/fn-transform-XSLT");
         alwaysOn.add("available_documents");
         alwaysOn.add("ordinal_scheme_name");
         alwaysOn.add("default_calendar_in_date_formatting_functions");
@@ -964,19 +1055,23 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         alwaysOn.add("default_output_encoding");
         alwaysOn.add("unparsed_text_encoding");
         alwaysOn.add("recognize_id_as_uri_fragment");
+        alwaysOn.add("feature/XPath_3.1");
+
+        needsPE.add("feature/backwards_compatibility");
+        needsPE.add("feature/Saxon-PE");
+        needsPE.add("feature/dynamic_evaluation");
 
         needsEE.add("languages_for_numbering");
         needsEE.add("feature/streaming");
         needsEE.add("feature/schema_aware");
-        needsEE.add("feature/Saxon-PE");
         needsEE.add("feature/Saxon-EE");
-        needsEE.add("feature/XSD_1.1");
-        needsEE.add("feature/XPath_3.1");
-        needsEE.add("feature/dynamic_evaluation");
+        //needsEE.add("feature/XSD_1.1");
+
+
         needsEE.add("feature/xquery_invocation");
         needsEE.add("feature/higher_order_functions");
 
-        alwaysOff.add("detect_accumulator_cycles");
+        alwaysOn.add("detect_accumulator_cycles");
     }
 
     /**
@@ -1008,8 +1103,12 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
         if (alwaysOff.contains(type) || alwaysOff.contains(tv)) {
             return !needed;
         }
+        String edition = env.processor.getSaxonEdition();
+        if (needsPE.contains(type) || needsPE.contains(tv)) {
+            return (edition.equals("PE") || edition.equals("EE")) == needed;
+        }
         if (needsEE.contains(type) || needsEE.contains(tv)) {
-            return env.processor.getSaxonEdition().equals("HE") != needed;
+            return edition.equals("EE") == needed;
         }
 
         if ("spec".equals(type)) {
@@ -1036,10 +1135,27 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 } else {
                     return false;
                 }
+            } else if ("XSD_1.1".equals(value)) {
+                    String requiredVersion = inverse ? "1.0" : "1.1";
+                    final String oldVersion = (String)env.processor.getConfigurationProperty(FeatureKeys.XSD_VERSION);
+                    if (!oldVersion.equals(requiredVersion)) {
+                        env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, requiredVersion);
+                        env.resetActions.add(new Environment.ResetAction() {
+                            @Override
+                            public void reset(Environment env) {
+                                env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, oldVersion);
+                            }
+                        });
+                    }
+                    return true;
             } else if ("higher_order_functions".equals(value)) {
-                return !inverse;
+                return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
             } else if ("simple-uca-fallback".equals(value)) {
                 return !inverse;
+            } else if ("advanced-uca-fallback".equals(value)) {
+                return (edition.equals("PE") || edition.equals("EE")) ^ inverse;
+            } else if ("streaming-fallback".equals(value)) {
+                return inverse;
             } else {
                 System.err.println("*** Unknown feature in HE: " + value);
                 return env.processor.getSaxonEdition().equals("HE") ? false : null;
@@ -1066,7 +1182,21 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
                 }
             });
             return true;
-
+        } else if ("extension-function".equals(type)) {
+            if (value.equals("Q{http://relaxng.org/ns/structure/1.0}schema-report#1")) {
+                try {
+                    Configuration config = env.processor.getUnderlyingConfiguration();
+                    Object sf = config.getInstance("net.cfoster.saxonjing.SchemaFunction", null);
+                    env.processor.registerExtensionFunction((ExtensionFunctionDefinition)sf);
+                    Object sfd = config.getInstance("net.cfoster.saxonjing.SchemaReportFunction", null);
+                    env.processor.registerExtensionFunction((ExtensionFunctionDefinition) sfd);
+                    return true;
+                } catch (XPathException err) {
+                    System.err.println("Failed to load Saxon-Jing extension functions");
+                    return false;
+                }
+            }
+            return false;
         } else if ("year_component_values".equals(type)) {
             if ("support year zero".equals(value)) {
                 if (env != null) {
@@ -1118,6 +1248,8 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
             return env.processor.getSaxonEdition().equals("HE") ? false : null;
         } else if ("sweep_and_posture".equals(type)) {
             return env.processor.getSaxonEdition().equals("HE") ? inverse : null;
+        } else if ("unicode-version".equals(type)) {
+            return value.equals("6.0"); // Avoid running Unicode 9.0 tests - they are slow!
         } else {
             println("**** dependency not recognized for HE: " + type);
             return false;
@@ -1126,12 +1258,12 @@ public class Xslt30TestSuiteDriverHE extends TestDriver {
 
     /**
      * Return a QNamed value from an attribute. This can handle active namespace prefix bindings or Clark notations in the
-     * attribute styring values
+     * attribute string values
      *
      * @param xpath         XPath compiler
      * @param contextItem   Context item
      * @param attributePath Path to the required (singleton?) attribute
-     * @return
+     * @return the value of the attribute as a QName
      * @throws SaxonApiException
      */
     protected static QName getQNameAttribute(XPathCompiler xpath, XdmItem contextItem, String attributePath) throws SaxonApiException {
