@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -18,6 +18,7 @@ import net.sf.saxon.lib.Logger;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.lib.SaxonOutputKeys;
 import net.sf.saxon.om.*;
+import net.sf.saxon.serialize.charcode.UTF16CharacterSet;
 import net.sf.saxon.trans.XPathException;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.type.BuiltInAtomicType;
@@ -31,7 +32,7 @@ import java.util.*;
 
 /**
  * This class handles the display of an abstract expression tree in an XML format
- * with some slight resemblence to XQueryX
+ * with some slight resemblance to XQueryX
  */
 public class ExpressionPresenter {
 
@@ -43,7 +44,14 @@ public class ExpressionPresenter {
     private Stack<Expression> expressionStack = new Stack<Expression>();
     private Stack<String> nameStack = new Stack<String>();
     private String defaultNamespace;
-    private Map<String, String> options = new HashMap<String,String>();
+    private Map<String, String> options = new HashMap<String, String>();
+    private boolean relocatable = false;
+
+    /**
+     * Make an uncommitted ExpressionPresenter. This must be followed by a call on init()
+     */
+
+    public ExpressionPresenter() {}
 
     /**
      * Make an ExpressionPresenter that writes indented output to the standard error output
@@ -71,17 +79,38 @@ public class ExpressionPresenter {
      * Make an ExpressionPresenter that writes indented output to a specified output stream,
      * with checksumming
      *
-     * @param config the Saxon configuration
-     * @param out    the output destination
+     * @param config   the Saxon configuration
+     * @param out      the output destination
+     * @param checksum true if a checksum is to be written at the end of the file
      */
 
     public ExpressionPresenter(Configuration config, StreamResult out, boolean checksum) {
+        init(config, out, checksum);
+    }
+
+    /**
+     * Make an ExpressionPresenter that writes indented output to a specified output stream,
+     * with checksumming
+     *
+     * @param config   the Saxon configuration
+     * @param out      the output destination
+     * @param checksum true if a checksum is to be written at the end of the file
+     */
+
+    public void init(Configuration config, StreamResult out, boolean checksum) {
         Properties props = makeDefaultProperties(config);
+        if (config.getXMLVersion() == Configuration.XML11) {
+            if ("JS".equals(getOption("target"))) {
+                config.getLogger().warning("For target=JS, the SEF file will use XML 1.0, which disallows control characters");
+            } else {
+                props.setProperty(OutputKeys.VERSION, "1.1");
+            }
+        }
         try {
             receiver = config.getSerializerFactory().getReceiver(
-                out,
-                config.makePipelineConfiguration(),
-                props);
+                    out,
+                    config.makePipelineConfiguration(),
+                    props);
             receiver = new NamespaceReducer(receiver);
             if (checksum) {
                 receiver = new CheckSumFilter(receiver);
@@ -100,6 +129,33 @@ public class ExpressionPresenter {
         }
     }
 
+    /**
+     * Make an ExpressionPresenter that writes indented output to a specified output stream,
+     * with checksumming
+     *
+     * @param config   the Saxon configuration
+     * @param out      the output destination
+     * @param checksum true if a checksum is to be written at the end of the file
+     */
+
+    public void init(Configuration config, Receiver out, boolean checksum) {
+
+            receiver = out;
+            receiver = new NamespaceReducer(receiver);
+            if (checksum) {
+                receiver = new CheckSumFilter(receiver);
+            }
+
+        this.config = config;
+        try {
+            receiver.open();
+            receiver.startDocument(0);
+        } catch (XPathException err) {
+            err.printStackTrace();
+            throw new InternalError(err.getMessage());
+        }
+    }
+
 
     /**
      * Make an ExpressionPresenter that writes indented output to a specified output stream
@@ -135,6 +191,7 @@ public class ExpressionPresenter {
     /**
      * Set the default namespace, used for subsequent calls on startElement.
      * Must be consistent throughout the whole document
+     *
      * @param namespace the default namespace
      */
 
@@ -144,7 +201,8 @@ public class ExpressionPresenter {
 
     /**
      * Set an option (a keyword=value pair, where both are strings)
-     * @param key the option name
+     *
+     * @param key   the option name
      * @param value the option value
      */
 
@@ -154,6 +212,7 @@ public class ExpressionPresenter {
 
     /**
      * Get the value of an option
+     *
      * @param key the option name
      * @return the option value, or null if the option has not been set
      */
@@ -162,6 +221,51 @@ public class ExpressionPresenter {
         return options.get(key);
     }
 
+    /**
+     * Get the value of an integer-valued option
+     *
+     * @param key the option name
+     * @param defaultValue the default value to be used if the option is not present or if it cannot
+     *                be parsed as an integer
+     * @return the option value as an integer, or -1 if the option
+     */
+
+    public int getIntOption(String key, int defaultValue) {
+        String s = options.get(key);
+        if (s != null) {
+            try {
+                return Integer.parseInt(s);
+            } catch (NumberFormatException err) {
+                return defaultValue;
+            }
+        } else {
+            return defaultValue;
+        }
+    }
+
+    /**
+     * Ask whether the package can be deployed to a different location, with a different base URI
+     *
+     * @return if true then static-base-uri() represents the deployed location of the package,
+     * rather than its compile time location
+     */
+
+    public boolean isRelocatable() {
+        return relocatable;
+    }
+
+    /**
+     * Say whether the package can be deployed to a different location, with a different base URI
+     *
+     * @param relocatable if true then static-base-uri() represents the deployed location of the package,
+     *                    rather than its compile time location
+     */
+
+    public void setRelocatable(boolean relocatable) {
+        this.relocatable = relocatable;
+    }
+
+
     /**
      * Make a receiver, using default output properties, with serialized output going
      * to a specified OutputStream
@@ -195,13 +299,9 @@ public class ExpressionPresenter {
         props.setProperty(OutputKeys.INDENT, "yes");
         props.setProperty(SaxonOutputKeys.INDENT_SPACES, "1");
         props.setProperty(SaxonOutputKeys.LINE_LENGTH, "4096");
-        if (config.getXMLVersion() == Configuration.XML11) {
-            props.setProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
-            props.setProperty(OutputKeys.VERSION, "1.1");
-        } else {
-            props.setProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
-            props.setProperty(OutputKeys.VERSION, "1.0");
-        }
+        props.setProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
+        props.setProperty(OutputKeys.ENCODING, "utf-8");
+        props.setProperty(OutputKeys.VERSION, "1.0");
         return props;
     }
 
@@ -211,7 +311,7 @@ public class ExpressionPresenter {
      * @param name the name of the element
      * @param expr the expression represented
      * @return the depth of the tree before this element: for diagnostics, this can be compared
-     *         with the value returned by endElement
+     * with the value returned by endElement
      */
 
     public int startElement(String name, Expression expr) {
@@ -223,12 +323,12 @@ public class ExpressionPresenter {
             if (expr.getRetainedStaticContext() == null) {
                 throw new AssertionError("Export failure: no retained static context on " + expr.toShortString());
             } else {
-                emitRetainedStaticContext(expr.getRetainedStaticContext(), parent==null ? null : parent.getRetainedStaticContext());
+                emitRetainedStaticContext(expr.getRetainedStaticContext(), parent == null ? null : parent.getRetainedStaticContext());
             }
         }
-        String mod = expr.getSystemId();
-        if (mod != null && parent != null && (parent.getSystemId()==null || !parent.getSystemId().equals(mod))) {
-            emitAttribute("module", mod);
+        String mod = expr.getLocation().getSystemId();
+        if (mod != null && parent != null && (parent.getLocation().getSystemId() == null || !parent.getLocation().getSystemId().equals(mod))) {
+            emitAttribute("module", truncatedModuleName(mod));
         }
         int lineNr = expr.getLocation().getLineNumber();
         if (parent == null || (parent.getLocation().getLineNumber() != lineNr && lineNr != -1)) {
@@ -237,12 +337,24 @@ public class ExpressionPresenter {
         return n;
     }
 
+    private String truncatedModuleName(String module) {
+        if (!relocatable) {
+            return module;
+        } else {
+            // If not exporting the base URI, cut the filename used for diagnostic location of errors down to its last component
+            String parts[] = module.split("/");
+            for (int p = parts.length - 1; p >= 0; p--) {
+                if (!parts[p].isEmpty()) {
+                    return parts[p];
+                }
+            }
+            return module;
+        }
+    }
+
     public void emitRetainedStaticContext(RetainedStaticContext sc, RetainedStaticContext parentSC) {
         try {
-            if (sc.getXPathVersion() != 30 && (parentSC == null || parentSC.getXPathVersion() != sc.getXPathVersion())) {
-                emitAttribute("vn", sc.getXPathVersion() + "");
-            }
-            if (sc.getStaticBaseUri() != null && (parentSC == null || !sc.getStaticBaseUri().equals(parentSC.getStaticBaseUri()))) {
+            if (!relocatable && sc.getStaticBaseUri() != null && (parentSC == null || !sc.getStaticBaseUri().equals(parentSC.getStaticBaseUri()))) {
                 emitAttribute("baseUri", sc.getStaticBaseUriString());
             }
             if (!sc.getDefaultCollationName().equals(NamespaceConstant.CODEPOINT_COLLATION_URI) &&
@@ -262,7 +374,14 @@ public class ExpressionPresenter {
                     String p = iter.next();
                     fsb.append(p);
                     fsb.append("=");
-                    fsb.append(sc.getURIForPrefix(p, true));
+                    String uri = sc.getURIForPrefix(p, true);
+                    if (Whitespace.containsWhitespace(uri)) {
+                        throw new XPathException("Cannot export a stylesheet if namespaces contain whitespace: '" + uri + "'");
+                    }
+                    if (uri.equals(NamespaceConstant.getUriForConventionalPrefix(p))) {
+                        uri = "~";
+                    }
+                    fsb.append(uri);
                     fsb.append(" ");
                 }
                 emitAttribute("ns", Whitespace.trim(fsb));
@@ -280,7 +399,7 @@ public class ExpressionPresenter {
      * with the value returned by endElement
      */
 
-    public int startElement (String name) {
+    public int startElement(String name) {
         nameStack.push(name);
         return _startElement(name);
     }
@@ -313,6 +432,7 @@ public class ExpressionPresenter {
 
     /**
      * Set the role of the next element to be output
+     *
      * @param role the value of the role output to be used
      */
 
@@ -329,6 +449,9 @@ public class ExpressionPresenter {
 
     public void emitAttribute(String name, String value) {
         if (value != null) {
+            if (name.equals("module")) {
+                value = truncatedModuleName(value);
+            }
             try {
                 receiver.attribute(new NoNamespaceName(name), BuiltInAtomicType.UNTYPED_ATOMIC, value, ExplicitLocation.UNKNOWN_LOCATION, 0);
             } catch (XPathException err) {
@@ -356,11 +479,11 @@ public class ExpressionPresenter {
                 }
                 receiver.namespace(new NamespaceBinding(prefix, value.getURI()), 0);
                 receiver.attribute(new NoNamespaceName(name),
-                        BuiltInAtomicType.UNTYPED_ATOMIC, prefix + ":" + value.getLocalPart(), ExplicitLocation.UNKNOWN_LOCATION, 0);
+                                   BuiltInAtomicType.UNTYPED_ATOMIC, prefix + ":" + value.getLocalPart(), ExplicitLocation.UNKNOWN_LOCATION, 0);
             } else {
                 receiver.namespace(new NamespaceBinding(value.getPrefix(), value.getURI()), 0);
                 receiver.attribute(new NoNamespaceName(name),
-                        BuiltInAtomicType.UNTYPED_ATOMIC, value.getDisplayName(), ExplicitLocation.UNKNOWN_LOCATION, 0);
+                                   BuiltInAtomicType.UNTYPED_ATOMIC, value.getDisplayName(), ExplicitLocation.UNKNOWN_LOCATION, 0);
             }
         } catch (XPathException err) {
             err.printStackTrace();
@@ -380,8 +503,8 @@ public class ExpressionPresenter {
             if (c == uri.length() - 1) {
                 c = uri.substring(0, uri.length() - 1).lastIndexOf('/');
             }
-            if (c >= 0 && c+3 < uri.length()) {
-                String prefix = uri.substring(c+1, c+3);
+            if (c >= 0 && c + 3 < uri.length()) {
+                String prefix = uri.substring(c + 1, c + 3);
                 if (NameChecker.isValidNCName(prefix)) {
                     return prefix;
                 }
@@ -392,8 +515,9 @@ public class ExpressionPresenter {
 
     /**
      * Output a namespace declaration
+     *
      * @param prefix the namespace prefix
-     * @param uri the namespace URI
+     * @param uri    the namespace URI
      */
 
     public void namespace(String prefix, String uri) {
@@ -409,7 +533,7 @@ public class ExpressionPresenter {
      * End an element in the expression tree
      *
      * @return the depth of the tree after ending this element. For diagnostics, this can be compared with the
-     *         value returned by startElement()
+     * value returned by startElement()
      */
 
     public int endElement() {
@@ -468,7 +592,6 @@ public class ExpressionPresenter {
     }
 
 
-
     /**
      * Get the Saxon configuration
      *
@@ -498,5 +621,55 @@ public class ExpressionPresenter {
     public TypeHierarchy getTypeHierarchy() {
         return config.getTypeHierarchy();
     }
+
+    /**
+     * Static method to escape a string using Javascript escaping conventions
+     * @param in the string to be escaped
+     */
+
+    public static String jsEscape(String in) {
+        FastStringBuffer out = new FastStringBuffer(in.length());
+        for (int i = 0; i < in.length(); i++) {
+            char c = in.charAt(i);
+            switch (c) {
+                case '\'':
+                    out.append("\\\'");
+                    break;
+                case '"':
+                    out.append("\\\"");
+                    break;
+                case '\b':
+                    out.append("\\b");
+                    break;
+                case '\f':
+                    out.append("\\f");
+                    break;
+                case '\n':
+                    out.append("\\n");
+                    break;
+                case '\r':
+                    out.append("\\r");
+                    break;
+                case '\t':
+                    out.append("\\t");
+                    break;
+                case '\\':
+                    out.append("\\\\");
+                    break;
+                default:
+                    if (c < 32 || (c > 127 && c < 160) || c > UTF16CharacterSet.SURROGATE1_MIN) {
+                        out.append("\\u");
+                        String hex = Integer.toHexString(c).toUpperCase();
+                        while (hex.length() < 4) {
+                            hex = "0" + hex;
+                        }
+                        out.append(hex);
+                    } else {
+                        out.append(c);
+                    }
+            }
+        }
+        return out.toString();
+    }
 }
 
