@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -20,6 +20,7 @@ import net.sf.saxon.pattern.*;
 import net.sf.saxon.style.StylesheetModule;
 import net.sf.saxon.style.StylesheetPackage;
 import net.sf.saxon.trace.ExpressionPresenter;
+import net.sf.saxon.trans.rules.*;
 import net.sf.saxon.tree.util.Navigator;
 import net.sf.saxon.type.*;
 import net.sf.saxon.value.AtomicValue;
@@ -39,20 +40,20 @@ public class SimpleMode extends Mode {
     // TODO:PERF the data structure does not cater well for a stylesheet making heavy use of
     // match="schema-element(X)". We should probably expand the substitution group.
 
-    protected Rule genericRuleChain = null;
-    protected Rule atomicValueRuleChain = null;
-    protected Rule functionItemRuleChain = null;
-    protected Rule documentRuleChain = null;
-    protected Rule textRuleChain = null;
-    protected Rule commentRuleChain = null;
-    protected Rule processingInstructionRuleChain = null;
-    protected Rule namespaceRuleChain = null;
-    protected Rule unnamedElementRuleChain = null;
-    protected Rule unnamedAttributeRuleChain = null;
-    protected IntHashMap<Rule> namedElementRuleChains = new IntHashMap<Rule>(32);
-    protected IntHashMap<Rule> namedAttributeRuleChains = new IntHashMap<Rule>(8);
-    protected Map<StructuredQName, Rule> qNamedElementRuleChains;
-    protected Map<StructuredQName, Rule> qNamedAttributeRuleChains;
+    protected RuleChain genericRuleChain = new RuleChain();
+    protected RuleChain atomicValueRuleChain = new RuleChain();
+    protected RuleChain functionItemRuleChain = new RuleChain();
+    protected RuleChain documentRuleChain = new RuleChain();
+    protected RuleChain textRuleChain = new RuleChain();
+    protected RuleChain commentRuleChain = new RuleChain();
+    protected RuleChain processingInstructionRuleChain = new RuleChain();
+    protected RuleChain namespaceRuleChain = new RuleChain();
+    protected RuleChain unnamedElementRuleChain = new RuleChain();
+    protected RuleChain unnamedAttributeRuleChain = new RuleChain();
+    protected IntHashMap<RuleChain> namedElementRuleChains = new IntHashMap<RuleChain>(32);
+    protected IntHashMap<RuleChain> namedAttributeRuleChains = new IntHashMap<RuleChain>(8);
+    protected Map<StructuredQName, RuleChain> qNamedElementRuleChains;
+    protected Map<StructuredQName, RuleChain> qNamedAttributeRuleChains;
 
     private BuiltInRuleSet builtInRuleSet = TextOnlyCopyRuleSet.getInstance();
     private Rule mostRecentRule;
@@ -64,6 +65,8 @@ public class SimpleMode extends Mode {
     private Map<String, Integer> explicitPropertyPrecedences = new HashMap<String, Integer>();
     private Map<String, String> explicitPropertyValues = new HashMap<String, String>();
 
+
+
     /**
      * Default constructor - creates a Mode containing no rules
      *
@@ -114,30 +117,68 @@ public class SimpleMode extends Mode {
      */
 
     public void checkForConflictingProperties() throws XPathException {
-        final String typed = explicitPropertyValues.get("typed");
-        mustBeTyped = "yes".equals(typed) || "strict".equals(typed) || "lax".equals(typed);
-        mustBeUntyped = "no".equals(typed);
         for (Map.Entry<String, String> entry : getActivePart().explicitPropertyValues.entrySet()) {
-            if (entry.getValue().equals("##conflict##")) {
-                throw new XPathException("There are conflicting values for xsl:mode/@" + entry.getKey() +
-                    " at the same import precedence", "XTSE0545");
+            String prop = entry.getKey();
+            String value = entry.getValue();
+            if (value.equals("##conflict##")) {
+                throw new XPathException(
+                        "For " + getLabel() +
+                                ", there are conflicting values for xsl:mode/@" +
+                                prop +
+                                " at the same import precedence", "XTSE0545");
+            }
+            if (prop.equals("typed")) {
+                mustBeTyped = "yes".equals(value) || "strict".equals(value) || "lax".equals(value);
+                mustBeUntyped = "no".equals(value);
+            } else if (prop.equals("on-no-match")) {
+                BuiltInRuleSet base = null;
+                if (value.equals("text-only-copy")) {
+                    base = TextOnlyCopyRuleSet.getInstance();
+                } else if (value.equals("shallow-copy")) {
+                     base = ShallowCopyRuleSet.getInstance();
+                } else if (value.equals("deep-copy")) {
+                     base = DeepCopyRuleSet.getInstance();
+                } else if (value.equals("shallow-skip")) {
+                     base = ShallowSkipRuleSet.getInstance();
+                } else if (value.equals("deep-skip")) {
+                    base = DeepSkipRuleSet.getInstance();
+                } else if (value.equals("fail")) {
+                    base = FailRuleSet.getInstance();
+                } else {
+                    // already validated
+                }
+                if ("yes".equals(explicitPropertyValues.get("warning-on-no-match"))) {
+                    base = new RuleSetWithWarnings(base);
+                }
+                setBuiltInRuleSet(base);
             }
         }
     }
 
+    /**
+     * Get an identifier for the mode for use in error messages
+     * @return either the string "the unnamed mode" or the string "mode NNNN" where "NNNN"
+     * is the display form of the mode's name.
+     */
+    public String getLabel() {
+        return isUnnamedMode() ?
+                "the unnamed mode" :
+                "mode " + modeName.getDisplayName();
+    }
+
     /**
      * Construct a new Mode, copying the contents of an existing Mode
      *
      * @param from the existing mode.
-     * @param to the name of the new mode to be created
+     * @param to   the name of the new mode to be created
      */
 
     public static void copyRules(final SimpleMode from, final SimpleMode to) {
         try {
             from.processRules(new RuleAction() {
                 public void processRule(Rule r) throws XPathException {
-                    Rule r2 = new Rule(r, false);
-                    to.addRule(r2.pattern, r2);
+                    Rule r2 = r.copy(false);
+                    to.addRule(r2.getPattern(), r2);
                 }
             });
         } catch (XPathException e) {
@@ -153,7 +194,7 @@ public class SimpleMode extends Mode {
      *
      * @return a new object capable of holding the state of a search for a rule
      */
-    protected RuleSearchState makeRuleSearchState() {
+    protected RuleSearchState makeRuleSearchState(RuleChain chain, XPathContext context) {
         return new RuleSearchState(this);
     }
 
@@ -237,6 +278,7 @@ public class SimpleMode extends Mode {
 
     /**
      * Add a rule to the Mode.
+     *
      * @param pattern      a Pattern
      * @param action       the Object to return from getRule() when the supplied node matches this Pattern
      * @param module       the stylesheet module containing the rule
@@ -318,44 +360,54 @@ public class SimpleMode extends Mode {
         if (uType.equals(UType.ELEMENT)) {
             int fp = pattern.getFingerprint();
             if (fp == -1) {
-                unnamedElementRuleChain = addRuleToList(newRule, unnamedElementRuleChain);
+                addRuleToList(newRule, unnamedElementRuleChain);
             } else {
-                Rule chain = namedElementRuleChains.get(fp);
-                namedElementRuleChains.put(fp, addRuleToList(newRule, chain));
+                RuleChain chain = namedElementRuleChains.get(fp);
+                if (chain == null) {
+                    chain = new RuleChain(newRule);
+                    namedElementRuleChains.put(fp, chain);
+                } else {
+                    addRuleToList(newRule, chain);
+                }
             }
         } else if (uType.equals(UType.ATTRIBUTE)) {
             int fp = pattern.getFingerprint();
             if (fp == -1) {
-                unnamedAttributeRuleChain = addRuleToList(newRule, unnamedAttributeRuleChain);
+                addRuleToList(newRule, unnamedAttributeRuleChain);
             } else {
-                Rule chain = namedAttributeRuleChains.get(fp);
-                namedAttributeRuleChains.put(fp, addRuleToList(newRule, chain));
+                RuleChain chain = namedAttributeRuleChains.get(fp);
+                if (chain == null) {
+                    chain = new RuleChain(newRule);
+                    namedAttributeRuleChains.put(fp, chain);
+                } else {
+                    addRuleToList(newRule, chain);
+                }
             }
         } else if (uType.equals(UType.DOCUMENT)) {
-            documentRuleChain = addRuleToList(newRule, documentRuleChain);
+            addRuleToList(newRule, documentRuleChain);
         } else if (uType.equals(UType.TEXT)) {
-            textRuleChain = addRuleToList(newRule, textRuleChain);
+            addRuleToList(newRule, textRuleChain);
         } else if (uType.equals(UType.COMMENT)) {
-            commentRuleChain = addRuleToList(newRule, commentRuleChain);
+            addRuleToList(newRule, commentRuleChain);
         } else if (uType.equals(UType.PI)) {
-            processingInstructionRuleChain = addRuleToList(newRule, processingInstructionRuleChain);
+            addRuleToList(newRule, processingInstructionRuleChain);
         } else if (uType.equals(UType.NAMESPACE)) {
-            namespaceRuleChain = addRuleToList(newRule, namespaceRuleChain);
+            addRuleToList(newRule, namespaceRuleChain);
         } else if (UType.ANY_NODE.subsumes(uType)) {
-            genericRuleChain = addRuleToList(newRule, genericRuleChain);
+            addRuleToList(newRule, genericRuleChain);
         } else {
             // TODO: we really shouldn't have to enumerate all possible kinds of pattern here!
-            if (pattern instanceof NodeTestPattern || pattern instanceof PatternWithPredicate) {
+            if (pattern instanceof NodeTestPattern || pattern instanceof BasePatternWithPredicate) {
                 ItemType type = pattern.getItemType();
                 if (type instanceof AtomicType) {
-                    atomicValueRuleChain = addRuleToList(newRule, atomicValueRuleChain);
+                    addRuleToList(newRule, atomicValueRuleChain);
                 } else if (type instanceof FunctionItemType) {
-                    functionItemRuleChain = addRuleToList(newRule, functionItemRuleChain);
+                    addRuleToList(newRule, functionItemRuleChain);
                 } else {
                     throw new UnsupportedOperationException("XSLT 3.0 '.[]' syntax not recognized with node type or external object types");
                 }
             } else if (pattern instanceof BooleanExpressionPattern || pattern instanceof UniversalPattern) {
-                genericRuleChain = addRuleToList(newRule, genericRuleChain);
+                addRuleToList(newRule, genericRuleChain);
             } else {
                 throw new UnsupportedOperationException("Unrecognized pattern " + pattern.getClass());
             }
@@ -372,34 +424,34 @@ public class SimpleMode extends Mode {
      */
 
 
-    private Rule addRuleToList(Rule newRule, Rule list) {
-        if (list == null) {
-            return newRule;
-        }
-        int precedence = newRule.getPrecedence();
-        double priority = newRule.getPriority();
-        Rule rule = list;
-        Rule prev = null;
-        while (rule != null) {
-            if ((rule.getPrecedence() < precedence) ||
-                    (rule.getPrecedence() == precedence && rule.getPriority() <= priority)) {
-                newRule.setNext(rule);
-                if (prev == null) {
-                    return newRule;
+    private void addRuleToList(Rule newRule, RuleChain list) {
+        if (list.head() == null) {
+            list.setHead(newRule);
+        } else {
+            int precedence = newRule.getPrecedence();
+            double priority = newRule.getPriority();
+            Rule rule = list.head();
+            Rule prev = null;
+            while (rule != null) {
+                if ((rule.getPrecedence() < precedence) ||
+                        (rule.getPrecedence() == precedence && rule.getPriority() <= priority)) {
+                    newRule.setNext(rule);
+                    if (prev == null) {
+                        list.setHead(newRule);
+                    } else {
+                        prev.setNext(newRule);
+                    }
+                    break;
                 } else {
-                    prev.setNext(newRule);
+                    prev = rule;
+                    rule = rule.getNext();
                 }
-                break;
-            } else {
-                prev = rule;
-                rule = rule.getNext();
+            }
+            if (rule == null) {
+                prev.setNext(newRule);
+                newRule.setNext(null);
             }
         }
-        if (rule == null) {
-            prev.setNext(newRule);
-            newRule.setNext(null);
-        }
-        return list;
     }
 
     /**
@@ -438,12 +490,9 @@ public class SimpleMode extends Mode {
             context = makeNewContext(context);
         }
 
-        // Get the rule search state object - this could be reusable within a rule chain.
-        RuleSearchState ruleSearchState = makeRuleSearchState();
-
         // search the specific list for this node type / node name
 
-        Rule unnamedNodeChain;
+        RuleChain unnamedNodeChain;
         Rule bestRule = null;
 
         if (item instanceof NodeInfo) {
@@ -455,27 +504,27 @@ public class SimpleMode extends Mode {
 
                 case Type.ELEMENT: {
                     unnamedNodeChain = unnamedElementRuleChain;
-                    Rule namedNodeChain;
-                    if (node instanceof FingerprintedNode) {
-                        namedNodeChain = namedElementRuleChains.get(((FingerprintedNode) node).getFingerprint());
+                    RuleChain namedNodeChain;
+                    if (node.hasFingerprint()) {
+                        namedNodeChain = namedElementRuleChains.get(node.getFingerprint());
                     } else {
                         namedNodeChain = getNamedRuleChain(context, Type.ELEMENT, node.getURI(), node.getLocalPart());
                     }
                     if (namedNodeChain != null) {
-                        bestRule = searchRuleChain(node, context, null, namedNodeChain, ruleSearchState);
+                        bestRule = searchRuleChain(node, context, null, namedNodeChain);
                     }
                     break;
                 }
                 case Type.ATTRIBUTE: {
                     unnamedNodeChain = unnamedAttributeRuleChain;
-                    Rule namedNodeChain;
-                    if (node instanceof FingerprintedNode) {
-                        namedNodeChain = namedAttributeRuleChains.get(((FingerprintedNode) node).getFingerprint());
+                    RuleChain namedNodeChain;
+                    if (node.hasFingerprint()) {
+                        namedNodeChain = namedAttributeRuleChains.get(node.getFingerprint());
                     } else {
                         namedNodeChain = getNamedRuleChain(context, Type.ATTRIBUTE, node.getURI(), node.getLocalPart());
                     }
                     if (namedNodeChain != null) {
-                        bestRule = searchRuleChain(node, context, null, namedNodeChain, ruleSearchState);
+                        bestRule = searchRuleChain(node, context, null, namedNodeChain);
                     }
                     break;
                 }
@@ -498,35 +547,32 @@ public class SimpleMode extends Mode {
             // search the list for unnamed nodes of a particular kind
 
             if (unnamedNodeChain != null) {
-                bestRule = searchRuleChain(node, context, bestRule, unnamedNodeChain, ruleSearchState);
+                bestRule = searchRuleChain(node, context, bestRule, unnamedNodeChain);
             }
 
             // Search the list for rules for nodes of unknown node kind
 
             if (genericRuleChain != null) {
-                bestRule = searchRuleChain(node, context, bestRule, genericRuleChain, ruleSearchState);
+                bestRule = searchRuleChain(node, context, bestRule, genericRuleChain);
             }
 
         } else if (item instanceof AtomicValue) {
             if (atomicValueRuleChain != null) {
-                bestRule = searchRuleChain(item, context, bestRule, atomicValueRuleChain, ruleSearchState);
+                bestRule = searchRuleChain(item, context, bestRule, atomicValueRuleChain);
             }
             if (genericRuleChain != null) {
-                bestRule = searchRuleChain(item, context, bestRule, genericRuleChain, ruleSearchState);
+                bestRule = searchRuleChain(item, context, bestRule, genericRuleChain);
             }
 
         } else if (item instanceof Function) {
             if (functionItemRuleChain != null) {
-                bestRule = searchRuleChain(item, context, bestRule, functionItemRuleChain, ruleSearchState);
+                bestRule = searchRuleChain(item, context, bestRule, functionItemRuleChain);
             }
             if (genericRuleChain != null) {
-                bestRule = searchRuleChain(item, context, bestRule, genericRuleChain, ruleSearchState);
+                bestRule = searchRuleChain(item, context, bestRule, genericRuleChain);
             }
         }
 
-       /* if (bestRule != null) { // project:preconditions
-            RuleSearchState.stats(bestRule, ruleSearchState);
-        }*/
         return bestRule;
     }
 
@@ -541,51 +587,53 @@ public class SimpleMode extends Mode {
      * to consider
      */
 
-    protected Rule getNamedRuleChain(XPathContext c, int kind, String uri, String local) {
+    protected RuleChain getNamedRuleChain(XPathContext c, int kind, String uri, String local) {
         // If this is the first attempt to match a non-fingerprinted node, build indexes
         // to the rule chains based on StructuredQName rather than fingerprint
-        if (qNamedElementRuleChains == null) {
-            synchronized (this) {
-                qNamedElementRuleChains = new HashMap<StructuredQName, Rule>(namedElementRuleChains.size());
-                qNamedAttributeRuleChains = new HashMap<StructuredQName, Rule>(namedAttributeRuleChains.size());
+        synchronized(this) {
+            if (qNamedElementRuleChains == null) {
+                qNamedElementRuleChains = new HashMap<StructuredQName, RuleChain>(namedElementRuleChains.size());
+                qNamedAttributeRuleChains = new HashMap<StructuredQName, RuleChain>(namedAttributeRuleChains.size());
                 NamePool pool = c.getNamePool();
-                IntIterator eIter = namedElementRuleChains.keyIterator();
-                while (eIter.hasNext()) {
-                    int fp = eIter.next();
-                    Rule eChain = namedElementRuleChains.get(fp);
-                    StructuredQName name = pool.getStructuredQName(fp);
-                    qNamedElementRuleChains.put(name, eChain);
-                }
-                IntIterator aIter = namedAttributeRuleChains.keyIterator();
-                while (aIter.hasNext()) {
-                    int fp = aIter.next();
-                    Rule eChain = namedAttributeRuleChains.get(fp);
-                    StructuredQName name = pool.getStructuredQName(fp);
-                    qNamedAttributeRuleChains.put(name, eChain);
-                }
+                indexByQName(pool, namedElementRuleChains, qNamedElementRuleChains);
+                indexByQName(pool, namedAttributeRuleChains, qNamedAttributeRuleChains);
             }
         }
         return (kind == Type.ELEMENT ? qNamedElementRuleChains : qNamedAttributeRuleChains)
                 .get(new StructuredQName("", uri, local));
     }
 
+    private static void indexByQName(NamePool pool, IntHashMap<RuleChain> indexByFP, Map<StructuredQName, RuleChain> indexByQN) {
+        IntIterator ii = indexByFP.keyIterator();
+        while (ii.hasNext()) {
+            int fp = ii.next();
+            RuleChain eChain = indexByFP.get(fp);
+            StructuredQName name = pool.getStructuredQName(fp);
+            indexByQN.put(name, eChain);
+        }
+    }
+
     /**
      * Search a chain of rules
      *
      * @param item            the item being matched
      * @param context         XPath dynamic context
      * @param bestRule        the best rule so far in terms of precedence and priority (may be null)
-     * @param head            the rule at the head of the chain to be searched
-     * @param ruleSearchState An appropriate rule state object
+     * @param chain           the chain to be searched
      * @return the best match rule found in the chain, or the previous best rule, or null
      * @throws XPathException if an error occurs matching a pattern
      */
 
-    protected Rule searchRuleChain(Item item, XPathContext context, /*@Nullable*/ Rule bestRule, Rule head, RuleSearchState ruleSearchState) throws XPathException {
+    protected Rule searchRuleChain(Item item, XPathContext context, /*@Nullable*/ Rule bestRule, RuleChain chain) throws XPathException {
+
         while (!(context instanceof XPathContextMajor)) {
             context = context.getCaller();
         }
 
+        // Get the rule search state object - this could be reusable within a rule chain.
+        RuleSearchState ruleSearchState = makeRuleSearchState(chain, context);
+
+        Rule head = chain == null ? null : chain.head();
         while (head != null) {
             if (bestRule != null) {
                 int rank = head.compareRank(bestRule);
@@ -612,11 +660,11 @@ public class SimpleMode extends Mode {
             } else if (ruleMatches(head, item, (XPathContextMajor) context, ruleSearchState)) {
                 bestRule = head;
                 if (getRecoveryPolicy() == Configuration.RECOVER_SILENTLY) {
-                    ruleSearchState.count();
+                    //ruleSearchState.count();
                     break;   // choose the first match; rules within a chain are in order of rank
                 }
             }
-            ruleSearchState.count();// Keep tab of the number of checks
+            //ruleSearchState.count();// Keep tab of the number of checks
             head = head.getNext();
         }
 
@@ -664,12 +712,12 @@ public class SimpleMode extends Mode {
         }
 
         // Get the rule search state object
-        RuleSearchState ruleSearchState = makeRuleSearchState();
+        RuleSearchState ruleSearchState;
 
         // search the specific list for this node type / node name
 
         Rule bestRule = null;
-        Rule unnamedNodeChain;
+        RuleChain unnamedNodeChain;
 
 
         // Search the list for unnamed nodes of a particular kind
@@ -682,24 +730,30 @@ public class SimpleMode extends Mode {
                     break;
                 case Type.ELEMENT: {
                     unnamedNodeChain = unnamedElementRuleChain;
-                    Rule namedNodeChain;
-                    if (node instanceof FingerprintedNode) {
-                        namedNodeChain = namedElementRuleChains.get(((FingerprintedNode) node).getFingerprint());
+                    RuleChain namedNodeChain;
+                    if (node.hasFingerprint()) {
+                        namedNodeChain = namedElementRuleChains.get(node.getFingerprint());
                     } else {
                         namedNodeChain = getNamedRuleChain(context, Type.ELEMENT, node.getURI(), node.getLocalPart());
                     }
-                    bestRule = searchRuleChain(item, context, null, namedNodeChain, ruleSearchState, filter);
+                    if (namedNodeChain != null) {
+                        ruleSearchState = makeRuleSearchState(namedNodeChain, context);
+                        bestRule = searchRuleChain(item, context, null, namedNodeChain, ruleSearchState, filter);
+                    }
                     break;
                 }
                 case Type.ATTRIBUTE: {
                     unnamedNodeChain = unnamedAttributeRuleChain;
-                    Rule namedNodeChain;
-                    if (node instanceof FingerprintedNode) {
-                        namedNodeChain = namedAttributeRuleChains.get(((FingerprintedNode) node).getFingerprint());
+                    RuleChain namedNodeChain;
+                    if (node.hasFingerprint()) {
+                        namedNodeChain = namedAttributeRuleChains.get(node.getFingerprint());
                     } else {
                         namedNodeChain = getNamedRuleChain(context, Type.ATTRIBUTE, node.getURI(), node.getLocalPart());
                     }
-                    bestRule = searchRuleChain(item, context, null, namedNodeChain, ruleSearchState, filter);
+                    if (namedNodeChain != null) {
+                        ruleSearchState = makeRuleSearchState(namedNodeChain, context);
+                        bestRule = searchRuleChain(item, context, null, namedNodeChain, ruleSearchState, filter);
+                    }
                     break;
                 }
                 case Type.TEXT:
@@ -718,26 +772,32 @@ public class SimpleMode extends Mode {
                     throw new AssertionError("Unknown node kind");
             }
 
+            ruleSearchState = makeRuleSearchState(unnamedNodeChain, context);
             bestRule = searchRuleChain(item, context, bestRule, unnamedNodeChain, ruleSearchState, filter);
 
             // Search the list for rules for nodes of unknown node kind
 
+            ruleSearchState = makeRuleSearchState(genericRuleChain, context);
             return searchRuleChain(item, context, bestRule, genericRuleChain, ruleSearchState, filter);
 
         } else if (item instanceof AtomicValue) {
             if (atomicValueRuleChain != null) {
+                ruleSearchState = makeRuleSearchState(atomicValueRuleChain, context);
                 bestRule = searchRuleChain(item, context, bestRule, atomicValueRuleChain, ruleSearchState, filter);
             }
             if (genericRuleChain != null) {
+                ruleSearchState = makeRuleSearchState(genericRuleChain, context);
                 bestRule = searchRuleChain(item, context, bestRule, genericRuleChain, ruleSearchState, filter);
             }
             return bestRule;
 
         } else if (item instanceof Function) {
             if (functionItemRuleChain != null) {
+                ruleSearchState = makeRuleSearchState(functionItemRuleChain, context);
                 bestRule = searchRuleChain(item, context, bestRule, functionItemRuleChain, ruleSearchState, filter);
             }
             if (genericRuleChain != null) {
+                ruleSearchState = makeRuleSearchState(genericRuleChain, context);
                 bestRule = searchRuleChain(item, context, bestRule, genericRuleChain, ruleSearchState, filter);
             }
             return bestRule;
@@ -754,7 +814,7 @@ public class SimpleMode extends Mode {
      * @param item            the item being matched
      * @param context         XPath dynamic context
      * @param bestRule        the best rule so far in terms of precedence and priority (may be null)
-     * @param head            the rule at the head of the chain to be searched
+     * @param chain           the chain to be searched
      * @param ruleSearchState An appropriate ruleState in this mode
      * @param filter          filter used to select which rules are candidates to be searched
      * @return the best match rule found in the chain, or the previous best rule, or null
@@ -762,7 +822,8 @@ public class SimpleMode extends Mode {
      */
 
     protected Rule searchRuleChain(Item item, XPathContext context,
-                                   Rule/*@Nullable*/ bestRule, Rule head, RuleSearchState ruleSearchState, RuleFilter filter) throws XPathException {
+                                   Rule/*@Nullable*/ bestRule, RuleChain chain, RuleSearchState ruleSearchState, RuleFilter filter) throws XPathException {
+        Rule head = chain == null ? null : chain.head();
         while (!(context instanceof XPathContextMajor)) {
             context = context.getCaller();
         }
@@ -803,7 +864,6 @@ public class SimpleMode extends Mode {
     }
 
 
-
     /**
      * Report an ambiguity, that is, the situation where two rules of the same
      * precedence and priority match the same node
@@ -823,7 +883,7 @@ public class SimpleMode extends Mode {
             return;
         }
         String path;
-        String errorCode = c.getController().getExecutable().isAllowXPath30() ? "XTDE0540" : "XTRE0540";
+        String errorCode = "XTDE0540";
 
         if (item instanceof NodeInfo) {
             path = Navigator.getPath((NodeInfo) item);
@@ -837,12 +897,15 @@ public class SimpleMode extends Mode {
         String message;
         if (r1.getAction() == r2.getAction()) {
             message = "Ambiguous rule match for " + path + ". " +
-                    "Matches \"" + showPattern(pat1) + "\" on line " + pat1.getLocation().getLineNumber() + " of " + pat1.getSystemId() +
+                    "Matches \"" + showPattern(pat1) + "\" on line " + pat1.getLocation().getLineNumber() +
+                    " of " + pat1.getLocation().getSystemId() +
                     ", a rule which appears in the stylesheet more than once, because the containing module was included more than once";
         } else {
             message = "Ambiguous rule match for " + path + '\n' +
-                    "Matches both \"" + showPattern(pat1) + "\" on line " + pat1.getLocation().getLineNumber() + " of " + pat1.getSystemId() +
-                    "\nand \"" + showPattern(pat2) + "\" on line " + pat2.getLocation().getLineNumber() + " of " + pat2.getSystemId();
+                    "Matches both \"" + showPattern(pat1) + "\" on line " + pat1.getLocation().getLineNumber() +
+                    " of " + pat1.getLocation().getSystemId() +
+                    "\nand \"" + showPattern(pat2) + "\" on line " + pat2.getLocation().getLineNumber() +
+                    " of " + pat2.getLocation().getSystemId();
         }
 
         XPathException err = new XPathException(message, errorCode);
@@ -860,6 +923,7 @@ public class SimpleMode extends Mode {
 
     /**
      * Prepare for possible streamability - null here, but can be subclassed
+     *
      * @throws XPathException
      */
     public void prepareStreamability() throws XPathException {
@@ -873,20 +937,26 @@ public class SimpleMode extends Mode {
      */
 
     public void allocateAllBindingSlots(final StylesheetPackage pack) {
-        if (getDeclaringComponent().getDeclaringPackage() == pack) {
+        if (getDeclaringComponent().getDeclaringPackage() == pack && !bindingSlotsAllocated) {
             forceAllocateAllBindingSlots(pack, this, getDeclaringComponent().getComponentBindings());
+            bindingSlotsAllocated = true;
         }
     }
 
     public static void forceAllocateAllBindingSlots(
-        final StylesheetPackage pack, final SimpleMode mode, final List<ComponentBinding> bindings) {
+            final StylesheetPackage pack, final SimpleMode mode, final List<ComponentBinding> bindings) {
         final Set<TemplateRule> rulesProcessed = new HashSet<TemplateRule>();
+        final IdentityHashMap<Pattern, Boolean> patternsProcessed = new IdentityHashMap<Pattern, Boolean>();
         try {
             mode.processRules(new RuleAction() {
-                public void processRule (Rule r){
-                    allocateBindingSlotsRecursive(pack, mode, r.getPattern(), bindings);
-                    TemplateRule tr = (TemplateRule)r.getAction();
-                    if (!rulesProcessed.contains(tr)) {
+                public void processRule(Rule r) {
+                    Pattern pattern = r.getPattern();
+                    if (!patternsProcessed.containsKey(pattern)) {
+                        allocateBindingSlotsRecursive(pack, mode, pattern, bindings);
+                        patternsProcessed.put(pattern, true);
+                    }
+                    TemplateRule tr = (TemplateRule) r.getAction();
+                    if (tr.getBody() != null && !rulesProcessed.contains(tr)) {
                         // A rule can appear twice if it uses a union pattern; only process it once
                         allocateBindingSlotsRecursive(pack, mode, tr.getBody(), bindings);
                         rulesProcessed.add(tr);
@@ -897,8 +967,7 @@ public class SimpleMode extends Mode {
             throw new AssertionError(e);
         }
     }
-
-
+    
     /**
      * Compute the streamability of all template rules. No action in Saxon-HE.
      */
@@ -978,9 +1047,9 @@ public class SimpleMode extends Mode {
         // when a template is present in more than one mode.
         RuleAction action = new RuleAction() {
             public void processRule(Rule r) throws XPathException {
-               // if (processedRules.add(r.getAction())) {
-                    r.export(out, isDeclaredStreamable());
-               // }
+                // if (processedRules.add(r.getAction())) {
+                r.export(out, isDeclaredStreamable());
+                // }
             }
         };
 
@@ -1036,7 +1105,7 @@ public class SimpleMode extends Mode {
     }
 
 
-    public void processRuleChains(IntHashMap<Rule> chains, RuleAction action, RuleGroupAction group) throws XPathException {
+    public void processRuleChains(IntHashMap<RuleChain> chains, RuleAction action, RuleGroupAction group) throws XPathException {
         if (chains.size() > 0) {
             if (group != null) {
                 group.start();
@@ -1047,7 +1116,7 @@ public class SimpleMode extends Mode {
                 if (group != null) {
                     group.start(i);
                 }
-                Rule r = chains.get(i);
+                RuleChain r = chains.get(i);
                 processRuleChain(r, action, null);
                 if (group != null) {
                     group.end();
@@ -1065,7 +1134,8 @@ public class SimpleMode extends Mode {
      * @param action an action that is to be applied to all the rules in this Mode
      * @throws XPathException if an error occurs processing any of the rules
      */
-    public void processRuleChain(Rule r, RuleAction action) throws XPathException {
+    public void processRuleChain(RuleChain chain, RuleAction action) throws XPathException {
+        Rule r = chain == null ? null : chain.head();
         while (r != null) {
             action.processRule(r);
             r = r.getNext();
@@ -1073,7 +1143,8 @@ public class SimpleMode extends Mode {
     }
 
 
-    public void processRuleChain(Rule r, RuleAction action, RuleGroupAction group) throws XPathException {
+    public void processRuleChain(RuleChain chain, RuleAction action, RuleGroupAction group) throws XPathException {
+        Rule r = chain == null ? null : chain.head();
         if (r != null) {
             if (group != null) {
                 group.start();
@@ -1106,8 +1177,8 @@ public class SimpleMode extends Mode {
         public int max = 0;
 
         public void processRule(Rule r) {
-            if (r.precedence > max) {
-                max = r.precedence;
+            if (r.getPrecedence() > max) {
+                max = r.getPrecedence();
             }
         }
     }
@@ -1116,6 +1187,7 @@ public class SimpleMode extends Mode {
      * Compute a rank for each rule, as a combination of the precedence and priority, to allow
      * rapid comparison.  This method also checks that there are no conflicts for
      * property values in different xsl:mode declarations
+     *
      * @param start the lowest rank to use
      * @throws XPathException if an error occurs processing the rules
      */
@@ -1225,7 +1297,7 @@ public class SimpleMode extends Mode {
      * @throws XPathException if anything goes wrong
      */
 
-    public void allocateAllPatternSlots() throws XPathException {
+    public void allocateAllPatternSlots() {
         final List<Integer> count = new ArrayList<Integer>(1);  // used to allow inner class to have side-effects
         count.add(0);
         final SlotManager slotManager = new SlotManager();
@@ -1236,7 +1308,11 @@ public class SimpleMode extends Mode {
                 count.set(0, max);
             }
         };
-        processRules(slotAllocator);
+        try {
+            processRules(slotAllocator);
+        } catch (XPathException e) {
+            throw new AssertionError(e);
+        }
         stackFrameSlotsNeeded = count.get(0);
     }
 
@@ -1250,7 +1326,5 @@ public class SimpleMode extends Mode {
     }
 
 
-
-
 }
 
