@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -19,25 +19,13 @@ import net.sf.saxon.om.SequenceIterator;
 public interface UnfailingIterator extends SequenceIterator {
 
     /**
-     * Get the next item in the sequence. <BR>
+     * Get the next item in the sequence.
      *
-     * @return the next Item. If there are no more nodes, return null.
+     * @return the next Item. If there are no more items, return null.
      */
 
     Item next();
 
 
-    /**
-     * Get another iterator over the same sequence of items, positioned at the
-     * start of the sequence. It must be possible to call this method at any time, whether
-     * none, some, or all of the items in the original iterator have been read. The method
-     * is non-destructive: it does not change the state of the original iterator.
-     *
-     * @return a new iterator over the same sequence
-     */
-
-    /*@NotNull*/
-    UnfailingIterator getAnother();
-
 }
 
