@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -30,11 +30,9 @@ import java.util.Iterator;
 import java.util.Set;
 
 /**
- * ElementImpl implements an element with no attributes or namespace declarations.<P>
- * This class is an implementation of NodeInfo. For elements with attributes or
- * namespace declarations, class ElementWithAttributes is used.
- *
- * @author Michael H. Kay
+ * ElementImpl implements an element node in the Linked tree model. Subclasses of ElementImpl,
+ * for holding particular kinds of element, can be defined by using a {@link NodeFactory}
+ * registered with the {@link LinkedTreeBuilder} used to construct the tree.
  */
 
 
@@ -43,7 +41,7 @@ public class ElementImpl extends ParentNodeImpl implements NamespaceResolver {
     private NodeName nodeName;
     private SchemaType type = Untyped.getInstance();
     private AttributeCollection attributeList;      // this excludes namespace attributes
-    /*@Nullable*/ private NamespaceBinding[] namespaceList = null;             // list of namespace codes
+    private NamespaceBinding[] namespaceList = null;             // list of namespace codes
 
     /**
      * Construct an empty ElementImpl
@@ -790,17 +788,6 @@ public class ElementImpl extends ParentNodeImpl implements NamespaceResolver {
         return attributeList == null ? null : attributeList.getValue(uri, localName);
     }
 
-    /**
-     * Get the value of the attribute with a given fingerprint.
-     *
-     * @param fp the fingerprint of the required attribute
-     * @return the string value of the attribute if present, or null if absent
-     */
-    @Override
-    public String getAttributeValue(int fp) {
-        return attributeList == null ? null : attributeList.getValueByFingerprint(fp);
-    }
-
     /**
      * Determine whether this node has the is-id property
      *
