@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -10,11 +10,16 @@ package net.sf.saxon.tree.linked;
 import net.sf.saxon.om.NodeInfo;
 import net.sf.saxon.pattern.AnyNodeTest;
 import net.sf.saxon.pattern.NodeTest;
-import net.sf.saxon.tree.iter.*;
+import net.sf.saxon.tree.iter.ArrayIterator;
+import net.sf.saxon.tree.iter.AxisIterator;
+import net.sf.saxon.tree.iter.EmptyIterator;
+import net.sf.saxon.tree.iter.SingleNodeIterator;
 import net.sf.saxon.tree.util.FastStringBuffer;
 import net.sf.saxon.tree.util.Navigator;
 import net.sf.saxon.type.Type;
 
+import java.util.Arrays;
+
 /**
  * ParentNodeImpl is an implementation of a non-leaf node (specifically, an Element node
  * or a Document node)
@@ -37,7 +42,7 @@ abstract class ParentNodeImpl extends NodeImpl {
      * consecutive. In the current implementation, parent nodes (elements and document nodes) have a zero
      * least-significant word, while namespaces, attributes, text nodes, comments, and PIs have
      * the top word the same as their owner and the bottom half reflecting their relative position.
-     * For nodes added by XQUery Update, the sequence number is -1L
+     * For nodes added by XQuery Update, the sequence number is -1L
      *
      * @return the sequence number if there is one, or -1L otherwise.
      */
@@ -109,8 +114,7 @@ abstract class ParentNodeImpl extends NodeImpl {
             }
         } else {
             if (test == null || test == AnyNodeTest.getInstance()) {
-                return new AxisIteratorOverSequence(
-                        new ArrayIterator((NodeImpl[]) children));
+                return new ArrayIterator.OfNodes((NodeImpl[]) children);
             } else {
                 return new ChildEnumeration(this, test);
             }
@@ -126,9 +130,13 @@ abstract class ParentNodeImpl extends NodeImpl {
 
     /*@Nullable*/
     public final NodeImpl getFirstChild() {
-        if (children == null) return null;
-        if (children instanceof NodeImpl) return (NodeImpl) children;
-        return ((NodeImpl[]) children)[0];
+        if (children == null) {
+            return null;
+        } else if (children instanceof NodeImpl) {
+            return (NodeImpl) children;
+        } else {
+            return ((NodeImpl[]) children)[0];
+        }
     }
 
     /**
@@ -230,9 +238,7 @@ abstract class ParentNodeImpl extends NodeImpl {
         if (j == c2.length) {
             return c2;
         } else {
-            NodeImpl[] c3 = new NodeImpl[j];
-            System.arraycopy(c2, 0, c3, 0, j);
-            return c3;
+            return Arrays.copyOf(c2, j);
         }
     }
 
@@ -287,9 +293,7 @@ abstract class ParentNodeImpl extends NodeImpl {
             c = (NodeImpl[]) children;
         }
         if (index >= c.length) {
-            NodeImpl[] kids = new NodeImpl[c.length * 2];
-            System.arraycopy(c, 0, kids, 0, c.length);
-            c = kids;
+            c = Arrays.copyOf(c, c.length*2);
         }
         c[index] = node;
         node.setRawParent(this);
@@ -376,9 +380,7 @@ abstract class ParentNodeImpl extends NodeImpl {
                     ((TextImpl) source[adjacent]).replaceStringValue(
                             ((TextImpl) children).getStringValue() + source[adjacent].getStringValue());
                 }
-                NodeImpl[] n2 = new NodeImpl[source.length];
-                System.arraycopy(source, 0, n2, 0, source.length);
-                children = n2;
+                children = Arrays.copyOf(source, source.length);
             } else {
                 NodeImpl[] n2 = new NodeImpl[source.length + 1];
                 if (index == 0) {
@@ -484,9 +486,7 @@ abstract class ParentNodeImpl extends NodeImpl {
                 children = ((NodeImpl[]) children)[0];
             }
         } else {
-            NodeImpl[] kids = new NodeImpl[size];
-            System.arraycopy(children, 0, kids, 0, size);
-            children = kids;
+            children = Arrays.copyOf((NodeImpl[])children, size);
         }
     }
 
