@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,7 +8,6 @@
 package net.sf.saxon.tree.linked;
 
 import net.sf.saxon.pattern.NodeTest;
-import net.sf.saxon.tree.iter.AxisIterator;
 
 final class PrecedingSiblingEnumeration extends TreeEnumeration {
 
@@ -21,14 +20,5 @@ final class PrecedingSiblingEnumeration extends TreeEnumeration {
         next = next.getPreviousSibling();
     }
 
-    /**
-     * Get another enumeration of the same nodes
-     */
-
-    /*@NotNull*/
-    public AxisIterator getAnother() {
-        return new PrecedingSiblingEnumeration(start, nodeTest);
-    }
-
 }
 
