@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -9,7 +9,6 @@ package net.sf.saxon.tree.tiny;
 
 import net.sf.saxon.om.NodeInfo;
 import net.sf.saxon.pattern.NodeTest;
-import net.sf.saxon.tree.iter.AxisIterator;
 import net.sf.saxon.tree.iter.AxisIteratorImpl;
 
 /**
@@ -17,13 +16,13 @@ import net.sf.saxon.tree.iter.AxisIteratorImpl;
  * starting at a given node. The start node will never be the root.
  */
 
-final class AncestorIterator extends AxisIteratorImpl {
+final public class AncestorIterator extends AxisIteratorImpl {
 
-    private TinyNodeImpl startNode;
+    private NodeInfo startNode;
     private NodeInfo current;
     private NodeTest test;
 
-    public AncestorIterator(TinyNodeImpl node, NodeTest nodeTest) {
+    public AncestorIterator(NodeInfo node, NodeTest nodeTest) {
         test = nodeTest;
         startNode = node;
         current = startNode;
@@ -41,14 +40,5 @@ final class AncestorIterator extends AxisIteratorImpl {
         return current = node;
     }
 
-    /**
-     * Get another enumeration of the same nodes
-     */
-
-    /*@NotNull*/
-    public AxisIterator getAnother() {
-        return new AncestorIterator(startNode, test);
-    }
-
 }
 
