@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -27,15 +27,15 @@ import java.util.*;
  * A data structure to hold the contents of a tree. As the name implies, this implementation
  * of the data model is optimized for size, and for speed of creation: it minimizes the number
  * of Java objects used.
- *
+ * <p>
  * <p>It can be used to represent a tree that is rooted at a document node, or one that is rooted
  * at an element node.</p>
- *
+ * <p>
  * <p>From Saxon 9.7, as a consequence of bug 2220, it is used only to hold a single tree, whose
  * root is always node number zero.</p>
  */
 
-public final class TinyTree extends GenericTreeInfo {
+public final class TinyTree extends GenericTreeInfo implements NodeVectorTree {
 
     /*@NotNull*/
     private static final String[] EMPTY_STRING_ARRAY = new String[0];
@@ -159,10 +159,21 @@ public final class TinyTree extends GenericTreeInfo {
     // a boolean that is set to true if the document declares a namespace other than the XML namespace
     protected boolean usesNamespaces = false;
 
+    protected PrefixPool prefixPool = new PrefixPool();
+
     private TinyDocumentImpl root;
     private HashMap<String, NodeInfo> idTable;
     protected HashMap<String, String[]> entityTable;
 
+    private NodeInfo copiedFrom;
+
+    /**
+     * Temporary flag introduced in Saxon 9.8.0.5 to enable or disable fast-path code for copying
+     * element nodes from one TinyTree to another. This code is currently disabled by default until
+     * it has been more thoroughly tested.
+     */
+    public static boolean useFastCopy = false;
+
     /**
      * Create a tree with a specified initial size
      *
@@ -171,8 +182,6 @@ public final class TinyTree extends GenericTreeInfo {
      */
 
     public TinyTree(/*@NotNull*/ Configuration config, Statistics statistics) {
-        //System.err.println("TinyTree.new() (initial size " + nodes + ", treesCreated = " + treesCreated + ")");
-
         super(config);
 
         int nodes = (int) statistics.getAverageNodes() + 1;
@@ -215,110 +224,97 @@ public final class TinyTree extends GenericTreeInfo {
         addNamespace(0, NamespaceBinding.XML);
     }
 
-    private void ensureNodeCapacity(short kind) {
-        if (nodeKind.length < numberOfNodes + 1) {
+    private void ensureNodeCapacity(short kind, int needed) {
+        if (nodeKind.length < numberOfNodes + needed) {
             //System.err.println("Number of nodes = " + numberOfNodes);
-            int k = kind == Type.STOPPER ? numberOfNodes + 1 : numberOfNodes * 2;
+            int k = kind == Type.STOPPER ? numberOfNodes + 1 : Math.max(numberOfNodes * 2, numberOfNodes + needed);
 
-            byte[] nodeKind2 = new byte[k];
-            int[] next2 = new int[k];
-            short[] depth2 = new short[k];
-            int[] alpha2 = new int[k];
-            int[] beta2 = new int[k];
-            int[] nameCode2 = new int[k];
-
-            System.arraycopy(nodeKind, 0, nodeKind2, 0, numberOfNodes);
-            System.arraycopy(next, 0, next2, 0, numberOfNodes);
-            System.arraycopy(depth, 0, depth2, 0, numberOfNodes);
-            System.arraycopy(alpha, 0, alpha2, 0, numberOfNodes);
-            System.arraycopy(beta, 0, beta2, 0, numberOfNodes);
-            System.arraycopy(nameCode, 0, nameCode2, 0, numberOfNodes);
-
-            nodeKind = nodeKind2;
-            next = next2;
-            depth = depth2;
-            alpha = alpha2;
-            beta = beta2;
-            nameCode = nameCode2;
+            nodeKind = Arrays.copyOf(nodeKind, k);
+            next = Arrays.copyOf(next, k);
+            depth = Arrays.copyOf(depth, k);
+            alpha = Arrays.copyOf(alpha, k);
+            beta = Arrays.copyOf(beta, k);
+            nameCode = Arrays.copyOf(nameCode, k);
 
             if (typeArray != null) {
-                SchemaType[] typeArray2 = new SchemaType[k];
-                System.arraycopy(typeArray, 0, typeArray2, 0, numberOfNodes);
-                typeArray = typeArray2;
+                typeArray = Arrays.copyOf(typeArray, k);
             }
 
             if (typedValueArray != null) {
-                @SuppressWarnings({"unchecked"})
-                AtomicSequence[] typedValueArray2 = new AtomicSequence[k];
-                //noinspection SuspiciousSystemArraycopy
-                System.arraycopy(typedValueArray, 0, typedValueArray2, 0, numberOfNodes);
-                typedValueArray = typedValueArray2;
+                typedValueArray = Arrays.copyOf(typedValueArray, k);
             }
 
             if (lineNumbers != null) {
-                int[] lines2 = new int[k];
-                System.arraycopy(lineNumbers, 0, lines2, 0, numberOfNodes);
-                lineNumbers = lines2;
-                int[] columns2 = new int[k];
-                System.arraycopy(columnNumbers, 0, columns2, 0, numberOfNodes);
-                columnNumbers = columns2;
+                lineNumbers = Arrays.copyOf(lineNumbers, k);
+                columnNumbers = Arrays.copyOf(columnNumbers, k);
             }
         }
     }
 
-    private void ensureAttributeCapacity() {
-        if (attParent.length < numberOfAttributes + 1) {
-            int k = numberOfAttributes * 2;
+    private void ensureAttributeCapacity(int needed) {
+        if (attParent.length < numberOfAttributes + needed) {
+            int k = Math.max(numberOfAttributes + needed, numberOfAttributes * 2);
             if (k == 0) {
-                k = 10;
+                k = 10 + needed;
             }
 
-            int[] attParent2 = new int[k];
-            int[] attCode2 = new int[k];
-            CharSequence[] attValue2 = new String[k];
-
-            System.arraycopy(attParent, 0, attParent2, 0, numberOfAttributes);
-            System.arraycopy(attCode, 0, attCode2, 0, numberOfAttributes);
-            System.arraycopy(attValue, 0, attValue2, 0, numberOfAttributes);
-
-            attParent = attParent2;
-            attCode = attCode2;
-            attValue = attValue2;
+            attParent = Arrays.copyOf(attParent, k);
+            attCode = Arrays.copyOf(attCode, k);
+            attValue = Arrays.copyOf(attValue, k);
 
             if (attType != null) {
-                SimpleType[] attType2 = new SimpleType[k];
-                System.arraycopy(attType, 0, attType2, 0, numberOfAttributes);
-                attType = attType2;
+                attType = Arrays.copyOf(attType, k);
             }
 
             if (attTypedValue != null) {
-                @SuppressWarnings({"unchecked"})
-                AtomicSequence[] attTypedValue2 = new AtomicSequence[k];
-                System.arraycopy(attTypedValue, 0, attTypedValue2, 0, numberOfAttributes);
-                attTypedValue = attTypedValue2;
+                attTypedValue = Arrays.copyOf(attTypedValue, k);
             }
         }
     }
 
-    private void ensureNamespaceCapacity() {
-        if (namespaceParent.length < numberOfNamespaces + 1) {
-            int k = numberOfNamespaces * 2;
+    private void ensureNamespaceCapacity(int needed) {
+        if (namespaceParent.length < numberOfNamespaces + needed) {
+            int k = Math.max(numberOfNamespaces * 2, numberOfNamespaces + needed);
             if (k == 0) {
                 k = 10;
             }
-
-            int[] namespaceParent2 = new int[k];
-            NamespaceBinding[] namespaceCode2 = new NamespaceBinding[k];
-
-            System.arraycopy(namespaceParent, 0, namespaceParent2, 0, numberOfNamespaces);
-            System.arraycopy(namespaceBinding, 0, namespaceCode2, 0, numberOfNamespaces);
-
-            namespaceParent = namespaceParent2;
-            namespaceBinding = namespaceCode2;
+            namespaceParent = Arrays.copyOf(namespaceParent, k);
+            namespaceBinding = Arrays.copyOf(namespaceBinding, k);
         }
     }
 
+    /**
+     * Get the prefix pool
+     *
+     * @return the prefix pool
+     */
 
+    public PrefixPool getPrefixPool() {
+        return prefixPool;
+    }
+
+    /**
+     * Declare that this tree was produced as a copy of another tree, and identify
+     * the root node of that tree
+     *
+     * @param copiedFrom the root of the tree from which this one was copied
+     */
+
+    public void setCopiedFrom(NodeInfo copiedFrom) {
+        this.copiedFrom = copiedFrom;
+    }
+
+    /**
+     * Declare that this tree was produced as a copy of another tree, and identify
+     * the root node of that tree
+     *
+     * @return the root of the tree from which this one was copied, or null if this
+     * does not apply
+     */
+
+    public NodeInfo getCopiedFrom() {
+        return this.copiedFrom;
+    }
 
     /**
      * Add a document node to the tree. The data structure since 9.7 can only hold one tree, therefore at
@@ -346,7 +342,7 @@ public final class TinyTree extends GenericTreeInfo {
      */
     int addNode(short kind, int depth, int alpha, int beta, int nameCode) {
 
-        ensureNodeCapacity(kind);
+        ensureNodeCapacity(kind, 1);
         nodeKind[numberOfNodes] = (byte) kind;
         this.depth[numberOfNodes] = (short) depth;
         this.alpha[numberOfNodes] = alpha;
@@ -404,20 +400,14 @@ public final class TinyTree extends GenericTreeInfo {
      * Condense the tree: release unused memory. This is done after the full tree has been built.
      * The method makes a pragmatic judgement as to whether it is worth reclaiming space; this is
      * only done when the constructed tree is very small compared with the space allocated.
+     *
      * @param statistics represents the family of trees to which this tree belongs; statistics for the
-     * size of the tree are recorded in this Statistics object
+     *                   size of the tree are recorded in this Statistics object
      */
 
     void condense(Statistics statistics) {
         //System.err.println("TinyTree.condense() " + this + " roots " + rootIndexUsed + " nodes " + numberOfNodes + " capacity " + nodeKind.length);
 
-        // If there are already two trees in this forest, the chances are that more will be added. In this
-        // case we don't want to condense the arrays because we will only have to expand them again, which gets
-        // increasingly expensive as they grow larger.
-
-//        if (rootIndexUsed > 1) {
-//            return;
-//        }
         if (numberOfNodes * 3 < nodeKind.length ||
                 (nodeKind.length - numberOfNodes > 20000)) {
 
@@ -467,6 +457,8 @@ public final class TinyTree extends GenericTreeInfo {
             namespaceBinding = Arrays.copyOf(namespaceBinding, numberOfNamespaces);
         }
 
+        prefixPool.condense();
+
         statistics.updateStatistics(numberOfNodes, numberOfAttributes, numberOfNamespaces, charBuffer.length());
 //        System.err.println("STATS: " + averageNodes + ", " + averageAttributes + ", "
 //                + averageNamespaces + ", " + averageCharacters);
@@ -482,8 +474,8 @@ public final class TinyTree extends GenericTreeInfo {
     /**
      * Set the type annotation of an element node
      *
-     * @param nodeNr   the node whose type annotation is to be set
-     * @param type the type annotation
+     * @param nodeNr the node whose type annotation is to be set
+     * @param type   the type annotation
      */
 
     void setElementAnnotation(int nodeNr, SchemaType type) {
@@ -532,9 +524,8 @@ public final class TinyTree extends GenericTreeInfo {
      *
      * @param element the element node
      * @return the typed value of the node (a Value whose items are AtomicValue instances)
-     * @throws net.sf.saxon.trans.XPathException
-     *          if a dynamic error occurs, for example if the node is
-     *          an element annotated with a type that has element-only content
+     * @throws net.sf.saxon.trans.XPathException if a dynamic error occurs, for example if the node is
+     *                                           an element annotated with a type that has element-only content
      */
 
     /*@Nullable*/
@@ -574,8 +565,7 @@ public final class TinyTree extends GenericTreeInfo {
      *
      * @param nodeNr the node number of the element node
      * @return the typed value of the node
-     * @throws net.sf.saxon.trans.XPathException
-     *          if the eement has no typed value
+     * @throws net.sf.saxon.trans.XPathException if the eement has no typed value
      */
 
     /*@Nullable*/
@@ -618,8 +608,7 @@ public final class TinyTree extends GenericTreeInfo {
      *               will be materialized only if it is needed.
      * @param nodeNr the node number of the attribute node
      * @return the typed value of the node
-     * @throws net.sf.saxon.trans.XPathException
-     *          if an error is found
+     * @throws net.sf.saxon.trans.XPathException if an error is found
      */
 
     public AtomicSequence getTypedValueOfAttribute(/*@Nullable*/ TinyAttributeImpl att, int nodeNr) throws XPathException {
@@ -681,6 +670,36 @@ public final class TinyTree extends GenericTreeInfo {
         return nameCode[nodeNr];
     }
 
+    /**
+     * Get the fingerprint for a given node, which must be a document, element,
+     * text, comment, or processing instruction node
+     *
+     * @param nodeNr the node number
+     * @return the name code
+     */
+
+    public int getFingerprint(int nodeNr) {
+        int nc = nameCode[nodeNr];
+        return nc == -1 ? -1 : nc & NamePool.FP_MASK;
+    }
+
+
+    /**
+     * Get the prefix for a given element node
+     *
+     * @param nodeNr the node number
+     * @return the prefix. Return "" for an unprefixed element node. Result is
+     * undefined for a non-element node.
+     */
+
+    public String getPrefix(int nodeNr) {
+        int code = nameCode[nodeNr] >> 20;
+        if (code <= 0) {
+            return code == 0 ? "" : null;
+        }
+        return prefixPool.getPrefix(code);
+    }
+
     /**
      * On demand, make an index for quick access to preceding-sibling nodes
      */
@@ -716,10 +735,10 @@ public final class TinyTree extends GenericTreeInfo {
 
 
     void addAttribute(/*@NotNull*/ NodeInfo root, int parent, int nameCode, SimpleType type, CharSequence attValue, int properties) {
-        ensureAttributeCapacity();
+        ensureAttributeCapacity(1);
         attParent[numberOfAttributes] = parent;
         attCode[numberOfAttributes] = nameCode;
-        this.attValue[numberOfAttributes] = attValue;
+        this.attValue[numberOfAttributes] = attValue.toString();
 
         if (!type.equals(BuiltInAtomicType.UNTYPED_ATOMIC)) {
             initializeAttributeTypeCodes();
@@ -819,6 +838,7 @@ public final class TinyTree extends GenericTreeInfo {
 
     /**
      * Mark an attribute as resulting from expansion of attribute defaults
+     *
      * @param attNr the attribute number
      */
 
@@ -831,6 +851,7 @@ public final class TinyTree extends GenericTreeInfo {
 
     /**
      * Ask whether an attribute results from expansion of attribute defaults
+     *
      * @param attNr the attribute number
      * @return true if this attribute resulted from expansion of default or fixed values defined
      * in a schema. Note that this property will only be set if both the configuration properties
@@ -845,8 +866,8 @@ public final class TinyTree extends GenericTreeInfo {
     /**
      * Index an element of type xs:ID
      *
-     * @param root    the root node of the document
-     * @param nodeNr  the element of type xs:ID
+     * @param root   the root node of the document
+     * @param nodeNr the element of type xs:ID
      */
 
     public void indexIDElement(/*@NotNull*/ NodeInfo root, int nodeNr) {
@@ -857,6 +878,21 @@ public final class TinyTree extends GenericTreeInfo {
         }
     }
 
+    /**
+     * Ask whether, somewhere in the tree, there is an attribute xml:space="preserve"
+     *
+     * @return true if some element in the tree has an xml:space attribute with the value preserve
+     */
+
+    public boolean hasXmlSpacePreserveAttribute() {
+        for (int i = 0; i < numberOfAttributes; i++) {
+            if ((attCode[i] & NamePool.FP_MASK) == StandardNames.XML_SPACE && "preserve".equals(attValue[i].toString())) {
+                return true;
+            }
+        }
+        return false;
+    }
+
     /**
      * Add a namespace node to the current element
      *
@@ -865,7 +901,7 @@ public final class TinyTree extends GenericTreeInfo {
      */
     void addNamespace(int parent, /*@NotNull*/ NamespaceBinding binding) {
 
-        ensureNamespaceCapacity();
+        ensureNamespaceCapacity(1);
         namespaceParent[numberOfNamespaces] = parent;
         namespaceBinding[numberOfNamespaces] = binding;
 
@@ -888,7 +924,7 @@ public final class TinyTree extends GenericTreeInfo {
     public final TinyNodeImpl getNode(int nr) {
         switch (nodeKind[nr]) {
             case Type.DOCUMENT:
-                return (TinyDocumentImpl)getRootNode();
+                return (TinyDocumentImpl) getRootNode();
             case Type.ELEMENT:
                 return new TinyElementImpl(this, nr);
             case Type.TEXT:
@@ -1099,19 +1135,6 @@ public final class TinyTree extends GenericTreeInfo {
         return false;
     }
 
-    /**
-     * Ask whether, somewhere in the tree, there is an attribute xml:space="preserve"
-     * @return true if some element in the tree has an xml:space attribute with the value preserve
-     */
-
-    public boolean hasXmlSpacePreserveAttribute() {
-        for (int i=0; i<numberOfAttributes; i++) {
-            if ((attCode[i] & NamePool.FP_MASK) == StandardNames.XML_SPACE && "preserve".equals(attValue[i].toString())) {
-                return true;
-            }
-        }
-        return false;
-    }
 
     /**
      * Set the system id of an element in the document. This identifies the external entity containing
@@ -1219,7 +1242,7 @@ public final class TinyTree extends GenericTreeInfo {
         if (columnNumbers != null) {
             // find the nearest preceding node that has a known column number, and return it
             for (int i = sequence; i >= 0; i--) {
-                int c = columnNumbers[sequence];
+                int c = columnNumbers[i];
                 if (c > 0) {
                     return c;
                 }
@@ -1230,6 +1253,7 @@ public final class TinyTree extends GenericTreeInfo {
 
     /**
      * Set an element node to be marked as nilled
+     *
      * @param nodeNr the node number to be marked as nilled
      */
 
@@ -1356,14 +1380,22 @@ public final class TinyTree extends GenericTreeInfo {
      * Produce diagnostic print of main tree arrays
      */
 
+    @SuppressWarnings("UseOfSystemOutOrSystemErr")
     public void diagnosticDump() {
         NamePool pool = getNamePool();
         System.err.println("    node    kind   depth    next   alpha    beta    name    type");
         for (int i = 0; i < numberOfNodes; i++) {
+            String eqName = "";
+            if (nameCode[i] != -1) {
+                try {
+                    eqName = pool.getEQName(nameCode[i]);
+                } catch (Exception err) {
+                    eqName = "#" + nameCode[1];
+                }
+            }
             System.err.println(n8(i) + n8(nodeKind[i]) + n8(depth[i]) + n8(next[i]) +
-                    n8(alpha[i]) + n8(beta[i]) + n8(nameCode[i]) +
-                    n8(getTypeAnnotation(i)) +
-                    (nameCode[i] == -1 ? "" : " " + pool.getDisplayName(nameCode[i])));
+                                       n8(alpha[i]) + n8(beta[i]) + n8(nameCode[i]) +
+                                       n8(getTypeAnnotation(i)) + " " + eqName);
         }
         System.err.println("    attr  parent    name    value");
         for (int i = 0; i < numberOfAttributes; i++) {
@@ -1410,7 +1442,7 @@ public final class TinyTree extends GenericTreeInfo {
 
     public void showSize() {
         System.err.println("Tree size: " + numberOfNodes + " nodes, " + charBuffer.length() + " characters, " +
-                numberOfAttributes + " attributes");
+                                   numberOfAttributes + " attributes");
     }
 
     /**
@@ -1508,7 +1540,7 @@ public final class TinyTree extends GenericTreeInfo {
      * Get the array holding alpha information
      *
      * @return an array of integers, whose meaning depends on the node kind. For elements it is a pointer
-     *         to the first attribute, for text, comment, and processing instruction nodes it is a pointer to the content
+     * to the first attribute, for text, comment, and processing instruction nodes it is a pointer to the content
      */
 
     public int[] getAlphaArray() {
@@ -1519,7 +1551,7 @@ public final class TinyTree extends GenericTreeInfo {
      * Get the array holding beta information
      *
      * @return an array of integers, whose meaning depends on the node kind. For elements it is a pointer
-     *         to the first namespace declaration
+     * to the first namespace declaration
      */
 
     public int[] getBetaArray() {
@@ -1609,6 +1641,149 @@ public final class TinyTree extends GenericTreeInfo {
         return namespaceParent;
     }
 
+    /**
+     * Ask whether the tree contains any namespace declarations
+     *
+     * @return true if there is a namespace declaration (other than the XML namespace) anywhere
+     * in the tree
+     */
+
+    public boolean isUsesNamespaces() {
+        return usesNamespaces;
+    }
+
+    /**
+     * Bulk copy an element node from another TinyTree. Type annotations will always be
+     * stripped.
+     * @param source        the source tree
+     * @param nodeNr        the element node to be deep-copied
+     */
+
+    public void bulkCopy(TinyTree source, int nodeNr, int currentDepth) {
+        //System.err.println(" **** doing bulk copy **** ");
+        int end = source.next[nodeNr];
+        while (end < nodeNr && end >= 0) {
+            end = source.next[end];
+        }
+        if (end == -1) {
+            end = source.numberOfNodes - 1;
+        }
+        int length = end - nodeNr;
+        ensureNodeCapacity(Type.ELEMENT, length);
+        System.arraycopy(source.nodeKind, nodeNr, nodeKind, numberOfNodes, length);
+        int depthDiff = currentDepth - source.depth[nodeNr];
+        for (int i = 0; i < length; i++) {
+            int from = nodeNr + i;
+            int to = numberOfNodes + i;
+            depth[to] = (short) (source.depth[from] + depthDiff);
+            next[to] = source.next[from] + (to - from);
+            switch (source.nodeKind[from]) {
+                case Type.ELEMENT: {
+                    nameCode[to] = (source.nameCode[from] & NamePool.FP_MASK) |
+                            (prefixPool.obtainPrefixCode(source.getPrefix(from)) << 20);
+                    int firstAtt = source.alpha[from];
+                    if (firstAtt >= 0) {
+                        int lastAtt = firstAtt;
+                        while (lastAtt < source.numberOfAttributes && source.attParent[lastAtt] == from) {
+                            lastAtt++;
+                        }
+                        int atts = lastAtt - firstAtt;
+                        ensureAttributeCapacity(atts);
+                        int aFrom = source.alpha[from];
+                        int aTo = numberOfAttributes;
+                        alpha[to] = aTo;
+                        System.arraycopy(source.attValue, firstAtt, attValue, aTo, atts);
+                        Arrays.fill(attParent, aTo, aTo + atts, to);
+                        for (int a = 0; a < atts; a++, aFrom++, aTo++) {
+                            int attNameCode = attCode[aTo] = source.attCode[aFrom];
+                            if (NamePool.isPrefixed(attNameCode)) {
+                                String prefix = source.prefixPool.getPrefix(attNameCode >> 20);
+                                attCode[aTo] |= (prefixPool.obtainPrefixCode(prefix) << 20);
+                            }
+                            if (source.isIdAttribute(aFrom)) {
+                                registerID(getNode(to), source.attValue[aFrom].toString());
+                            }
+                            if (source.isIdrefAttribute(aFrom)) {
+                                if (idRefAttributes == null) {
+                                    idRefAttributes = new IntHashSet();
+                                }
+                                idRefAttributes.add(aTo);
+                            }
+                        }
+                        numberOfAttributes += atts;
+                    } else {
+                        alpha[to] = -1;
+                    }
+                    int firstNS = source.beta[from];
+                    if (firstNS >= 0) {
+                        int lastNS = firstNS;
+                        while (lastNS < source.numberOfNamespaces && source.namespaceParent[lastNS] == from) {
+                            lastNS++;
+                        }
+                        int nrOfNamespaces = lastNS - firstNS;
+                        ensureNamespaceCapacity(nrOfNamespaces);
+                        int nTo = numberOfNamespaces;
+                        beta[to] = nTo;
+                        System.arraycopy(source.namespaceBinding, firstNS, namespaceBinding, nTo, nrOfNamespaces);
+                        Arrays.fill(namespaceParent, nTo, nTo + nrOfNamespaces, to);
+                        numberOfNamespaces += nrOfNamespaces;
+                    } else {
+                        beta[to] = -1;
+                    }
+                    break;
+                }
+                case Type.TEXT: {
+                    int start = source.alpha[from];
+                    int len = source.beta[from];
+                    nameCode[to] = -1;
+                    alpha[to] = charBuffer.length();
+                    appendChars(source.charBuffer.subSequence(start, start + len));
+                    beta[to] = len;
+                    break;
+                }
+                case Type.WHITESPACE_TEXT: {
+                    nameCode[to] = -1;
+                    alpha[to] = source.alpha[from];
+                    beta[to] = source.beta[from];
+                    break;
+                }
+                case Type.COMMENT: {
+                    int start = source.alpha[from];
+                    int len = source.beta[from];
+                    nameCode[to] = -1;
+                    CharSequence text = source.commentBuffer.subSequence(start, start+len);
+                    if (commentBuffer == null) {
+                        commentBuffer = new FastStringBuffer(FastStringBuffer.C256);
+                    }
+                    alpha[to] = commentBuffer.length();
+                    commentBuffer.append(text);
+                    beta[to] = len;
+                    break;
+                }
+                case Type.PROCESSING_INSTRUCTION:
+                    int start = source.alpha[from];
+                    int len = source.beta[from];
+                    nameCode[to] = source.nameCode[from];
+                    CharSequence text = source.commentBuffer.subSequence(start, start + len);
+                    if (commentBuffer == null) {
+                        commentBuffer = new FastStringBuffer(FastStringBuffer.C256);
+                    }
+                    alpha[to] = commentBuffer.length();
+                    commentBuffer.append(text);
+                    beta[to] = len;
+                    break;
+
+                case Type.PARENT_POINTER:
+                    nameCode[to] = -1;
+                    alpha[to] = source.alpha[from] + (to - from);
+                    beta[to] = -1;
+                    break;
+                default:
+                    break;
+            }
+        }
+        numberOfNodes += length;
+    }
 
 }
 
