@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -111,7 +111,7 @@ public final class Navigator {
         } else if (localName == null) {
             return new NamespaceTest(pool, nodeKind, uri);
         } else {
-            int fp = pool.allocate("", uri, localName);
+            int fp = pool.allocateFingerprint(uri, localName);
             return new NameTest(nodeKind, fp, pool);
         }
     }
@@ -289,8 +289,8 @@ public final class Navigator {
      */
 
     public static boolean haveSameName(NodeInfo n1, NodeInfo n2) {
-        if (n1 instanceof FingerprintedNode && n2 instanceof FingerprintedNode) {
-            return ((FingerprintedNode)n1).getFingerprint() == ((FingerprintedNode) n2).getFingerprint();
+        if (n1.hasFingerprint() && n2.hasFingerprint()) {
+            return n1.getFingerprint() == n2.getFingerprint();
         } else {
             return n1.getLocalPart().equals(n2.getLocalPart()) &&
                 n1.getURI().equals(n2.getURI());
@@ -423,7 +423,7 @@ public final class Navigator {
         // we've found the ancestor to count from
 
         SequenceIterator preceding =
-                target.iterateAxis(AxisInfo.PRECEDING_SIBLING, (NodeTest) count.getItemType());
+                target.iterateAxis(AxisInfo.PRECEDING_SIBLING, getNodeTestForPattern(count));
         // pass the filter condition down to the axis enumeration where possible
         boolean alreadyChecked = count instanceof NodeTestPattern;
         int i = 1;
@@ -498,7 +498,7 @@ public final class Navigator {
         // Pass part of the filtering down to the axis iterator if possible
         NodeTest filter;
         if (from == null) {
-            filter = (NodeTest) count.getItemType();
+            filter = getNodeTestForPattern(count);
         } else if (from.getUType() == UType.ELEMENT && count.getUType() == UType.ELEMENT) {
             filter = NodeKindTest.ELEMENT;
         } else {
@@ -595,6 +595,21 @@ public final class Navigator {
         return v;
     }
 
+    /**
+     * Get a NodeTest to use as a filter for nodes, given a pattern.
+     */
+
+    private static NodeTest getNodeTestForPattern(Pattern pattern) {
+        ItemType type = pattern.getItemType();
+        if (type instanceof NodeTest) {
+            return (NodeTest) type;
+        } else if (pattern.getUType().overlaps(UType.ANY_NODE)) {
+            return AnyNodeTest.getInstance();
+        } else {
+            return ErrorType.getInstance();
+        }
+    }
+
     /**
      * Generic (model-independent) implementation of deep copy algorithm for nodes.
      * This is available for use by any node implementations that choose to use it.
@@ -1057,10 +1072,12 @@ public final class Navigator {
      * @param max      the maximum number of nodes to be counted
      * @return the number of preceding siblings that satisfy the node test, plus one, unless the
      *         number exceeds max, in which case return some number greater than or equal to max.
-     * @throws net.sf.saxon.trans.XPathException if a dynamic error occurs
      */
 
-    public static int getSiblingPosition(NodeInfo node, NodeTest nodeTest, int max) throws XPathException {
+    public static int getSiblingPosition(NodeInfo node, NodeTest nodeTest, int max) {
+        if (node instanceof SiblingCountingNode && nodeTest instanceof AnyNodeTest) {
+            return ((SiblingCountingNode)node).getSiblingPosition();
+        }
         AxisIterator prev = node.iterateAxis(AxisInfo.PRECEDING_SIBLING, nodeTest);
         int count = 1;
         while (prev.next() != null) {
@@ -1108,10 +1125,6 @@ public final class Navigator {
             }
         }
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new AxisFilter(base.getAnother(), nodeTest);
-        }
     }
 
     /**
@@ -1146,10 +1159,6 @@ public final class Navigator {
             }
         }
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new EmptyTextFilter(base.getAnother());
-        }
     }
 
 
@@ -1188,11 +1197,6 @@ public final class Navigator {
             return current = current == null ? null : current.getParent();
         }
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new AncestorEnumeration(start, includeSelf);
-        }
-
     } // end of class AncestorEnumeration
 
     /**
@@ -1272,17 +1276,13 @@ public final class Navigator {
                         if (children instanceof ReversibleIterator) {
                             children = (AxisIterator) ((ReversibleIterator) children).getReverseIterator();
                         } else {
-                            List<NodeInfo> list = new ArrayList<NodeInfo>(20);
+                            LinkedList<NodeInfo> list = new LinkedList<NodeInfo>();
                             AxisIterator forwards = start.iterateAxis(AxisInfo.CHILD);
-                            while (true) {
-                                NodeInfo n = forwards.next();
-                                if (n == null) {
-                                    break;
-                                }
-                                list.add(n);
+                            NodeInfo n;
+                            while ((n = forwards.next()) != null) {
+                                list.addFirst(n);
                             }
-                            children = new AxisIteratorOverSequence(
-                                    new ReverseListIterator(list));
+                            children = new ListIterator.OfNodes(list);
                         }
                     }
                 } else {
@@ -1300,11 +1300,6 @@ public final class Navigator {
 
         }
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new DescendantEnumeration(start, includeSelf, forwards);
-        }
-
     } // end of class DescendantEnumeration
 
     /**
@@ -1389,11 +1384,6 @@ public final class Navigator {
             }
         }
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new FollowingEnumeration(start);
-        }
-
     } // end of class FollowingEnumeration
 
     /**
@@ -1478,11 +1468,6 @@ public final class Navigator {
         }
 
 
-        /*@NotNull*/
-        public AxisIterator getAnother() {
-            return new PrecedingEnumeration(start, includeAncestors);
-        }
-
     } // end of class PrecedingEnumeration
 
 
