@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -10,6 +10,7 @@ package net.sf.saxon.tree.util;
 import net.sf.saxon.Configuration;
 import net.sf.saxon.event.Builder;
 import net.sf.saxon.event.Receiver;
+import net.sf.saxon.event.ReceiverOptions;
 import net.sf.saxon.expr.parser.Location;
 import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.NodeTest;
@@ -40,8 +41,7 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
     /*@Nullable*/ private NodeName nodeName = null;
     private CharSequence stringValue;
     private SchemaType typeAnnotation = null;
-    private boolean isId;
-    private boolean isIdref;
+    private int options = 0; // Bit-settings defined in ReceiverOptions
 
     /**
      * Create an Orphan node
@@ -133,7 +133,19 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
      */
 
     public void setIsId(boolean id) {
-        this.isId = id;
+        setOption(ReceiverOptions.IS_ID, id);
+    }
+
+    private void setOption(int option, boolean on) {
+        if (on) {
+            this.options |= option;
+        } else {
+            this.options &= ~option;
+        }
+    }
+
+    private boolean isOption(int option) {
+        return (this.options & option) != 0;
     }
 
     /**
@@ -143,7 +155,16 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
      */
 
     public void setIsIdref(boolean idref) {
-        this.isIdref = idref;
+        setOption(ReceiverOptions.IS_IDREF, idref);
+    }
+
+    /**
+     * Set the disable-output-escaping property
+     * @param doe true if the property is to be set
+     */
+
+    public void setDisableOutputEscaping(boolean doe) {
+        setOption(ReceiverOptions.DISABLE_ESCAPING, doe);
     }
 
     /**
@@ -156,6 +177,41 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
         return kind;
     }
 
+    /**
+     * Get fingerprint. The fingerprint is a coded form of the expanded name
+     * of the node: two nodes
+     * with the same name code have the same namespace URI and the same local name.
+     * The fingerprint contains no information about the namespace prefix. For a name
+     * in the null namespace, the fingerprint is the same as the name code.
+     *
+     * @return an integer fingerprint; two nodes with the same fingerprint have
+     * the same expanded QName. For unnamed nodes (text nodes, comments, document nodes,
+     * and namespace nodes for the default namespace), returns -1.
+     * @throws UnsupportedOperationException if this kind of node does not hold
+     *                                       namepool fingerprints (specifically, if {@link #hasFingerprint()} returns false).
+     * @since 8.4 (moved into FingerprintedNode at some stage; then back into NodeInfo at 9.8).
+     */
+    @Override
+    public int getFingerprint() {
+        throw new UnsupportedOperationException();
+    }
+
+    /**
+     * Ask whether this NodeInfo implementation holds a fingerprint identifying the name of the
+     * node in the NamePool. If the answer is true, then the {link #getFingerprint} method must
+     * return the fingerprint of the node. If the answer is false, then the {link #getFingerprint}
+     * method should throw an {@link UnsupportedOperationException}. In the case of unnamed nodes
+     * such as text nodes, the result can be either true (in which case getFingerprint() should
+     * return -1) or false (in which case getFingerprint may throw an exception).
+     *
+     * @return true if the implementation of this node provides fingerprints.
+     * @since 9.8; previously Saxon relied on using <code>FingerprintedNode</code> as a marker interface.
+     */
+    @Override
+    public boolean hasFingerprint() {
+        return false;
+    }
+
     /**
      * Get the typed value.
      * @return the typed value.
@@ -356,40 +412,6 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
         return stringValue;
     }
 
-    /**
-     * Get name code. The name code is a coded form of the node name: two nodes
-     * with the same name code have the same namespace URI, the same local name,
-     * and the same prefix. By masking the name code with &0xfffff, you get a
-     * fingerprint: two nodes with the same fingerprint have the same local name
-     * and namespace URI.
-     *
-     * @see net.sf.saxon.om.NamePool#allocate allocate
-     */
-
-//    public int getNameCode() {
-//        NodeName nn = nodeName;
-//        if (nn == null) {
-//            return -1;
-//        } else {
-//            return nn.allocateNameCode(getNamePool());
-//        }
-//    }
-
-    /**
-     * Get fingerprint. The fingerprint is a coded form of the expanded name
-     * of the node: two nodes
-     * with the same name code have the same namespace URI and the same local name.
-     * A fingerprint of -1 should be returned for a node with no name.
-     */
-
-//    public int getFingerprint() {
-//        if (nodeName == null) {
-//            return -1;
-//        } else {
-//            return nodeName.allocateNameCode(getNamePool()) & NamePool.FP_MASK;
-//        }
-//    }
-
     /**
      * Get the local part of the name of this node. This is the name after the ":" if any.
      *
@@ -605,27 +627,27 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
     }
 
     /**
-     * Determine whether this node has the is-id property
+     * Ask whether this node has the is-id property
      *
      * @return true if the node is an ID
      */
 
     public boolean isId() {
-        return isId || (kind == Type.ATTRIBUTE && nodeName.equals(StandardNames.XML_ID_NAME));
+        return isOption(ReceiverOptions.IS_ID) || (kind == Type.ATTRIBUTE && nodeName.equals(StandardNames.XML_ID_NAME));
     }
 
     /**
-     * Determine whether this node has the is-idref property
+     * Ask whether this node has the is-idref property
      *
      * @return true if the node is an IDREF or IDREFS element or attribute
      */
 
     public boolean isIdref() {
-        return isIdref;
+        return isOption(ReceiverOptions.IS_IDREF);
     }
 
     /**
-     * Determine whether the node has the is-nilled property
+     * Ask whether the node has the is-nilled property
      *
      * @return true if the node has the is-nilled property
      */
@@ -634,6 +656,15 @@ public final class Orphan extends GenericTreeInfo implements MutableNodeInfo {
         return false;
     }
 
+    /**
+     * Ask whether the node has the disable-output-escaping property
+     * @return true if the node has the disable-output-escaping property
+     */
+
+    public boolean isDisableOutputEscaping() {
+        return isOption(ReceiverOptions.DISABLE_ESCAPING);
+    }
+
     /**
      * Insert copies of a sequence of nodes as children of this node.
      * <p/>
