@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -10,6 +10,7 @@ package net.sf.saxon.type;
 import com.saxonica.ee.schema.UserSimpleType;
 import net.sf.saxon.expr.Expression;
 import net.sf.saxon.expr.Literal;
+import net.sf.saxon.expr.StaticProperty;
 import net.sf.saxon.expr.instruct.ValueOf;
 import net.sf.saxon.lib.ConversionRules;
 import net.sf.saxon.lib.NamespaceConstant;
@@ -25,7 +26,7 @@ import java.util.Set;
  * (such as xs:decimal or xs:anyURI) or a derived type (such as xs:ID or xs:dayTimeDuration).
  */
 
-public class BuiltInAtomicType implements AtomicType {
+public class BuiltInAtomicType implements AtomicType, ItemType.WithSequenceTypeCache {
 
     private int fingerprint;
     private int baseFingerprint;
@@ -33,6 +34,10 @@ public class BuiltInAtomicType implements AtomicType {
     private UType uType;
     private boolean ordered = false;
     public StringConverter stringConverter; // may be null for types where conversion rules can vary
+    private SequenceType _one;
+    private SequenceType _oneOrMore;
+    private SequenceType _zeroOrOne;
+    private SequenceType _zeroOrMore;
 
     public final static BuiltInAtomicType ANY_ATOMIC =
             makeAtomicType(StandardNames.XS_ANY_ATOMIC_TYPE, AnySimpleType.getInstance(), true);
@@ -273,6 +278,61 @@ public class BuiltInAtomicType implements AtomicType {
         );
     }
 
+    /**
+     * Get a sequence type representing exactly one instance of this atomic type
+     * @return a sequence type representing exactly one instance of this atomic type
+     * @since 9.8.0.2
+     */
+
+    public SequenceType one() {
+        if (_one == null) {
+            _one = new SequenceType(this, StaticProperty.EXACTLY_ONE);
+        }
+        return _one;
+    }
+
+    /**
+     * Get a sequence type representing zero or one instances of this atomic type
+     *
+     * @return a sequence type representing zero or one instances of this atomic type
+     * @since 9.8.0.2
+     */
+
+    public SequenceType zeroOrOne() {
+        if (_zeroOrOne == null) {
+            _zeroOrOne = new SequenceType(this, StaticProperty.ALLOWS_ZERO_OR_ONE);
+        }
+        return _zeroOrOne;
+    }
+
+    /**
+     * Get a sequence type representing one or more instances of this atomic type
+     *
+     * @return a sequence type representing one or more instances of this atomic type
+     * @since 9.8.0.2
+     */
+
+    public SequenceType oneOrMore() {
+        if (_oneOrMore == null) {
+            _oneOrMore = new SequenceType(this, StaticProperty.ALLOWS_ONE_OR_MORE);
+        }
+        return _oneOrMore;
+    }
+
+    /**
+     * Get a sequence type representing one or more instances of this atomic type
+     *
+     * @return a sequence type representing one or more instances of this atomic type
+     * @since 9.8.0.2
+     */
+
+    public SequenceType zeroOrMore() {
+        if (_zeroOrMore == null) {
+            _zeroOrMore = new SequenceType(this, StaticProperty.ALLOWS_ZERO_OR_MORE);
+        }
+        return _zeroOrMore;
+    }
+
     /**
      * Get the redefinition level. This is zero for a component that has not been redefined;
      * for a redefinition of a level-0 component, it is 1; for a redefinition of a level-N
@@ -439,7 +499,7 @@ public class BuiltInAtomicType implements AtomicType {
      */
 
     public final boolean isPrimitiveType() {
-        return Type.isPrimitiveType(fingerprint);
+        return Type.isPrimitiveAtomicType(fingerprint);
     }
 
     /**
@@ -539,17 +599,6 @@ public class BuiltInAtomicType implements AtomicType {
         return primitiveFingerprint;
     }
 
-    /**
-     * Determine whether this type is supported in a basic XSLT 2.0 processor (all types are
-     * allowed in a basic XSLT 3.0 processor)
-     *
-     * @return true if this type is permitted in a basic XSLT 2.0 processor
-     */
-
-    public boolean isAllowedInBasicXSLT20() {
-        return isPrimitiveType() && getFingerprint() != StandardNames.XS_NOTATION;
-    }
-
     /**
      * Determine whether this type is supported when using XSD 1.0
      *
@@ -564,6 +613,20 @@ public class BuiltInAtomicType implements AtomicType {
         return getDisplayName();
     }
 
+    /**
+     * Return a string representation of this SequenceType suitable for use in export (SEF) files.
+     * The difference from the toString() method is that the representation will not contain
+     * references to anonymous types.
+     *
+     * @return the string representation as an instance of the XPath
+     * SequenceType construct, using only names that will be in-scope if the target environment
+     * imports the same schema as the source environment
+     */
+    @Override
+    public String toExportString() {
+        return toString();
+    }
+
     /**
      * Get the item type of the atomic values that will be produced when an item
      * of this type is atomized
@@ -870,7 +933,7 @@ public class BuiltInAtomicType implements AtomicType {
         if (converter == null) {
             converter = getStringConverter(node.getTreeInfo().getConfiguration().getConversionRules());
             if (isNamespaceSensitive()) {
-                converter = (StringConverter)converter.setNamespaceResolver(new InscopeNamespaceResolver(node));
+                converter = (StringConverter) converter.setNamespaceResolver(new InscopeNamespaceResolver(node));
             }
         }
         return converter.convertString(stringValue).asAtomic();
@@ -903,7 +966,7 @@ public class BuiltInAtomicType implements AtomicType {
         }
         StringConverter converter = getStringConverter(rules);
         if (isNamespaceSensitive()) {
-            converter = (StringConverter)converter.setNamespaceResolver(resolver);
+            converter = (StringConverter) converter.setNamespaceResolver(resolver);
         }
         return converter.convertString(value).asAtomic();
     }
@@ -1136,6 +1199,7 @@ public class BuiltInAtomicType implements AtomicType {
     /**
      * Generate Javascript code to test whether an item conforms to this item type
      * @param knownToBe a type that the item is known to conform to
+     * @param targetVersion 1 or 2 depending on the version of Saxon-JS we are targeting
      * @return a Javascript instruction or sequence of instructions, which can be used as the body
      * of a Javascript function, and which returns a boolean indication whether the value of the
      * variable "item" is an instance of this item type.
@@ -1144,11 +1208,17 @@ public class BuiltInAtomicType implements AtomicType {
      *
      */
     @Override
-    public String generateJavaScriptItemTypeTest(ItemType knownToBe) throws XPathException {
+    public String generateJavaScriptItemTypeTest(ItemType knownToBe, int targetVersion) throws XPathException {
         if (this == BuiltInAtomicType.NOTATION) {
             return "return false;";
         }
-        return "return SaxonJS.U.Atomic." + getName() + ".matches(item);";
+        if (targetVersion == 1) {
+            return "return SaxonJS.U.Atomic." + getName() + ".matches(item);";
+        } else if (targetVersion == 2) {
+            return "return SaxonJS.U.hasAtomicType(item, '" + getName() + "');";
+        } else {
+            throw new IllegalArgumentException();
+        }
     }
 
     /**
@@ -1156,13 +1226,20 @@ public class BuiltInAtomicType implements AtomicType {
      * if conversion is possible, or throw an error otherwise.
      *
      * @param errorCode the error to be thrown if conversion is not possible
+     * @param targetVersion 1 or 2 depending on the version of Saxon-JS we are targeting
      * @return a Javascript instruction or sequence of instructions, which can be used as the body
      * of a Javascript function, and which returns the result of conversion to this type, or throws
      * an error if conversion is not possible. The variable "val" will hold the supplied Javascript
      * value.
      */
-    public String generateJavaScriptItemTypeAcceptor(String errorCode) throws XPathException {
-        return "return SaxonJS.U.Atomic." + getName() + ".cast(val);";
+    public String generateJavaScriptItemTypeAcceptor(String errorCode, int targetVersion) throws XPathException {
+        if (targetVersion == 1) {
+            return "return SaxonJS.U.Atomic." + getName() + ".cast(val);";
+        } else if (targetVersion == 2) {
+            return "return SaxonJS.U.cast(val, '" + getName() + "');";
+        } else {
+            throw new IllegalArgumentException(targetVersion+"");
+        }
     }
 
     //#ifdefined  SCHEMA
