@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,13 +8,13 @@
 package net.sf.saxon.type;
 
 import com.saxonica.ee.schema.UserSimpleType;
-import net.sf.saxon.expr.*;
+import net.sf.saxon.expr.Expression;
+import net.sf.saxon.expr.MappingFunction;
+import net.sf.saxon.expr.MappingIterator;
 import net.sf.saxon.lib.ConversionRules;
 import net.sf.saxon.lib.NamespaceConstant;
 import net.sf.saxon.om.*;
 import net.sf.saxon.trans.XPathException;
-import net.sf.saxon.tree.iter.UnfailingIterator;
-import net.sf.saxon.value.AtomicValue;
 import net.sf.saxon.value.StringValue;
 import net.sf.saxon.value.Whitespace;
 
@@ -467,20 +467,17 @@ public class BuiltInListType implements ListType {
     public ValidationFailure validateContent(
             /*@NotNull*/ CharSequence value, /*@Nullable*/ NamespaceResolver nsResolver, /*@NotNull*/ ConversionRules rules) {
         SimpleType base = getItemType();
-        StringTokenIterator iter = new StringTokenIterator(value.toString());
-        int count = 0;
-        while (true) {
-            StringValue val = iter.next();
-            if (val == null) {
-                break;
-            }
-            count++;
+        Whitespace.Tokenizer iter = new Whitespace.Tokenizer(value);
+        boolean found = false;
+        StringValue val;
+        while ((val = iter.next()) != null) {
+            found = true;
             ValidationFailure v = base.validateContent(val.getStringValue(), nsResolver, rules);
             if (v != null) {
                 return v;
             }
         }
-        if (count == 0) {
+        if (!found) {
             return new ValidationFailure("The built-in list type " +
                     StandardNames.getDisplayName(fingerprint) +
                     " does not allow a zero-length list");
@@ -499,7 +496,7 @@ public class BuiltInListType implements ListType {
 
     /*@NotNull*/
     public AtomicSequence getTypedValue(/*@NotNull*/ CharSequence value, NamespaceResolver resolver, ConversionRules rules) throws ValidationException {
-        UnfailingIterator iter = new StringTokenIterator(value.toString());
+        Whitespace.Tokenizer iter = new Whitespace.Tokenizer(value);
         ListTypeMappingFunction map = new ListTypeMappingFunction();
         map.resolver = resolver;
         map.atomicType = (AtomicType) getItemType();
@@ -518,7 +515,7 @@ public class BuiltInListType implements ListType {
         return t;
     }
 
-    private static class ListTypeMappingFunction implements MappingFunction<StringValue, AtomicValue> {
+    private static class ListTypeMappingFunction implements MappingFunction {
 
         public NamespaceResolver resolver;
         /*@Nullable*/ public AtomicType atomicType;
@@ -531,9 +528,9 @@ public class BuiltInListType implements ListType {
          * For details see {@link net.sf.saxon.expr.MappingFunction}
          */
 
-        public SequenceIterator map(/*@NotNull*/ StringValue item) throws XPathException {
+        public SequenceIterator map(Item item) throws XPathException {
             try {
-                return atomicType.getTypedValue(item.getStringValue(), resolver, rules).iterate();
+                return atomicType.getTypedValue(item.getStringValueCS(), resolver, rules).iterate();
             } catch (ValidationException err) {
                 throw new XPathException(err);
             }
