@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -111,9 +111,9 @@ public abstract class Converter {
             throws ValidationException {
         Converter converter = rules.getConverter(value.getPrimitiveType(), targetType);
         if (converter == null) {
-            ValidationException ve = new ValidationException("Cannot convert value from " + value.getPrimitiveType() + " to " + targetType);
+            ValidationFailure ve = new ValidationFailure("Cannot convert value from " + value.getPrimitiveType() + " to " + targetType);
             ve.setErrorCode("FORG0001");
-            throw ve;
+            throw ve.makeException();
         }
         return converter.convert(value).asAtomic();
     }
@@ -172,17 +172,6 @@ public abstract class Converter {
         return conversionRules;
     }
 
-    /**
-     * Ask if this converter implements a conversion that requires XPath 3.0 (or XQuery 3.0 etc)
-     * to be enabled
-     *
-     * @return true if XPath 3.0 support is required
-     */
-
-    public boolean isXPath30Conversion() {
-        return false;
-    }
-
     /**
      * Ask if this converter will always succeed
      *
@@ -203,8 +192,8 @@ public abstract class Converter {
      */
 
     public Converter setNamespaceResolver(NamespaceResolver resolver) {
-        // no action
         return this;
+        // no action
     }
 
     /**
@@ -430,9 +419,9 @@ public abstract class Converter {
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
             try {
-                return new DecimalValue(((DoubleValue) input).getDoubleValue());
+                return new BigDecimalValue(((DoubleValue) input).getDoubleValue());
             } catch (ValidationException e) {
-                return new ValidationFailure(e);
+                return e.getValidationFailure();
             }
         }
     }
@@ -445,9 +434,9 @@ public abstract class Converter {
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
             try {
-                return new DecimalValue(((FloatValue) input).getFloatValue());
+                return new BigDecimalValue(((FloatValue) input).getFloatValue());
             } catch (ValidationException e) {
-                return new ValidationFailure(e);
+                return e.getValidationFailure();
             }
         }
     }
@@ -460,9 +449,9 @@ public abstract class Converter {
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
             if (input instanceof Int64Value) {
-                return new DecimalValue(((Int64Value) input).longValue());
+                return new BigDecimalValue(((Int64Value) input).longValue());
             } else {
-                return new DecimalValue(((BigIntegerValue) input).asDecimal());
+                return new BigDecimalValue(((BigIntegerValue) input).asDecimal());
             }
         }
 
@@ -480,9 +469,9 @@ public abstract class Converter {
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
             try {
                 BigDecimal decimal = ((NumericValue) input).getDecimalValue();
-                return new DecimalValue(decimal);
-            } catch (XPathException e) {
-                return new ValidationFailure(e);
+                return new BigDecimalValue(decimal);
+            } catch (ValidationException e) {
+                return e.getValidationFailure();
             }
         }
     }
@@ -494,7 +483,7 @@ public abstract class Converter {
     public static class BooleanToDecimal extends Converter {
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
-            return ((BooleanValue) input).getBooleanValue() ? DecimalValue.ONE : DecimalValue.ZERO;
+            return ((BooleanValue) input).getBooleanValue() ? BigDecimalValue.ONE : BigDecimalValue.ZERO;
         }
 
         public boolean isAlwaysSuccessful() {
@@ -536,7 +525,7 @@ public abstract class Converter {
             if (input instanceof IntegerValue) {
                 return input;
             }
-            return BigIntegerValue.makeIntegerValue(((DecimalValue) input).getDecimalValue().toBigInteger());
+            return BigIntegerValue.makeIntegerValue(((BigDecimalValue) input).getDecimalValue().toBigInteger());
         }
 
         public boolean isAlwaysSuccessful() {
@@ -561,8 +550,8 @@ public abstract class Converter {
                 } else {
                     return BigIntegerValue.makeIntegerValue(((NumericValue) input).getDecimalValue().toBigInteger());
                 }
-            } catch (XPathException e) {
-                return new ValidationFailure(e);
+            } catch (ValidationException e) {
+                return e.getValidationFailure();
             }
         }
     }
@@ -783,7 +772,7 @@ public abstract class Converter {
             try {
                 return new Base64BinaryValue(input);
             } catch (XPathException e) {
-                return new ValidationFailure(e);
+                return ValidationFailure.fromException(e);
             }
         }
     }
@@ -808,10 +797,6 @@ public abstract class Converter {
      */
 
     public static class NotationToQName extends Converter {
-        @Override
-        public boolean isXPath30Conversion() {
-            return true;
-        }
 
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
@@ -829,11 +814,6 @@ public abstract class Converter {
 
     public static class QNameToNotation extends Converter {
 
-        @Override
-        public boolean isXPath30Conversion() {
-            return true;
-        }
-
         /*@NotNull*/
         public ConversionResult convert(/*@NotNull*/ AtomicValue input) {
             return new NotationValue(((QNameValue) input).getStructuredQName(), BuiltInAtomicType.NOTATION);
