@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -364,7 +364,7 @@ public final class ErrorType extends NodeTest implements AtomicType, UnionType,
 
     /*@NotNull*/
     public AtomicSequence getTypedValue(CharSequence value, NamespaceResolver resolver, ConversionRules rules) throws ValidationException {
-        throw new ValidationException("Cast to xs:error always fails");
+        throw new ValidationFailure("Cast to xs:error always fails").makeException();
     }
 
     /**
@@ -394,11 +394,6 @@ public final class ErrorType extends NodeTest implements AtomicType, UnionType,
         return new ValidationFailure("No content is ever valid against the type xs:error");
     }
 
-    @Override
-    public ValidationFailure checkAgainstFacets(AtomicValue value, ConversionRules rules) {
-        return null;
-    }
-
     /**
      * Test whether this type represents namespace-sensitive content
      *
@@ -551,6 +546,20 @@ public final class ErrorType extends NodeTest implements AtomicType, UnionType,
         return "xs:error";
     }
 
+    /**
+     * Return a string representation of this SequenceType suitable for use in export (SEF) files.
+     * The difference from the toString() method is that the representation will not contain
+     * references to anonymous types.
+     *
+     * @return the string representation as an instance of the XPath
+     * SequenceType construct, using only names that will be in-scope if the target environment
+     * imports the same schema as the source environment
+     */
+    @Override
+    public String toExportString() {
+        return toString();
+    }
+
     /**
      * Validate that a primitive atomic value is a valid instance of a type derived from the
      * same primitive type.
@@ -609,6 +618,21 @@ public final class ErrorType extends NodeTest implements AtomicType, UnionType,
         return new StructuredQName("xs", NamespaceConstant.SCHEMA, "error");
     }
 
+    /**
+     * Validate an atomic value, which is known to be an instance of one of the member types of the
+     * union, against any facets (pattern facets or enumeration facets) defined at the level of the
+     * union itself.
+     *
+     * @param value the Atomic Value to be checked. This must be an instance of a member type of the
+     *              union
+     * @param rules the ConversionRules for the Configuration
+     * @return a ValidationFailure if the value is not valid; null if it is valid.
+     */
+    @Override
+    public ValidationFailure checkAgainstFacets(AtomicValue value, ConversionRules rules) {
+        return null;
+    }
+
     /**
      * Generate Javascript code to test whether an item conforms to this item type
      *
@@ -618,9 +642,10 @@ public final class ErrorType extends NodeTest implements AtomicType, UnionType,
      * @throws XPathException if JS code cannot be generated for this item type, for example because
      *                        the test is schema-aware.
      * @param knownToBe
+     * @param targetVersion
      */
     @Override
-    public String generateJavaScriptItemTypeTest(ItemType knownToBe) throws XPathException {
+    public String generateJavaScriptItemTypeTest(ItemType knownToBe, int targetVersion) throws XPathException {
         return "return false;";
     }
 
