@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -200,7 +200,7 @@ public abstract class StringConverter extends Converter {
             try {
                 in = phaseTwo.getTargetType().preprocess(in);
             } catch (ValidationException err) {
-                return new ValidationFailure(err);
+                return err.getValidationFailure();
             }
             ConversionResult temp = phaseOne.convertString(in);
             if (temp instanceof ValidationFailure) {
@@ -214,7 +214,7 @@ public abstract class StringConverter extends Converter {
             try {
                 input = phaseTwo.getTargetType().preprocess(input);
             } catch (ValidationException err) {
-                return new ValidationFailure(err);
+                return err.getValidationFailure();
             }
             ConversionResult temp = phaseOne.convertString(input);
             if (temp instanceof ValidationFailure) {
@@ -236,7 +236,7 @@ public abstract class StringConverter extends Converter {
             try {
                 input = phaseTwo.getTargetType().preprocess(input);
             } catch (ValidationException err) {
-                return new ValidationFailure(err);
+                return err.getValidationFailure();
             }
             ConversionResult temp = phaseOne.convertString(input);
             if (temp instanceof ValidationFailure) {
@@ -491,7 +491,7 @@ public abstract class StringConverter extends Converter {
             try {
                 cs = targetType.preprocess(cs);
             } catch (ValidationException err) {
-                return new ValidationFailure(err);
+                return err.getValidationFailure();
             }
             StringValue sv = new StringValue(cs);
             ValidationFailure f = targetType.validate(sv, cs, getConversionRules());
@@ -505,6 +505,11 @@ public abstract class StringConverter extends Converter {
 
         public ValidationFailure validate(/*@NotNull*/ CharSequence input) {
             CharSequence cs = Whitespace.applyWhitespaceNormalization(whitespaceAction, input);
+            try {
+                cs = targetType.preprocess(cs);
+            } catch (ValidationException err) {
+                return err.getValidationFailure();
+            }
             return targetType.validate(new StringValue(cs), cs, getConversionRules());
         }
     }
@@ -535,7 +540,7 @@ public abstract class StringConverter extends Converter {
             try {
                 cs = targetType.preprocess(cs);
             } catch (ValidationException err) {
-                return new ValidationFailure(err);
+                return err.getValidationFailure();
             }
             StringValue sv = new StringValue(cs);
             f = targetType.validate(sv, cs, getConversionRules());
@@ -601,13 +606,13 @@ public abstract class StringConverter extends Converter {
     public static class StringToDecimal extends StringConverter {
         /*@NotNull*/
         public ConversionResult convertString(/*@NotNull*/ CharSequence input) {
-            return DecimalValue.makeDecimalValue(input, true);
+            return BigDecimalValue.makeDecimalValue(input, true);
         }
 
         /*@Nullable*/
         @Override
         public ValidationFailure validate(/*@NotNull*/ CharSequence input) {
-            if (DecimalValue.castableAsDecimal(input)) {
+            if (BigDecimalValue.castableAsDecimal(input)) {
                 return null;
             } else {
                 return new ValidationFailure("Cannot convert string to decimal: " + input.toString());
@@ -864,7 +869,7 @@ public abstract class StringConverter extends Converter {
             try {
                 return new HexBinaryValue(input);
             } catch (XPathException e) {
-                return new ValidationFailure(e);
+                return ValidationFailure.fromException(e);
             }
         }
     }
@@ -881,15 +886,10 @@ public abstract class StringConverter extends Converter {
             super(rules);
         }
 
-        @Override
-        public boolean isXPath30Conversion() {
-            return true;
-        }
-
         public Converter setNamespaceResolver(NamespaceResolver resolver) {
-            StringToQName copy = new StringToQName(getConversionRules());
-            copy.nsResolver = resolver;
-            return copy;
+            StringToQName c = new StringToQName(getConversionRules());
+            c.nsResolver = resolver;
+            return c;
         }
 
         @Override
@@ -940,9 +940,9 @@ public abstract class StringConverter extends Converter {
 
         @Override
         public Converter setNamespaceResolver(NamespaceResolver resolver) {
-            StringToNotation copy = new StringToNotation(getConversionRules());
-            copy.nsResolver = resolver;
-            return copy;
+            StringToNotation c = new StringToNotation(getConversionRules());
+            c.nsResolver = resolver;
+            return c;
         }
 
         @Override
@@ -1063,9 +1063,8 @@ public abstract class StringConverter extends Converter {
         public ConversionResult convertString(/*@NotNull*/ CharSequence input) {
             try {
                 return targetType.getTypedValue(input, null, rules).head();
-            } catch (XPathException err) {
-                //noinspection ThrowableResultOfMethodCallIgnored
-                return new ValidationFailure(ValidationException.makeXPathException(err));
+            } catch (ValidationException err) {
+                return err.getValidationFailure();
             }
         }
     }
