@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -16,7 +16,9 @@ import net.sf.saxon.om.*;
 import net.sf.saxon.pattern.*;
 import net.sf.saxon.tree.iter.AxisIterator;
 import net.sf.saxon.value.AtomicValue;
+import net.sf.saxon.value.Cardinality;
 import net.sf.saxon.value.ObjectValue;
+import net.sf.saxon.value.SequenceType;
 
 
 /**
@@ -120,7 +122,10 @@ public abstract class Type {
     }
 
     /**
-     * Get the ItemType of an Item
+     * Get the ItemType of an Item. This method is used almost entirely for diagnostics, to report
+     * the actual type of an item when it differs from the required type. The actual result is
+     * not completely well-defined, since there are many item types that a given type conforms
+     * to, and it is not always the case that one of them is a subtype of all the others.
      *
      * @param item the item whose type is required
      * @param th   the type hierarchy cache. If null, the returned type may be less precise
@@ -206,9 +211,9 @@ public abstract class Type {
         } else if (item instanceof ObjectValue) {
             return ((ObjectValue) item).getItemType(th);
         } else if (item instanceof MapItem) {
-            return new MapType(((MapItem) item).getKeyType(), ((MapItem) item).getValueType());
+            return ((MapItem)item).getItemType(th);
         } else if (item instanceof ArrayItem) {
-            return new ArrayItemType(((ArrayItem) item).getMemberType());
+            return new ArrayItemType(((ArrayItem) item).getMemberType(th));
         } else { //if (item instanceof FunctionItem) {
             return ((Function) item).getFunctionItemType();
         }
@@ -262,7 +267,7 @@ public abstract class Type {
     }
 
     /**
-     * Get the ItemType object for a built-in atomic type
+     * Get the ItemType object for a built-in atomic or list type
      *
      * @param namespace the namespace URI of the type
      * @param localName the local name of the type
@@ -347,6 +352,25 @@ public abstract class Type {
             Class c2 = ((JavaExternalObjectType) t2).getJavaClass();
             return new JavaExternalObjectType(leastCommonSuperClass(c1, c2));
         }
+        if (t1 instanceof MapType && t2 instanceof MapType) {
+            if (t1 == MapType.EMPTY_MAP_TYPE) {
+                return t2;
+            }
+            if (t2 == MapType.EMPTY_MAP_TYPE) {
+                return t1;
+            }
+            ItemType keyType = getCommonSuperType(((MapType) t1).getKeyType(), ((MapType)t2).getKeyType());
+            AtomicType k;
+            if (keyType instanceof AtomicType) {
+                k = (AtomicType)keyType;
+            } else {
+                k = keyType.getAtomizedItemType().getPrimitiveItemType();
+            }
+            SequenceType v = SequenceType.makeSequenceType(
+                    getCommonSuperType(((MapType) t1).getValueType().getPrimaryType(), ((MapType) t2).getValueType().getPrimaryType()),
+                    Cardinality.union(((MapType) t1).getValueType().getCardinality(), ((MapType) t2).getValueType().getCardinality()));
+            return new MapType(k, v);
+        }
         int r = th.relationship(t1, t2);
         if (r == TypeHierarchy.SAME_TYPE) {
             return t1;
@@ -433,22 +457,32 @@ public abstract class Type {
 
 
     /**
-     * Determine whether this type is a primitive type. The primitive types are
+     * Determine whether a given atomic type is a primitive type. The primitive types are
      * the 19 primitive types of XML Schema, plus xs:integer, xs:dayTimeDuration and xs:yearMonthDuration;
      * xs:untypedAtomic; the 7 node kinds; and all supertypes of these (item(), node(), xs:anyAtomicType,
      * xs:numeric, ...)
      *
-     * @param code the item type code to be tested
+     * @param fingerprint the item type code to be tested
      * @return true if the type is considered primitive under the above rules
      */
-    public static boolean isPrimitiveType(int code) {
-        return code >= 0 && (code <= StandardNames.XS_INTEGER ||
-                code == StandardNames.XS_NUMERIC ||
-                code == StandardNames.XS_UNTYPED_ATOMIC ||
-                code == StandardNames.XS_ANY_ATOMIC_TYPE ||
-                code == StandardNames.XS_DAY_TIME_DURATION ||
-                code == StandardNames.XS_YEAR_MONTH_DURATION ||
-                code == StandardNames.XS_ANY_SIMPLE_TYPE);
+    public static boolean isPrimitiveAtomicType(int fingerprint) {
+        return fingerprint >= 0 && (fingerprint <= StandardNames.XS_INTEGER ||
+                fingerprint == StandardNames.XS_NUMERIC ||
+                fingerprint == StandardNames.XS_UNTYPED_ATOMIC ||
+                fingerprint == StandardNames.XS_ANY_ATOMIC_TYPE ||
+                fingerprint == StandardNames.XS_DAY_TIME_DURATION ||
+                fingerprint == StandardNames.XS_YEAR_MONTH_DURATION ||
+                fingerprint == StandardNames.XS_ANY_SIMPLE_TYPE);
+    }
+
+    /**
+     * Determine whether this type is a type that corresponds exactly to a UType
+     *
+     * @param fingerprint the item type code to be tested
+     * @return true if the type is considered primitive under the above rules
+     */
+    public static boolean isPrimitiveAtomicUType(int fingerprint) {
+        return fingerprint >= 0 && fingerprint <= StandardNames.XS_INTEGER;
     }
 
     /**
