@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -8,7 +8,6 @@
 package net.sf.saxon.value;
 
 import net.sf.saxon.expr.sort.AtomicMatchKey;
-import net.sf.saxon.expr.sort.CodepointCollator;
 import net.sf.saxon.lib.StringCollator;
 import net.sf.saxon.regex.BMPString;
 import net.sf.saxon.regex.LatinString;
@@ -125,6 +124,22 @@ public class StringValue extends AtomicValue {
         }
     }
 
+    /**
+     * Utility method to test whether a CharSequence is empty
+     * @param string the input CharSequence
+     * @return true if the CharSequence is empty
+     */
+
+    public static boolean isEmpty(CharSequence string) {
+        if (string instanceof String) {
+            return ((String) string).isEmpty();
+        } else if (string instanceof UnicodeString) {
+            return ((UnicodeString)string).uLength() == 0;
+        } else {
+            return string.length() == 0;
+        }
+    }
+
 //    public static StringValue makeStringValue(UnicodeString value) {
 //        if (value == null || value.uLength() == 0) {
 //            return StringValue.EMPTY_STRING;
@@ -370,7 +385,19 @@ public class StringValue extends AtomicValue {
      */
 
     public boolean codepointEquals(/*@NotNull*/ StringValue other) {
-        return CodepointCollator.getInstance().comparesEqual(value, other.value);
+        if (value instanceof String) {
+            return ((String) value).contentEquals(other.value);
+        } else if (other.value instanceof String) {
+            return ((String)other.value).contentEquals(value);
+        } else if (value instanceof UnicodeString) {
+            if (!(other.value instanceof UnicodeString)) {
+                other.makeUnicodeString();
+            }
+            return value.equals(other.value);
+        } else {
+            // Avoid conversion to String unless the lengths are equal
+            return value.length() == other.value.length() && value.toString().equals(other.value.toString());
+        }
     }
 
     /**
@@ -490,11 +517,6 @@ public class StringValue extends AtomicValue {
         public void close() {
         }
 
-        /*@NotNull*/
-        public UnfailingIterator getAnother() {
-            return new CharacterIterator(value);
-        }
-
         /**
          * Get properties of this iterator, as a bit-significant integer.
          *
@@ -533,11 +555,6 @@ public class StringValue extends AtomicValue {
         public void close() {
         }
 
-        /*@NotNull*/
-        public UnfailingIterator getAnother() {
-            return new UnicodeCharacterIterator(uValue);
-        }
-
         /**
          * Get properties of this iterator, as a bit-significant integer.
          *
