@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -156,7 +156,7 @@ public final class TimeValue extends CalendarValue implements Comparable {
         }
 
         int tz = 0;
-
+        boolean negativeTz = false;
         int state = 0;
         while (tok.hasMoreElements()) {
             if (state == 9) {
@@ -171,6 +171,9 @@ public final class TimeValue extends CalendarValue implements Comparable {
                     return badTime("decimal point must be followed by digits", s);
                 }
                 part = (String) tok.nextElement();
+                if (part.length() > 9 && part.matches("^[0-9]+$")) {
+                    part = part.substring(0, 9);
+                }
                 value = DurationValue.simpleInteger(part);
                 if (value < 0) {
                     return badTime("Non-numeric fractional seconds component", s);
@@ -208,9 +211,8 @@ public final class TimeValue extends CalendarValue implements Comparable {
                 if (tz > 14 * 60) {
                     return badTime("timezone hour is out of range", s);
                 }
-                //if (tz > 12 * 60) return badTime("Because of Java limitations, Saxon currently limits the timezone to +/- 12 hours");
                 if ("-".equals(delim)) {
-                    tz = -tz;
+                    negativeTz = true;
                 }
             } else if (":".equals(delim)) {
                 if (state != 2) {
@@ -229,10 +231,11 @@ public final class TimeValue extends CalendarValue implements Comparable {
                 if (tzminute > 59) {
                     return badTime("timezone minute is out of range", s);
                 }
-                if (tz < 0) {
-                    tzminute = -tzminute;
-                }
+
                 tz += tzminute;
+                if (negativeTz) {
+                    tz = -tz;
+                }
                 tv.setTimezoneInMinutes(tz);
             } else {
                 return badTime("timezone format is incorrect", s);
@@ -472,9 +475,9 @@ public final class TimeValue extends CalendarValue implements Comparable {
                 return Int64Value.makeIntegerValue(minute);
             case SECONDS:
                 BigDecimal d = BigDecimal.valueOf(microsecond);
-                d = d.divide(DecimalValue.BIG_DECIMAL_ONE_MILLION, 6, BigDecimal.ROUND_HALF_UP);
+                d = d.divide(BigDecimalValue.BIG_DECIMAL_ONE_MILLION, 6, BigDecimal.ROUND_HALF_UP);
                 d = d.add(BigDecimal.valueOf(second));
-                return new DecimalValue(d);
+                return new BigDecimalValue(d);
             case WHOLE_SECONDS: //(internal use only)
                 return Int64Value.makeIntegerValue(second);
             case MICROSECONDS:
