@@ -1,5 +1,5 @@
 ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// Copyright (c) 2015 Saxonica Limited.
+// Copyright (c) 2017 Saxonica Limited.
 // This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 // If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 // This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
@@ -18,7 +18,7 @@ public interface IntIterator {
      * @return true if there are more integers to come
      */
 
-    public boolean hasNext();
+    boolean hasNext();
 
     /**
      * Return the next integer in the sequence. The result is undefined unless hasNext() has been called
@@ -27,5 +27,5 @@ public interface IntIterator {
      * @return the next integer in the sequence
      */
 
-    public int next();
+    int next();
 }
